Package: prodlim
Title: Product-Limit Estimation for Censored Event History Analysis
Version: 2026.03.11
Authors@R: person(given = c("Thomas", "A."),
                    family = "Gerds",
                    role = c("aut", "cre"),
                    email = "tag@biostat.ku.dk")
Description: Fast and user friendly implementation of nonparametric estimators
    for censored event history (survival) analysis. Kaplan-Meier and
    Aalen-Johansen method.
Depends: R (>= 4.1.0)
Imports: Rcpp (>= 0.11.5), stats, rlang, data.table, grDevices,
        ggplot2, scales, graphics, diagram, survival, KernSmooth, lava
Suggests: tibble, ggthemes, riskRegression, testthat, etm, survey
LinkingTo: Rcpp
Maintainer: Thomas A. Gerds <tag@biostat.ku.dk>
BugReports: https://github.com/tagteam/prodlim/issues
License: GPL (>= 2)
Packaged: 2026-03-11 12:23:03 UTC; tag
RoxygenNote: 7.3.2
NeedsCompilation: yes
Author: Thomas A. Gerds [aut, cre]
Repository: CRAN
Date/Publication: 2026-03-11 17:30:02 UTC
Built: R 4.6.0; x86_64-apple-darwin20; 2026-04-28 13:39:00 UTC; unix
Archs: prodlim.so.dSYM

# Generated by roxygen2: do not edit by hand

S3method(as.data.frame,EventHistory.frame)
S3method(as.data.frame,prodlim)
S3method(as.data.table,prodlim)
S3method(getStates,Hist)
S3method(getStates,prodlim)
S3method(lifeTab,competing.risks)
S3method(lifeTab,survival)
S3method(plot,Hist)
S3method(plot,prodlim)
S3method(predict,prodlim)
S3method(print,Hist)
S3method(print,IntIndex)
S3method(print,neighborhood)
S3method(print,potential.followup)
S3method(print,prodlim)
S3method(print,quantile.prodlim)
S3method(print,summary.prodlim)
S3method(quantile,prodlim)
S3method(summary,Hist)
S3method(summary,prodlim)
export('[.Hist')
export(EventHistory.frame)
export(Hist)
export(List2Matrix)
export(PercentAxis)
export(SimCompRisk)
export(SimSurv)
export(SmartControl)
export(as.data.frame.prodlim)
export(as.data.table.prodlim)
export(as_tibble.prodlim)
export(atRisk)
export(backGround)
export(checkCauses)
export(confInt)
export(crModel)
export(dimColor)
export(followup)
export(geom_prodlim)
export(getEvent)
export(getStates)
export(ggprodlim)
export(head.Hist)
export(is.na.Hist)
export(jackknife)
export(jackknife.competing.risks)
export(jackknife.survival)
export(leaveOneOut)
export(leaveOneOut.competing.risks)
export(leaveOneOut.survival)
export(lines.prodlim)
export(markTime)
export(meanNeighbors)
export(model.design)
export(neighborhood)
export(parseFormula)
export(parseSpecialNames)
export(plot.Hist)
export(plot.prodlim)
export(plotCompetingRiskModel)
export(plotIllnessDeathModel)
export(predictSurvIndividual)
export(prodlim)
export(quantile.prodlim)
export(redist)
export(row.match)
export(sindex)
export(stopTime)
export(str.Hist)
export(strip.terms)
export(subset.Hist)
export(summary.prodlim)
export(survModel)
export(tail.Hist)
import(data.table)
import(lava)
importFrom(Rcpp,sourceCpp)
importFrom(grDevices,rainbow)
importFrom(graphics,abline)
importFrom(graphics,axis)
importFrom(graphics,grconvertX)
importFrom(graphics,grconvertY)
importFrom(graphics,lines)
importFrom(graphics,mtext)
importFrom(graphics,par)
importFrom(graphics,plot)
importFrom(graphics,points)
importFrom(graphics,polygon)
importFrom(graphics,rect)
importFrom(graphics,segments)
importFrom(graphics,strheight)
importFrom(graphics,strwidth)
importFrom(graphics,text)
importFrom(stats,.getXlevels)
importFrom(stats,delete.response)
importFrom(stats,drop.terms)
importFrom(stats,formula)
importFrom(stats,get_all_vars)
importFrom(stats,median)
importFrom(stats,model.frame)
importFrom(stats,model.matrix)
importFrom(stats,model.response)
importFrom(stats,na.omit)
importFrom(stats,pchisq)
importFrom(stats,predict)
importFrom(stats,qnorm)
importFrom(stats,quantile)
importFrom(stats,reformulate)
importFrom(stats,terms)
importFrom(stats,update)
importFrom(stats,update.formula)
importFrom(survival,Surv)
importFrom(survival,cluster)
importFrom(survival,survdiff)
useDynLib(prodlim, .registration=TRUE)

R/`prodlim`: Product limit method for event history analysis including the Kaplan-Meier and the Aalen-Johansen estimator
========================================================================================================
[![CRAN\_Status\_Badge](https://www.r-pkg.org/badges/version/prodlim)](https://CRAN.R-project.org/package=prodlim) [![](https://cranlogs.r-pkg.org/badges/prodlim)](https://CRAN.R-project.org/package=prodlim) 
[![R-CMD-check](https://github.com/tagteam/prodlim/actions/workflows/R-CMD-check.yaml/badge.svg)](https://github.com/tagteam/prodlim/actions/workflows/R-CMD-check.yaml)

### as.data.frame.prodlim.R --- 
#----------------------------------------------------------------------
## Author: Thomas Alexander Gerds
## Created: Mar  3 2025 (12:57) 
## Version: 
## Last-Updated: Mar  5 2025 (16:14) 
##           By: Thomas Alexander Gerds
##     Update #: 41
#----------------------------------------------------------------------
## 
### Commentary: 
## 
### Change Log:
#----------------------------------------------------------------------
## 
### Code:
##' Collect results of a fitted prodlim object in a data.frame
##'
##' By default object contains results for all fitted time points and all strata.
##' Use arguments times and newdata of \code{\link{summary.prodlim}} to subset.
##' @title Turn prodlim objects into a data.frame
##' @name as.data.frame.prodlim
##' @param x object obtained with function \code{\link{prodlim}}
##' @param ... passed to \code{\link{summary.prodlim}}
##' @return A data.table with the results of the prodlim object 
##' @seealso \code{\link{prodlim}}
##' @examples
##' set.seed(8)
##' d <- SimCompRisk(17)
##' fit <- prodlim(Hist(time,event)~X1,data=d)
##' as.data.frame.prodlim(fit)
##' as.data.frame.prodlim(fit)
##' 
##'@export as.data.frame.prodlim
##'@export 
##'@author Thomas A. Gerds <tag@@biostat.ku.dk>
as.data.frame.prodlim <- function(x,...){
    args <- list(...)
    if ("newdata"%in%names(args)) X <- args$newdata else X <- x$X 
    if ("times"%in%names(args)) ttt <- args$times else ttt <- sort(unique(c(0,x$time)))
    if ("cause"%in%names(args)) ccc <- args$cause else ccc <- attr(x$model.response,"states")
    args <- c(list(object = x,
                   times = ttt,
                   newdata = X,
                   cause = ccc,
                   format = "df"),args)
    out <- do.call("summary.prodlim",args[!duplicated(names(args))])
    class(out) = "data.frame"
    names(out) = sub("cuminc","absolute_risk",names(out))
    out[]
}


######################################################################
### as.data.frame.prodlim.R ends here

# {{{ SmartControl 
#' Function to facilitate the control of arguments passed to subroutines.
#' 
#' Many R functions need to pass several arguments to several different
#' subroutines. Such arguments can are given as part of the three magic dots
#' "...". The function SmartControl reads the dots together with a list of
#' default values and returns for each subroutine a list of arguments.
#' 
#' 
#' @param call A list of named arguments, as for example can be obtained via
#' \code{list(...)}.
#' @param keys A vector of names of subroutines.
#' @param ignore A list of names which are removed from the argument
#' \code{call} before processing.
#' @param defaults A named list of default argument lists for the subroutines.
#' @param forced A named list of forced arguments for the subroutines.
#' @param split Regular expression used for splitting keys from arguments.
#' Default is \code{"\."}.
#' @param ignore.case If \code{TRUE} then all matching and splitting is not
#' case sensitive.
#' @param replaceDefaults If \code{TRUE} default arguments are replaced by
#' given arguments. Can also be a named list with entries for each subroutine.
#' @param verbose If \code{TRUE} warning messages are given for arguments in
#' \code{call} that are not ignored via argument \code{ignore} and that do not
#' match any \code{key}.
#' @author Thomas Alexander Gerds <tag@@biostat.ku.dk>
#' @seealso \code{\link{plot.prodlim}}
#' @keywords Graphics
#' @examples
#' 
#' 
#' myPlot = function(...){
#'  ## set defaults
#'  plot.DefaultArgs=list(x=0,y=0,type="n")
#'  lines.DefaultArgs=list(x=1:10,lwd=3)
#'  ## apply smartcontrol
#'  x=SmartControl(call=list(...),
#'            defaults=list("plot"=plot.DefaultArgs, "lines"=lines.DefaultArgs),
#'         ignore.case=TRUE,keys=c("plot","axis2","lines"),
#'              forced=list("plot"=list(axes=FALSE),"axis2"=list(side=2)))
#'  ## call subroutines
#'  do.call("plot",x$plot)
#'  do.call("lines",x$lines)
#'  do.call("axis",x$axis2)
#' }
#' myPlot(plot.ylim=c(0,5),plot.xlim=c(0,20),lines.lty=3,axis2.At=c(0,3,4))
#'
#' @export
SmartControl <- function(call,
                         keys,
                         ignore,
                         defaults,
                         forced,
                         split,
                         ignore.case=TRUE,
                         replaceDefaults,
                         verbose=TRUE)
  # }}}
{
  if (missing(split)) split <- "\\."
  # {{{ set up argument list 
  SmartArgs <- as.list(call)
  SmartArgs <- SmartArgs[names(SmartArgs)!=""]
  if (ignore.case==TRUE){
    names(SmartArgs) <- tolower(names(SmartArgs))
  }
  # }}}
  # {{{remove ignorable arguments
  if (!missing(ignore) && is.character(ignore)){
    if (ignore.case==TRUE){
      ignore <- tolower(ignore)
    }
    SmartArgs <- SmartArgs[match(names(SmartArgs),
                                 ignore,
                                 nomatch=0)==0]
  }
  if (verbose==TRUE){
    allKeysRegexp <- paste("^",keys,split,sep="",collapse="|")
    notIgnored <- grep(allKeysRegexp,names(SmartArgs),value=FALSE,ignore.case=TRUE)
    Ignored <- names(SmartArgs)[-notIgnored]
    SmartArgs <- SmartArgs[notIgnored]
    if (length(Ignored)>0){
      paste(Ignored,collapse=", ")
      warning(paste("The following argument(s) are not smart and therefore ignored: ",paste(Ignored,collapse=", ")))
    }
  }
  # }}}
  # {{{ default arguments
  DefaultArgs <- vector(mode="list",length=length(keys))
  names(DefaultArgs) <- keys
  if (!missing(defaults)){
    whereDefault <- match(names(defaults),names(DefaultArgs),nomatch=0)
    if (all(whereDefault))
      DefaultArgs[whereDefault] <- defaults
    else
      stop("Could not find the following default arguments: ",paste(names(defaults[0==whereDefault]),","))
  }
  if (!missing(replaceDefaults)){
    if (length(replaceDefaults)==1){
      replaceDefaults <- rep(replaceDefaults,length(keys))
      names(replaceDefaults) <- keys
    }
    else {
      stopifnot(length(replaceDefaults)==length(keys))
      stopifnot(all(match(names(replaceDefaults),keys)))
      replaceDefaults <- replaceDefaults[keys]
    }
  }
  else{
    replaceDefaults <- rep(FALSE,length(keys))
    names(replaceDefaults) <- keys
  }
  # }}}
  # {{{ forced arguments
  keyForced <- vector(mode="list",length=length(keys))
  names(keyForced) <- keys
  if (!missing(forced)){
    whereDefault <- match(names(forced),names(keyForced),nomatch=0)
    if (all(whereDefault))
      keyForced[whereDefault] <- forced
    else stop("Not all forced arguments found.")
  }
  # }}}
  # {{{ loop over keys
  keyArgList <- lapply(keys,function(k){
    keyRegexp <- paste("^",k,split,sep="")
    foundArgs <- grep(keyRegexp,names(SmartArgs),value=TRUE,ignore.case=TRUE)
    if (length(foundArgs)>0){
      keyArgs <- SmartArgs[foundArgs]
      if (ignore.case)
        argNames <- sapply(strsplit(tolower(names(keyArgs)),tolower(keyRegexp)),function(x)x[[2]])
      else
        argNames <- sapply(strsplit(names(keyArgs),keyRegexp),function(x)x[[2]])
      
      keyArgs <- lapply(keyArgs,function(x){
        ## expressions for arrow labels in plot.Hist
        ## cannot be evaluated at this point
        ## if the expression is communicated
        ## more than one level higher
        maybeFail <- try(e <- eval(x),silent=TRUE)
        if (inherits(maybeFail,"try-error"))
          x
        else
          eval(x)
      })
      names(keyArgs) <- argNames
    }
    else{
      keyArgs <- NULL
    }
    # }}}
    # {{{ prepending the forced arguments-----------------
    if (length(keyForced[[k]])>0){
      keyArgs <- c(keyForced[[k]],keyArgs)
    }
    # }}}
    # {{{ appending default arguments
    if (length(DefaultArgs[[k]])>0 && replaceDefaults[k]==FALSE){
      keyArgs <- c(keyArgs,DefaultArgs[[k]])
    }
    # }}}
    # {{{ removing duplicates
    if (!is.null(names(keyArgs))){
      keyArgs[!duplicated(names(keyArgs))]
    }
  })
  
  names(keyArgList) <- keys
  keyArgList
  # }}}
}

### followup.R --- 
#----------------------------------------------------------------------
## author: Thomas Alexander Gerds
## created: Sep 22 2015 (10:29) 
## Version: 
## last-updated: Apr 25 2025 (08:17) 
##           By: Thomas Alexander Gerds
##     Update #: 37
#----------------------------------------------------------------------
## 
### Commentary: 
## 
### Change Log:
#----------------------------------------------------------------------
## 
### Code:
##' The reverse Kaplan-Meier method estimates the median potential followup time. 
##'
##' This is merely a wrapper for \code{prodlim} with argument reverse. 
##' @title Estimation of the median potential followup time.
##' @param formula A formula whose left hand side is a \code{Hist} or a 
##' \code{\link[survival]{Surv}} object specifying the event time
##' and the event type where by default 0=censored, 1=event, 2=competing risk (if any).
##' Use \code{cens.code} to change the value for censored.
##' @param cens.code Value of the event 
##' @param data A data.frame in which all the variables of
##' \code{formula} can be interpreted.
##' @param ... Arguments passed to \code{prodlim.quantile}. 
##' @return
##' The estimated median potential followup time with inter quartile ranges. 
##' @seealso \code{\link{prodlim}}
##' @examples
##' set.seed(8)
##' d <- SimCompRisk(117)
##'
##' # overall
##' followup(Hist(time,event)~1,data=d)
##' 
##' # in strata defined by variable X1 
##' followup(Hist(time,event)~X1,data=d)
##' @references Michael Schemper and Terry L. Smith. A note on quantifying follow-up in studies of failure time. Controlled Clinical Trials, 17(4):343--346, 1996.
##' @export 
##' @author Thomas A. Gerds <tag@@biostat.ku.dk>
followup <- function(formula,cens.code = 0,data,...){
    G <- prodlim(formula,data,reverse=TRUE)
    x <- quantile(G,q = c(0.25,0.5,0.75),...)
    x <- data.table::copy(x)
    class(x) = c("data.table","data.frame")
    data.table::set(x,j = "lower",value = NULL)
    data.table::set(x,j = "upper",value = NULL)
    class(x) <- c("potential.followup","data.table","data.frame")
    byvars <- c(G$discrete.predictors,G$continuous.predictors)
    if (length(byvars) == 0) {
        byvars <- "X"
        x <- cbind(X = "Overall",x)
    }
    ff <- formula(paste0(paste0(byvars,collapse = "+"),"~q"))
    x <- data.table::dcast(data = x,formula = ff,value.var = "quantile")
    setnames(x,c("0.25","0.5","0.75"),c("Q1","median","Q3"))
    attr(x,"covariates") <- byvars
    class(x) = c("potential.followup","data.table","data.frame")
    x
}
#----------------------------------------------------------------------
### followup.R ends here

##' Extract the states of a multi-state model
##'
##' Applying this function to the fit of prodlim means to apply
##' it to \code{fit$model.response}.
##' @title States of a multi-state model
##' @param object Object of class \code{prodlim} or \code{Hist} .
##' @param ... not used
##' @return A character vector with the states of the model.
##' @author Thomas A. Gerds
#' @export
getStates <- function(object,...){
  UseMethod("getStates",object)
}
#' @export 
getStates.Hist <- function(object,...){
  attr(object,"states")
}

#' @export 
getStates.prodlim <- function(object,...){
  attr(object$model.response,"states")
}

findArrow <- function(Box1,
                      Box2,
                      Box1Dim,
                      Box2Dim,
                      verbose=FALSE){
  left1 <- Box1[1]
  bottom1 <- Box1[2]
  left2 <- Box2[1]
  bottom2 <- Box2[2]
  width1 <- Box1Dim[1]
  height1 <- Box1Dim[2]
  width2 <- Box2Dim[1]
  height2 <- Box2Dim[2]
  
  #    ############################
  #    #p3         p4           p5#
  #    #                          #
  #    #                          #
  #    #p2                      p6# 
  #    #                          #
  #    #                          #
  #    #p1          p8          p7#
  #    ############################

  box1 <- list(left=as.numeric(left1),
               right=as.numeric(left1+width1),
               mid.horizontal=as.numeric(left1+width1/2),
               bottom=as.numeric(bottom1),
               top=as.numeric(bottom1+height1),
               mid.vertical=as.numeric(bottom1+height1/2))
  
  box1$p1 <- c(x=box1$left,y=box1$bottom)
  box1$p2 <- c(x=box1$left,y=box1$mid.vertical)
  box1$p3 <- c(x=box1$left,y=box1$top)
  box1$p4 <- c(x=box1$mid.horizontal,y=box1$top)
  box1$p5 <- c(x=box1$right,y=box1$top)
  box1$p6 <- c(x=box1$right,y=box1$mid.vertical) 
  box1$p7 <- c(x=box1$right,y=box1$bottom) 
  box1$p8 <- c(x=box1$mid.horizontal,y=box1$bottom) 
  
  box2 <- list(left=as.numeric(left2),
               right=as.numeric(left2+width2),
               mid.horizontal=as.numeric(left2+width2/2),
               bottom=as.numeric(bottom2),
               top=as.numeric(bottom2+height2),
               mid.vertical=as.numeric(bottom2+height2/2))
  box2$p1 <- c(x=box2$left,y=box2$bottom)
  box2$p2 <- c(x=box2$left,y=box2$mid.vertical)
  box2$p3 <- c(x=box2$left,y=box2$top)
  box2$p4 <- c(x=box2$mid.horizontal,y=box2$top)
  box2$p5 <- c(x=box2$right,y=box2$top)
  box2$p6 <- c(x=box2$right,y=box2$mid.vertical) 
  box2$p7 <- c(x=box2$right,y=box2$bottom) 
  box2$p8 <- c(x=box2$mid.horizontal,y=box2$bottom) 
  ##   boxwidth <- abs(box1$left-box1$right)
  ##   boxheight <- abs(box1$top-box1$bottom)
  direction <- 1
  if (box2$mid.horizontal <box1$mid.horizontal){  
    if (verbose) print("change boxes")
    direction <- 2
    tmpBox <- box1
    box1 <- box2
    box2 <- tmpBox
  }
  
  ##   points(box1$left,box1$bottom)
  ##   points(box2$left,box2$bottom,col=2)
  ##   print(list(box1,box2))
  
  if (box1$mid.horizontal==box2$mid.horizontal){
    if (box1$bottom<box2$bottom){
      #########################
      ####      2        
      ####      |
      ####      1
      ######################### 
      if (verbose==TRUE) print("case 0a: top -> bottom")
      out <- list(from=box1$p4,to=box2$p8)
    }
    else{
      #########################
      ####      1      
      ####      |
      ####      2
      ######################### 
      if (verbose==TRUE) print("case 0: bottom -> top")
      out <- list(from=box1$p8,to=box2$p4)
    }
  }
  else{
    ##     if (box1$right<=box2$left){        
    if (box1$bottom<=box2$bottom){
      if (box1$top >= box2$bottom){
        #########################
        ####          2
        ####  1   ->  
        ####
        #########################
        if (verbose==TRUE) {
          print("case 2: mid.left -> mid.right")
          print(c(from=box1$p6,to=box2$p2))
        }
        out <- list(from=box1$p6,to=box2$p2)
        #########################
        ## THIS IS A SPECIAL CASE
        ####
        ####  1   ->  2
        ####
        #########################
      }
      else{ # box1$top < box2$bottom
        if ((box2$bottom-box1$top) <= (box2$left-box1$right)){
          if ((box2$bottom-box1$top) <= .5*(box2$left-box1$right)){
            #########################
            ####    -> 2
            ####   /   
            ####  1    
            #########################
            if (verbose==TRUE) print("case 3a: corner.left.top -> mid.right")
            out <- list(from=box1$p5,to=box2$p2)
          }
          else{
            #########################
            ####     2
            ####   /   
            ####  1    
            #########################
            if (verbose==TRUE) print("case 3b: corner.left.top -> corner.right.bottom")
            out <- list(from=box1$p5,to=box2$p1)
          }
        }
        else{
          #########################
          ####    2
          ####   / 
          ####  | 
          ####  1    
          #########################
          if (verbose==TRUE) print("case 4: top.left -> bottom.right")
          out <- list(from=box1$p4,to=box2$p8)
        }
      }
    }
    ##   }
    else{ ## box1$bottom>box2$bottom
      if (box2$top>=box1$bottom){
        #########################
        ####          
        ####  1   ->  
        ####          2
        #########################
        if (verbose==TRUE) {
          print("case 5: mid.left -> mid.right")
          print(c(from=box1$p6,to=box2$p2))
        }
        
        out <- list(from=box1$p6,to=box2$p2)
      }
      else{
        if ((box1$bottom-box2$top) <= (box2$left-box1$right)){
          ## print((box1$bottom-box2$top) <= .5*(box2$left-box1$right))
          if ((box1$bottom-box2$top) <= .5*(box2$left-box1$right)){
            #########################
            ####  1        
            ####    \
            ####      2
            #########################
            if (verbose==TRUE) print("case 6a: corner.left.bottom -> mid.right")
            out <- list(from=box1$p7,to=box2$p2)
          }
          else{
            #########################
            ####  1        
            ####    \
            ####      2
            #########################
            if (verbose==TRUE) print("case 6b: corner.left.bottom -> corner.right.top")
            out <- list(from=box1$p7,to=box2$p3)
          }
        }
        else{
          if (box1$bottom>=box2$top){
            #########################
            ####  1        
            ####     \->   2
            ####
            #########################
            if (verbose==TRUE) print("case 7: top.left -> bottom.right")
            out <- list(from=box1$p8,to=box2$p4)
          }
        }
      }
    }
  }
  if (direction==2){
    names(out) <- c("to","from")
  }
  out
}

### stopTime.R --- 
#----------------------------------------------------------------------
## author: Thomas Alexander Gerds
## created: Nov 28 2015 (10:07) 
## Version: 
## last-updated: Dec  4 2015 (06:57) 
##           By: Thomas Alexander Gerds
##     Update #: 23
#----------------------------------------------------------------------
## 
### Commentary: 
## 
### Change Log:
#----------------------------------------------------------------------
## 
### Code:
##' All event times are stopped at a given time point and
##' corresponding events are censored
##'
##' @title Stop the time of an event history object
##' @param object Event history object as obtained with \code{Hist}
##' @param stop.time Time point at which to stop the event history object
##' @return Stopped event history object where all times are censored 
##'         at \code{stop.time}. All observations with times greater than \code{stop.time}
##'         are set to \code{stop.time} and the event status is set to \code{attr(object,"cens.code")}.
##'         A new column \code{"stop.time"} is equal to \code{1} for stopped observations
##'         and equal to \code{0} for the other observations.
##' @seealso Hist 
##' @examples
##'
##' set.seed(29)
##' d <- SimSurv(10)
##' h <- with(d,Hist(time,status))
##' h
##' stopTime(h,8)
##' stopTime(h,5)
##'
##' ## works also with Surv objects
##' library(survival)
##' s <- with(d,Surv(time,status))
##' stopTime(s,5)
##'
##' ## competing risks
##' set.seed(29)
##' dr <- SimCompRisk(10)
##' hr <- with(dr,Hist(time,event))
##' hr
##' stopTime(hr,8)
##' stopTime(hr,5)
##' 
##' @export 
##' @author Thomas A. Gerds <tag@@biostat.ku.dk>
stopTime <- function(object,stop.time){
    if (missing(stop.time)) stop("Argument stop.time missing. Need a time point at which to stop the event history.")
    if (length(stop.time)>1) {
        warning("Argument stop.time is a vector. Proceed with the first element.")
        stop.time <- stop.time[[1]]
    }
    cc <- class(object)[[1]]
    stopifnot(cc%in% c("Hist","Surv"))
    if (cc=="Surv"){
        model <- "survival"
    }else{
         model <- attr(object,"model")
         if(!(model %in% c("survival","competing.risks")))
             stop(paste("Don't know (not yet) how to stop this type of model:",model))
     }
    stopped <- object[,"time"] >= stop.time
    sobject <- cbind(object,"stopped"=1*stopped)
    sobject[,"status"][stopped] <- 0
    if(model=="competing.risks") 
        sobject[,"event"][stopped] <- length(attr(object,"states"))+1
    sobject[,"time"][stopped] <- stop.time
    attr(sobject,"stop.time") <- stop.time
    attr(sobject,"class") <- attr(object,"class")
    if (cc=="Surv"){
        attr(sobject,"type") <- attr(object,"type")
    }
    attr(sobject,"states") <- attr(object,"states")
    attr(sobject,"model") <- attr(object,"model")
    attr(sobject,"cens.type") <- attr(object,"cens.type")
    attr(sobject,"cens.code") <- attr(object,"cens.code")
    attr(sobject,"entry.type") <- attr(object,"entry.type")
    sobject
}
#----------------------------------------------------------------------
### stopTime.R ends here

##' This function calls first \code{\link{col2rgb}} on a color name and then 
##' uses \code{\link{rgb}} to adjust the intensity of the result.
##' 
##' @title Dim a given color to a specified density
##' @param col Color name or number passed to \code{\link{col2rgb}}.
##' @param density Integer value passed as alpha coefficient to
##' \code{\link{rgb}} between 0 and 255
##' @return A character vector with the color code. See \code{rgb} for details.
##' @seealso rgb col2rgb
##' @examples 
##' dimColor(2,33)
##' dimColor("green",133)
##' @export 
##' @author Thomas A. Gerds <tag@@biostat.ku.dk>
dimColor <- function(col,density=55){
    ccrgb=as.list(grDevices::col2rgb(col,alpha=TRUE))
    names(ccrgb) <- c("red","green","blue","alpha")
    ccrgb$alpha=density
    do.call(grDevices::rgb,c(ccrgb,list(max=255)))
}

##' Simulate right censored competing risks data with two covariates X1 and X2. Both covariates have effect exp(1) on the hazards of event 1 and zero effect on the hazard of event 2.
##'
##' This function calls \code{crModel}, then adds covariates and finally calls \code{sim.lvm}.
##' @title Simulate competing risks data
##' @param N sample size
##' @param ... do nothing.
##' @return data.frame with simulated data
##' @author Thomas Alexander Gerds
##' @examples
##' 
##' SimCompRisk(10)
##'             
##' @export
SimCompRisk <- function(N, ...){
    ## require(lava)
    m <- crModel()
    regression(m,from="X1",to="eventtime1") <- 1
    regression(m,from="X2",to="eventtime1") <- 1
    distribution(m,"X1") <- binomial.lvm()
    out <- sim(m,N)
    ## for backward compatibility
    out$cause <- out$event
    out
}
 

library(testthat)
library(prodlim)
library(data.table)
context("Prodlim")
test_that("Competing risks: asking for case for event that does not exist in data",{
    ##
    set.seed(10)
    d <- SimSurv(10)
    setDT(d)
    d[,event:=factor(event,levels=c(0,1),labels=c("0","2"))]
    f <- prodlim(Hist(time,event)~X1,data=d)
    predict(f,cause="2",times=8,newdata=data.frame(X1=1))
    expect_error(predict(f,cause="1",times=8,newdata=data.frame(X1=1)))
    set.seed(10)
    dd <- SimCompRisk(20)
    F <- prodlim(Hist(time,event)~X1,data=dd)
    predict(F,cause="1",times=4,newdata=data.frame(X1=0:1))
    expect_equal(lapply(predict(F,cause=2,times=4,newdata=data.frame(X1=0:1)),round,4),list(`X1=0`=0.0714,`X1=1`=0))
    expect_error(predict(F,cause=3,times=4,newdata=data.frame(X1=0:1)))
    expect_error(summary(F,cause=3))
    expect_error(plot(F,cause=3))
})

test_that("strata",{
    ## bug in version 1.5.1
    d <- data.frame(time=1:3,status=c(1,0,1),a=c(1,9,9),b=factor(c(0,1,0)))
    expect_output(print(prodlim(Hist(time,status)~b+factor(a),data=d)))
})


test_that("prodlim: print and summary  ",{
    library(lava)
    library(riskRegression)
    library(survival)
    # lava some data
    m <- crModel()
    addvar(m) <- ~X1+X2+X3+X4+X5+X6
    distribution(m,"X3") <- binomial.lvm()
    distribution(m,"X4") <- normal.lvm(mean=50,sd=10)
    distribution(m,"eventtime1") <- coxWeibull.lvm(scale=1/200)
    distribution(m,"censtime") <- coxWeibull.lvm(scale=1/1000)
    m <- categorical(m,K=4,eventtime1~X5,beta=c(1,0,0,0),p=c(0.1,0.2,0.3))
    m <- categorical(m,K=3,eventtime1~X1,beta=c(2,1,0),p=c(0.3,0.2))
    regression(m,to="eventtime1",from=c("X2","X4")) <- c(0.3,0)
    regression(m,to="eventtime2",from=c("X2","X4")) <- c(0.6,-0.07)
    set.seed(17)
    d <- sim(m,200)
    d$X1 <- factor(d$X1,levels=c(0,1,2),labels=c("low survival","medium survival","high survival"))
    d$X5 <- factor(d$X5,levels=c("0","1","2","3"),labels=c("one","two","three","four"))
    d$Event <- factor(d$event,levels=c("0","1","2"),labels=c("0","cause-1","cause-2"))
    d$status <- 1*(d$event!=0)
    # marginal Kaplan-Meier
    s0 <- prodlim(Hist(time,status)~1,data=d)
    expect_output(print(s0))
    a <- summary(s0,intervals=TRUE)
    b <- summary(s0,intervals=TRUE,format="list")$table
    attributes(a) <- attributes(b)
    expect_equal(a,b)
    stats::predict(s0,times=1:10)
    # 
    # stratified Kaplan-Meier vs subsetting
    #
    su <- prodlim(Hist(time,status)~1,data=d,subset=d$X1=="medium survival")
    s1 <- prodlim(Hist(time,status)~X1,data=d)
    S1 <- summary(s1,intervals=0,newdata=data.frame(X1=c("medium survival")))
    S2 <- summary(s1,intervals=0,newdata=data.frame(X1=c("medium survival","high survival","low survival")))
    S1a <- S2[S2$X1=="medium survival",]
    rownames(S1a) <- 1:NROW(S1a)
    expect_equal(S1,S1a)
    stats::predict(s1,times=0:10,newdata=data.frame(X1=c("medium survival","low survival","high survival")))
    suppressWarnings(A <- summary(su,intervals=0,times=S1$time))
    S1$X1 <- NULL
    attributes(A) <- attributes(S1)
    all.equal(A,S1)
    #
    # Beran estimator
    #
    s2 <- prodlim(Hist(time,status)~X2,data=d)
    expect_output(print(s2))
    suppressWarnings(summary(s2,intervals=TRUE))
    stats::predict(s2,times=0:10,newdata=data.frame(X2=quantile(d$X2)))
    #
    # Kaplan-Meier: two strata variable
    #
    s1a <- prodlim(Hist(time,status)~X1+X3,data=d)
    expect_output(print(s1a))
    stats::predict(s1a,times=0:10,newdata=expand.grid(X1=levels(d$X1),X3=unique(d$X3)))
    summary(s1a,intervals=TRUE,newdata=s1a$X[c(2,3),])
    #
    # Stratified Beran estimator
    #
    s3 <- prodlim(Hist(time,status)~X1+X2,data=d)
    expect_output(print(s3))
    stats::predict(s3,times=0:10,newdata=expand.grid(X1=levels(d$X1),X2=c(quantile(d$X2,0.05),median(d$X2))))
    suppressWarnings(summary(s3,intervals=TRUE))
    #
    # marginal Aalen-Johansen competing risks
    #
    f0 <- prodlim(Hist(time,event)~1,data=d)
    expect_output(print(f0))
    summary(f0,intervals=TRUE)
    stats::predict(f0,times=1:10)
    #
    # stratified Aalen-Johansen competing risks
    #
    f1 <- prodlim(Hist(time,event)~X1,data=d)
    expect_output(print(f1))
    summary(f1,intervals=TRUE,newdata=data.frame(X1=c("medium survival","high survival","low survival")))
    summary(f1,intervals=TRUE,cause=2,newdata=data.frame(X1=c("medium survival","low survival")))
    stats::predict(f1,times=0:10,newdata=data.frame(X1=c("medium survival","low survival","high survival")))
    #
    # Beran-Aalen-Johansen competing risks
    #
    f2 <- prodlim(Hist(time,event)~X2,data=d)
    expect_output(print(f2))
    suppressWarnings(summary(f2,intervals=TRUE))
    stats::predict(f2,times=0:10,newdata=data.frame(X2=quantile(d$X2)))
    #
    # 2-strata Aalen-Johansen competing risks
    #    
    f1a <- prodlim(Hist(time,event)~X1+X3,data=d)
    expect_output(print(f1a))
    summary(f1a,intervals=TRUE)
    stats::predict(f1a,times=0:10,newdata=expand.grid(X1=levels(d$X1),X3=unique(d$X3)))
    #
    # stratified Beran-Aalen-Johansen competing risks
    #    
    f3 <- prodlim(Hist(time,event)~X1+X2,data=d)
    expect_output(print(f3))
    suppressWarnings(summary(f3,intervals=TRUE))
    stats::predict(f3,times=0:10,newdata=expand.grid(X1=levels(d$X1),X2=c(quantile(d$X2,0.05),median(d$X2))))
})

test_that("prodlim vs survfit",{
    data(pbc)
    prodlim.0 <- prodlim(Hist(time,status!=0)~1,data=pbc)
    survfit.0 <- survfit(Surv(time,status!=0)~1,data=pbc)
    ttt <- sort(unique(pbc$time)[pbc$status!=0])
    ttt <- ttt[-length(ttt)]
    sum0.s <- summary(survfit.0,times=ttt)
    ## There is arounding issue with survfit:
    #---------------------------------------------------------------------------------------
    ## testdata <- data.frame(time=c(16.107812,3.657545,1.523978),event=c(0,1,1))
    ## sum0 <- summary(survfit(Surv(time,event)~1,data=testdata),times=sort(testdata$time))
    ## testdata$timeR <- round(testdata$time,1)
    ## sum1 <- summary(survfit(Surv(timeR,event)~1,data=testdata),times=sort(testdata$time))
    ## sum0
    ## sum1
    ## sum0 != sum1
    ## summary(survfit.0,times=c(0,0.1,0.2,0.3))
    #---------------------------------------------------------------------------------------
    result.survfit <- data.frame(time=sum0.s$time,n.risk=sum0.s$n.risk,n.event=sum0.s$n.event,surv=sum0.s$surv,std.err=sum0.s$std.err,lower=sum0.s$lower,upper=sum0.s$upper)
    result.prodlim <- data.frame(summary(prodlim.0,times=ttt)[,c("time","n.risk","n.event","n.lost","surv","se.surv","lower","upper")])
    ## cbind(result.survfit[,c("time","n.risk","n.event","surv")],result.prodlim[,c("time","n.risk","n.event","surv")])
    a <- round(result.survfit$surv,8)
    b <- round(result.prodlim$surv[!is.na(result.prodlim$se.surv)],8)
    expect_equal(a,b)
    expect_equal(round(result.survfit$std.err,8),round(result.prodlim$se.surv[!is.na(result.prodlim$se.surv)],8))
    pbc <- pbc[order(pbc$time,-pbc$status),]
    set.seed(17)
    boot <- sample(1:NROW(pbc),size=NROW(pbc),replace=TRUE)
    boot.weights <- table(factor(boot,levels=1:NROW(pbc)))
    s1 <- prodlim(Hist(time,status>0)~1,data=pbc,caseweights=boot.weights)
    ## plot(s1,col=1,confint=FALSE,lwd=8)
    s2 <- prodlim(Hist(time,status>0)~1,data=pbc[sort(boot),])
    ## plot(s2,add=TRUE,col=2,confint=FALSE,lwd=3)
    expect_equal(summary(s1,intervals=1,times=seq(0,3500,500)),summary(s2,intervals=1,times=seq(0,3500,500)))
})


test_that("weigths, subset and smoothing",{
    d <- SimSurv(100)
    f1 <- prodlim(Hist(time,status)~X2,data=d)
    f2 <- prodlim(Hist(time,status)~X2,data=d,caseweights=rep(1,100))
    expect_equal(f1$surv,f2$surv)
    d <- SimSurv(100)
    d <- data.frame(d, group = c(rep(1, 70), rep(0,30)))
    f1a <- prodlim(Hist(time,status)~X2,data=d, caseweights = rep(1, 100), subset = d$group==1,bandwidth=0.1)
    f1b <- prodlim(Hist(time,status)~X2,data=d[d$group==1, ], caseweights = rep(1, 100)[d$group==1], bandwidth=0.1)
    f1a$call <- f1b$call
    expect_equal(f1a,f1b)
    f1 <- prodlim(Hist(time,status)~X1,data=d, subset = d$group==1)
    f2 <- prodlim(Hist(time,status)~X1,data=d,caseweights=d$group)
    expect_equal(unique(f1$surv),unique(f2$surv))
    expect_equal(predict(f1,newdata = d[1, ], times = 5),
                 predict(f2, newdata = d[1, ], times = 5))
})

test_that("weights and delay",{
    library(survival)
    library(survey)
    ## library(SmoothHazard)
    library(etm)
    pbc <- pbc[order(pbc$time,-pbc$status),]
    ## pbc$randprob<-fitted(biasmodel)
    ## pbc$randprob <- as.numeric(pbc$sex=="m")+0.1
    set.seed(17)
    pbc$randprob <- abs(rnorm(NROW(pbc)))
    dpbc <- svydesign(id=~id, weights=~randprob, strata=NULL, data=pbc)
    survey.1<-svykm(Surv(time,status>0)~1, design=dpbc)
    ## plot(survey.1,lwd=8)
    prodlim.1 <- prodlim(Hist(time,status>0)~1,data=pbc,caseweights=pbc$randprob)
    ## plot(prodlim.1,add=TRUE,col=2,confint=FALSE)
    pbc$entry <- round(pbc$time/5)
    survfit.delay <- survfit(Surv(entry,time,status!=0)~1,data=pbc)
    prodlim.delay <- prodlim(Hist(time,status!=0,entry=entry)~1,data=pbc)
    a <- summary(survfit.delay)
    b <- summary(prodlim.delay,times=a$time)
    expect_equal(a$surv,b$surv)
    ## plot(survfit.delay,lwd=8)
    ## plot(prodlim.delay,lwd=4,col=2,add=TRUE,confint=FALSE)
    pbc0 <- pbc
    pbc0$entry <- round(pbc0$time/5)
    survfit.delay.edema <- survfit(Surv(entry,time,status!=0)~edema,data=pbc0)
    ## survfit.delay.edema.0.5 <- survfit(Surv(entry,time,status!=0)~1,data=pbc0[pbc0$edema==0.5,])
    prodlim.delay.edema <- prodlim(Hist(time,status!=0,entry=entry)~edema,data=pbc0)
    ## prodlim.delay.edema.0.5 <- prodlim(Hist(time,status!=0,entry=entry)~1,data=pbc0[pbc0$edema==0.5,])
    ## plot(survfit.delay.edema,conf.int=FALSE,col=1:3,lwd=8)
    ## plot(prodlim.delay.edema,add=TRUE,confint=FALSE,col=c("gray88","orange",5),lwd=4)
    ## a <- summary(survfit.delay.edema)
    ## b <- summary(prodlim.delay.edema,times=a$time)
    ## expect_equal(a$surv,b$surv)    
    data(abortion)
    cif.ab.etm <- etmCIF(Surv(entry, exit, cause != 0) ~ 1,abortion,etype = cause,failcode = 3)
    cif.ab.prodlim <- prodlim(Hist(time=exit, event=cause,entry=entry) ~ 1,data=abortion)
    ## plot(cif.ab.etm,lwd=8,col=3)
    ## plot(cif.ab.prodlim,add=TRUE,lwd=4,col=5,cause=3)
    x <- prodlim(Hist(time=exit, event=cause,entry=entry) ~ 1,data=abortion)
    x0 <- etmCIF(Surv(entry, exit, cause != 0) ~ 1,data=abortion,etype = cause)
    ## graphics::par(mfrow=c(2,2))
    cif.ab.etm <- etmCIF(Surv(entry, exit, cause != 0) ~ 1,abortion,etype = cause,failcode = 3)
    cif.ab.prodlim <- prodlim(Hist(time=exit, event=cause,entry=entry) ~ 1,data=abortion)
                                        # cause 3
    ## plot(cif.ab.etm, ci.type = "bars", pos.ci = 24, col = c(1, 2), lty = 1,which.cif=3,lwd=8)
    ## plot(cif.ab.prodlim,add=TRUE,cause=3,confint=TRUE,col=2)
                                        # cause 2
    ## plot(cif.ab.etm, ci.type = "bars", pos.ci = 24, col = c(1, 2), lty = 1,which.cif=2,lwd=8)
    ## plot(cif.ab.prodlim,add=TRUE,cause=2,confint=TRUE,col=2)
                                        # cause 1
    ## plot(cif.ab.etm, ci.type = "bars", pos.ci = 24, col = c(1, 2), lty = 1,which.cif=1,lwd=8)
    ## plot(cif.ab.prodlim,add=TRUE,cause=1,confint=TRUE,col=2)
    cif.ab.etm <- etmCIF(Surv(entry, exit, cause != 0) ~ group,abortion,etype = cause,failcode = 3)
    names(cif.ab.etm[[1]])
    head(cbind(cif.ab.etm[[1]]$time,cif.ab.etm[[1]]$n.risk))
    cif.ab.prodlim <- prodlim(Hist(time=exit, event=cause,entry=entry) ~ group,data=abortion)
    ## plot(cif.ab.etm, ci.type = "bars", pos.ci = 24, col = c(1, 2), lty = 1, curvlab = c("Control", "Exposed"),lwd=8)
    ## plot(cif.ab.prodlim,add=TRUE,cause=3,confint=FALSE,col="yellow")
    testdata <- data.frame(entry=c(1,5,2,8,5),exit=c(10,6,4,12,33),event=c(0,1,0,1,0))
    cif.test.etm <- etmCIF(Surv(entry, exit, event) ~ 1,data=testdata,etype = event,failcode = 1)
    cif.test.survival <- survfit(Surv(entry, exit, event) ~ 1,data=testdata)
    cif.test.prodlim <- prodlim(Hist(exit,event,entry=entry)~1,data=testdata)
    ## plot(cif.test.etm, ci.type = "bars", pos.ci = 24, lwd=5)
    ## plot(cif.test.etm, ci.type = "bars", pos.ci = 24, lwd=5)
    ## plot(cif.test.prodlim,add=TRUE,cause=2,col=2,confint=TRUE,type="cuminc")
    ## simulate data from an illness-death model
    ## mod <- idmModel(K=10,schedule=0,punctuality=1)
    ## regression(mod,from="X",to="lifetime") <- log(2)
    ## regression(mod,from="X",to="waittime") <- log(2)
    ## regression(mod,from="X",to="illtime") <- log(2)
    ## set.seed(137)
    ## we round the event times to have some ties
    ## testdata <- round(sim(mod,250),1)
    ## the data enter with delay into the intermediate state (ill)
    ## thus, to estimate the absolute risk cumulative incidence of
    ## the absorbing state (death) after illness we 
    ## have left-truncated data
    ## illdata <- testdata[testdata$illstatus==1,]
    ## illdata <- illdata[order(illdata$lifetime,-illdata$seen.exit),]
    ## sindex(jump.times=illdata$illtime,eval.times=illdata$lifetime)
    ## F <- prodlim(Hist(lifetime,status,entry=illtime)~1,data=illdata[1:5,])
    ## f <- survfit(Surv(illtime,lifetime,status)~1,data=illdata[1:5,],type="kaplan-meier")
    ## survfit.delayed.ill <- survfit(Surv(illtime,lifetime,seen.exit)~1,data=illdata)
    ## prodlim.delayed.ill <- prodlim(Hist(lifetime,seen.exit,entry=illtime)~1,data=illdata)
    ## plot(survfit.delayed.ill,lwd=5)
    ## plot(prodlim.delayed.ill,lwd=2,col=2,add=TRUE)
})

test_that("left truncation: survival",{
    library(prodlim)
    library(data.table)
    library(survival)
    dd <- data.table(entry=c(1,1,56,1,1,225,277,1647,1,1),
                     time=c(380,46,217,107,223,277,1638,2164,45,40),
                     status=c(1,0,1,1,0,0,0,1,0,1))
    ## --------------------------------------------------------------
    ## by convention in case of ties 
    ## entry happens after events and after censoring
    ## --------------------------------------------------------------
    prodlim.delayed <- prodlim(Hist(time,status,entry=entry)~1,data=dd)
    data.table(time=prodlim.delayed$time,n.risk=prodlim.delayed$n.risk,n.event=prodlim.delayed$n.event,n.lost=prodlim.delayed$n.lost)
    summary(prodlim.delayed,times=c(0,10,56,267,277,1000,2000))    
    survfit.delayed <- survfit(Surv(entry,time,status)~1,data=dd)
    summary(prodlim.delayed,times=c(0,10,40),intervals=TRUE)
    summary(survfit.delayed,times=c(0,1,10,40,50))
    summary.survfit.delayed <- summary(survfit.delayed,times=c(0,10,56,267,277,1000,2000))
    summary.prodlim.delayed <- summary(prodlim.delayed,times=c(0,10,56,267,277,1000,2000),intervals=1)
    expect_equal(as.numeric(summary.survfit.delayed$surv),
                 as.numeric(unlist(summary.prodlim.delayed[,"surv"])))
    ## FIXME: lifetab does not handle delayed entry
    ##        and shows wrong numbers at risk before the
    ##        first event time
    ## expect_equal(as.numeric(summary.survfit.delayed$n.risk),
    ## as.numeric(summary.prodlim.delayed[,"n.risk"]))
    
})


context("Clustered survival data")
test_that("clustersurv",{
    library(prodlim)
    ## if (!is.function("cluster")) cluster <- function(x)x
    clusterTestData <- data.frame(midtimeX=1:8,eventX=c(0,"pn","pn",0,0,0,0,0),patientid=c(1,1,2,2,3,3,4,4),AnyCrownFracture=c(1,1,1,1,2,2,2,2))
    a <- prodlim(Hist(midtimeX,eventX=="pn")~cluster(patientid)+AnyCrownFracture,data=clusterTestData)
    b <- prodlim(Hist(midtimeX,eventX=="pn")~cluster(patientid),data=clusterTestData[clusterTestData$AnyCrownFracture==1,])
    c <- prodlim(Hist(midtimeX,eventX=="pn")~cluster(patientid),data=clusterTestData,subset=clusterTestData$AnyCrownFracture==1)
    d <- prodlim(Hist(midtimeX,eventX=="pn")~1,data=clusterTestData[clusterTestData$AnyCrownFracture==2,])
    expect_equal(round(as.numeric(summary(a)[AnyCrownFracture==1][["se.surv"]]),5),c(0,0.20951,0.10476,0.10476,NA,NA,NA,NA))
    expect_equal(summary(b), summary(c))
})

#' Helper function to obtain running means for prodlim objects.
#' 
#' Compute average values of a variable according to neighborhoods.
#' 
#' 
#' @param x Object of class \code{"neighborhood"}.
#' @param y Vector of numeric values.
#' @param \dots Not used.
#' @author Thomas Alexander Gerds <tag@@biostat.ku.dk>
#' @seealso \code{\link{neighborhood}}
#' @keywords survival
#' @examples
#' 
#' meanNeighbors(x=1:10,y=c(1,10,100,1000,1001,1001,1001,1002,1002,1002))
#'
#' @export
meanNeighbors <- function(x,y,...){
  nnn=neighbors(x,y,...)
  out <- data.frame(x=nnn$nbh$values,
                    y=sapply(nnn$list,mean))
  names(out) <- c("uniqueX","averageY")
  out
}

#' Competing risks model for simulation
#' 
#' Create a competing risks model with to causes to simulate a right censored event time data without
#' covariates
#' 
#' This function requires the \code{lava} package.
#' @title Competing risks model for simulation
#' @return A structural equation model initialized with four variables: the
#' latent event times of two causes, the latent right censored time, and the observed
#' right censored event time.
#' @author Thomas A. Gerds
#' @examples
#' library(lava)
#' m <- crModel()
#' d <- sim(m,6)
#' print(d)
#'
#' @export
crModel <- function(){
    # require(lava)
    crm <- lava::lvm()
    lava::distribution(crm,"eventtime1") <- lava::coxWeibull.lvm(scale=1/100)
    lava::distribution(crm,"eventtime2") <- lava::coxWeibull.lvm(scale=1/100)
    lava::distribution(crm,"censtime") <- lava::coxWeibull.lvm(scale=1/100)
    crm <- lava::eventTime(crm,time~min(eventtime1=1,eventtime2=2,censtime=0),"event")
    crm
}

#' Identifying rows in a matrix or data.frame
#' 
#' Function for finding matching rows between two matrices or data.frames.
#' First the matrices or data.frames are vectorized by row wise pasting
#' together the elements. Then it uses the function match.  Thus the function
#' returns a vector with the row numbers of (first) matches of its first
#' argument in its second.
#' 
#' 
#' @param x Vector or matrix whose rows are to be matched
#' @param table Matrix or data.frame that contain the rows to be matched
#' against.
#' @param nomatch the value to be returned in the case when no match is found.
#' Note that it is coerced to 'integer'.
#' @return A vector of the same length as 'x'.
#' @author Thomas A. Gerds
#' @seealso \code{match}
#' @keywords misc
#' @examples
#' 
#' tab <- data.frame(num=1:26,abc=letters)
#' x <- c(3,"c")
#' row.match(x,tab)
#' x <- data.frame(n=c(3,8),z=c("c","h"))
#' row.match(x,tab)
#'
#' @export
"row.match" <-
  function(x, table, nomatch=NA){
    if (inherits(table,"matrix")) table <- as.data.frame(table)
    if (is.null(dim(x))) x <- as.data.frame(matrix(x,nrow=1))
    cx <- do.call("paste",c(x[,,drop=FALSE],sep="\r"))
    ct <- do.call("paste",c(table[,,drop=FALSE],sep="\r"))
    match(cx,ct,nomatch=nomatch)
  }

#' Check availability of a cause in competing risk settings
#'
#' For competing risk settings, check if the requested cause is known to the object
#' @param cause cause of interest
#' @param object object either obtained with \code{Hist} or \code{prodlim} 
#' @export
checkCauses <- function(cause,object){
    if (!is.null(cause)){
        cause <- unique(cause)
        if (!is.character(cause)) cause <- as.character(cause)
        fitted.causes <- prodlim::getStates(object)
        if (!is.null(fitted.causes)){
            if (!(all(cause %in% fitted.causes))){
                stop(paste0("Cannot predict the requested cause of the event:\n",
                            "Requested: ",
                            paste0(cause,collapse=", "),
                            "\nFitted: ",
                            paste(fitted.causes,collapse=", "),"\n"))
            }
        }
        cause
    }
}
#----------------------------------------------------------------------
### checkCauses.R ends here

#' @export
lifeTab.competing.risks <- function(object,
                                    times,
                                    cause,
                                    newdata,
                                    stats,
                                    intervals=FALSE,
                                    percent=TRUE,
                                    format,...){
    # {{{---------get the indices--------------------------
    IndeX <- predict(object,newdata=newdata,level.chaos=0,times=times,type="list")
    # }}}
    # {{{--------------times-------------------------------
    times <- IndeX$times
    Ntimes <- IndeX$dimensions$time
    pindex <- IndeX$indices$time
    delayed <- attr(object$model.response,"entry.type")=="leftTruncated"
    # }}}
    # {{{---------covariate strata--------------------------
    Nstrata <- IndeX$dimensions$strata
    findex <- IndeX$indices$strata
    # }}}
    # {{{---------competing causes--------------------------
    if (missing(cause)){
        causes <- attributes(object$model.response)$states
    } else{
        causes <- checkCauses(cause,object)
    }
    # }}}
    # {{{--------------stats-------------------------------
    if (missing(stats) || (!missing(stats) && is.null(stats)))
        stats <- list(c("n.event",0),c("n.lost",0))
    else
        stats <- c(list(c("n.event",0),c("n.lost",0)),stats)
    #
    # }}}
    # {{{---------loop over causes--------------------------
    #
    outList <- lapply(causes,function(cc){
        # ---no. at atrisk, events, and censored------------------
        if (intervals==FALSE){
            if (is.null(object$clustervar)){
                ## only one column for n.risk
                xxx <- .C("summary_prodlim",pred.nrisk=integer(Ntimes*Nstrata),pred.nevent=integer(Ntimes*Nstrata),pred.nlost=integer(Ntimes*Nstrata),nrisk=as.integer(object$n.risk),nevent=as.integer(object$n.event[[cc]]),nlost=as.integer(object$n.lost),as.double(times),as.double(object$time),as.integer(object$first.strata[findex]),as.integer(object$size.strata[findex]),as.integer(Nstrata),as.integer(Ntimes),NAOK=FALSE,PACKAGE="prodlim")
                out <- data.frame(cause=cc,n.risk=xxx$pred.nrisk,n.event=xxx$pred.nevent,n.lost=xxx$pred.nlost)
                ## out <- data.frame(n.risk=xxx$pred.nrisk)
            }
            else{
                xxx <- .C("summary_prodlim",pred.nrisk=integer(Ntimes*Nstrata),pred.nevent=integer(Ntimes*Nstrata),pred.nlost=integer(Ntimes*Nstrata),nrisk=as.integer(object$n.risk[,1]),nevent=as.integer(object$n.event[[cc]][,1]),nlost=as.integer(object$n.lost),as.double(times),as.double(object$time),as.integer(object$first.strata[findex]),as.integer(object$size.strata[findex]),as.integer(Nstrata),as.integer(Ntimes),NAOK=FALSE,PACKAGE="prodlim")
                out <- data.frame(cause=cc,n.risk=xxx$pred.nrisk,n.event=xxx$pred.nevent,n.lost=xxx$pred.nlost)
                ## out <- data.frame(n.risk=xxx$pred.nrisk)
                for (cv in 1:length(object$clustervar))
                    yyy <- .C("summary_prodlim",pred.nrisk=integer(Ntimes*Nstrata),pred.nevent=integer(Ntimes*Nstrata),pred.nlost=integer(Ntimes*Nstrata),nrisk=as.integer(object$n.risk[,1+cv]),nevent=as.integer(object$n.event[[cc]][,1+cv]),nlost=as.integer(object$n.lost[,1+cv]),as.double(times),as.double(object$time),as.integer(object$first.strata[findex]),as.integer(object$size.strata[findex]),as.integer(Nstrata),as.integer(Ntimes),NAOK=FALSE,PACKAGE="prodlim")
                outCV <- data.frame(cause=cc,n.risk=yyy$pred.nrisk,n.event=yyy$pred.nevent,n.lost=yyy$pred.nlost)
                ## outCV <- data.frame(n.risk=yyy$pred.nrisk)
                names(outCV) <- paste(object$clustervar,names(outCV),sep=".")
                out <- cbind(out,outCV)
            }
        }
        # }}}
        # {{{-------Intervals---------------------------
        else{
            #,----
            #|      get the no. at risk at the left limit of the interval
            #|      and count events and censored excluding the left limit
            #`----
            start <- min(min(object$time),0)-.1
            lower <- c(start,times[-length(times)])
            upper <- times
            lagTimes <- c(min(min(object$time),0)-.1 , times[-length(times)])
            if (is.null(object$clustervar)){
                ## only one column in n.event and n.risk 
                xxx <- .C("life_table",
                          pred.nrisk=integer(Ntimes*Nstrata),
                          pred.nevent=integer(Ntimes*Nstrata),
                          pred.nlost=integer(Ntimes*Nstrata),
                          nrisk=as.integer(object$n.risk),
                          nevent=as.integer(object$n.event[[cc]]),
                          nlost=as.integer(object$n.lost),
                          as.double(lower),
                          as.double(upper),
                          as.double(object$time),
                          as.integer(object$first.strata[findex]),
                          as.integer(object$size.strata[findex]),
                          as.integer(Nstrata),
                          as.integer(Ntimes),
                          as.integer(delayed),
                          NAOK=FALSE,
                          PACKAGE="prodlim")
                out <- data.frame(cause=cc,n.risk=xxx$pred.nrisk,n.event=xxx$pred.nevent,n.lost=xxx$pred.nlost)
            }
            else{
                xxx <- .C("life_table",
                          pred.nrisk=integer(Ntimes*Nstrata),
                          pred.nevent=integer(Ntimes*Nstrata),
                          pred.nlost=integer(Ntimes*Nstrata),
                          nrisk=as.integer(object$n.risk[,1]),
                          nevent=as.integer(object$n.event[[cc]][,1]),
                          nlost=as.integer(object$n.lost[,1]),
                          as.double(lower),
                          as.double(upper),
                          as.double(object$time),
                          as.integer(object$first.strata[findex]),
                          as.integer(object$size.strata[findex]),
                          as.integer(Nstrata),
                          as.integer(Ntimes),
                          as.integer(delayed),
                          NAOK=FALSE,
                          PACKAGE="prodlim")
                out <- data.frame(cause=cc,n.risk=xxx$pred.nrisk,n.event=xxx$pred.nevent,n.lost=xxx$pred.nlost)
                lagxxx <- .C("life_table",
                             pred.nrisk=integer(Ntimes*Nstrata),
                             pred.nevent=integer(Ntimes*Nstrata),
                             pred.nlost=integer(Ntimes*Nstrata),
                             nrisk=as.integer(object$n.risk[,1]),
                             nevent=as.integer(object$n.event[[cc]][,1]),
                             nlost=as.integer(object$n.lost[,1]),
                             as.double(lagTimes),
                             as.double(object$time),
                             as.integer(object$first.strata[findex]),
                             as.integer(object$size.strata[findex]),
                             as.integer(Nstrata),
                             as.integer(Ntimes),
                             as.integer(delayed),
                             intervals=as.integer(TRUE),
                             NAOK=FALSE,
                             PACKAGE="prodlim")
                out$n.risk <- lagxxx$pred.nrisk
                for (cv in 1:length(object$clustervar)){
                    yyy <- .C("life_table",
                              pred.nrisk=integer(Ntimes*Nstrata),
                              pred.nevent=integer(Ntimes*Nstrata),
                              pred.nlost=integer(Ntimes*Nstrata),
                              nrisk=as.integer(object$n.risk[,1+cv]),
                              nevent=as.integer(object$n.event[[cc]][,1+cv]),
                              nlost=as.integer(object$n.lost[,1+cv]),
                              as.double(lower),
                              as.double(upper),
                              as.double(object$time),
                              as.integer(object$first.strata[findex]),
                              as.integer(object$size.strata[findex]),
                              as.integer(Nstrata),
                              as.integer(Ntimes),
                              as.integer(delayed),
                              NAOK=FALSE,
                              PACKAGE="prodlim")
                    outCV <- data.frame(n.risk=yyy$pred.nrisk,n.event=yyy$pred.nevent,n.lost=yyy$pred.nlost)
                    names(outCV) <- paste(object$clustervar,names(outCV),sep=".")
                    out <- cbind(out,outCV)
                }
            }
        }
        # }}}
        # {{{ percent
        if (!is.null(stats)){
            statsList <- lapply(stats,function(x){
                name.x <- x[1]
                if (name.x=="risk") name.x="cuminc"
                if (percent==TRUE && (match(x[1],c("n.event","n.lost","n.risk"),nomatch=0)==0)){
                    if (x[1]=="surv") { # only one for all causes
                        100*as.numeric(c(x[2],object[[name.x]])[pindex+1])
                    } else{
                        100*as.numeric(c(x[2],object[[name.x]][[cc]])[pindex+1])
                    }
                }
                else{
                    if (x[1]%in%c("surv","n.lost")) {# only one for all causes
                        as.numeric(c(x[2],object[[name.x]])[pindex+1])
                    } else{
                        as.numeric(c(x[2],object[[name.x]][[cc]])[pindex+1])
                    }
                }
            })
            names(statsList) <- sapply(stats,function(x)x[[1]])
            add <- do.call("cbind",statsList)
            add <- add[,match(colnames(add),colnames(out),nomatch=FALSE)==0,drop=FALSE]
            if (NROW(out)==1)
                out <- data.frame(cbind(out,add))
            else
                out <- cbind(out,add)
        }
        # }}}
        # {{{ split according to covariate strata----------------
        if (intervals==TRUE){
            tt <- data.table::data.table(time0=c(0,round(times[-length(times)],2)),time1=times)
        }else{
            tt <- data.table::data.table(time=times)
        }
        if (!is.null(newdata) || Nstrata > 1) {
            if (format[[1]]=="list"){
                split.cova <- rep(1:Nstrata,rep(Ntimes,Nstrata))
                out <- split(out,split.cova)
                names(out) <- IndeX$names.strata
                out <- lapply(out,function(x){
                    x <- cbind(as.matrix(tt),as.matrix(x))
                    rownames(x) <- 1:NROW(x)
                    x
                })
            }else{
                X <- IndeX$predictors[rep(1:Nstrata,rep(Ntimes,Nstrata)),,drop=FALSE]
                out <- cbind(X,tt,out)
                data.table::setDT(out)
                data.table::setkeyv(out,colnames(X))
                out
            }
        } else{
            out <- cbind(tt,out)
            rownames(out) <- 1:NROW(out)
            out
        }
    }) 
    # }}}
    if (format[[1]]!="list") {
        key <- c("cause",data.table::key(outList[[1]]))
        outList <- data.table::rbindlist(outList)
        data.table::setkeyv(outList,key)
    } else{
        names(outList) <- causes
    }
    outList
}

# {{{ header

#' Summary method for prodlim objects.
#' 
#' Summarizing the result of the product limit method in life-table format.
#' Calculates the number of subjects at risk and counts events and censored
#' observations at specified times or in specified time intervals.
#' 
#' For cluster-correlated data the number of clusters at-risk are are also
#' given. Confidence intervals are displayed when they are part of the fitted
#' object.
#' 
#' @param object An object with class `prodlim' derived with
#' \code{\link{prodlim}}
#' @param times Vector of times at which to return the estimated
#' probabilities.
#' @param newdata A data frame with the same variable names as those
#' that appear on the right hand side of the 'prodlim' formula.
#' Defaults to \code{object$X}.
#' @param max.tables Integer. If \code{newdata} is not given the value
#' of \code{max.tables} decides about the maximal number of tables to
#' be shown.  Defaults to 20.
#' @param surv Logical. If FALSE report event probabilities instead of
#' survival probabilities. Only available for
#' \code{object$model=="survival"}.
#' @param cause For competing risk models. The event of interest for which predictions of the absolute risks are obtained by evaluating the cause-specific cumulative
#' incidence functions at  \code{times}.
#' @param intervals Logical. If TRUE count events and censored in
#' intervals between the values of \code{times}.
#' @param percent Logical. If TRUE all estimated values are multiplied
#' by 100 and thus interpretable on a percent scale.
#' @param format Control format of output. Since May 2021,
#' the result is a data.table and data.frame with attributes. When there are multiple
#' covariate strata or competing risks, these are indicated by columns.
#' Set format to \code{"list"} to get the old behaviour.
#' @param ... Further arguments that are passed to the print
#' function.
#' @return A data.frame with the relevant information.
#' @author Thomas A. Gerds \email{tag@@biostat.ku.dk}
#' @seealso \code{\link{prodlim}}, \code{\link{summary.Hist}}
#' 
#' @keywords survival
##' @examples
##' 
##' library(lava)
##' set.seed(17)
##' m <- survModel()
##' distribution(m,~age) <- uniform.lvm(30,80)
##' distribution(m,~sex) <- binomial.lvm()
##' m <- categorical(m,~z,K=3)
##' regression(m,eventtime~age) <- 0.01
##' regression(m,eventtime~sex) <- -0.4
##' d <- sim(m,50)
##' d$sex <- factor(d$sex,levels=c(0,1),labels=c("female","male"))
##' d$Z <- factor(d$z,levels=c(1,0,2),labels=c("B","A","C"))
##' 
##' # Univariate Kaplan-Meier
##' # -----------------------------------------------------------------------------------------
##' fit0 <- prodlim(Hist(time,event)~1,data=d)
##' summary(fit0)
##' 
##' ## show survival probabilities as percentage and
##' ## count number of events within intervals of a
##' ## given time-grid:
##' summary(fit0,times=c(1,5,10,12),percent=TRUE,intervals=TRUE)
##' 
##' ## the result of summary has a print function
##' ## which passes ... to print and print.listof
##' sx <- summary(fit0,times=c(1,5,10,12),percent=TRUE,intervals=TRUE)
##' print(sx,digits=3)
##' 
##' ## show absolute risks, i.e., cumulative incidences (1-survival)
##' summary(fit0,times=c(1,5,10,12),surv=FALSE,percent=TRUE,intervals=TRUE)
##' 
##' # Stratified Kaplan-Meier
##' # -----------------------------------------------------------------------------------------
##' 
##' fit1 <- prodlim(Hist(time,event)~sex,data=d)
##' print(summary(fit1,times=c(1,5,10),intervals=TRUE,percent=TRUE),digits=3)
##' # old behaviour
##' print(summary(fit1,times=c(1,5,10),intervals=TRUE,percent=TRUE,format="list"),digits=3)
##' 
##' summary(fit1,times=c(1,5,10),intervals=TRUE,percent=TRUE)
##' 
##' fit2 <- prodlim(Hist(time,event)~Z,data=d)
##' print(summary(fit2,times=c(1,5,10),intervals=TRUE,percent=TRUE),digits=3)
##' 
##' ## Continuous strata (Beran estimator)
##' # -----------------------------------------------------------------------------------------
##' fit3 <- prodlim(Hist(time,event)~age,data=d)
##' print(summary(fit3,
##'               times=c(1,5,10),
##'               newdata=data.frame(age=c(20,50,70)),
##'               intervals=TRUE,
##'               percent=TRUE),digits=3)
##' 
##' ## stratified Beran estimator
##' # -----------------------------------------------------------------------------------------
##' fit4 <- prodlim(Hist(time,event)~age+sex,data=d)
##' print(summary(fit4,
##'               times=c(1,5,10),
##'               newdata=data.frame(age=c(20,50,70),sex=c("female","male","male")),
##'               intervals=TRUE,
##'               percent=TRUE),digits=3)
##' 
##' print(summary(fit4,
##'               times=c(1,5,10),
##'               newdata=data.frame(age=c(20,50,70),sex=c("female","male","male")),
##'               intervals=TRUE,
##'               percent=TRUE),digits=3)
##' 
##' ## assess results from summary
##' x <- summary(fit4,times=10,newdata=expand.grid(age=c(60,40,50),sex=c("male","female")))
##' cbind(names(x$table),do.call("rbind",lapply(x$table,round,2)))
##' 
##' x <- summary(fit4,times=10,newdata=expand.grid(age=c(60,40,50),sex=c("male","female")))
##' 
##' ## Competing risks: Aalen-Johansen
##' # -----------------------------------------------------------------------------------------
##' d <- SimCompRisk(30)
##' crfit <- prodlim(Hist(time,event)~X1,data=d)
##' summary(crfit,times=c(1,2,5))
##' summary(crfit,times=c(1,2,5),cause=1,intervals=TRUE)
##' summary(crfit,times=c(1,2,5),cause=1)
##' summary(crfit,times=c(1,2,5),cause=1:2)
##' 
##' 
##' # extract the actual tables from the summary 
##' sumfit <- summary(crfit,times=c(1,2,5),print=FALSE)
##' sumfit$table[[1]] # cause 1
##' sumfit$table[[2]] # cause 2
##' 
##' 
##' # '
#' @export summary.prodlim
#' @export
summary.prodlim <- function(object,
                            times,
                            newdata,
                            max.tables=20,
                            surv=TRUE,
                            cause,
                            intervals=FALSE,
                            percent=FALSE,
                            format="df",
                            ...) {
    if ("showTime" %in% names(match.call())) stop("Argument showTime is not supported anymore: you should remove argument showTime from the call.")
    if ("asMatrix" %in% names(match.call())) stop("Argument asMatrix is not supported anymore: you should remove argument asMatrix from the call. Use new argument 'format' instead.")

    # }}}
    # {{{ classify the situation
    cens.type <- object$cens.type         # uncensored, right or interval censored
    model <- object$model                 # survival, competing risks or multi-state
    ## cluster <- object$clustervar          # clustered data?
    cotype <- object$covariate.type       # no, discrete, continuous or both
    # }}}
    # {{{ times
    jump.times <- object$time
    if (missing(times) && (length(times <- jump.times) > 50)) 
        times <- quantile(sort(unique(jump.times)))
    times <- sort(unique(times))
    if (any(times>max(jump.times)))
        warning(call.=TRUE,
                immediate.=TRUE,
                paste("\n","Time(s) ",paste(times[times>max(jump.times)],collapse=", "),
                      " are beyond the maximal follow-up time ",max(jump.times),"\n"))
    ntimes <- length(times)
    # }}}
    # {{{ interval-censored
    if (cens.type=="intervalCensored"){
        ltab <- data.frame(time=paste("(",paste(signif(object$time[1,],2),
                                                signif(object$time[2,],2),
                                                sep="-"),"]",sep=""),
                           n.risk=signif(object$n.risk,2),
                           n.event=signif(object$n.event,2),
                           ##    n.lost=object$n.lost,
                           surv=object$surv)
    }
    else{
        # }}}
    # {{{ with covariates
        if (cotype>1){
            if (missing(newdata) || length(newdata)==0){
                X <- object$X
                if (NROW(X)>max.tables){
                    warning(call.=TRUE,immediate.=TRUE,paste("\nLife tables are available for",
                                                             NROW(X),
                                                             "different covariate constellations.\n",
                                                             "Shown are the table corresponding to the first row in object$X,",
                                                             "corresponding to the middle row (median of the number of rows in object$X) ",
                                                             "and corresponding to the last row in object$X ...\n",
                                                             "to see more tables use arguments `newdata' and `max.tables'\n"))
                    X <- X[c(1,round(median(1:NROW(X))),NROW(X)),,drop=FALSE]
                }
            } else{
                X <- unique.data.frame(newdata)
                if (NROW(X) < NROW(newdata))
                    warning("Returned is only one summary for each unique value in newdata.")
            }
        } else {
            X <- NULL
        }
        if (!missing(cause)){
            cause <- checkCauses(cause=cause,object=object)
        } else{ ## show all causes
            cause <- attr(object$model.response,"states")
        }
        if (model=="survival") {
            stats <- list(c("surv",1),c("se.surv",0))
            if (!is.null(object$conf.int))
                stats <- c(stats,list(c("lower",1),c("upper",1)))
            if (surv==FALSE){
                object$cuminc <- 1-object$surv
                object$se.cuminc <- object$se.surv
                cuminc.upper <- 1-object$lower
                cuminc.lower <- 1-object$upper
                object$lower <- cuminc.lower
                object$upper <- cuminc.upper
                stats <- list(c("cuminc",0),c("se.cuminc",0))
                if (!is.null(object$conf.int))
                    stats <- c(stats,list(c("lower",1),c("upper",1)))
            }
        }
        if (model=="competing.risks"){
            stats <- list(c("cuminc",0),c("se.cuminc",0))
            if (!is.null(object$conf.int))
                stats <- c(stats,list(c("lower",0),c("upper",0)))
            out <- lifeTab(object=object,
                           times=times,
                           cause=cause,
                           newdata=X,
                           stats=stats,
                           intervals=intervals,
                           percent=percent,
                           format=format)
        }else{ ## survival model
            out <- lifeTab(object=object,
                           times=times,
                           newdata=X,
                           stats=stats,
                           intervals=intervals,
                           percent=percent,format=format)
        }
    }
    # }}}
    # {{{ preparation of output
    if (format=="list"){
        out <- list(table=out,model=model,cotype=cotype,percent=percent)
        if (model=="competing.risks"){
            out <- c(out,list(cause=cause))
        }
    }else{
        # reduce list of competing risks 
        if (inherits(out,"list")) out <- do.call("rbind",out)
        attributes(out) <- c(attributes(out),list(model=model,cotype=cotype,percent=percent))
    }
    # }}}
    class(out) <- c("summary.prodlim",class(out))
    out
}

plotIntervals <- function(object,rightCensored=FALSE,xlim,ylim,ylab,xlab,...){
  x <- object[order(object[,"L"]),]
  L <- x[,"L"]
  R <- x[,"R"]
  if (rightCensored==FALSE)
    x <- x[!is.na(R)&!is.infinite(R),]
  if (missing(ylim))
    ylim <- c(0,NROW(x)+1)
  if (missing(xlim))
    xlim <- c(0,max(R[!is.na(R)&!is.infinite(R)]))
  if (missing(xlab))
    xlab <- "Time"
  if (missing(ylab))
    ylab <- "Observed intervals"
  plot(0,0,type="n",xlim=xlim,ylim=ylim,xlab=xlab,ylab=ylab,...)
  nix <- lapply(1:NROW(x),function(f){
    x <- unlist(x[f,c("L","R"),drop=TRUE])
    x[is.infinite(x)] <- max(R[!is.na(R)&!is.infinite(R)])
    segments(x0=x[1],y0=f,x1=x[2],y1=f,lwd=2)})
  invisible(x)
}

### as.data.table.prodlim.R --- 
#----------------------------------------------------------------------
## Author: Thomas Alexander Gerds
## Created: Mar  3 2025 (12:57) 
## Version: 
## Last-Updated: Mar  5 2025 (11:58) 
##           By: Thomas Alexander Gerds
##     Update #: 23
#----------------------------------------------------------------------
## 
### Commentary: 
## 
### Change Log:
#----------------------------------------------------------------------
## 
### Code:
##' Collect results of a fitted prodlim object in a data.table
##'
##' By default object contains results for all fitted time points and all strata.
##' Use arguments times and newdata of \code{\link{summary.prodlim}} to subset.
##' @title Turn prodlim objects into a \code{\link[data.table]{data.table}}
##' @name as.data.table.prodlim
##' @param x object obtained with function \code{\link{prodlim}}
##' @param keep.rownames Not used
##' @param ... passed to \code{\link{summary.prodlim}}
##' @return A data.table with the results of the prodlim object 
##' @seealso \code{\link{prodlim}}, \code{\link[data.table]{data.table}}
##' @examples
##' library(data.table)
##' set.seed(8)
##' d <- SimCompRisk(17)
##' fit <- prodlim(Hist(time,event)~X1,data=d)
##' as.data.table(fit)
##'
##'@export as.data.table.prodlim
##'@export 
##'@author Thomas A. Gerds <tag@@biostat.ku.dk>
as.data.table.prodlim <- function(x,keep.rownames = FALSE,...){
    data.table::as.data.table(as.data.frame.prodlim(x,...))
}


######################################################################
### as.data.table.prodlim.R ends here

##' Simulate right censored survival data with two covariates X1 and X2, both have effect exp(1) on the hazard of the unobserved event time.
##'
##' This function calls \code{survModel}, then adds  covariates and finally calls \code{sim.lvm}.
##' @title Simulate survival data
##' @param N sample size
##' @param ... do nothing
##' @return data.frame with simulated data
##' @references Bender, Augustin & Blettner. Generating survival times to simulate Cox proportional hazards models. Statistics in Medicine, 24: 1713-1723, 2005.
##' @author Thomas Alexander Gerds
##' @examples
##' 
##' SimSurv(10)
##'
##' @export
SimSurv <- function(N, ...){
    m <- survModel()
    regression(m,from="X1",to="eventtime") <- 1
    regression(m,from="X2",to="eventtime") <- 1
    distribution(m,"X1") <- binomial.lvm()
    m <- eventTime(m,time~min(eventtime=1,censtime=0),"status")
    sim(m,N)
}

## SimSurvInternalIntervalCensored <- function(N,
                                            ## unit,
                                            ## lateness,
                                            ## compliance,
                                            ## withdraw.time,
                                            ## event.time){
  ## Intervals <- do.call("rbind",lapply(1:N,function(i){
    ## schedule <- seq(0,withdraw.time[i],unit)
    ## M <- length(schedule)
    ## g <- c(0,rep(unit,M))
    ## # introduce normal variation of the visit times
    ## g <- g+c(abs(rnorm(1,0,lateness)),rnorm(M,0,lateness))
    ## grid <- c(0,cumsum(g))
    ## # remove visits after the end of follow-up time
    ## grid <- grid[grid<withdraw.time[i]]
    ## # remove intermediate visits
    ## if (compliance<1){
      ## stopifnot(compliance>0)
      ## missed <- rbinom(length(grid),1,compliance)==0
      ## grid <- grid[missed==FALSE]
    ## }
    ## if (length(grid)==0){
      ## L <- 0
      ## R <- Inf
    ## }
    ## else{
      ## posTime <- sindex(jump.times=grid,
                        ## eval.times=event.time[i])
      ## L <- grid[posTime]
      ## R <- grid[posTime+1]
      ## if (is.na(R)){
        ## R <- Inf
      ## }
    ## }
    ## c(L=L,R=R)
  ## }))
  ## out <- data.frame(Intervals)
  ## out
## }
# }}}
# {{{ find.baseline
## find.baseline <- function(x=.5,
                          ## setting,
                          ## verbose=FALSE){
  ## N <- setting$N
  ## f <- function(y){
    ## setting$cens.baseline <- y
    ## ncens <- sum(do.call("SimSurv",replace(setting,"verbose",verbose))$status==0)
    ## x-ncens/N
  ## }
  ## base.cens <- uniroot(f,c(exp(-50),1000000),tol=.0000001,maxiter=100)$root
  ## new.setting <- setting
  ## new.setting$cens.baseline <- base.cens
  ## do.call("SimSurv",replace(new.setting,"verbose",TRUE))
  ## new.setting
## }
# }}}
# {{{quantile.SimSurv
## quantile.SimSurv <- function(x,B=10,na.rm=FALSE,probs=.9){
  ## callx <- attr(x,"call")
  ## nix <- do.call("rbind",lapply(1:B,function(b){
    ## quantile(eval(callx)$time,probs)
  ## }))
  ## nix <- colMeans(nix)
  ## nix
## }
# }}}

#' Nearest neighborhoods for kernel smoothing
#' 
#' Nearest neighborhoods for the values of a continuous predictor. The result
#' is used for the conditional Kaplan-Meier estimator and other conditional
#' product limit estimators.
#' 
#' 
#' @param x Numeric vector -- typically the observations of a continuous random
#' variate.
#' @param bandwidth Controls the distance between neighbors in a neighborhood.
#' It can be a decimal, i.e.\ the bandwidth, or the string `"smooth"', in which
#' case \code{N^{-1/4}} is used, \code{N} being the sample size, or \code{NULL}
#' in which case the \code{\link[KernSmooth]{dpik}} function of the package KernSmooth is
#' used to find the optimal bandwidth.
#' @param kernel Only the rectangular kernel ("box") is implemented.
#' @return An object of class 'neighborhood'.  The value is a list that
#' includes the unique values of `x' (\code{values}) for which a neighborhood,
#' consisting of the nearest neighbors, is defined by the first neighbor
#' (\code{first.nbh}) of the usually very long vector \code{neighbors} and the
#' size of the neighborhood (\code{size.nbh}).
#' 
#' Further values are the arguments \code{bandwidth}, \code{kernel}, the total
#' sample size \code{n} and the number of unique values \code{nu}.
#' @author Thomas Gerds
#' @seealso \code{\link[KernSmooth]{dpik}}, \code{\link{prodlim}}
#' @references Stute, W. "Asymptotic Normality of Nearest Neighbor Regression
#' Function Estimates", \emph{The Annals of Statistics}, 1984,12,917--926.
#' @keywords smooth
#' @examples
#'
#' d <- SimSurv(20)
#' neighborhood(d$X2)
#' @export
"neighborhood" <- function(x,bandwidth=NULL,kernel="box"){
    if (any(is.na(x))) stop("Missing values in x")
    N <- length(x)
    if (N<2) stop("Not enough observations for kernel smoothing.")
    orderx <- order(x)
    values <- sort(unique(x))
    NU <- length(values)
    workx <- factor(x,labels=1:NU)
    tabu <- tabulate(workx)
    cumtabu <- cumsum(tabu)
    cumtabx <- rep(cumtabu,tabu)
    tabx <- rep(tabu,tabu)
    if (!length(bandwidth)){ ## need a bandwidth (dpik is from KernSmooth)
        ## require(KernSmooth)
        bandwidth <- KernSmooth::dpik(cumtabx/N,kernel="box")
    }
    else
        if (bandwidth=="smooth") bandwidth <- N^{-1/4}
    radius <- floor(bandwidth*N)
  
    nbh <- .C("neighborhoodSRC",
              first=integer(NU),
              size=integer(NU),
              as.integer(cumtabu),
              as.integer(cumtabx),
              as.integer(tabx),
              as.integer(radius),
              as.integer(NU),
              as.integer(N),
              PACKAGE="prodlim")
    nall <- sum(nbh$size)
    nbors <- .C("neighborsSRC",
                first=nbh$first,
                size=nbh$size,
                as.integer(orderx),
                neighbors=integer(nall),
                as.integer(NU),
                PACKAGE="prodlim")$neighbors
  
  out <- list(values=values,
              first.nbh=nbh$first,
              size.nbh=nbh$size,
              neighbors=nbors,
              bandwidth=bandwidth,
              kernel=kernel,
              nu=NU,
              n=N)
    
  class(out) <- "neighborhood"
  out
}

listNbh <- function(object,y,val){
  stopifnot(inherits(object,"neighborhood"))
  if (missing(y)) y=object$neighbors
  else{
    stopifnot(length(y)==object$n)
    y=y[object$neighbors]
  }
  if (missing(val)) val <- object$values
  posVal <- match(val,object$values,nomatch=FALSE)
  stopifnot(all(posVal!=0))
  out <- split(y,rep(1:object$nu,object$size.nbh))[posVal]
  out
}

#' Compute jackknife pseudo values.
#' 
#' Compute jackknife pseudo values based on marginal Kaplan-Meier estimate of
#' survival, or based on marginal Aalen-Johansen estimate of the absolute risks, i.e., the cumulative
#' incidence function.
#' 
#' @title Compute jackknife pseudo values.
#' @aliases jackknife jackknife.survival jackknife.competing.risks
#' @param object Object of class \code{"prodlim"}.
#' @param times Time points at which to compute pseudo values.
#' @param cause Character (other classes are converted with \code{as.character}).
#' For competing risks the cause of failure.
#' @param keepResponse If \code{TRUE} add the model response,
#' i.e. event time, event status, etc. to the result.
#' @param ... not used
#' @note The R-package pseudo does a similar job, and appears to be a little faster in small samples, but much slower in large samples. See examples.
#' @author Thomas Alexander Gerds <tag@@biostat.ku.dk>
#' @seealso \code{\link{prodlim}}
#' @references Andersen PK & Perme MP (2010). Pseudo-observations in survival
#' analysis Statistical Methods in Medical Research, 19(1), 71-99.
#' @keywords survival
##' @examples
##' 
##' 
##' ## pseudo-values for survival models
##' 
##' d=SimSurv(20) 
##' f=prodlim(Hist(time,status)~1,data=d) 
##' jackknife(f,times=c(3,5))
##' 
##' ## in some situations it may be useful to attach the
##' ## the event time history to the result
##' jackknife(f,times=c(3,5),keepResponse=TRUE)
##' 
##' # pseudo-values for competing risk models
##' set.seed(15)
##' d=SimCompRisk(15) 
##' f=prodlim(Hist(time,event)~1,data=d) 
##' jackknife(f,times=c(3,5),cause=1)
##' jackknife(f,times=c(1,3,5),cause=2)
##' 
#' @export
jackknife <- function(object,times,cause,keepResponse=FALSE,...){
    if (object$model=="survival")
        jackknife.survival(object=object,times=times,keepResponse=keepResponse,...)
    else if (object$model=="competing.risks"){
        if (!missing(cause)) cause <- checkCauses(cause,object)
        else cause <- attr(object$model.response,which="states")[[1]]
        jackknife.competing.risks(object=object,
                                  times=times,
                                  cause=cause,
                                  keepResponse=keepResponse,
                                  ...)
    }
    else stop("No method for jackknifing this object.")
}

#' @export
jackknife.survival <- function(object,times,keepResponse=FALSE,...){
    S <- predict(object,times=times,newdata=object$model.response)
    Sk <- leaveOneOut.survival(object,times,...)
    N <- NROW(Sk)
    Jk <- t(N*S-t((N-1)*Sk))
    colnames(Jk) <- paste("t",times,sep=".")
    if (keepResponse==TRUE){
        Jk <- cbind(object$model.response,Jk)
    }
    ## re-order the pseudo-values  
    Jk <- Jk[object$originalDataOrder,,drop=FALSE]
    Jk
}
#' @export
jackknife.competing.risks <- function(object,times,cause,keepResponse=FALSE,...){
    F <- predict(object,times=times,newdata=object$model.response,cause=cause)
    Fk <- leaveOneOut.competing.risks(object,times,cause=cause,...)
    N <- NROW(Fk)
    Jk <- t(N*F-t((N-1)*Fk))
    colnames(Jk) <- paste("t",times,sep=".")
    if (keepResponse==TRUE){
        Jk <- cbind(object$model.response,Jk)
        colnames(Jk)[(NCOL(Jk)-length(times)+1):NCOL(Jk)] <- paste("t",times,sep=".")
    }
    ## re-order the pseudo-values
    Jk <- Jk[object$originalDataOrder,,drop=FALSE]
    Jk
}

#' Box-arrow diagrams for multi-state models.
#' 
#' Automated plotting of the states and transitions that characterize a multi
#' states model.
#' 
#' 
#' @param x An object of class \code{Hist}.
#' @param nrow the number of graphic rows
#' @param ncol the number of graphic columns
#' @param box.width the widths of the boxes on the scale from 0 to 100
#' @param box.height the heights of the boxes on the scale from 0 to 100
#' @param box.padding how much room there should be between the label and the border of a box.
#' Two values on the scale from 0 to 100: the first for the horizontal
#' x-direction and the second for the vertical y-direction padding. 
#' @param xbox.position the x box positions (left lower corner) on the scale from 0 to 100.
#' @param ybox.position the y box positions (left lower corner) on the scale from 0 to 100.
#' @param stateLabels Vector of names to appear in the boxes (states).
#'     Defaults to attr(x,"state.names").  The boxes can also be
#'     individually labeled by smart arguments of the form
#'     \code{box3.label="diseased"}, see examples.
#' @param arrowLabels Vector of labels to appear in the boxes
#'     (states). One for each arrow.  The arrows can also be
#'     individually labeled by smart arguments of the form
#'     \code{arrow1.label=paste(expression(eta(s,u)))}, see examples.
#' @param arrowLabelStyle Either "symbolic" for automated symbolic
#'     arrow labels, or "count" for arrow labels that reflect the
#'     number of transitions in the data.
#' @param arrowLabelSymbol Symbol for automated symbolic arrow
#'     labels. Defaults to "lambda".
#' @param changeArrowLabelSide A vector of mode logical (TRUE,FALSE)
#'     one for each arrow to change the side of the arrow on which the
#'     label is placed.
#' @param curved The curvature of curved arrows via diagram::curvedarrow. Experimental. Values between 0 (no curvature) and 1 are meaningful. 
#' @param tagBoxes Logical. If TRUE the boxes are numbered in the
#'     upper left corner. The size can be controlled with smart
#'     argument boxtags.cex. The default is boxtags.cex=1.28.
#' @param startCountZero Control states numbers for symbolic arrow
#'     labels and box tags.
#' @param oneFitsAll If \code{FALSE} then boxes have individual size,
#'     depending on the size of the label, otherwise all boxes have
#'     the same size dependent on the largest label.
#' @param margin Set the figure margin via
#'     \code{par(mar=margin)}. Less than 4 values are repeated.
#' @param cex Initial cex value for the state and the arrow
#'     \code{labels}.
#' @param rasta For construction purposes.
#' @param verbose If TRUE echo various things.
#' @param \dots Smart control of arguments for the subroutines text
#'     (box label), rect (box), arrows, text (arrow label). Thus the
#'     three dots can be used to draw individual boxes with individual
#'     labels, arrows and arrow labels. E.g. arrow2.label="any label"
#'     changes the label of the second arrow.  See examples.
#' @note Use the functionality of the unix program `dot'
#'     http://www.graphviz.org/About.php via R package Rgraphviz to
#'     obtain more complex graphs.
#' @author Thomas A Gerds \email{tag@@biostat.ku.dk}
#' @seealso \code{\link{Hist}}\code{\link{SmartControl}}
#' @keywords survival
##' @examples
##' 
##' 
##' ## A simple survival model
##' 
##' SurvFrame <- data.frame(time=1:10,status=c(0,1,1,0,0,1,0,0,1,0))
##' SurvHist <- with(SurvFrame,Hist(time,status))
##' plot(SurvHist)
##' plot(SurvHist,box2.col=2,box2.label="experienced\nR user")
##' plot(SurvHist,
##'      box2.col=2,
##'      box1.label="newby",
##'      box2.label="experienced\nR user",
##'      oneFitsAll=FALSE,
##'      arrow1.length=.5,
##'      arrow1.label="",
##'      arrow1.lwd=4)
##' 
##' ## change the cex of all box labels:
##' plot(SurvHist,
##'      box2.col=2,
##'      box1.label="newby",
##'      box2.label="experienced\nR user",
##'      oneFitsAll=FALSE,
##'      arrow1.length=.5,
##'      arrow1.label="",
##'      arrow1.lwd=4,
##'      label.cex=1)
##' 
##' ## change the cex of single box labels:
##' plot(SurvHist,
##'      box2.col=2,
##'      box1.label="newby",
##'      box2.label="experienced\nR user",
##'      oneFitsAll=FALSE,
##'      arrow1.length=.5,
##'      arrow1.label="",
##'      arrow1.lwd=4,
##'      label1.cex=1,
##'      label2.cex=2)
##' 
##' 
##' ## The pbc data set from the survival package
##' library(survival)
##' data(pbc)
##' plot(with(pbc,Hist(time,status)),
##'      stateLabels=c("randomized","transplant","dead"),
##'      arrowLabelStyle="count")
##' 
##' ## two competing risks
##' comprisk.model <- data.frame(time=1:3,status=1:3)
##' CRHist <- with(comprisk.model,Hist(time,status,cens.code=2))
##' plot(CRHist)
##' plot(CRHist,arrow1.label=paste(expression(eta(s,u))))
##' 
##' plot(CRHist,box2.label="This\nis\nstate 2",arrow1.label=paste(expression(gamma[1](t))))
##' plot(CRHist,box3.label="Any\nLabel",arrow2.label="any\nlabel")
##' 
##' ## change the layout
##' plot(CRHist,
##'      box1.label="Alive",
##'      box2.label="Dead\n cause 1",
##'      box3.label="Dead\n cause 2",
##'      arrow1.label=paste(expression(gamma[1](t))),
##'      arrow2.label=paste(expression(eta[2](t))),
##'      box1.col=2,
##'      box2.col=3,
##'      box3.col=4,
##'      nrow=2,
##'      ncol=3,
##'      box1.row=1,
##'      box1.column=2,
##'      box2.row=2,
##'      box2.column=1,
##'      box3.row=2,
##'      box3.column=3)
##' 
##' ## more competing risks
##' comprisk.model2 <- data.frame(time=1:4,status=1:4)
##' CRHist2 <- with(comprisk.model2,Hist(time,status,cens.code=2))
##' plot(CRHist2,box1.row=2)
##' 
##' ## illness-death models
##' illness.death.frame <- data.frame(time=1:4,
##' 				  from=c("Disease\nfree",
##'                                       "Disease\nfree",
##'                                       "Diseased",
##'                                       "Disease\nfree"),
##' 				  to=c("0","Diseased","Dead","Dead"))
##' IDHist <- with(illness.death.frame,Hist(time,event=list(from,to)))
##' plot(IDHist)
##' 
##' ## illness-death with recovery
##' illness.death.frame2 <- data.frame(time=1:5,
##' from=c("Disease\nfree","Disease\nfree","Diseased","Diseased","Disease\nfree"),
##' to=c("0","Diseased","Disease\nfree","Dead","Dead"))
##' IDHist2 <- with(illness.death.frame2,Hist(time,event=list(from,to)))
##' plot(IDHist2)
##' 
##' ## 4 state models
##' x=data.frame(from=c(1,2,1,3,4),to=c(2,1,3,4,1),time=1:5)
##' y=with(x,Hist(time=time,event=list(from=from,to=to)))
##' plot(y)
##' 
##' ## moving the label of some arrows
##' 
##' d <- data.frame(time=1:5,from=c(1,1,1,2,2),to=c(2,3,4,3,4))
##' h <- with(d,Hist(time,event=list(from,to)))
##' plot(h,box.padding=c(5,2),
##' tagBoxes=TRUE,
##' stateLabels=c("Remission\nwithout\nGvHD",
##'     "Remission\nwith\nGvHD",
##'     "Relapse",
##'     "Death\nwithout\nrelapse"),
##' arrowLabelSymbol='alpha',
##' arrowlabel3.x=35,
##' arrowlabel3.y=53,
##' arrowlabel4.y=54,
##' arrowlabel4.x=68)
##' 
##' ##'
#' @export plot.Hist
#' @export 
plot.Hist <- function(x,
                      nrow,
                      ncol,
                      box.width,
                      box.height,
                      box.padding,
                      xbox.position,
                      ybox.position,
                      stateLabels,
                      arrowLabels,
                      arrowLabelStyle="symbolic",
                      arrowLabelSymbol='lambda',
                      changeArrowLabelSide,
                      curved,
                      tagBoxes=FALSE,
                      startCountZero=TRUE,                      
                      oneFitsAll,
                      margin,
                      cex,
                      rasta=FALSE,
                      verbose=FALSE,
                      ...){
    # {{{ margin 
    oldmar <- par()$mar
    oldoma <- par()$oma
    par(oma=c(0,0,0,0))
    oldxpd <- par()$xpd
    # {{{ reset margin
    on.exit(par(mar=oldmar,xpd=oldxpd,oma=oldoma))
    # }}}

    if (!missing(margin)){
        par(mar=rep(margin,length.out=4),xpd=TRUE)
    }
    else
        par(mar=c(0,0,0,0),xpd=TRUE)
    # }}}
    # {{{ find states 
    model.type <- attr(x,"model")
    states <- attr(x,"states")
    origStates <- states
    if (model.type!="multi.states"){ ## need an initial state
        states <- c("initial", states)
    }
    NS <- length(states)
    if (missing(stateLabels)){
        if (all(as.character(as.numeric(as.factor(origStates)))==origStates))  ## make nice state boxlabels if states are integers
            stateLabs <- switch(model.type,"survival"=paste(c("","Event"),states),"competing.risks"=paste(c("",rep("Cause",NS-1)),states),paste("State",states))
        else
            stateLabs <- states
    }
    else{
        if(length(stateLabels)==NS-1){
            stateLabs <- c("initial",stateLabels)
        }
        else{
            if (length(stateLabels)==NS){
                stateLabs <- stateLabels
            }
            else{
                stop("Wrong number of state names.")
            }
        }
    }
    ## forcedLabels
    thecall <- match.call(expand.dots=TRUE)
    labelhits <- match(paste("box",1:NS,".label",sep=""),names(thecall),nomatch=0)
    for (i in 1:NS){
        if (labelhits[i]!=0)
            ## may be language: thecall[[labelhits[i]]]
            ## if user specifies box2.label=c("Event 1")
            ## instead of box2.label="Event 1"
            stateLabs[i] <- eval(thecall[[labelhits[i]]])[1]
    }
    numstates <- as.numeric(as.character(factor(states,levels=states,labels=1:NS)))
    startCountZero <- TRUE
    if (length(tagBoxes)==1){
        if (startCountZero)
            numstateLabels <- numstates-1
        else
            numstateLabels <- numstates
    }else{
        if (length(tagBoxes)==NS){
            numstateLabels <- tagBoxes
            tagBoxes <- TRUE
        }else{
            stop(paste0("The length of argument 'tagBoxes' is ",length(tagBoxes)," does not match number of states which is ",NS,"."))
        }
    }
    # {{{  find transitions between the states

  ## first remove the censored lines from the transition matrix
  ## x <- x[x[,"status"]!=attr(x,"cens.code"),,drop=FALSE]
  x <- x[x[,"status"]!=0,,drop=FALSE]
  if (NROW(x)==0) stop("No uncensored transitions.")
  sumx <- summary(x,verbose=verbose)
  notCensored <- sumx$trans.frame$to!="unknown"
  sumx$trans.frame <- sumx$trans.frame[notCensored,]
  sumx$transitions <- sumx$transitions[notCensored]
  transitions <- sumx$trans.frame
  ordered.transitions <- unique(transitions)
  N <- NROW(ordered.transitions)
  # }}}

  # }}}
    # {{{ default layout: arranging the boxes

    state.types <- sumx$states
    state.types <- state.types[state.types>0]
    if (missing(nrow))
        if (model.type=="multi.states")
            nrow <- NS
        else
            if (ceiling(NS/2)==floor(NS/2))
                nrow <- NS-1
            else
                nrow <- NS
    if (missing(ncol))
        if (model.type=="multi.states")
            ncol <- NS
        else
            ncol <- 2
    ## placing boxes in rows and columns
    if (model.type=="multi.states"){
        adjustRowsInColumn <- rep(0,ncol)
        adjustColsInRow <- rep(0,nrow)
        box.col <- switch(as.character(NS),
                          "2"=c(1,ncol),
                          "3"=c(1,2,ncol),
                          "4"=c(1,1,ncol,ncol),
                          "5"=c(1,1,ceiling((ncol-1)/2),ncol,ncol),
                          "6"=c(1,3,3,5,6,6),rep(1:ncol,length.out=NS))
        box.row <- switch(as.character(NS),
                          "2"=c(1,1),
                          "3"=c(nrow,1,nrow),
                          "4"=c(1,nrow,1,nrow),
                          "5"=c(1,nrow,ceiling(nrow/2),1,nrow),
                          "6"=c(3,1,6,4,1,6),
                          rep(1:nrow,length.out=NS))
    }
    else{ # survival or competing risks
        ## adjustRowsInColumn <- rep(1,ncol)
        ## adjustColsInRow <- rep(1,nrow)
        if (ceiling(NS/2)==floor(NS/2)){ ## equal number of states and unequal number of absorbing states
            box.col <- c(1,rep(ncol,NS-1))
            box.row <- c(NS/2,1:(NS-1))
        } else{ 
            box.col <- c(1,rep(ncol,NS-1))
            box.row <- c((NS+1)/2,(1:NS)[-(NS+1)/2])
        }
    }
    if (is.null(box.row) || is.null(box.col))
        stop(paste0("Please specify the layout for this ",NS," state model in either of the following two ways:\n",
                    paste0("box",1:NS,".row=")
                    ))
    layoutDefaults <- data.frame(name=paste("box",1:NS,sep=""),
                                 row=box.row,
                                 column=box.col,
                                 stringsAsFactors=FALSE)
    layoutDefaultList <- lapply(1:NS,function(x)layoutDefaults[x,-1,drop=FALSE])
    names(layoutDefaultList) <- layoutDefaults$name
    layout <- SmartControl(list(...),
                           keys=c(layoutDefaults$name),
                           defaults=c(layoutDefaultList),
                           ignore.case=TRUE,
                           replaceDefaults=FALSE,
                           verbose=FALSE)

  # }}}
    # {{{ draw empty frame

  # plot
  Xlim <- 100
  Ylim <- 100
  plot(0,0,type="n",xlim=c(0,Xlim),ylim=c(0,Ylim),xlab="",ylab="",axes=FALSE)
  ## backGround(c(0,100),c(0,100),bg="yellow")

  # }}}
    # {{{ default values

    if (missing(cex))
        theCex <- 2
    else
        theCex <- cex
    if (found <- match("arrowLabel.cex",names(thecall),nomatch=0))
        arrowLabel.cex <- thecall[[found]]
    else
        arrowLabel.cex <- rep(theCex,N)
    ## boxes
    boxDefaults <- data.frame(name=paste("box",1:NS,sep=""),
                              xpd=TRUE,
                              stringsAsFactors=FALSE)
    ## box labels
    boxLabelDefaults <- data.frame(name=paste("label",1:NS,sep=""),stringsAsFactors=FALSE,label=stateLabs)
    ## arrows
    arrowDefaults <- data.frame(name=paste("arrow",1:N,sep=""),
                                code=2,
                                lwd=1,
                                headoffset=strwidth("ab",cex=arrowLabel.cex),
                                length=.13,
                                stringsAsFactors=FALSE)
    arrowDefaults <- cbind(arrowDefaults,ordered.transitions)
    ## arrowlabels
    if (missing(changeArrowLabelSide))
        changeArrowLabelSide <- rep(FALSE,N)
    arrowlabelDefaults <- data.frame(name=paste("arrowlabel",1:N,sep=""),
                                     label=arrowLabelStyle,
                                     x=NA,
                                     y=NA,
                                     stringsAsFactors=FALSE,
                                     cex=arrowLabel.cex)
    arrowlabelDefaults <- cbind(arrowlabelDefaults,ordered.transitions)
    arrowlabelDefaults$numfrom <- factor(arrowlabelDefaults$from,levels=states,labels=numstateLabels)
    arrowlabelDefaults$numto <- factor(arrowlabelDefaults$to,levels=states,labels=numstateLabels)
    if (missing(arrowLabels)){
        arrowLabels <- NULL
    }
    arrowLabels.p <- TRUE
    if (length(arrowLabels)>0 &&is.logical(arrowLabels) && arrowLabels==FALSE){
        arrowLabels <- rep("",N)
        arrowLabels.p <- FALSE
    }
    else{
        if (length(arrowLabels)==0){
            arrowLabels <- lapply(1:N,function(i){
                bquote(paste(expression(.(as.name(arrowLabelSymbol))[.(paste(as.character(arrowlabelDefaults$numfrom[i]),
                                                                             as.character(arrowlabelDefaults$numto[i]),
                                                                             sep=""))](t))))
            })
        } else{
            stopifnot(length(arrowLabels)==N)
        }
    }
    arrowlabelhits <- match(paste("arrow",1:N,".label",sep=""),names(thecall),nomatch=0)
    for (i in 1:N){
        if (arrowlabelhits[i]!=0){
            arrowLabels[[i]] <- thecall[[arrowlabelhits[i]]]
        }
    }

    # }}}
    # {{{ compute box dimensions relative to cex of box labels
    
    ## to find the cex for the box labels, first initialize
    boxLabelCex <- rep(theCex,NS)
    ## then look for label.cex
    if (theLabelCex <- match("label.cex",names(thecall),nomatch=0)){
        boxLabelCex <- rep(thecall[[theLabelCex]],NS)
    }
    # finally adjust for box individual values 
    if (any(iLabelCex <- match(paste("label",1:NS,".cex",sep=""),names(thecall),nomatch=0))){
        for (i in 1:NS){
            if ((argi <- iLabelCex[i])!=0)
                boxLabelCex[i] <- thecall[[argi]]
        }
    }

    if (length(boxLabelCex)<length(stateLabs))
        boxLabelCex <- rep(boxLabelCex,length.out=length(stateLabs))
    state.width <- sapply(1:length(stateLabs),function(i){strwidth(stateLabs[i],cex=boxLabelCex[i])})
    state.height <- sapply(1:length(stateLabs),function(i){strheight(stateLabs[i],cex=boxLabelCex[i])})
    if (missing(oneFitsAll)){
        oneFitsAll <- length(unique(boxLabelCex))==1
    }
    estimate.box.width <- missing(box.width)
    estimate.box.height <- missing(box.height)
    estimate.box.padding <- missing(box.padding)
    if (estimate.box.padding)
        box.padding <- rep(strwidth("ab",cex=max(boxLabelCex)),2)
        ## box.padding <- rep(0,2)
    else
        box.padding <- rep(box.padding,length.out=2)
    if (estimate.box.width){
        state.width <- pmin(Xlim/ncol,state.width)
        box.width <- if (oneFitsAll==TRUE) rep(max(state.width),NS) + box.padding[[1]] else state.width + box.padding[[1]]
    }else{
        state.height <- pmin(Xlim/ncol,state.height)
        box.width <- rep(box.width,length.out=NS)
    }
    if (estimate.box.height){
        box.height <- if (oneFitsAll==TRUE) rep(max(state.height),NS) + box.padding[[2]] else state.height + box.padding[[2]]
    }else{
        box.height <- rep(box.height,length.out=NS)
    }
    if ((estimate.box.height||estimate.box.width||estimate.box.padding) && is.null(attr(.Device,"filepath"))){
        warning("The dimension of the boxes may depend on the current graphical device",
                "\nin the sense that the layout and centering of text may change when you resize the",
                " graphical device and call the same plot.")
    }
    names(box.height) <- paste0("box",1:NS)
    names(box.width) <- paste0("box",1:NS)
    thecall$box.height=round(box.height,4)
    thecall$box.width=round(box.width,4)
    # }}}
    # {{{ arrange the boxes in the layout
    boxCol <- sapply(layout,function(x){x$column})
    if (any(boxCol>ncol)) ncol <- max(boxCol)
    boxRow <- sapply(layout,function(x){x$row})
    if (any(boxRow>ncol)) nrow <- max(boxRow)
    if (missing(ybox.position)){
        ybox.position <- numeric(NS)
        names(ybox.position) <- paste("box",numstates,sep="")
        # {{{y box positions
        for (x in 1:ncol){
            ## For each column find y positions for boxes
            boxesInColumn <- names(boxCol)[boxCol==x]
            boxesInColumnNumbers <- as.numeric(sapply(strsplit(boxesInColumn,"box"),function(x)x[[2]]))
            if (length(boxesInColumn)>0){
                yPossible <- centerBoxes(Ylim,box.height[boxesInColumnNumbers],nrow,boxRow[boxesInColumn])
                for (b in 1:length(boxesInColumn)){
                    ybox.position[boxesInColumn[b]] <- yPossible[b]
                }
            }
        }
        ## row 1 is on top but the y-axis starts at the bottom
        ## therefore need to transform
        ybox.position <- 100-(ybox.position+box.height)
    }else{
        stopifnot(length(ybox.position)==NS)
        names(ybox.position) <- paste("box",numstates,sep="")
    }
    # }}}
    # {{{x box positions
    if (missing(xbox.position)){
        xbox.position <- numeric(NS)
        names(xbox.position) <- paste("box",numstates,sep="")
        for (x in 1:nrow){
            ## For each row find x positions for boxes
            boxesInRow <- names(boxRow)[boxRow==x]
            boxesInRowNumbers <- as.numeric(sapply(strsplit(boxesInRow,"box"),function(x)x[[2]]))
            if (length(boxesInRow)>0){
                if (sum(box.width[boxesInRowNumbers])>Xlim)
                    stop(paste("Sum of box widths in row",x,"exceed limit",Xlim))
                xpossible <- centerBoxes(Xlim,box.width[boxesInRowNumbers],ncol,boxCol[boxesInRow])
                for (b in 1:length(boxesInRow)){
                    xbox.position[boxesInRow[b]] <- xpossible[b]
                }
            }
        }
 
    }else{
        stopifnot(length(xbox.position)==NS)
        names(xbox.position) <- paste("box",numstates,sep="")
    }
    # }}}
    
    thecall$xbox.position=round(xbox.position,4)
    thecall$ybox.position=round(ybox.position,4)
    xtext.position <- xbox.position + pmax(0,box.width - state.width)/2
    ytext.position <- ybox.position + pmax(0,box.height - state.height)/2
    names(xtext.position) <- paste("box",1:NS,sep="")
    names(ytext.position) <- paste("box",1:NS,sep="") 
    thecall$xtext.position=round(xtext.position,4)
    thecall$ytext.position=round(ytext.position,4)
    
    if (verbose){
        cat("\n\nBoxlabel data:\n\n")
        print(data.frame(stateLabs,
                         boxCol,
                         boxRow,
                         x.pos=round(xbox.position,2),
                         y.pos=round(ybox.position,2),
                         width=round(box.width,2),
                         label.width=round(state.width,2),
                         label.height=round(state.height,2),
                         boxLabelCex))
    }
    boxDefaults <- cbind(boxDefaults,xleft=xbox.position,ybottom=ybox.position,xright=xbox.position+box.width,ytop=ybox.position+box.height)
    boxLabelDefaults <- cbind(boxLabelDefaults,
                              x=xtext.position,
                              y=ytext.position,
                              cex=boxLabelCex)

  # }}}
    # {{{ compute arrow positions

    doubleArrow <- match(paste(arrowDefaults[,"to"],arrowDefaults[,"from"]),paste(arrowDefaults[,"from"],arrowDefaults[,"to"]),nomatch=0)
    arrowDefaults <- cbind(arrowDefaults,doubleArrow)
    arrowList <- for (trans in 1:N){
                     from.state <- factor(ordered.transitions[trans,1],levels=states,labels=numstates)
                     to.state <- factor(ordered.transitions[trans,2],levels=states,labels=numstates)
                     ArrowPositions <- findArrow(Box1=c(round(xbox.position[from.state],4),round(ybox.position[from.state],4)),
                                                 Box2=c(round(xbox.position[to.state],4),round(ybox.position[to.state],4)),
                                                 Box1Dim=c(box.width[from.state],box.height[from.state]),
                                                 Box2Dim=c(box.width[to.state],box.height[to.state]),
                                                 verbose=FALSE)
                     Len <- function(x){sqrt(sum(x^2))}
                     from <- ArrowPositions$from
                     to <- ArrowPositions$to
                     ArrowDirection <- to-from
                     ArrowDirection <- ArrowDirection/Len(ArrowDirection)
                     ## perpendicular direction
                     PerDir <- rev(ArrowDirection)*c(1,-1)/Len(ArrowDirection)
                     ## shift double arrows
                     dd <- arrowDefaults[trans,"doubleArrow"]
                     if (dd!=0){
                         dist <- strwidth(".",cex=arrowLabel.cex)
                         arrowDefaults[trans,"headoffset"]+dist
                         if (dd>trans){
                             from <- from + sign(PerDir) * c(dist,dist)
                             to <- to + sign(PerDir) * c(dist,dist)
                         }
                         else{
                             from <- from + sign(PerDir) * c(dist,dist)
                             to <- to + sign(PerDir) * c(dist,dist)
                         }
                     }
                     # shift the start and end points of arrows by ArrowHeadOffset
                     ArrowHeadOffset <- arrowDefaults[trans,"headoffset"]
                     from <- from+sign(ArrowDirection)*c(ArrowHeadOffset,ArrowHeadOffset)*abs(ArrowDirection)
                     to <- to-sign(ArrowDirection)*c(ArrowHeadOffset,ArrowHeadOffset)*abs(ArrowDirection)
                     arrowDefaults[trans,"x0"] <- from[1]
                     arrowDefaults[trans,"x1"] <- to[1]
                     arrowDefaults[trans,"y0"] <- from[2]
                     arrowDefaults[trans,"y1"] <- to[2]
                     ## shift arrow label perpendicular (left) to arrow direction
                     offset <- strwidth(".",cex=arrowLabel.cex)
                     ArrowMid <- (to+from)/2
                     ## points(x=ArrowMid[1],y=ArrowMid[2],col=3,pch=16)
                     if (changeArrowLabelSide[trans]==TRUE)
                         ArrowLabelPos <- ArrowMid - sign(PerDir) * c(offset,offset)
                     else
                         ArrowLabelPos <- ArrowMid + sign(PerDir) * c(offset,offset)
                     try1 <- try(mode((arrowLabels[[trans]])[2])[[1]]=="call",silent=TRUE)
                     ## try2 <- try(as.character(arrowLabels[[trans]])[[1]]=="paste",silent=TRUE)
                     labIsCall <- (!inherits(try1,"try-error") && try1)
                     if (labIsCall){ # symbolic label
                         arrowLabels[[trans]] <- ((arrowLabels[[trans]])[2])[[1]][[2]]
                     }
                     ## relative label height
                     lab <- arrowLabels[[trans]]
                     labelHeight <- strheight(lab,cex=arrowlabelDefaults[trans,"cex"])
                     ## relative label width 
                     labelWidth <-  strwidth(lab,cex=arrowlabelDefaults[trans,"cex"])
                     ## shift further according to label height and width in perpendicular direction
                     if (changeArrowLabelSide[trans]==TRUE)
                         ArrowLabelPos <- ArrowLabelPos-sign(PerDir)*c(labelWidth/2,labelHeight/2)
                     else
                         ArrowLabelPos <- ArrowLabelPos+sign(PerDir)*c(labelWidth/2,labelHeight/2)
                     arrowlabelDefaults[trans,"x"] <- ArrowLabelPos[1] 
                     arrowlabelDefaults[trans,"y"] <- ArrowLabelPos[2]
                 }

  # }}}
    # {{{ Smart argument control

    boxDefaultList <- lapply(1:NS,function(x)boxDefaults[x,-1,drop=FALSE])
    names(boxDefaultList) <- boxDefaults$name
    boxLabelDefaultList <- lapply(1:NS,function(x)boxLabelDefaults[x,-1,drop=FALSE])
    names(boxLabelDefaultList) <- boxLabelDefaults$name
    arrowDefaultList <- lapply(1:N,function(x)arrowDefaults[x,-1,drop=FALSE])
    names(arrowDefaultList) <- as.character(arrowDefaults$name)
    arrowlabelDefaultList <- lapply(1:N,function(x)arrowlabelDefaults[x,-1,drop=FALSE])
    names(arrowlabelDefaultList) <- as.character(arrowlabelDefaults$name)
    boxTagsDefaultList <- list(labels=numstateLabels,cex=1.28,adj=c(-.5,1.43))
    smartArgs <- SmartControl(list(...),
                              keys=c(boxDefaults$name,
                                     boxLabelDefaults$name,
                                     as.character(arrowDefaults$name),
                                     as.character(arrowlabelDefaults$name),
                                     "boxtags"),
                              defaults=c(boxLabelDefaultList,arrowDefaultList,arrowlabelDefaultList,boxDefaultList,list("boxtags"=boxTagsDefaultList)),
                              ignore.case=TRUE,
                              replaceDefaults=FALSE,
                              verbose=verbose)
    
    # }}}
    # {{{
    if (rasta[[1]]){
        abline(v=seq(0,100,5),col="gray55")
        abline(h=seq(0,100,5),col="gray55")
        text(seq(0,100,5),x=seq(0,100,5),y=100)
        text(seq(0,100,5),x=seq(0,100,5),y=0)
        text(seq(0,100,5),y=seq(0,100,5),x=100)
        text(seq(0,100,5),y=seq(0,100,5),x=0)
    }
    # }}}
    # {{{  draw the boxes
  
  for (i in 1:NS) {
    suppressWarnings(do.call("rect",smartArgs[[paste("box",i,sep="")]]))
  }

  # }}}
    # {{{  label the boxes
  
  for (i in 1:NS) {
    suppressWarnings(do.call("text",c(list(adj=c(0,0)),smartArgs[[paste("label",i,sep="")]])))
  }

    # }}}
    # {{{  draw the arrows
    for (i in 1:N){
        alist <- c(smartArgs[[paste("arrow",i,sep="")]])
        if (!missing(curved) && requireNamespace("diagram",quietly=TRUE)){
            dd <- strwidth("ab",cex=max(boxLabelCex))
            alist <- c(with(alist,list(from=c(x0,y0),
                                       to=c(x1,y1),curve=curved,segment=c(dd,100-dd)/100)))
            carrow <- diagram::curvedarrow
            suppressWarnings(do.call("carrow",alist))
        }else{
            suppressWarnings(do.call("arrows",alist))
        }
    }
    # }}}
    # {{{  label the arrows
    if (verbose) arrowLabel.data <- NULL
    if (arrowLabels.p==TRUE){
        for (i in 1:N){
            labelList <- smartArgs[[paste("arrowlabel",i,sep="")]]
            if (verbose) arrowLabel.data <- rbind(arrowLabel.data,cbind("arrowLabel"=i,data.frame(labelList)))
            switch(labelList$label,"symbolic"={
                try1 <- try(mode((arrowLabels[[i]])[2])[[1]]=="call",silent=TRUE)
                labIsCall <- (!inherits(try1,"try-error") && try1)
                suppressWarnings(do.call("text",c(list(labels=bquote(arrowLabels[[i]])),labelList)))        
            }, "count"={
                tabTrans <- as.matrix(table(transitions))
                lab <- paste("n=",tabTrans[as.character(labelList$from),as.character(labelList$to)])
                suppressWarnings(do.call("text",c(list(labels=quote(lab)),labelList)))
            })
        }
    }
    if (verbose) {
        cat("\n\nArrow label data:\n\n")
        print(arrowLabel.data)
    }
    # }}}
    # {{{  put numbers in the upper left corner of the boxes (if wanted)

  if (tagBoxes[[1]]==TRUE){
    tagList <- smartArgs$boxtags
    nix <- lapply(1:NS,function(b) {
      lab <- tagList[b]
      text(x=xbox.position[b],
           y=ybox.position[b]+box.height,
           labels=tagList$labels[b],
           cex=tagList$cex,
           adj=tagList$adj)})
  }
    # }}}
    if (verbose){
        cat("To change the order of the boxes,\nrelevel the factor 'event' in the dataset which defines the Hist object.\n")
    }
    invisible(list(call=thecall,parameters=smartArgs))
}


position.finder <- function(border,len,n){
## distribute the boxes of lenght len uniformly
## over [0,border]
 if (n==1)
    (border - len)/2
  else{
    seq(0,border-.5*len,len + (border-(n * len))/(n-1))
  }  
}

centerBoxes <- function(border,len,ncell,pos){
    ## box i has length len[i] and is centered in cell pos[i]
    ## return the position in [0,border] of the lower
    ## border of the boxes
    cellwidth <- border/ncell
    nboxes <- length(len)
    if ((luft <- border-sum(len))<0) stop("sum of box dimensions exceeds limits")
    if (nboxes>ncell) stop("too many boxes in one row")
    ## case: all boxes fit into given cell width
    ## if (all(len<cellwidth)){
    box.pos <- seq(from=0,to=border,by=cellwidth)[pos] + pmax(0,sapply(len,function(l) {(cellwidth - l)/2}))
    ## spread as far as possible
    boxPos <- sapply(1:length(box.pos),function(b){
        bp <- box.pos[b]
        if (ncell>1 && pos[b]==1) # at the left/lower border
            bp <- min(0,abs(box.pos[b]))
        if (ncell> 1 && pos[b]==ncell)# at the right/upper border
            bp <- max(border-len[b],box.pos[b])
        bp
    })
    boxPos
}

#' Product-limit geoms for survival and competing risks curves
#'
#' Draw Kaplan-Meier or Aalen-Johansen curves with optional confidence bands.
#' For competing risks with multiple causes, the default display shows one curve
#' per cause. When `cause = "stacked"`, causes are drawn as stacked filled
#' rectangles instead of step curves with confidence shadows.
#'
#' @param mapping Set of aesthetic mappings created by [ggplot2::aes()].
#' @param data Data frame.
#' @param position Position adjustment.
#' @param na.rm If `FALSE`, missing values are removed with a warning.
#' @param show.legend Logical. Should this layer be included in the legends?
#' @param inherit.aes If `FALSE`, overrides the default aesthetics.
#' @param type Passed to [prodlim::prodlim()], usually `"risk"` or `"surv"`.
#' @param cause Cause(s) for competing risks. Can be a vector of causes or the
#'   string `"stacked"`.
#' @param conf_int Logical. Draw confidence intervals.
#' @param conf_int_alpha Alpha level for confidence shadows.
#' @param percent Logical. Passed to [prodlim::summary.prodlim()].
#' @param timeconverter Optional time conversion string.
#' @param times Optional evaluation times passed to
#'   [prodlim::summary.prodlim()].
#' @param cens.code Censoring code passed to [prodlim::Hist()].
#' @param ... Further arguments passed to [ggplot2::layer()].
#'
#' @details
#' When multiple causes are specified, `fill`/`colour` aesthetics are ignored
#' and replaced by a cause-based mapping so that all causes are shown in a
#' single legend. When `cause = "stacked"`, causes are stacked on top of each
#' other and confidence shadows are not drawn.
#'
#' @examples
#' library(riskRegression)
#' library(data.table)
#' library(ggplot2)
#' data(Melanoma)
#' # Kaplan-Meier
#' ggplot(data = Melanoma,aes(x = time, event = 1*(status != 0)))+
#'        geom_prodlim(type = "surv")
#' # stratified Kaplan-Meier inherited aes
#' ggplot(data = Melanoma,aes(x = time, event = 1*(status != 0),fill = sex,color = sex))+
#'        geom_prodlim(type = "surv")
#' # stratified Kaplan-Meier geom aes
#' ggplot(data = Melanoma,aes(x = time, event = 1*(status != 0)))+
#'        geom_prodlim(aes(fill = sex,color = sex),type = "surv")
#' # facet
#' ggplot(data = Melanoma,aes(x = time, event = 1*(status != 0)))+
#'        geom_prodlim(type = "surv")+facet_grid(~sex)
#' # stratified and facet
#' ggplot(data = Melanoma,aes(x = time, event = 1*(status != 0),fill = epicel,color = epicel))+
#'        geom_prodlim(type = "surv")+facet_grid(~sex)
#'
#' # Aalen-Johansen
#' ggplot(data = Melanoma,aes(x = time, event = status))+
#'        geom_prodlim(type = "surv",cause = 1:2)
#' # stratified Aalen-Johansen inherited aes
#' ggplot(data = Melanoma,aes(x = time, event = status,fill = sex,color = sex))+
#'        geom_prodlim(type = "surv")
#' # stratified Aalen-Johansen geom aes
#' ggplot(data = Melanoma,aes(x = time, event = status))+
#'        geom_prodlim(aes(fill = sex,color = sex),type = "surv")
#' # facet
#' ggplot(data = Melanoma,aes(x = time, event = status))+
#'        geom_prodlim(type = "surv")+facet_grid(~sex)
#' # stratified and facet
#' ggplot(data = Melanoma,aes(x = time, event = status,fill = epicel,color = epicel))+
#'        geom_prodlim(type = "surv")+facet_grid(~sex)
#' # stacked
#' ggplot(data = Melanoma,aes(x = time, event = status))+
#'        geom_prodlim(cause = "stacked")+facet_grid(~sex)
#' # stacked with cens.code option
#' ggplot(data = Melanoma,aes(x = time, event = event))+
#'        geom_prodlim(cens.code="censored",cause = "stacked")+facet_grid(~sex)
#' @export
geom_prodlim <- function(mapping = NULL, data = NULL,
                         position = "identity", na.rm = FALSE,
                         show.legend = NA,
                         inherit.aes = TRUE,
                         type = "risk",
                         cause = NULL,
                         conf_int = TRUE,
                         conf_int_alpha = 0.2,
                         percent = TRUE,
                         timeconverter = NULL,
                         times = NULL,
                         cens.code = "0",
                         ...) {
    Event <- NULL
    modify_aes <- function(a, b) {
        if (is.null(a)) return(b)
        la <- as.list(a)
        lb <- as.list(b)
        out <- utils::modifyList(la, lb)
        class(out) <- class(a)
        out
    }

    multi_cause <- identical(cause, "stacked") || length(cause) > 1
    user_mapping <- if (is.null(mapping)) character(0) else names(as.list(mapping))

    if (multi_cause) {
        mapping <- modify_aes(
            mapping,
            ggplot2::aes(
                group  = ggplot2::after_stat(Event),
                colour = ggplot2::after_stat(Event),
                fill   = ggplot2::after_stat(Event)
            )
        )
        if (length(intersect(user_mapping, c("colour", "color", "fill"))) > 0) {
            warning(
                "colour/fill aesthetics are ignored when multiple causes are specified",
                call. = FALSE
            )
        }
    }

    ggplot2::layer(
        stat = StatProdlim,
        geom = GeomProdlim,
        data = data,
        mapping = mapping,
        position = position,
        show.legend = show.legend,
        inherit.aes = inherit.aes,
        params = list(
            na.rm = na.rm,
            type = type,
            cause = cause,
            conf_int = conf_int,
            conf_int_alpha = conf_int_alpha,
            percent = percent,
            timeconverter = timeconverter,
            times = times,
            cens.code = cens.code,
            ...
        )
    )
}

### parseFormula.R --- 
#----------------------------------------------------------------------
## Author: Thomas Alexander Gerds
## Created: mar  4 2026 (11:28) 
## Version: 
## Last-Updated: mar  4 2026 (14:33) 
##           By: Thomas Alexander Gerds
##     Update #: 23
#----------------------------------------------------------------------
## 
### Commentary: 
## 
### Change Log:
#----------------------------------------------------------------------
## 
### Code:
##' Parse a formula to extract terms and handle special functions
##'
##' This function parses a model formula, extracts terms, and processes specified 'special' functions and their arguments.
##' It allows defining default values for special functions and specifying aliases for those names.
##'
##' @title Parse Formula
##' @param formula A model formula (e.g., Y ~ X + Z or Y ~ fun(X, param = value)).
##' @param data An optional data frame containing the variables referenced in the formula.
##' @param specials A character vector of special function names that should be specifically handled.
##' @param special.default.values A named list of named lists specifying default argument values for each special.
##' @param alias.names A named list mapping special names to their alias(es) (e.g., `list("special" = "alias")`).
##' @return A list where each element corresponds to a special function in the formula and contains the parsed terms and their arguments.
##' @seealso [stats::terms()], [strip.terms()]
##' @examples
##' # Basic usage with specials
##' parseFormula(Y ~ fun(X, a = 1), specials = "fun")
##'
##' # With defaults and aliases
##' parseFormula(Y ~ gfun(X, a = 7), 
##'              specials = "fun", 
##'              alias.names = list("fun" = "gfun"),
##'              special.default.values = list("fun" = list("a" = 0)))
##'
##' # Complex use case with multiple specials
##' parseFormula(Surv(time, status) ~ age + strata(sex, test = 1) + prop(albumin),
##'              specials = c("prop", "strata"), 
##'              special.default.values = list("prop" = list(power = 0), "strata" = list(test = 1)))
##' @export
##' @author Thomas A. Gerds <tag@@biostat.ku.dk>
##' @return A named list with an element for each special that contains a named list where each element
##'         corresponds to a formula term that is affected by the special that contains
##'         the value(s) of the additional argument(s) for this term.
##' @seealso model.design strip.terms
##' @examples
##' parseFormula(Y~fun(X,a=1),specials="fun",special.default.values=list("fun"=list("a"=0)))
##' parseFormula(Y~fun(X),specials="fun",special.default.values=list("fun"=list("a"=0)))
##' parseFormula(Y~gfun(X,a=7),specials="fun",
##'              alias.names=list("fun"="gfun"),
##'              special.default.values=list("fun"=list("a"=0)))
##' parseFormula(formula = Surv(time,status)~age
##'                             +const(factor(edema))
##'                             +strata(sex,test=0)
##'                             +prop(alb)+prop(bili,power=1)
##'                             +tp(albumin),
##'                   specials = c("prop","timevar","strata","tp","const"),
##'                   special.default.values =list("prop" = list(power = 0),
##'                                                "strata" = list(test = 1)))
##' parseFormula(formula = Surv(time,status)~age
##'                             +const(factor(edema))
##'                             +strata(sex,test=0)
##'                             +prop(alb)+prop(bili,power=1)
##'                             +tp(albumin),
##'                   specials = c("prop","timevar","strata","tp"),
##'                   special.default.values =list("prop" = list(power = 0),
##'                                                "strata" = list(test = 1)),
##'                   alias.names=list("prop"="const"))
##' @export 
##' @author Thomas A. Gerds <tag@@biostat.ku.dk>
parseFormula <- function(formula,
                         data = NULL,
                         specials,
                         special.default.values = NULL,
                         alias.names = NULL){
    Terms <- stats::terms(x=formula,specials=specials,data=data)
    Terms <- strip.terms(Terms,
                         specials=specials,
                         arguments=special.default.values,
                         alias.names=alias.names,
                         unspecials = NULL)
    result <- list()
    for (special in specials) {
        strippedArgs <- attr(Terms, "stripped.arguments")[[special]]
        if (is.null(strippedArgs)) {
            result[[special]] <- NULL
        } else {
            parsedElements <- list()
            for (element in names(strippedArgs)) {
                elementArgs <- strippedArgs[[element]]
                if (is.null(elementArgs)) {
                    parsedElements[[element]] <- NULL
                } else {
                    parsedElements[[element]] <- lapply(elementArgs, function(arg) {
                        if (!is.na(as.numeric(arg))) as.numeric(arg) else arg
                    })
                }
            }
            result[[special]] <- parsedElements
        }
    }
    result
}



######################################################################
### parseFormula.R ends here

##' Extract from a vector of character strings the names of special functions and auxiliary arguments
##'
##' Signals an error if an element has more arguments than specified by argument arguments.
##' @title Parse special terms 
##' @param x Vector of character strings.
##' @param special A character string: the name of the special argument.
##' @param arguments A vector which contains the arguments of the special function
##' @return A named list of parsed arguments. The names of the list are the special variable names, the elements
##' are lists of arguments.
##' @seealso model.design
##' @examples
##' 
##' ## ignore arguments
##' parseSpecialNames("treat(Z)",special="treat")
##' ## set default to 0
##' parseSpecialNames(c("log(Z)","a","log(B)"),special="log",arguments=list("base"=0))
##' ## set default to 0
##' parseSpecialNames(c("log(Z,3)","a","log(B,base=1)"),special="log",arguments=list("base"=0))
##' ## different combinations of order and names
##' parseSpecialNames(c("log(Z,3)","a","log(B,1)"),
##'                   special="log",
##'                   arguments=list("base"=0))
##' parseSpecialNames(c("log(Z,1,3)","a","log(B,u=3)"),
##'                   special="log",
##'                   arguments=list("base"=0,"u"=1))
##' parseSpecialNames(c("log(Z,u=1,base=3)","a","log(B,u=3)"),
##'                   special="log",
##'                   arguments=list("base"=0,"u"=1))
##' parseSpecialNames(c("log(Z,u=1,base=3)","a","log(B,base=8,u=3)"),
##'                   special="log",
##'                   arguments=list("base"=0,"u"=1))
##' parseSpecialNames("treat(Z,u=2)",
##'                   special="treat",
##'                   arguments=list("u"=1,"k"=1))
##' parseSpecialNames(c("treat(Z,1,u=2)","treat(B,u=2,k=3)"),
##'                   special="treat",
##'                   arguments=list("u"=NA,"k"=NULL))
##' ## does not work to set default to NULL:
##' parseSpecialNames(c("treat(Z,1,u=2)","treat(B,u=2)"),
##'                   special="treat",
##'                   arguments=list("u"=NA,"k"=NULL))
##' @author Thomas A. Gerds <tag@@biostat.ku.dk>
##' @export
parseSpecialNames <- function(x,special,arguments){
    if (missing(arguments)) {
        argnames <- NULL
    } else {
        argnames <- names(arguments)
    }
    ## it would be possible to vectorize the function with the regexp:
    ## paste("(",paste(special,collapse="|"),")\\(|)$",sep="")
    ## but this causes some
    ## confusion and extra work
    specialRegexp <- paste("^",special,"\\(|)$",sep="")
    posSpecial <- grep(specialRegexp,x,value=FALSE)
    if (length(posSpecial)>0){
        specialTerms <- strsplit(x[posSpecial],specialRegexp)
        ## if length is 1 then term is unspecial
        ## isSpecial <- sapply(listTerms,length)
        # check for further arguments
        termsWithArguments <- unlist(lapply(specialTerms,function(x){
            if (length(x)<2) NULL
            else strsplit(x[[2]],"[ ]*,[ ]*")}), recursive=FALSE)
        varnames <- lapply(termsWithArguments,function(x){x[[1]]})
        ## attr(varnames,"special.position") <- posSpecial
        ## only fish arguments if this is desired
        if (is.null(argnames)){
            out <- vector(mode="list",length(varnames))
            names(out) <- varnames
            return(out)
        }else{
            varnames <- unlist(varnames)
            if (length(problem <- grep("=",varnames,value=TRUE))>0)
                stop(paste("Problematic variable name '",problem,"'. Variable names used in special may not contain '='.",sep=""))
            givenArguments <- lapply(termsWithArguments,function(x){
                if (length(x)==1) NULL else x[2:length(x)]
            })
            names(givenArguments) <- varnames
            # {{{  parse arguments
            specialArgumentList <- lapply(givenArguments,function(args){
                if (!is.null(args)){
                    fullvalue <- strsplit(args,"=")
                    fullvalue <- lapply(fullvalue,function(x){ ## remove whitespace
                        gsub(" ","",x)
                    })
                    givennames <- sapply(fullvalue,function(x){
                        if (length(x)==1)
                            ""
                        else
                            x[[1]]
                    })
                    values <- lapply(fullvalue,function(x){
                        if (length(x)==1)
                            x[[1]]
                        else
                            x[[2]]
                    })
                    if(length(argnames)<length(args)) stop("Too many arguments for special function '",special,"'.")
                    realnames <- givennames[givennames!=""]
                    thismatch <- match(realnames,argnames,nomatch=0)
                    if (length(realnames)>0)
                        if (!all(thismatch))
                            stop("Argument(s) '",
                                 paste(realnames,collapse=", "),
                                 "' is not an argument of '",
                                 special,
                                 "'. Valid argument(s): '",
                                 paste(argnames,collapse=", "),"'.")
                    names(values) <- givennames
                    nadd <- length(argnames)-length(values)
                    if (nadd>0){
                        values <- c(values,rep(NA,nadd))
                    }
                    thatmatch <- match(argnames,names(values),nomatch=0)
                    names(values)[names(values)==""] <- argnames[thatmatch==0]
                    values <- values[argnames]
                    ## set defaults
                    values[is.na(values)] <- unlist(arguments)[is.na(values)]
                    values
                } else {
                    ## use defaults
                    arguments
                }
            })
            # }}}
            names(specialArgumentList) <- names(givenArguments)
            ## attr(specialArgumentList,"special.position") <- posSpecial
            specialArgumentList
        }
    } else{NULL}
}

##' product limit method
##' 
##' Nonparametric estimation in event history analysis. Featuring fast
##' algorithms and user friendly syntax adapted from the survival package.  The
##' product limit algorithm is used for right censored data; the
##' self-consistency algorithm for interval censored data.
##' 
##' 
##' The response of \code{formula} (ie the left hand side of the `~' operator)
##' specifies the model.
##' 
##' In two-state models -- the classical survival case -- the standard
##' Kaplan-Meier method is applied.  For this the response can be specified as a
##' \code{\link[survival]{Surv}} or as a \code{\link{Hist}} object. The \code{\link{Hist}}
##' function allows you to change the code for censored observations, e.g.
##' \code{Hist(time,status,cens.code="4")}.
##' 
##' Besides a slight gain of computing efficiency, there are some extensions
##' that are not included in the current version of the survival package:
##' 
##' (0) The Kaplan-Meier estimator for the censoring times \code{reverse=TRUE}
##' is correctly estimated when there are ties between event and censoring
##' times.
##' 
##' (1) A conditional version of the kernel smoothed Kaplan-Meier estimator for at most one
##' continuous predictors using nearest neighborhoods (Beran 1981,
##' Stute 1984, Akritas 1994).
##' 
##' (2) For cluster-correlated data the right hand side of \code{formula} may
##' identify a \code{\link[survival]{cluster}} variable. In that case Greenwood's variance
##' formula is replaced by the formula of Ying and Wei (1994).
##' 
##' (3) Competing risk models can be specified via \code{\link{Hist}} response
##' objects in \code{formula}.
##' 
##' The Aalen-Johansen estimator is applied for estimating the absolute risk of the competing causes, i.e., the cumulative
##' incidence functions. 
##' 
##' Under construction:
##' 
##' (U0) Interval censored event times specified via \code{\link{Hist}} are used
##' to find the nonparametric maximum likelihood estimate. Currently this works
##' only for two-state models and the results should match with those from the
##' package `Icens'.
##' 
##' (U1) Extensions to more complex multi-states models
##' 
##' (U2) The nonparametric maximum likelihood estimate for interval censored
##' observations of competing risks models.
##' 
##' @param formula A formula whose left hand side is a \code{Hist}
##' object. In some special cases it can also be a \code{\link[survival]{Surv}}
##' response object, see the details section. The right hand side is
##' as usual a linear combination of covariates which may contain at
##' most one continuous factor. Whether or not a covariate is
##' recognized as continuous or discrete depends on its class and on
##' the argument \code{discrete.level}. The right hand side may also
##' be used to specify clusters, see the details section.
##' @param data A data.frame in which all the variables of
##' \code{formula} can be interpreted.
##' @param subset Passed as argument \code{subset} to function
##' \code{subset} which applied to \code{data} before the formula is
##' processed.
##' @param na.action All lines in data with any missing values in the
##' variables of formula are removed.
##' @param reverse For right censored data, if reverse=TRUE then the
##' censoring distribution is estimated.
##' @param conf.int The level (between 0 and 1) for two-sided
##' pointwise confidence intervals. Defaults to 0.95. Remark: only
##' plain Wald-type confidence limits are available.
##' @param bandwidth Smoothing parameter for nearest neighborhoods
##' based on the values of a continuous covariate. See function
##' \code{neighborhood} for details.
##' @param caseweights Weights applied to the contribution of each
##' subject to change the number of events and the number at
##' risk. This can be used for bootstrap and survey analysis. Should
##' be a vector of the same length and the same order as \code{data}.
##' @param discrete.level Numeric covariates are treated as factors
##' when their number of unique values exceeds not
##' \code{discrete.level}. Otherwise the product limit method is
##' applied, in overlapping neighborhoods according to the bandwidth.
##' @param x logical value: if \code{TRUE}, the full covariate matrix
##' with is returned in component \code{model.matrix}.  The reduced
##' matrix contains unique rows of the full covariate matrix and is
##' always returned in component \code{X}.
##' @param method For interval censored data only.  If equal to
##' \code{"npmle"} (the default) use the usual Turnbull algorithm,
##' else the product limit version of the self-consistent estimate.
##' @param exact If TRUE the grid of time points used for estimation
##' includes all the L and R endpoints of the observed intervals.
##' @param maxiter For interval censored data only.  Maximal number of
##' iterations to obtain the nonparametric maximum likelihood
##' estimate.  Defaults to 1000.
##' @param grid For interval censored data only. When method=one.step
##' grid for one-step product limit estimate. Defaults to sorted list
##' of unique left and right endpoints of the observed intervals.
##' @param tol For interval censored data only. Numeric value whose
##' negative exponential is used as convergence criterion for finding
##' the nonparametric maximum likelihood estimate.  Defaults to 7
##' meaning exp(-7).
##' @param type In two state models either \code{"surv"} for the Kaplan-Meier estimate of the survival
##' function or \code{"risk"} for 1-Kaplan-Meier. Default is \code{"surv"} when \code{reverse==FALSE} and \code{"risk"} when \code{reverse==TRUE}.
##' In competing risks models it has to be \code{"risk"}
##' Aalen-Johansen estimate of the cumulative incidence function.
##' @return Object of class "prodlim". See \code{\link{print.prodlim}}, \code{\link{predict.prodlim}}, predict,
##' \code{\link{summary.prodlim}}, \code{\link{plot.prodlim}}.
##' @author Thomas A. Gerds \email{tag@@biostat.ku.dk}
##' @seealso \code{\link{predictSurv}}, \code{\link{predictSurvIndividual}},
##' \code{\link{predictAbsrisk}}, \code{\link{Hist}}, \code{\link{neighborhood}},
##' \code{\link[survival]{Surv}}, \code{\link[survival]{survfit}}, \code{\link[survival]{strata}},
##' @references Andersen, Borgan, Gill, Keiding (1993) Springer `Statistical
##' Models Based on Counting Processes'
##' 
##' Akritas (1994) The Annals of Statistics 22, 1299-1327 Nearest neighbor
##' estimation of a bivariate distribution under random censoring.
##' 
##' R Beran (1981) http://anson.ucdavis.edu/~beran/paper.html `Nonparametric
##' regression with randomly censored survival data'
##' 
##' Stute (1984) The Annals of Statistics 12, 917--926 `Asymptotic Normality of
##' Nearest Neighbor Regression Function Estimates'
##' 
##' Ying, Wei (1994) Journal of Multivariate Analysis 50, 17-29 The Kaplan-Meier
##' estimate for dependent failure time observations
##' @keywords survival nonparametric cluster
##' @examples
##' 
##' ##---------------------two-state survival model------------
##' dat <- SimSurv(30)
##' with(dat,plot(Hist(time,status)))
##' fit <- prodlim(Hist(time,status)~1,data=dat)
##' print(fit)
##' plot(fit)
##' summary(fit)
##' quantile(fit)
##'
##' ## Subset
##' fit1a <- prodlim(Hist(time,status)~1,data=dat,subset=dat$X1==1)
##' fit1b <- prodlim(Hist(time,status)~1,data=dat,subset=dat$X1==1 & dat$X2>0)
##' 
##' ## --------------------clustered data---------------------
##' library(survival)
##' cdat <- cbind(SimSurv(30),patnr=sample(1:5,size=30,replace=TRUE))
##' fit <- prodlim(Hist(time,status)~cluster(patnr),data=cdat)
##' print(fit)
##' plot(fit)
##' summary(fit)
##' 
##' 
##' ##-----------compare Kaplan-Meier to survival package---------
##' 
##' dat <- SimSurv(30)
##' pfit <- prodlim(Surv(time,status)~1,data=dat)
##' pfit <- prodlim(Hist(time,status)~1,data=dat) ## same thing
##' sfit <- survfit(Surv(time,status)~1,data=dat,conf.type="plain")
##' ##  same result for the survival distribution function 
##' all(round(pfit$surv,12)==round(sfit$surv,12))
##' summary(pfit,digits=3)
##' summary(sfit,times=quantile(unique(dat$time)))
##' 
##' ##-----------estimating the censoring survival function----------------
##' 
##' rdat <- data.frame(time=c(1,2,3,3,3,4,5,5,6,7),status=c(1,0,0,1,0,1,0,1,1,0))
##' rpfit <- prodlim(Hist(time,status)~1,data=rdat,reverse=TRUE)
##' rsfit <- survfit(Surv(time,1-status)~1,data=rdat,conf.type="plain")
##' ## When there are ties between times at which events are observed
##' ## times at which subjects are right censored, then the convention
##' ## is that events come first. This is not obeyed by the above call to survfit,
##' ## and hence only prodlim delivers the correct reverse Kaplan-Meier:
##' cbind("Wrong:"=rsfit$surv,"Correct:"=rpfit$surv)
##' ##----------------  quantiles of the potential followup time-----
##'
##' G=prodlim(Hist(time,status)~X1,data=dat,reverse=TRUE)
##' quantile(G)
##' 
##' ##-------------------stratified Kaplan-Meier---------------------
##' 
##' stratfit <- prodlim(Surv(time,status)~X1,data=dat)
##' summary(stratfit)
##' summary(stratfit,intervals=TRUE)
##' plot(stratfit)
##' 
##' ##----------continuous covariate: Stone-Beran estimate------------
##' 
##' SB=prodlim(Surv(time,status)~X2,data=dat)
##' summary(SB,newdata=data.frame(X2=c(-0.3,0,.5)))
##' 
##' ##-------------both discrete and continuous covariates------------
##' 
##' prodlim(Surv(time,status)~X2+X1,data=dat)
##' 
##' ##----------------------interval censored data----------------------
##' 
##' dat <- data.frame(L=1:10,R=c(2,3,12,8,9,10,7,12,12,12),status=c(1,1,0,1,1,1,1,0,0,0))
##' with(dat,Hist(time=list(L,R),event=status))
##' 
##' dat$event=1
##' npmle.fitml <- prodlim(Hist(time=list(L,R),event)~1,data=dat)
##' 
##' ##-------------competing risks-------------------
##' 
##' CompRiskFrame <- data.frame(time=1:100,event=rbinom(100,2,.5),X=rbinom(100,1,.5))
##' crFit <- prodlim(Hist(time,event)~X,data=CompRiskFrame)
##' summary(crFit)
##' plot(crFit)
##' summary(crFit,cause=2)
##' plot(crFit,cause=2)
##' 
##' 
##' # Changing the cens.code:
##' dat <- data.frame(time=1:10,status=c(1,2,1,2,5,5,1,1,2,2))
##' fit <- prodlim(Hist(time,status)~1,data=dat)
##' print(fit$model.response)
##' fit <- prodlim(Hist(time,status,cens.code="2")~1,data=dat)
##' print(fit$model.response)
##' plot(fit)
##' plot(fit,cause="5")
##' 
##' 
##' ##------------delayed entry----------------------
##' 
##' ## left-truncated event times with competing risk endpoint 
##' 
##' dat <- data.frame(entry=c(7,3,11,12,11,2,1,7,15,17,3),time=10:20,status=c(1,0,2,2,0,0,1,2,0,2,0))
##' fitd <- prodlim(Hist(time=time,event=status,entry=entry)~1,data=dat)
##' summary(fitd)
##' plot(fitd)
##' 
#' @export
"prodlim" <- function(formula,
                      data = parent.frame(),
                      subset,
                      na.action=NULL,
                      reverse=FALSE,
                      conf.int=.95,
                      bandwidth=NULL,
                      caseweights,
                      discrete.level=3,
                      x=TRUE,
                      maxiter=1000,
                      grid,
                      tol=7,
                      method=c("npmle","one.step","impute.midpoint","impute.right"),
                      exact=TRUE,
                      type){
    # {{{  find the data
    call <- match.call()
    if (!missing(subset)){
        data <- subset(data,subset=subset)
        if (!missing(caseweights)){
            caseweights <- subset(caseweights,subset=subset)
        }
    }
    EHF <- EventHistory.frame(formula=formula,
                              data=data,
                              unspecialsDesign=FALSE,
                              specials=c("Strata","strata","factor", "NN","cluster"),
                              stripSpecials=c("strata","cluster","NN"),
                              stripAlias=list("strata"=c("Strata","factor")),
                              stripArguments=list("strata"=NULL,"NN"=NULL,"cluster"=NULL),
                              specialsDesign=FALSE,
                              check.formula=TRUE)
    event.history <- EHF$event.history
    response <- EHF$event.history
    if (reverse==TRUE){ ## estimation of censoring distribution
        model.type <- 1
    }else{
        model.type <- match(attr(event.history,"model"),c("survival","competing.risks","multi.states"))
    }
    if (missing(type)) type <- switch(model.type,"survival"=ifelse(reverse,"risk","surv"),"risk")
    else {
        type <- tolower(type)
        stopifnot(match(type,c("surv","risk"),nomatch=0)!=0)
    }
    cens.type <- attr(response,"cens.type")

    # {{{ order according to event times
    if (cens.type!="intervalCensored"){
        event.time.order <- order(event.history[,"time"],-event.history[,"status"])
    }
    else{
        event.time.order <- order(event.history[,"L"],-event.history[,"status"])
    }
    # }}}
    # {{{  covariates
    
    covariates <- EHF[-1]
    ##  `factor' and 'Strata' are aliases for `strata'
    strata.pos <- match(c("strata","factor","Strata"),names(covariates),nomatch=0)
    if (sum(strata.pos)>0)
        strata <- do.call("cbind",covariates[strata.pos])
    else
        strata <- NULL
    ##  'NN'
    NN <- covariates[["NN"]]
    xlevels <- attr(strata,"levels")
    ## unspecified
    rest <- covariates$design
    xlevels <- c(attr(strata,"levels"),attr(rest,"levels"))
    if ((is.null(NN)+is.null(strata)+is.null(rest))==3){
        cotype <- 1
    } else{
        unspecified <- NULL
        if (!is.null(rest)){
            discrete.p <- sapply(colnames(rest),function(u){
                x <- rest[,u,drop=TRUE]
                !is.numeric(x) || !length(unique(x))>discrete.level
            })
            if (any(!discrete.p)){ ## continuous covariates
                NN <- if (is.null(NN))
                          rest[,!discrete.p,drop=FALSE]
                      else
                          cbind(NN,rest[,!discrete.p,drop=FALSE])
            }
            if (any(discrete.p)){ ## discrete covariates
                strata <- if (is.null(strata)){
                    rest[,discrete.p,drop=FALSE]
                } else{
                    cbind(strata,rest[,discrete.p,drop=FALSE])
                }
            }
        }
        if (NCOL(NN)>1) stop(paste("Currently we can not compute neighborhoods in",length(colnames(NN)),"continuous dimensions."))
        cotype <- 1 + (!is.null(strata))*1+(!is.null(NN))*2
    }
    ## use unique values as levels
    ## for variables that are not factors
    ## but treated as such
    if (any(found <- (match(colnames(strata),names(xlevels),nomatch=0)==0))){
        uniquelevels <- lapply(colnames(strata)[found],function(x){
            unique(strata[,x])
        })
        names(uniquelevels) <- colnames(strata)[found]
        xlevels <- c(xlevels,uniquelevels)
    }
    ## cotype
    # 1 : no covariates
    # 2 : only strata 
    # 3 : only continuous
    # 4 : strata AND continuous
    # }}}
    # {{{  disjunct strata (discrete covariates)
    if (cotype %in% c(2,4)){
        ## changed 09 Dec 2014 (16:57)-->
        ## S <- do.call("paste", c(data.frame(strata), sep = "\r"))
        S <- interaction(data.frame(strata), sep = ":",drop=TRUE)
        ## <-- changed 09 Dec 2014 (16:57) 
        NS <- length(unique(S))
        ## changed 09 Dec 2014 (16:57) -->
        Sfactor <- factor(S,levels=levels(S),labels=1:NS)
        ## <-- changed 09 Dec 2014 (16:57)
        if (cens.type!="intervalCensored"){
            sorted <- order(Sfactor, response[,"time"],-response[,"status"])
        }
        else{
            sorted <- order(Sfactor, response[,"L"],-response[,"status"])
        }
        Sfactor <- Sfactor[sorted]
    }
    else{
        sorted <- event.time.order
    }
  
    response <- response[sorted,] # sort each stratum
  
    # }}}

    # {{{ caseweights
    if (missing(caseweights)) {
        weighted <- 0
        caseweights <- NULL
    }
    else {
        weighted <- 1
        if(length(caseweights)!=NROW(response))
            stop(paste("The length of caseweights is: ",
                       length(caseweights),
                       "\nthis is not the same as the number of subjects\nwith no missing values, which is ",
                       NROW(response),
                       sep=""))
        ## wrong to order by event.time.order when there are covariates
        ## caseweights <- caseweights[event.time.order]
        ## this fixes bug in versions < 1.5.7 
        caseweights <- caseweights[sorted]
    }
    # }}}

    # {{{  overlapping neighborhoods (continuous covariates)
  
    if (cotype %in% c(3,4)){
        Z <- NN[sorted,,drop=TRUE]
        if (cotype==3){
            nbh <- neighborhood(Z,bandwidth=bandwidth) 
            nbh.list <- list(nbh)
            bandwidth <- nbh$bandwidth
            neighbors <- nbh$neighbors
        }
        else{ # nearest neighbors within each stratum
            nbh.list <- lapply(split(Z,Sfactor),neighborhood,bandwidth=bandwidth)
            bandwidth <- sapply(nbh.list,function(nbh)nbh$bandwidth)
            tabS <- c(0,cumsum(tabulate(Sfactor))[-NS])
            neighbors <- unlist(lapply(1:NS,function(l){ ## incrementing the neighbors by
                nbh.list[[l]]$neighbors+tabS[l]}),use.names=FALSE) ## the size of the previous strata
        }
        response <- response[neighbors,,drop=FALSE]
        if (weighted==TRUE) caseweights <- caseweights[neighbors]
    }
  
    # }}}
    # {{{ delay (left truncation)
    delayed <- attr(event.history,"entry.type")=="leftTruncated"
    if (!delayed) { ## either NULL or ""
        entrytime <- NULL
    }  else {
        entrytime <- response[,"entry"]
        if(!(all(entrytime>=0)))
            stop(paste("Not all entry times in dataset are greater or equal to zero."))
        if (any(entrytime==response[,"time"]))
            warning("For some subjects the entry time is equal to the event time.
  This could be a mistake in the data.")
    }
    # }}}

    # {{{  bound on the number of unique time points over all strata  
    switch(cotype,
           { # type=1
               size.strata <- NROW(response)
               NU <- 1
               if (cens.type!="intervalCensored")
                   N <- length(unique(response[,"time"]))
               else
                   N <- length(unique(response[,"L"]))
               ## if (delayed) N <- N + length(entrytime)
               if (delayed) N <- length(unique(c(entrytime,response[,"time"])))
           },
           {   # type=2
               size.strata <- tabulate(Sfactor)
               N <- NROW(response)
               NU <- length(size.strata)
               if (delayed) N <- 2*N
           },
           {   # type=3
               size.strata <- nbh$size.nbh
               N <- sum(size.strata)
               NU <- nbh$nu
               if (delayed) N <- 2*N
           },
           {   # type=4
               size.strata <- unlist(lapply(nbh.list,function(nbh)nbh$size.nbh),use.names=FALSE)
               N <- sum(size.strata)
               if (delayed) N <- 2*N
               n.unique.strata <- unlist(lapply(nbh.list,function(nbh)nbh$nu),use.names=FALSE)
               NU <- sum(n.unique.strata)
           })
  
  # }}}

  # {{{  characterizing the covariate space
  
    continuous.predictors <- colnames(NN)
    discrete.predictors <- colnames(strata)
    X <- switch(cotype,
    {#type=1
        NULL},
    { #type=2
        X <- data.frame(unique(strata[sorted,,drop=FALSE]))
        ## colnames(X) <- paste("strata",names(strata),sep=".")
        # colnames(X) <- names(strata)
        rownames(X) <- 1:NROW(X)
        X
    },
    { #type=3
        X <- unlist(lapply(nbh.list,function(x)x$values),use.names=FALSE)
        X <- data.frame(X)
        ## colnames(X) <- paste("NN",names(NN),sep=".")
        colnames(X) <- colnames(NN)
        rownames(X) <- 1:NROW(X)                
        X
    },
    { #type=4
        D <- data.frame(unique(strata[sorted,,drop=FALSE]))
        ## colnames(D) <- paste("strata",names(strata),sep=".")
        D <- data.frame(D[rep(1:NS,n.unique.strata),,drop=FALSE])
        C <- data.frame(unlist(lapply(nbh.list,function(x)x$values),use.names=FALSE))
        X <- cbind(D,C)
        ## colnames(X) <- c(paste("strata",names(strata),sep="."),paste("NN",names(NN),sep="."))
        colnames(X) <- c(colnames(strata),colnames(NN))
        rownames(X) <- 1:NROW(X)                
        X
    },
    { #type=5
        X=data.frame(pseudo="pseudo")
        rownames(X) <- 1:NROW(X)                
        X
    })
    if (x==TRUE)
        model.matrix <- switch(cotype,{NULL},strata,NN,cbind(strata,NN))[event.time.order,,drop=FALSE]
    else
        model.matrix <- NULL
    event.history <- event.history[event.time.order,,drop=FALSE]
    # }}}
    # {{{  cluster correlated data need an adjusted variance formula
    clustered <- (length(covariates$cluster)>0)
    if (clustered)
        clustervar <- colnames(covariates$cluster)
    else
        clustervar <- NULL
    if (clustered){
        cluster <- covariates$cluster[sorted,,drop=TRUE]
        if (cotype==1){
            NC <- length(unique(cluster))
            cluster <- factor(cluster,labels=1:NC)
        } else{
            if (cotype==2){
                NC <- unlist(tapply(cluster,Sfactor,function(x){length(unique(x))}))
                cluster <- as.numeric(unlist(tapply(cluster,Sfactor,function(x){
                    factor(x,labels=1:length(unique(x)))})))
            }
        }
    }
    # }}}
    # {{{  find the appropriate C routine
    # with respect to model.type, cens.type, cotype and clustered
    # the following cases are not yet available
    ## if (length(attr(event.history,"entry.type"))>1) stop("Prodlim: Estimation for left-truncated data not yet implemented.")
    if (delayed & weighted>0) stop("Prodlim: Estimation for left-truncated data with caseweights not implemented.")
    if (reverse && cens.type!="rightCensored") stop("Prodlim: Estimation of the censoring distribution works only for right censored data.")
    if (delayed && clustered) stop("Prodlim: Estimation with delayed entry and cluster-correlated observations not yet implemented.")
    if (reverse && clustered) stop("Prodlim: Estimation of censoring distribution with cluster-correlated observations not yet handled.")
    if (cens.type=="intervalCensored" && model.type>=2) stop("Prodlim: Interval censored observations only handled for two-state models")
    ##   if (cens.type=="intervalCensored" && model.type>2) stop("Interval censored observations only handled for two-state and competing risks models")
    if (clustered && model.type>1) stop("Prodlim: Cluster-correlated observations only handled for two-state models")
    if (clustered && cotype %in% c(3,4)) stop("Prodlim: Cluster-correlated observations not yet handled in presence of continuous covariates") #cluster <- cluster[neighbors]
    if (cotype>1 && cens.type=="intervalCensored") stop("Prodlim: Interval censored data and covariate strata not yet handled.")
    if (model.type==1){
        # }}}
        # {{{  two state model
        if (clustered){
            ## right censored clustered
            fit <- .C("prodlimSRC",as.double(response[,"time"]),as.double(response[,"status"]),integer(0),as.double(entrytime),as.double(caseweights),as.integer(cluster),as.integer(N),integer(0),as.integer(NC),as.integer(NU),as.integer(size.strata),time=double(N),nrisk=double(2*N),nevent=double(2*N),ncens=double(2*N),surv=double(N),risk=double(0),hazard=double(N),var.hazard=double(N+N),extra.double=double(4 * max(NC)),max.nc=as.integer(max(NC)),ntimes=integer(1),ntimes.strata=integer(NU),first.strata=integer(NU),reverse=integer(0),model=as.integer(0),independent=as.integer(0),delayed=as.integer(delayed),weighted=as.integer(weighted),PACKAGE="prodlim")
            NT <- fit$ntimes
            Cout <- list("time"=fit$time[1:NT],"n.risk"=matrix(fit$nrisk,ncol=2,byrow=FALSE,dimnames=list(NULL,c("n.risk","cluster.n.risk")))[1:NT,],"n.event"=matrix(fit$nevent,ncol=2,byrow=FALSE,dimnames=list(NULL,c("n.event","cluster.n.event")))[1:NT,],"n.lost"=matrix(fit$ncens,ncol=2,byrow=FALSE,dimnames=list(NULL,c("n.lost","cluster.n.lost")))[1:NT,],"surv"=fit$surv[1:NT],"se.surv"=fit$surv[1:NT]*sqrt(pmax(0,fit$var.hazard[N+(1:NT)])),"naive.se.surv"=fit$surv[1:NT]*sqrt(pmax(0,fit$var.hazard[1:NT])),"hazard"=fit$hazard[1:NT],"first.strata"=fit$first.strata,"size.strata"=fit$ntimes.strata,"model"="survival")
            Cout$maxtime <- max(Cout$time)
        }
        else{
            if (cens.type=="intervalCensored"){
                if (length(method)>1) method <- method[1]
                if (length(grep("impute",method))>0){
                    naiiveMethod <- strsplit(method,"impute.")[[1]][[2]]
                    if (naiiveMethod=="midpoint"){
                        naiveResponse <- data.frame(unclass(response))
                        naiveResponse$imputedTime <- (naiveResponse$L+naiveResponse$R)/2
                        naiveResponse[naiveResponse[,"status"]==0,"imputedTime"] <- naiveResponse[naiveResponse[,"status"]==0,"L"]
                        Cout <- prodlim(Hist(imputedTime,status!=0)~1,data=naiveResponse)
                        return(Cout)
                    }
                }
                else{
                    Cout <- prodlimIcensSurv(response,
                                             grid,
                                             tol=tol,
                                             maxiter=maxiter,
                                             ml=ifelse(method=="one.step",FALSE,TRUE),
                                             exact=exact)
                }
            }
            else{
                ## right censored not clustered
                fit <- .C("prodlimSRC",
                          as.double(response[,"time"]),
                          as.double(response[,"status"]),
                          integer(0),
                          as.double(entrytime),
                          as.double(caseweights),
                          integer(0),
                          as.integer(N),
                          integer(0),
                          integer(0),
                          as.integer(NU),
                          as.integer(size.strata),
                          time=double(N),
                          nrisk=double(N),
                          nevent=double(N),
                          ncens=double(N),
                          surv=double(N),
                          double(0),
                          hazard = double(N),
                          var.hazard=double(N),
                          extra.double=double(0),
                          max.nc=integer(0),
                          ntimes=integer(1),
                          ntimes.strata=integer(NU),
                          first.strata=integer(NU),
                          as.integer(reverse),
                          model=as.integer(0),
                          independent=as.integer(1),
                          delayed=as.integer(delayed),
                          weighted=as.integer(weighted),
                          PACKAGE="prodlim")
                NT <- fit$ntimes
                Cout <- list("time"=fit$time[1:NT],
                             "n.risk"=fit$nrisk[1:NT],
                             "n.event"=fit$nevent[1:NT],
                             "n.lost"=fit$ncens[1:NT],
                             "surv"=fit$surv[1:NT],
                             "se.surv"=fit$surv[1:NT]*sqrt(pmax(0,fit$var.hazard[1:NT])),
                             "hazard"=fit$hazard[1:NT],
                             "first.strata"=fit$first.strata,
                             "size.strata"=fit$ntimes.strata,
                             "model"="survival")
                Cout$maxtime <- max(Cout$time)
            }
        }
    }
    else{
        # }}}
        # {{{ competing.risks model
        if (model.type==2){
            states <- attr(response,"states")
            E <- response[,"event"]-1 # for the c routine
            D <- response[,"status"]
            ## NS <- length(unique(E[D!=0])) # number of different causes
            NS <- length(states)
            fit <- .C("prodlimSRC",
                      as.double(response[,"time"]),
                      as.double(D),
                      as.integer(E),
                      as.double(entrytime),
                      as.double(caseweights),
                      integer(0),
                      as.integer(N),
                      as.integer(NS),
                      integer(0),
                      as.integer(NU),
                      as.integer(size.strata),
                      time=double(N),
                      nrisk=double(N),
                      nevent=double(N * NS),
                      ncens=double(N),
                      surv=double(N),
                      risk=double(N * NS),
                      cause.hazard = double(N * NS),
                      var.hazard=double(N * NS),
                      extra.double=double(4 * NS),
                      max.nc=integer(0),
                      ntimes=integer(1),
                      ntimes.strata=integer(NU),
                      first.strata=integer(NU),
                      reverse=integer(0),
                      model=as.integer(1),
                      independent=as.integer(1),
                      delayed=as.integer(delayed),
                      weighted=as.integer(weighted),
                      PACKAGE="prodlim")
            NT <- fit$ntimes
            # changed Tue Sep 30 12:51:58 CEST 2008 from matrix to list
            # changed Tue 04 Mar 2025 (13:01) to allow events that do not occur in the data
            gatherC <- function(x,ntimes=fit$ntimes,nstrata=NS,names=states){
                out <- split(x[1:(ntimes*nstrata)],rep(1:nstrata,ntimes))
                names(out) <- names
                out
            }
            Cout <- list("time"=fit$time[1:NT],
                         "n.risk"=fit$nrisk[1:NT],
                         "n.event"=gatherC(x = fit$nevent),
                         "n.lost"=fit$ncens[1:NT],
                         "cuminc"=gatherC(fit$risk),
                         "var.cuminc"=gatherC(fit$var.hazard),
                         "se.cuminc"=gatherC(sqrt(pmax(0,fit$var.hazard))),
                         "surv"=fit$surv[1:NT],
                         "cause.hazard"=gatherC(fit$cause.hazard),
                         "first.strata"=fit$first.strata,
                         "size.strata"=fit$ntimes.strata,
                         "model"="competing.risks")
            Cout$maxtime <- max(Cout$time)
        }
        else {
            # multi.state model
            # --------------------------------------------------------------------    
            Cout <- prodlimMulti(response,size.strata,N,NU)
            Cout$maxtime <- max(Cout$time)
        }
    }
    if (conf.int==TRUE) conf.int <- 0.95
    # }}}
    # {{{  confidence intervals
    if (is.numeric(conf.int) && cens.type!="intervalCensored"){
        if (model.type==1){
            if (!(is.null(Cout$se.surv))){
                ## pointwise confidence intervals for survival probability
                zval <- qnorm(1- (1-conf.int)/2, 0,1)
                lower <- pmax(Cout$surv - zval * Cout$se.surv,0)
                # change 05 Mar 2025 (08:45) from [0,1] to [S(t),S(t)] in
                # in case of no variation
                no_variation <- Cout$se.surv==0
                no_variation[is.na(no_variation)] = TRUE
                lower[no_variation] <- Cout$surv[no_variation]
                upper <- pmin(Cout$surv + zval * Cout$se.surv,1)
                upper[no_variation] <- Cout$surv[no_variation]
                Cout <- c(Cout,list(lower=lower,upper=upper))
            }
        }
        else{
            if (is.numeric(conf.int)){
                if (!(0<conf.int && conf.int<1)) conf.int <- 0.95
                ## pointwise confidence intervals for absolute risks (cumulative incidence function)
                # variance for absolute risk (Korn & Dorey (1992), Stat in Med, Vol 11, page 815)
                zval <- qnorm(1- (1-conf.int)/2, 0,1)
                lower <- lapply(1:NS, function(state){
                    pmax(Cout$cuminc[[state]] - zval * Cout$se.cuminc[[state]],0)})
                upper <- lapply(1:NS, function(state){
                    pmin(Cout$cuminc[[state]] + zval * Cout$se.cuminc[[state]],1)})
                names(lower) <- states
                names(upper) <- states
                Cout <- c(Cout,list(lower=lower,upper=upper))
            }
        }
    }
    # }}}
    # {{{ return object of class "prodlim"
    out <- list("call"=call,
                "formula"=formula,
                "model.response"=event.history,
                "originalDataOrder"=order(event.time.order),
                "X"=X,
                "model.matrix"=model.matrix,
                "discrete.predictors"=discrete.predictors,
                "continuous.predictors"=continuous.predictors,
                "xlevels"=xlevels,
                "clustervar"=clustervar,
                "covariate.type"=cotype,
                "cens.type"=cens.type,
                "conf.int"=conf.int,
                "reverse"=reverse,
                "type"=type,
                "na.action"=attr(EHF,"na.action"))
    if (cotype %in% c(3,4)) out <- c(out,list("bandwidth"=bandwidth))
    out <- c(Cout,out)
    class(out) <-  "prodlim"
    return(out)
    # }}}
}

#' Percentage-labeled axis.
#' 
#' Use percentages instead of decimals to label the an axis with a probability
#' scale .
#' 
#' 
#' @param x Side of the axis
#' @param at Positions (decimals) at which to label the axis.
#' @param \dots Given to \code{axis}.
#' @author Thomas Alexander Gerds
#' @seealso \code{\link{plot.prodlim}}
#' @keywords survival
#' @examples
#' 
#'   plot(0,0,xlim=c(0,1),ylim=c(0,1),axes=FALSE)
#'   PercentAxis(1,at=seq(0,1,.25))
#'   PercentAxis(2,at=seq(0,1,.25))
#' 
#' @export
PercentAxis <- function(x,at,...){
  axis(x,at=at,labels=paste(100*at,"%"),...)
}

#' Predicting event probabilities from product limit estimates
#' 
#' Evaluation of estimated survival or event probabilities at given times and
#' covariate constellations.
#' 
#' Predicted (survival) probabilities are returned that can be plotted,
#' summarized and used for inverse of probability of censoring weighting.
#' 
#' @aliases predict.prodlim predictSurv predictAbsrisk predictCuminc
#' @param object A fitted object of class "prodlim".
#' @param times Vector of times at which to return the estimated probabilities (survival or absolute event risks).
#' @param newdata A data frame with the same variable names as those that
#' appear on the right hand side of the 'prodlim' formula.  If there are
#' covariates this argument is required.
#' @param level.chaos Integer specifying the sorting of the output: `0' sort by
#' time and newdata; `1' only by time; `2' no sorting at all
#' @param type Choice between "surv","risk","cuminc","list":
#' 
#' "surv": predict survival probabilities only survival models
#' 
#' "risk"/"cuminc": predict absolute risk, i.e., cumulative incidence function.
#' 
#' "list": find the indices corresponding to times and newdata. See value.
#' 
#' Defaults to "surv" for two-state models and to "risk" for competing risk
#' models.
#' @param mode Only for \code{type=="surv"} and \code{type=="risk"}. Can
#' either be "list" or "matrix". For "matrix" the predicted probabilities will
#' be returned in matrix form.
#' @param bytime Logical. If TRUE and \code{mode=="matrix"} the matrix with
#' predicted probabilities will have a column for each time and a row for each
#' newdata. Only when \code{object$covariate.type>1} and more than one time is
#' given.
#' @param cause Character (other classes are converted with \code{as.character}).
#' The cause for predicting the absolute risk of an event, i.e., the cause-specific cumulative
#' incidence function, in competing risk models. At any time after time zero this is the absolute risk of 
#' an event of type \code{cause} to occur between time zero and \code{times} . 
#' @param \dots Only for compatibility reasons.
#' @return \code{type=="surv"} A list or a matrix with survival probabilities
#' for all times and all newdata.
#' 
#' \code{type=="risk"} or \code{type=="cuminc"} A list or a matrix with cumulative incidences for all
#' times and all newdata.
#' 
#' \code{type=="list"} A list with the following components:
#' 
#' \item{times}{The argument \code{times} carried forward}
#' 
#' \item{predictors}{The relevant part of the argument \code{newdata}.}
#' \item{indices}{ A list with the following components
#' 
#' \code{time}: Where to find values corresponding to the requested times
#' \code{strata}: Where to find values corresponding to the values of the
#' variables in newdata.  Together time and strata show where to find the
#' predicted probabilities.  } \item{dimensions}{ a list with the following
#' components: \code{time} : The length of \code{times} \code{strata} : The
#' number of rows in \code{newdata} \code{names.strata} : Labels for the
#' covariate values.  }
#' @author Thomas Alexander Gerds <tag@@biostat.ku.dk>
#' @seealso \code{\link{predictSurvIndividual}}
#' @keywords survival
#' @examples
#' 
#' 
#' dat <- SimSurv(400)
#' fit <- prodlim(Hist(time,status)~1,data=dat)
#' 
#' ## predict the survival probs at selected times 
#' predict(fit,times=c(3,5,10))
#' 
#' ## NA is returned when the time point is beyond the
#' ## range of definition of the Kaplan-Meier estimator:
#' predict(fit,times=c(-1,0,10,100,1000,10000))
#' 
#' ## when there are strata, newdata is required 
#' ## or neighborhoods (i.e. overlapping strata)
#' mfit <- prodlim(Hist(time,status)~X1+X2,data=dat)
#' predict(mfit,times=c(-1,0,10,100,1000,10000),newdata=dat[18:21,])
#' 
#' ## this can be requested in matrix form
#' predict(mfit,times=c(-1,0,10,100,1000,10000),newdata=dat[18:21,],mode="matrix")
#' 
#' ## and even transposed
#' predict(mfit,times=c(-1,0,10,100,1000,10000),newdata=dat[18:21,],mode="matrix",bytime=TRUE)
#' 
#' @export 
"predict.prodlim" <- function(object,
                              times,
                              newdata,
                              level.chaos=1,
                              type=c("surv","risk","cuminc","list"),
                              mode="list",
                              bytime=FALSE,
                              cause,
                              ...){
    if (!missing(newdata)&&!is.null(newdata)&&inherits(newdata,"data.table"))
        data.table::setDF(newdata)
    if (missing(cause)){
        cause <- attr(object$model.response,"states")
    } else {
        cause <- checkCauses(cause,object)
    }
    if (length(times)==0) stop("Argument 'times' has length 0")
    if (missing(type))
        type <- switch(object$model,"survival"="surv","competing.risks"="risk","list")
    else
        type <- switch(type,"survival"="surv","surv"="surv","risk"="risk","cuminc"="risk","list")
  
  if (type=="surv"){
    predictSurv(object=object,
                times=times,
                newdata=newdata,
                level.chaos=level.chaos,
                mode=mode,
                bytime=bytime)
  }
  else{
    if (type=="risk"){
      predictAbsrisk(object=object,
                    times=times,
                    newdata=newdata,
                    level.chaos=level.chaos,
                    mode=mode,
                    cause=cause)
    }
    else{
      predictList(object=object,
                  times=times,
                  newdata=newdata,
                  level.chaos=level.chaos)
    }
  }
}

"predictList" <- function(object,times,newdata,level.chaos=1){
  if (missing(times)) stop("Argument times is missing.")
  NT <- length(times)
  order.times <- order(times)
  unsorted.times <- times
  times <- times[order.times]
  if (object$cens.type=="intervalCensored")
    jTimes <- object$time[2,]
  else
    jTimes <- object$time

  # no factors
  # --------------------------------------------------------------------
  if (object$covariate.type==1){
      tindex <- sindex(jump.times=jTimes,eval.times=times)
      tindex[times>object$maxtime] <- NA
      if (level.chaos==2)
          indices <- list(time=tindex[order(order.times)],strata=1)
      else
          indices <- list(time=tindex,strata=1)
      dimensions <- list(time=NT,strata=1)
      predictors <- NULL
      names.strata <- NULL
      requested.X <- NULL
  }
  else {
      # conditional on factors
      # --------------------------------------------------------------------
      if (missing(newdata)) stop("Argument newdata is missing.")
      ## deal with tibbles and data.tables
      newdata <- data.table::copy(newdata)      
      newdata <- as.data.frame(newdata)
      NX <- NROW(object$X)
      fit.X <- object$X
      ## strata.vars <- sapply(strsplit(grep("strata",names(fit.X),val=TRUE),"strata."),function(x)x[2])
      ## NN.vars <- sapply(strsplit(grep("NN",names(object$X),val=TRUE),"NN."),function(x)x[2])
      strata.vars <- object$discrete.predictors
      NN.vars <- object$continuous.predictors
      X.formula <- update(formula(object$formula),NULL~.)
      ## delete.response(terms(formula(object$formula)))
      iid <- is.null(object$clustervar)
      if (!iid){
          find.clu <- match(object$clustervar,all.vars(X.formula))
          X.formula <- drop.terms(terms(X.formula),find.clu)
      }
      if (!all(match(all.vars(X.formula),names(newdata),nomatch=FALSE)))
          stop("Arg newdata does not contain all the covariates used for fitting. \n\nfitted variables: ", paste(all.vars(X.formula),collapse=", "),"\nnewdata contains:",ifelse(length(names(newdata))==0," nothing",names(newdata)))
      requested.X <- newdata[,all.vars(X.formula),drop=FALSE]
      NR <- NROW(requested.X)
      requested.names <- extract.name.from.special(names(requested.X))
      names(requested.X) <- requested.names
      check.vars <- match(c(strata.vars,NN.vars),requested.names,nomatch=FALSE)
      if (length(strata.vars)==0){
          requested.strata <- rep(1,NR)
          fit.strata <- rep(1,NX)
          freq.strata <- NX
      } else{
          # strata
          # --------------------------------------------------------------------
          ## changed 09 Dec 2014 (16:44) -->
          ## requested.strata <- do.call("paste",c(requested.X[,strata.vars,drop=FALSE],sep="\r"))
          fit.strata <- interaction(fit.X[,strata.vars,drop=FALSE],sep=":",drop=TRUE)
          requested.strata <- interaction(requested.X[,strata.vars,drop=FALSE],sep=":",drop=TRUE)
          fit.levels <- as.character(unique(fit.strata))
          ## <-- changed 09 Dec 2014 (16:44)          
          ## before version 1.5.1
          ## fit.strata <- factor(do.call("paste",c(fit.X[,strata.vars,drop=FALSE],sep="\r")))
          ## fit.levels <- unique(fit.strata)
          if (!all(unique(requested.strata) %in% (fit.levels))){
              stop(paste("Not all values of newdata strata variables occur in fit:\nrequested:",
                         paste(unique(requested.strata),collapse=","),
                         "\nfitted:",
                         paste(fit.levels,collapse=",")))
          }
          NS <- length(fit.levels)
          ## fit.strata <- factor(fit.strata,levels=unique(fit.strata),labels=1:NS)
          fit.strata <- factor(fit.strata,levels=levels(fit.strata),labels=1:NS)
          requested.strata <- factor(requested.strata,levels=fit.levels,labels=1:NS)
          freq.strata <- cumsum(tabulate(fit.strata))
      }
      # neighborhoods
      # --------------------------------------------------------------------
      switch(length(NN.vars)+1,
      {requested.NN <- NULL
          fit.NN <- NULL
          new.order <- order(requested.strata)},
      {
          requested.NN <- requested.X[,NN.vars,drop=TRUE]
          fit.NN <- fit.X[,NN.vars,drop=TRUE]
          new.order <- order(requested.strata,requested.NN)
      },
      stop("Currently only one continuous covariate allowed."),
      stop("Currently only one continuous covariate allowed."))
      # findex identifies the individual strata neighborhood combination 
      # --------------------------------------------------------------------
      findex <- .C("findex",
                   index=integer(NR),
                   as.integer(as.integer(length(NN.vars)>0)),
                   as.integer(requested.strata[new.order]),
                   as.integer(freq.strata),
                   as.double(requested.NN[new.order]),
                   as.double(fit.NN),
                   as.integer(NR),
                   as.integer(NT),
                   NAOK=FALSE,
                   PACKAGE="prodlim")$index
      if (level.chaos==2) stop("Need to sort the times if there are strata.")
      if (level.chaos==1){# do NOT sort by factors
          predictors <- requested.X
          findex <- findex[order(new.order)]
      }
      else{
          predictors <- requested.X[new.order,,drop=FALSE]
      }
      # pindex identifies the predicted probabilities
      # --------------------------------------------------------------------
      pindex <- .C("pred_index",
                   index=integer(NT*NR),
                   as.double(times),
                   as.double(jTimes),
                   as.integer(object$first.strata[findex]),
                   as.integer(object$size.strata[findex]),
                   as.integer(NR),
                   as.integer(NT),
                   NAOK=FALSE,
                   PACKAGE="prodlim")$index
      pindex[pindex==-1] <- NA
      indices <- list(time=pindex,strata=findex)
      dimensions <- list(time=NT,strata=NR)
      ## bug fix (10 Oct 2013 (10:08)):
      ##        order of names needs to
      ##        obey level.chaos
      names.strata <- apply(do.call("cbind",lapply(names(requested.X),function(n){
          if(is.numeric(requested.X[,n]))
              paste(n,format(requested.X[,n],digits=2),sep="=")
          else 
              paste(n,requested.X[,n],sep="=")})),1,paste,collapse=", ")
      if (level.chaos==0) {names.strata <- names.strata[new.order]}
  }
  if (level.chaos==2) times <- unsorted.times
  else times <- times
  out <- list(times=times,
              predictors=predictors,
              indices=indices,
              dimensions=dimensions,
              strata=requested.X,
              names.strata=names.strata)
  out
}

predictSurv <- function(object,
                        times,
                        newdata,
                        level.chaos=1,
                        mode="list",
                        bytime=FALSE){
    p <- predict(object,
                 newdata=newdata,
                 level.chaos=level.chaos,
                 times=times,type="list")
    NT <- p$dimensions$time
    NR <- p$dimensions$strata
    pindex <- p$indices$time
    if (object$covariate.type==1){
        psurv <- c(1,object$surv)[pindex+1]
    }
    else{
        if (bytime==FALSE){
            psurv <- split(c(1,object$surv)[pindex+1],
                           rep(1:NR,rep(NT,NR)))
            names(psurv) <- p$names.strata
        }
        else{
            psurv <- split(c(1,object$surv)[pindex+1],rep(1:NT,NR))
            names(psurv) <- paste("t",times,sep="=")
        }
    }
    if (mode=="matrix" && NR>1) {
        ## psurv <- cbind(p$strata,do.call("rbind",psurv))
        psurv <- do.call("rbind",psurv)
    } 
    psurv
}

"predictAbsrisk" <- function(object,
                            times,
                            newdata,
                            level.chaos=1,
                            mode="list",
                            cause,
                            ...){
    #  if (object$model!="competing.risks") stop("This object is not a competing.risks model.")
    p <- predict(object,newdata=newdata,level.chaos=level.chaos,times=times,type="list")
    NT <- p$dimensions$time
    NR <- p$dimensions$strata
    pindex <- p$indices$time
    if (object$model=="survival"){
        object$cuminc <- list("1"=1-object$surv)
        cause <- 1
    }
    if (object$model=="competing.risks"){
        if (missing(cause))
            cause <- attributes(object$model.response)$states
        else
            cause <- checkCauses(cause,object)
    }
    out <- lapply(cause,function(thisCause){
        if (NR == 1){
            prisk <- c(0,object$cuminc[[thisCause]])[pindex+1]
            if (mode=="matrix")
                prisk <- matrix(prisk,nrow=1)
        }
        else{
            prisk <- split(c(0,object$cuminc[[thisCause]])[pindex+1],
                             rep(1:NR,rep(NT,NR)))
            names(prisk) <- p$names.strata
            if (mode=="matrix" && NR>1) {
                prisk <- do.call("rbind",prisk)
            }
        }
        prisk})
    if (length(cause)==1){
        out[[1]]
    } else{
        names(out) <- cause
        out
    }
}

#' Plotting a competing-risk-model.
#' 
#' Plotting a competing-risk-model.
#' 
#' 
#' @param stateLabels Labels for the boxes.
#' @param horizontal The orientation of the plot.
#' @param \dots Arguments passed to \code{\link{plot.Hist}}.
#' @author Thomas Alexander Gerds <tag@@biostat.ku.dk>
#' @seealso \code{\link{plotIllnessDeathModel}}, \code{\link{plot.Hist}}
#' @keywords survival
#' @examples
#' 
#' plotCompetingRiskModel()
#' plotCompetingRiskModel(labels=c("a","b"))
#' plotCompetingRiskModel(labels=c("a","b","c"))
#'
#' @export
plotCompetingRiskModel <- function(stateLabels,horizontal=TRUE,...){
  if (missing(stateLabels)) stateLabels <- c("Disease\nfree","Cause1","Cause2")
  nTrans <- length(stateLabels)-1
  if (horizontal==TRUE){
    comprisk.model <- data.frame(time=1:3,status=1:3)
    CRHist <- with(comprisk.model,Hist(time,status,cens.code=2))
    plot(CRHist,stateLabels=stateLabels,...)
  }
  else{
    crHist <- Hist(time=1:nTrans,event=list(from=rep("1",nTrans),to=stateLabels[-1]))
    nrow <- 3
    if (nTrans/2==round(nTrans/2)){
      ncol <- nTrans+1
      midCol <- ceiling(ncol/2)
      columns <- c(midCol,(1:ncol)[-midCol])
      names(columns) <- paste("box",1:length(stateLabels),".column",sep="")
      rows <- c(1,rep(3,nTrans))
      names(rows) <- paste("box",1:length(stateLabels),".row",sep="")
    }
    else{
      ncol <- nTrans
      columns <- c(nTrans+1/2,1:nTrans)
      names(columns) <- paste("box",1:length(stateLabels),".column",sep="")
      rows <- c(1,rep(3,nTrans))
      names(rows) <- paste("box",2:length(stateLabels),".row",sep="")
    }
    do.call("plot.Hist",c(list(x=crHist,stateLabels=stateLabels,nrow=nrow,ncol=ncol,...),columns,rows))
  }
}

##' Extract design matrix and data specials from a model.frame
##'
##' The function separates special terms from the unspecial terms and returns
##' a list of design matrices, one for unspecial terms and one for each special.
##' Some special specials cannot or should not be evaluated in
##' data. E.g., \code{y~a+dummy(x)+strata(v)} the function strata can and should be evaluated,
##' but in order to have \code{model.frame} also evaluate dummy(x) one would be to define
##' and export the function \code{dummy}. Still the term \code{dummy(x)} can be used
##' to identify a special treatment of the variable \code{x}. To deal with this case,
##' one can specify \code{stripSpecials="dummy"}. In addition,  the data
##' should include variables \code{strata(z)} and \code{x}, not \code{dummy(x)}.
##' See examples.
##' The function \code{untangle.specials} of the survival function does a similar job.
##' @title Extract a design matrix and specials from a model.frame 
##' @param terms terms object as obtained either with function \code{terms} or \code{strip.terms}.
##' @param data A data set in which terms are defined.
##' @param xlev a named list of character vectors giving the full set of levels to be assumed for the factors.
##' Can have less elements, in which case the other levels are learned from the \code{data}.
##' @param dropIntercept If TRUE drop intercept term from the design
##' matrix
##' @param maxOrder An error is produced if special variables are
##' involved in interaction terms of order higher than max.order.
##' @param unspecialsDesign A logical value: if \code{TRUE} apply
##' \code{\link{model.matrix}} to unspecial covariates. If
##' \code{FALSE} extract unspecial covariates from data.
##' @param specialsFactor A character vector containing special
##' variables which should be coerced into a single factor. If
##' \code{TRUE} all specials are treated in this way, if \code{FALSE}
##' none of the specials is treated in this way.
##' @param specialsDesign A character vector containing special
##' variables which should be transformed into a design matrix via
##' \code{\link{model.matrix}}.  If \code{TRUE} all specials are
##' treated in this way.
##' @return A list which contains
##'   - the design matrix with the levels of the variables stored in attribute 'levels' 
##'   - separate data.frames which contain the values of the special variables.
##' @seealso \code{\link{EventHistory.frame}} model.frame terms model.matrix .getXlevels  
##' @examples
##' # specials that are evaluated. here ID needs to be defined 
##' set.seed(8)
##' d <- data.frame(y=rnorm(5),x=factor(c("a","b","b","a","c")),z=c(2,2,7,7,7),v=sample(letters)[1:5])
##' d$z <- factor(d$z,levels=c(1:8))
##' ID <- function(x)x
##' f <- formula(y~x+ID(z))
##' t <- terms(f,special="ID",data=d)
##' mda <- model.design(terms(t),data=d,specialsFactor=TRUE)
##' mda$ID
##' mda$design
##' ## 
##' mdb <- model.design(terms(t),data=d,specialsFactor=TRUE,unspecialsDesign=FALSE)
##' mdb$ID
##' mdb$design
##'
##' # set x-levels
##' attr(mdb$ID,"levels")
##' attr(model.design(terms(t),data=d,xlev=list("ID(z)"=1:10),
##'      specialsFactor=TRUE)$ID,"levels")
##' 
##' # special specials (avoid define function SP)
##' f <- formula(y~x+SP(z)+factor(v))
##' t <- terms(f,specials="SP",data=d)
##' st <- strip.terms(t,specials="SP",arguments=NULL)
##' md2a <- model.design(st,data=d,specialsFactor=TRUE,specialsDesign="SP")
##' md2a$SP
##' md2b <- model.design(st,data=d,specialsFactor=TRUE,specialsDesign=FALSE)
##' md2b$SP
##' 
##' # special function with argument
##' f2 <- formula(y~x+treat(z,power=2)+treat(v,power=-1))
##' t2 <- terms(f2,special="treat")
##' st2 <- strip.terms(t2,specials="treat",arguments=list("treat"=list("power")))
##' model.design(st2,data=d,specialsFactor=FALSE)
##' model.design(st2,data=d,specialsFactor=TRUE)
##' model.design(st2,data=d,specialsDesign=TRUE)
##' 
##' library(survival)
##' data(pbc)
##' t3 <- terms(Surv(time,status!=0)~factor(edema)*age+strata(I(log(bili)>1))+strata(sex),
##'             specials=c("strata","cluster"))
##' st3 <- strip.terms(t3,specials=c("strata"),arguments=NULL)
##' md3 <- model.design(terms=st3,data=pbc[1:4,])
##' md3$strata
##' md3$cluster
##' 
##' f4 <- Surv(time,status)~const(age)+prop(factor(edema),power=2)+
##'                         strata(sex,test=0)+prop(bili,power=1)+tp(albumin)
##' t4 <- terms(f4,specials=c("prop","timevar","strata","tp","const"))
##' st4 <- strip.terms(t4,
##'                    specials=c("prop","timevar"),
##'                    unspecials="prop",
##'                    alias.names=list("timevar"="strata","prop"=c("const","tp")),
##'                    arguments=list("prop"=list("power"=0),"timevar"=list("test"=0)))
##' formula(st4)
##' md4 <- model.design(st4,data=pbc[1:4,],specialsDesign=TRUE)
##' md4$prop
##' md4$timevar
##' 
##' @author Thomas A. Gerds <tag@@biostat.ku.dk>
##' @export 
model.design <- function(terms,
                         data,
                         xlev=NULL,
                         dropIntercept=FALSE,
                         maxOrder=1,
                         unspecialsDesign=TRUE,
                         specialsFactor=FALSE,
                         specialsDesign=FALSE){
    # {{{ analyse the terms
    if (missing(terms))
        terms <- attr(data,"terms")
    if (!inherits(terms, "terms")) 
        stop(gettextf("'terms' must be an object of class %s", 
                      dQuote("terms")), domain = NA)
    response <- attr(terms,"response")
    if (response==1){
        terms <- delete.response(terms)
    }
    if (dropIntercept) attr(terms, "intercept") <- 1
    design <- attr(terms,"factors")
    varnames <- rownames(design)
    termsOrder <- attr(terms,"order")
    stripped.position <- attr(terms,"stripped.specials")
    stripped.arguments <- attr(terms,"stripped.arguments")
    stripped.position <- stripped.position[sapply(stripped.position,length)>0]
    stripped <- names(stripped.position)
    specials.position <- attr(terms,"specials")
    specials.position <- specials.position[sapply(specials.position,length)>0]
    specials <- c(names(specials.position),stripped)
    names(specials) <- specials
    if (is.logical(specialsDesign) && (specialsDesign==TRUE)){
        specialsDesign <- specials
    }
    if (is.logical(specialsFactor) && (specialsFactor==TRUE)){
        specialsFactor <- specials
    }
    # }}}
    if (length(specials)>0){
        # {{{ extract information about specials
        specialInfo <- lapply(specials,function(spc){
            if (match(spc,stripped,nomatch=0))
                ## delete.response does not know about stripped terms
                ## so, we need to adjust manually
                pos <- stripped.position[[spc]]-response
            else
                pos <- specials.position[[spc]]
            ## print(pos)
            ## print(class(design))
            ## print(design)
            ## print(NCOL(design))
            if (NCOL(design)>0 && NROW(design)>0){
                ff <- colSums(design[pos,,drop=FALSE])
            }
            else{
                ## stopifnot(pos==1)
                ## there is only one variable
                ff <- 1
            }
            terms <- seq(ff)[ff>0]
            if (any(termsOrder[terms]>maxOrder))
                stop(paste("Special '",spc,
                           "' cannot be used in an interaction of order higher than ",
                           maxOrder,
                           sep=""),call.=FALSE)
            ## extract additional arguments from term.labels 
            spc.vnames <- varnames[pos]
            list(vars=varnames[pos],terms=as.vector(terms))
        })
        specialTerms <- unlist(lapply(specialInfo,function(x)x$terms))
        termLabels <- attr(terms,"term.labels")
        ## only specials
        if (length(termLabels) == length(specialTerms)){
            unspecialTerms <- NULL
        }else{
            unspecialTerms <- drop.terms(terms,specialTerms)
        }
        # }}}
        # {{{ loop over specials
        specialFrames <- lapply(specials,function(sp){
            Info <- specialInfo[[sp]]
            sp.terms <- attr(terms, "term.labels")[Info$terms]
            spTerms <- terms[Info$terms]
            attr(spTerms,"specials") <- NULL
            if (length(xlev)>0){
                spLevels <- xlev[match(sp.terms,names(xlev),nomatch=0)]
                if (length(spLevels)>0)
                    spData <- model.frame(spTerms,data=data,xlev=spLevels)
                else
                    spData <- model.frame(spTerms,data)
            } else{
                spData <- model.frame(spTerms,data)
            }
            spLevels <- .getXlevels(spTerms,spData)
            if (match(sp,stripped,nomatch=0)){
                ## stripped specials may have arguments
                ## in which case we need to know which
                ## columns are affected
                vars <- names(stripped.arguments[[sp]])
                mterms <- lapply(vars,function(v){
                    if (match(v,names(spLevels),nomatch=0))
                        paste(v,spLevels[[v]],sep="")
                    else v})
                names(mterms) <- vars
                stripped.args <- stripped.arguments[[sp]]
                arg.names <- names(stripped.args[[1]])
                arguments.terms <- lapply(arg.names,function(a){
                    unlist(lapply(names(stripped.args),function(var){
                        val <- stripped.args[[var]][[a]]
                        if (length(val)==0) val <- NA
                        tmp <- rep(val,length(mterms[[var]]))
                        names(tmp) <- mterms[[var]]
                        tmp
                    }))})
                names(arguments.terms) <- arg.names
            }
            if (sp %in% specialsDesign){
                spMatrix <- model.matrix(spTerms,data=spData,xlev=spLevels)[,-1,drop=FALSE]
                attr(spMatrix,"levels") <- spLevels
                if (match(sp,stripped,nomatch=0)){
                    attr(spMatrix,"arguments") <- stripped.arguments[[sp]]
                    attr(spMatrix,"arguments.terms") <- arguments.terms
                    attr(spMatrix,"matrix.terms") <- mterms
                }
                spMatrix
            }else{
                if (sp %in% specialsFactor){
                    ## force into a single factor
                    ## in this case ignore any arguments 
                    if (NCOL(spData)>1) {
                        cnames <- colnames(spData)
                        spData <- data.frame(apply(spData,1,paste,collapse=", "))
                        names(spData) <- paste(cnames,collapse=", ")
                    }
                } else{
                    if (match(sp,stripped,nomatch=0)){
                        ## stripped specials may have arguments
                        attr(spData,"arguments") <- stripped.arguments[[sp]]
                        attr(spData,"arguments.terms") <- arguments.terms
                    }
                }
                attr(spData,"levels") <- spLevels
                spData
            }
        })
        # }}}
        # {{{ unspecials
        if (length(unspecialTerms)>0){
            if (length(xlev)>0){
                uLevels <- xlev[match(attr(unspecialTerms,"term.labels"),names(xlev),nomatch=0)]
                if (length(uLevels)>0)
                    X <- model.frame(unspecialTerms,data=data,xlev=uLevels)
                else
                    X <- model.frame(unspecialTerms,data=data)
            } else{
                X <- model.frame(unspecialTerms,data)
            }
            uLevels <- .getXlevels(unspecialTerms,X)
            if (unspecialsDesign==TRUE){
                X <- model.matrix(unspecialTerms,data,xlev=uLevels)
                if (dropIntercept) X <- X[,-1,drop=FALSE]
            }
        } else {
            X <- NULL
            uLevels <- NULL
        }
        attr(X,"levels") <- uLevels
        c(list(design=X),specialFrames)
        # }}}
    }else{
        # {{{ no specials
        if (length(xlev)>0){
            levels <- xlev[match(attr(terms,"term.labels"),names(xlev),nomatch=0)]
            if (length(levels)>0)
                X <- model.frame(terms,data=data,xlev=uLevels)
            else
                X <- model.frame(terms,data)
        } else{
            X <- model.frame(terms,data)
        }
        levels <- .getXlevels(terms,X)
        if (unspecialsDesign==TRUE){
            X <- model.matrix(terms,data,xlev=levels)
            if (dropIntercept) X <- X[,-1,drop=FALSE]
        }
        attr(X,"levels") <- levels        
        list(design=X)
        # }}}
    }
}

##' Extract event history data and design matrix including specials from call
##'
##' Obtain a list with the data used for event history regression analysis. This
##' function cannot be used directly on the user level but inside a function
##' to prepare data for survival analysis. 
##' @title Event history frame
##' @param formula Formula whose left hand side specifies the event
##' history, i.e., either via Surv() or Hist().
##' @param data Data frame in which the formula is interpreted
##' @param unspecialsDesign Passed as is to
##' \code{\link{model.design}}.
##' @param specials Character vector of special function names.
##' Usually the body of the special functions is function(x)x but
##' e.g., \code{\link[survival]{strata}} from the survival package does treat
##' the values
##' @param specialsFactor Passed as is to \code{\link{model.design}}.
##' @param specialsDesign Passed as is to \code{\link{model.design}}
##' @param stripSpecials Passed as \code{specials} to
##' \code{\link{strip.terms}}
##' @param stripArguments Passed as \code{arguments} to
##' \code{\link{strip.terms}}
##' @param stripAlias Passed as \code{alias.names} to
##' \code{\link{strip.terms}}
##' @param stripUnspecials Passed as \code{unspecials} to
##' \code{\link{strip.terms}}
##' @param dropIntercept Passed as is to \code{\link{model.design}}
##' @param check.formula If TRUE check if formula is a Surv or Hist
##' thing.
##' @param response If FALSE do not evaluate the left hand side of the formula and do not return the response (event.history).
##' @param na.action Decide what to do with missing values. 
##' @return A list which contains
##' - the event.history (see \code{\link{Hist}})
##' - the design matrix (see \code{\link{model.design}})
##' - one entry for each special (see \code{\link{model.design}})
##' @seealso model.frame model.design Hist
##' @examples
##' 
##' ## Here are some data with an event time and no competing risks
##' ## and two covariates X1 and X2.
##' ## Suppose we want to declare that variable X1 is treated differently
##' ## than variable X2. For example, X1 could be a cluster variable, or
##' ## X1 should have a proportional effect on the outcome.
##' dsurv <- data.frame(time=1:7,
##'                     status=c(0,1,1,0,0,0,1),
##'                     X2=c(2.24,3.22,9.59,4.4,3.54,6.81,5.05),
##'                     X3=c(1,1,1,1,0,0,1),
##'                     X4=c(44.69,37.41,68.54,38.85,35.9,27.02,41.84),
##'                     X1=factor(c("a","b","a","c","c","a","b"),
##'                         levels=c("c","a","b")))
##' ## We pass a formula and the data
##' e <- EventHistory.frame(Hist(time,status)~prop(X1)+X2+cluster(X3)+X4,
##'                         data=dsurv,
##'                         specials=c("prop","cluster"),
##'                         stripSpecials=c("prop","cluster"))
##' names(e)
##' ## The first element is the event.history which is result of the left hand
##' ## side of the formula:
##' e$event.history
##' ## same as
##' with(dsurv,Hist(time,status))
##' ## to see the structure do 
##' colnames(e$event.history)
##' unclass(e$event.history)
##' ## in case of competing risks there will be an additional column called event,
##' ## see help(Hist) for more details
##' 
##' ## The other elements are the design, i.e., model.matrix for the non-special covariates
##' e$design
##' ## and a data.frame for the special covariates
##' e$prop
##' ## The special covariates can be returned as a model.matrix
##' e2 <- EventHistory.frame(Hist(time,status)~prop(X1)+X2+cluster(X3)+X4,
##'                          data=dsurv,
##'                          specials=c("prop","cluster"),
##'                          stripSpecials=c("prop","cluster"),
##'                          specialsDesign=TRUE)
##' e2$prop
##' ## and the non-special covariates can be returned as a data.frame
##' e3 <- EventHistory.frame(Hist(time,status)~prop(X1)+X2+cluster(X3)+X4,
##'                          data=dsurv,
##'                          specials=c("prop","cluster"),
##'                          stripSpecials=c("prop","cluster"),
##'                          specialsDesign=TRUE,
##'                          unspecialsDesign=FALSE)
##' e3$design
##' 
##' ## the general idea is that the function is used to parse the combination of
##' ## formula and data inside another function. Here is an example with
##' ## competing risks
##' SampleRegression <- function(formula,data=parent.frame()){
##'     thecall <- match.call()
##'     ehf <- EventHistory.frame(formula=formula,
##'                               data=data,
##'                               stripSpecials=c("prop","cluster","timevar"),
##'                               specials=c("prop","timevar","cluster"))
##'     time <- ehf$event.history[,"time"]
##'     status <- ehf$event.history[,"status"]
##'     ## event as a factor
##'     if (attr(ehf$event.history,"model")=="competing.risks"){
##'         event <- ehf$event.history[,"event"]
##'         Event <- getEvent(ehf$event.history)
##'         list(response=data.frame(time,status,event,Event),X=ehf[-1])
##'     }
##'     else{ # no competing risks
##'         list(response=data.frame(time,status),X=ehf[-1])
##'     }
##' }
##' dsurv$outcome <- c("cause1","0","cause2","cause1","cause2","cause2","0")
##' SampleRegression(Hist(time,outcome)~prop(X1)+X2+cluster(X3)+X4,dsurv)
##' ehf <- EventHistory.frame(formula=Hist(time,outcome)~prop(X1)+X2+cluster(X3)+X4,
##'                           data=dsurv,
##'                           stripSpecials=c("prop","cluster","timevar"),
##'                           specials=c("prop","timevar","cluster"))
##' 
##' ## let's test if the parsing works
##' form1 <- Hist(time,outcome!="0")~prop(X1)+X2+cluster(X3)+X4
##' form2 <- Hist(time,outcome)~prop(X1)+cluster(X3)+X4
##' ff <- list(form1,form2)
##' lapply(ff,function(f){SampleRegression(f,dsurv)})
##' 
##' 
##' ## here is what the riskRegression package uses to
##' ## distinguish between covariates with
##' ## time-proportional effects and covariates with
##' ## time-varying effects:
##' \dontrun{
##' library(riskRegression)
##' data(Melanoma)
##' f <- Hist(time,status)~prop(thick)+strata(sex)+age+prop(ulcer,power=1)+timevar(invasion,test=1)
##' ## here the unspecial terms, i.e., the term age is treated as prop
##' ## also, strata is an alias for timvar
##' 
##' EHF <- prodlim::EventHistory.frame(formula,
##'                                    Melanoma[1:10],
##'                                    specials=c("timevar","strata","prop","const","tp"),
##'                                    stripSpecials=c("timevar","prop"),
##'                                    stripArguments=list("prop"=list("power"=0),
##'                                        "timevar"=list("test"=0)),
##'                                    stripAlias=list("timevar"=c("strata"),
##'                                        "prop"=c("tp","const")),
##'                                    stripUnspecials="prop",
##'                                    specialsDesign=TRUE,
##'                                    dropIntercept=TRUE)       
##' EHF$prop
##' EHF$timevar
##' }
##' @export 
##' @author Thomas A. Gerds <tag@@biostat.ku.dk>
EventHistory.frame <- function(formula,
                               data,
                               unspecialsDesign=TRUE,
                               specials,
                               specialsFactor=TRUE,
                               specialsDesign=FALSE,
                               stripSpecials=NULL,
                               stripArguments=NULL,
                               stripAlias=NULL,
                               stripUnspecials=NULL,
                               dropIntercept=TRUE,
                               check.formula=TRUE,
                               response=TRUE,
                               na.action=options()$na.action){
    # {{{  check if formula is a proper formula
    formula.names <- try(all.names(formula),silent=TRUE)
    if (isTRUE(response)){
        is_event_history <- ((any(match(c("survival::Surv","Surv","prodlim::Hist","Hist"),formula.names,nomatch=0)))
            || inherits(data[[formula.names[[2]]]],"Surv")
            || inherits(data[[formula.names[[2]]]],"Hist"))
    }else{
        is_event_history <- FALSE
    }
    if (isTRUE(response) && isTRUE(check.formula)){
        if (!(formula.names[1]=="~")
            ||
            (match("$",formula.names,nomatch=0)+match("[",formula.names,nomatch=0)>0)){
            stop("Invalid specification of formula. Perhaps forgotten right hand side?\nNote that any subsetting, i.e., data$var or data[,\"var\"], is not supported here.\nSpecify response=FALSE and check.formula=FALSE to avoid this error when using the design functionality without a response.")}
        else{
                if (!is_event_history) {
                    stop("formula is NOT a proper survival formula,\nwhich must have a `Surv' or `Hist' object as response.")
                }
            }
    }
    # }}}
    # {{{call model.frame
    ## data argument is used to resolve '.' see help(terms.formula)
    if (isTRUE(response)){
        Terms <- terms(x=formula,specials=specials,data=data)
    }else{
        Terms <- delete.response(terms(x=formula,specials=specials,data=data))
    }
    if (!is.null(stripSpecials)){
        ## Terms <- terms(x=formula, specials=specials)
        if (length(attr(Terms,"term.labels"))>0)
            Terms <- strip.terms(Terms,
                                 specials=stripSpecials,
                                 arguments=stripArguments,
                                 alias.names=stripAlias,
                                 unspecials=stripUnspecials)
    }
    # }}}
    # {{{ get all variables and remove missing values
    ## use the stripped formula because, otherwise
    ## it may be hard to distinguis the variable names from functions, e.g.,
    ## riskRegression::FGR may use cov2(var,tf=qfun) where qfun is a function
    mm <- do.call(na.action,list(object = get_all_vars(formula(Terms),data=data)))
    if (NROW(mm) == 0) stop("No (non-missing) observations")
    # }}}

    if (isTRUE(response) && attr(Terms,"response")!=0){
        if (is_event_history){
            event.history <- model.response(model.frame(update(formula,".~1"),data=mm,na.action=na.action))
        }else{
            event.history <- model.frame(update(formula,".~1"), data=mm,na.action=na.action)
        }
        # Fix for those who use `Surv' instead of `Hist' 
        if (is_event_history){
            if (inherits(event.history,"Surv") || inherits(data[[formula.names[[2]]]],"Surv")){
                event.history = as.matrix(event.history)
                class(event.history) = "Hist"
                attr(event.history,"model") <- "survival"
                attr(event.history,"cens.code") <- 0
                attr(event.history,"cens.type") <- "rightCensored"
                attr(event.history,"entry.type") <- ifelse(ncol(event.history)==2,"","leftTruncated")
                if (attr(event.history,"entry.type")=="leftTruncated")
                    colnames(event.history) <- c("entry","time","status")
            }
        }
    }else {
        event.history <- NULL
    }
    # {{{ design
    design <- model.design(Terms,
                           data=mm,
                           maxOrder=1,
                           dropIntercept=dropIntercept,
                           unspecialsDesign=unspecialsDesign,
                           specialsFactor=specialsFactor,
                           specialsDesign=specialsDesign)
    # }}}
    out <- c(list(event.history=event.history),
             design[sapply(design,length)>0])
    if (!is_event_history){
        names(out)[1] <- "response"
    }
    attr(out,"Terms") <- Terms
    attr(out,"na.action") <- attr(mm,"na.action")
    class(out) <- "EventHistory.frame"
    out
}
##' @export 
as.data.frame.EventHistory.frame <- function(x,...){
    Y <- data.frame(unclass(x$event.history))
    X <- do.call("cbind",x[-1])
    cbind(Y,X)
}

# {{{ Header
#' Plotting event probabilities over time
#' 
#' Function to plot survival probabilities or absolute risks (cumulative incidence function) against time.
#' 
#' From version 1.1.3 on the arguments legend.args, atrisk.args, confint.args
#' are obsolete and only available for backward compatibility. Instead
#' arguments for the invoked functions \code{atRisk}, \code{legend},
#' \code{confInt}, \code{markTime}, \code{axis} are simply specified as
#' \code{atrisk.cex=2}. The specification is not case sensitive, thus
#' \code{atRisk.cex=2} or \code{atRISK.cex=2} will have the same effect.  The
#' function \code{axis} is called twice, and arguments of the form
#' \code{axis1.labels}, \code{axis1.at} are used for the time axis whereas
#' \code{axis2.pos}, \code{axis1.labels}, etc. are used for the y-axis.
#' 
#' These arguments are processed via \code{\dots{}} of \code{plot.prodlim} and
#' inside by using the function \code{SmartControl}.  Documentation of these
#' arguments can be found in the help pages of the corresponding functions.
#' 
#' @aliases plot.prodlim lines.prodlim
#' @param x an object of class `prodlim' as returned by the
#'     \code{prodlim} function.
#' @param type Either \code{"surv"} or \code{"risk"} AKA \code{"cuminc"}. Controls what
#' part of the object is plotted.  Defaults to \code{object$type}.
#' @param cause For competing risk models. Character (other classes are converted with \code{as.character}).
#' The argument \code{cause} determines the event of interest. Currently one cause is allowed at a time, but you can
#' call the function again with \code{add=TRUE} to add the lines of the other
#' causes. Also, if \code{cause="stacked"} is specified the absolute risks of all causes are stacked.
#' @param select Select which lines to plot. This can be used when
#'     there are many strata or many competing risks to select a
#'     subset of the lines.  However, a more clean way to select
#'     covariate strata is to use the argument \code{newdata}. Another
#'     application is when there are several competing risks and the 
#'      stacked plot (\code{cause="stacked"}) should only show a selected subset
#'     of the available causes.
#' @param newdata a data frame containing covariate strata for which
#'     to show curves. When omitted element \code{X} of object
#'     \code{x} is used.
#' @param add if \code{TRUE} curves are added to an existing plot.
#' @param col color for curves. Default is \code{1:number(curves)}
#' @param lty line type for curves. Default is 1.
#' @param lwd line width for all curves. Default is 3.
#' @param ylim limits of the y-axis
#' @param xlim limits of the x-axis
#' @param ylab label for the y-axis
#' @param xlab label for the x-axis
#' @param num.digits Number of digits when rounding off numerical values for legend and at-risk tables.
#' @param timeconverter The following options are supported:
#'  "days2years" (conversion factor: 1/365.25)
#'  "months2years" (conversion factor: 1/12)
#'  "days2months" (conversion factor 1/30.4368499)
#'  "years2days" (conversion factor 365.25)
#'  "years2months" (conversion factor 12)
#'   "months2days" (conversion factor 30.4368499)
#' @param legend if TRUE a legend is plotted by calling the function
#'     legend.  Optional arguments of the function \code{legend} can
#'     be given in the form \code{legend.x=val} where x is the name of
#'     the argument and val the desired value. See also Details.
#' @param short.labels Logical. When \code{FALSE} construct labels as cause=1, var1=v1, var2=v2 else as 1, v1, v2.
#' @param logrank If TRUE, the logrank p-value will be extracted from
#'     a call to \code{survdiff} and added to the legend. This works
#'     only for survival models, i.e. Kaplan-Meier with discrete
#'     predictors.
#' @param marktime if TRUE the curves are tick-marked at right
#'     censoring times by invoking the function
#'     \code{markTime}. Optional arguments of the function
#'     \code{markTime} can be given in the form \code{confint.x=val}
#'     as with legend. See also Details.
#' @param confint Logical. If \code{TRUE} pointwise confidence intervals are plotted
#'     by invoking the function \code{confInt}. Optional arguments of
#'     the function \code{confInt} can be given in the form
#'     \code{confint.x=val} as with legend.  See also Details.
#' @param automar If TRUE the function trys to find suitable values
#'     for the figure margins around the main plotting region.
#' @param atrisk if TRUE display numbers of subjects at risk by
#'     invoking the function \code{atRisk}. Optional arguments of the
#'     function \code{atRisk} can be given in the form
#'     \code{atrisk.x=val} as with legend. See also Details.
#' @param timeOrigin Start of the time axis
#' @param axes If true axes are drawn. See details.
#' @param background If \code{TRUE} the background color and grid
#'     color can be controlled using smart arguments SmartControl,
#'     such as background.bg="yellow" or
#'     background.bg=c("gray66","gray88").  The following defaults are
#'     passed to \code{background} by \code{plot.prodlim}:
#'     horizontal=seq(0,1,.25), vertical=NULL, bg="gray77",
#'     fg="white".  See \code{background} for all arguments, and the
#'     examples below.
#' @param percent If true the y-axis is labeled in percent.
#' @param minAtrisk Integer. Show the curve only until the number
#'     at-risk is at least \code{minAtrisk}
#' @param limit When newdata is not specified and the number of lines
#'     in element \code{X} of object \code{x} exceeds limits, only the
#'     results for covariate constellations of the first, the middle
#'     and the last row in \code{X} are shown. Otherwise all lines of
#'     \code{X} are shown.
#' @param ... Parameters that are filtered by
#'     \code{\link{SmartControl}} and then passed to the functions
#'     \code{\link{plot}}, \code{\link{legend}}, \code{\link{axis}},
#'     \code{\link{atRisk}}, \code{\link{confInt}},
#'     \code{\link{markTime}}, \code{\link{backGround}}
#' @return The (invisible) object.
#' @author Thomas Alexander Gerds <tag@@biostat.ku.dk>
#' @seealso \code{\link{plot}}, \code{\link{legend}},
#'     \code{\link{axis}},
#'     \code{\link{prodlim}},\code{\link{plot.Hist}},\code{\link{summary.prodlim}},
#'     \code{\link{neighborhood}}, \code{\link{atRisk}},
#'     \code{\link{confInt}}, \code{\link{markTime}},
#'     \code{\link{backGround}}
#' @keywords survival
##' @examples
##' ## simulate right censored data from a two state model 
##' set.seed(100)
##' dat <- SimSurv(100)
##' # with(dat,plot(Hist(time,status)))
##' 
##' ### marginal Kaplan-Meier estimator
##' kmfit <- prodlim(Hist(time, status) ~ 1, data = dat)
##' plot(kmfit)
##' plot(kmfit,atrisk.show.censored=1L,atrisk.at=seq(0,12,3))
##' plot(kmfit,timeconverter="years2months")
##' 
##' # change time range
##' plot(kmfit,xlim=c(0,4))
##' 
##' # change scale of y-axis
##' plot(kmfit,percent=FALSE)
##' 
##' # mortality instead of survival
##' plot(kmfit,type="risk")
##' 
##' # change axis label and position of ticks
##' plot(kmfit,
##'      xlim=c(0,10),
##'      axis1.at=seq(0,10,1),
##'      axis1.labels=0:10,
##'      xlab="Years",
##'      axis2.las=2,
##'      atrisk.at=seq(0,10,2.5),
##'      atrisk.title="")
##' 
##' # change background color
##' plot(kmfit,
##'      xlim=c(0,10),
##'      confint.citype="shadow",
##'      col=1,
##'      axis1.at=0:10,
##'      axis1.labels=0:10,
##'      xlab="Years",
##'      axis2.las=2,
##'      atrisk.at=seq(0,10,2.5),
##'      atrisk.title="",
##'      background=TRUE,
##'      background.fg="white",
##'      background.horizontal=seq(0,1,.25/2),
##'      background.vertical=seq(0,10,2.5),
##'      background.bg=c("gray88"))
##' 
##' # change type of confidence limits
##' plot(kmfit,
##'      xlim=c(0,10),
##'      confint.citype="dots",
##'      col=4,
##'      background=TRUE,
##'      background.bg=c("white","gray88"),
##'      background.fg="gray77",
##'      background.horizontal=seq(0,1,.25/2),
##'      background.vertical=seq(0,10,2))
##' 
##' 
##' ### Kaplan-Meier in discrete strata
##' kmfitX <- prodlim(Hist(time, status) ~ X1, data = dat)
##' plot(kmfitX,atrisk.show.censored=1L)
##' # move legend
##' plot(kmfitX,legend.x="bottomleft",atRisk.cex=1.3,
##'      atrisk.title="No. subjects")
##' 
##' ## Control the order of strata
##' ## since version 1.5.1 prodlim does  obey the order of
##' ## factor levels
##' dat$group <- factor(cut(dat$X2,c(-Inf,0,0.5,Inf)),
##'                     labels=c("High","Intermediate","Low"))
##' kmfitG <- prodlim(Hist(time, status) ~ group, data = dat)
##' plot(kmfitG)
##' 
##' ## relevel 
##' dat$group2 <- factor(cut(dat$X2,c(-Inf,0,0.5,Inf)),
##'                      levels=c("(0.5, Inf]","(0,0.5]","(-Inf,0]"),
##'                      labels=c("Low","Intermediate","High"))
##' kmfitG2 <- prodlim(Hist(time, status) ~ group2, data = dat)
##' plot(kmfitG2)
##' 
##' # add log-rank test to legend
##' plot(kmfitX,
##'      atRisk.cex=1.3,
##'      logrank=TRUE,
##'      legend.x="topright",
##'      atrisk.title="at-risk")
##' 
##' # change atrisk labels
##' plot(kmfitX,
##'      legend.x="bottomleft",
##'      atrisk.title="Patients",
##'      atrisk.cex=0.9,
##'      atrisk.labels=c("X1=0","X1=1"))
##' 
##' # multiple categorical factors
##' 
##' kmfitXG <- prodlim(Hist(time,status)~X1+group2,data=dat)
##' plot(kmfitXG,select=1:2)
##' 
##' ### Kaplan-Meier in continuous strata
##' kmfitX2 <- prodlim(Hist(time, status) ~ X2, data = dat)
##' plot(kmfitX2,xlim=c(0,10))
##' 
##' # specify values of X2 for which to show the curves 
##' plot(kmfitX2,xlim=c(0,10),newdata=data.frame(X2=c(-1.8,0,1.2)))
##' 
##' ### Cluster-correlated data
##' library(survival)
##' cdat <- cbind(SimSurv(20),patnr=sample(1:5,size=20,replace=TRUE))
##' kmfitC <- prodlim(Hist(time, status) ~ cluster(patnr), data = cdat)
##' plot(kmfitC)
##' plot(kmfitC,atrisk.labels=c("Units","Patients"))
##' 
##' kmfitC2 <- prodlim(Hist(time, status) ~ X1+cluster(patnr), data = cdat)
##' plot(kmfitC2)
##' plot(kmfitC2,atrisk.labels=c("Teeth","Patients","Teeth","Patients"),
##'      atrisk.col=c(1,1,2,2))
##' 
##' 
##' ### Cluster-correlated data with strata
##' n = 50
##' foo = runif(n)
##' bar = rexp(n)
##' baz = rexp(n,1/2)
##' d = stack(data.frame(foo,bar,baz))
##' d$cl = sample(10, 3*n, replace=TRUE)
##' fit = prodlim(Surv(values) ~ ind + cluster(cl), data=d)
##' plot(fit)
##' 
##' 
##' ## simulate right censored data from a competing risk model 
##' datCR <- SimCompRisk(100)
##' with(datCR,plot(Hist(time,event)))
##' 
##' ### marginal Aalen-Johansen estimator
##' ajfit <- prodlim(Hist(time, event) ~ 1, data = datCR)
##' plot(ajfit) # same as plot(ajfit,cause=1)
##' plot(ajfit,atrisk.show.censored=1L)
##' 
##' # cause 2
##' plot(ajfit,cause=2)
##' 
##' # both in one
##' plot(ajfit,cause=1)
##' plot(ajfit,cause=2,add=TRUE,col=2)
##' 
##' ### stacked plot
##' 
##' plot(ajfit,cause="stacked",select=2)
##' 
##' ### stratified Aalen-Johansen estimator
##' ajfitX1 <- prodlim(Hist(time, event) ~ X1, data = datCR)
##' plot(ajfitX1)
##' 
##' ## add total number at-risk to a stratified curve
##' ttt = 1:10
##' plot(ajfitX1,atrisk.at=ttt,col=2:3)
##' plot(ajfit,add=TRUE,col=1)
##' atRisk(ajfit,newdata=datCR,col=1,times=ttt,line=3,labels="Total")
##' 
##'
##' ## stratified Aalen-Johansen estimator in nearest neighborhoods
##' ## of a continuous variable
##' ajfitX <- prodlim(Hist(time, event) ~ X1+X2, data = datCR)
##' plot(ajfitX,newdata=data.frame(X1=c(1,1,0),X2=c(4,10,10)))
##' plot(ajfitX,newdata=data.frame(X1=c(1,1,0),X2=c(4,10,10)),cause=2)
##' 
##' ## stacked plot
##' 
##' plot(ajfitX,
##'      newdata=data.frame(X1=0,X2=0.1),
##'      cause="stacked",
##'      legend.title="X1=0,X2=0.1",
##'      legend.legend=paste("cause:",getStates(ajfitX$model.response)),
##'      plot.main="Subject specific stacked plot")
##'  
#' @export plot.prodlim
#' @export 
plot.prodlim <- function(x,
                         type,
                         cause,
                         select,
                         newdata,
                         add = FALSE,
                         col,
                         lty,
                         lwd,
                         ylim,
                         xlim,
                         ylab,
                         xlab="Time",
                         num.digits=2,
                         timeconverter,
                         legend=TRUE,
                         short.labels=TRUE,
                         logrank=FALSE,
                         marktime=FALSE,
                         confint=TRUE,
                         automar,
                         atrisk=ifelse(add,FALSE,TRUE),
                         timeOrigin=0,
                         axes=TRUE,
                         background=TRUE,
                         percent=TRUE,
                         minAtrisk=0,
                         limit=10,
                         ...){
    # }}}
    # {{{ data.table NULL's
    risk <- surv <- lower.risk <- lower <- upper <- cuminc <- surv <- time <- NULL
    # }}}
    # {{{  backward compatibility
    ##   args=match.call(expand=TRUE)
    ##   args[[1]]=list
    allArgs <- match.call()
    if (missing(type)){
        type=allArgs[[match("what",names(allArgs))]]
    }
    # }}}
    # {{{  extracting a list of lines to draw

    cens.type <- x$cens.type    # uncensored, right or interval censored
    if (cens.type=="intervalCensored") {
        confint <- FALSE
        atrisk <- FALSE
    }
    model <- x$model                 # survival, competing risks or multi-state
    clusterp <- !is.null(x$clustervar)
    if (clusterp ==TRUE && logrank==TRUE){
        warning("Argument 'logrank' internally set to FALSE due to cluster variable.")
        logrank=FALSE
    }
    if (missing(type)||is.null(type)){
        type <- x$type
    }
    else
        type <- match.arg(type,c("surv","risk","cuminc","hazard"))
    if (type=="cuminc") type = "risk"
    if (model=="competing.risks" && type=="surv")
        stop("To plot the event-free survival curve, please fit a suitable model: prodlim(Hist(time,status!=0)~....")
  
    if (cens.type=="intervalCensored")
        plot.times <- sort(unique(x$time[2,]))
    else{
        plot.times <- sort(unique(x$time))
        if (plot.times[1]>timeOrigin) plot.times <- c(timeOrigin,plot.times)
        else plot.times <- c(timeOrigin,plot.times[plot.times>timeOrigin])
    }
    if (length(x$clustervar)>0)
        nRisk <- x$n.risk[,1]
    else
        nRisk <- x$n.risk
    if (minAtrisk>0 && any(nRisk<=minAtrisk)){
        if (all(nRisk<=minAtrisk)){
            return(plot(0,0,type="n",xlim=c(min(plot.times), max(plot.times)),ylim=c(0, 1),axes=FALSE))
        }
        criticalTime <- min(x$time[nRisk<=minAtrisk])
        plot.times <- plot.times[plot.times<criticalTime]
        ## if (sum(nEvent[nRisk>minAtrisk])<=1)
    }
    if (missing(newdata)) {
        newdata <- x$X
        if (NROW(newdata)>limit)
            newdata <- newdata[c(1,round(median(1:NROW(newdata))),NROW(newdata)),,drop=FALSE]          
    }
    ## restrict plot.times to xlim
    if (!missing(xlim)){
        if (xlim[1]>plot.times[1]) plot.times <- plot.times[plot.times>=xlim[1]]
        if (xlim[2]<plot.times[length(plot.times)]) plot.times <- unique(c(plot.times[plot.times<=xlim[2]],xlim[2]))
    }
    ## if (missing(newdata) && NROW(newdata)>limit)
    ## newdata <- newdata[c(1,round(median(1:NROW(newdata))),NROW(newdata)),,drop=FALSE]
    if (missing(cause)){
        cause <- attr(x$model.response,which="states")[1]
        stacked <- FALSE
    } else{
        stacked <- cause[1]=="stacked"
        if (stacked) ## all causes
            cause <- attributes(x$model.response)$states
        else
            cause <- checkCauses(cause,x)
    }
    if (stacked){
        confint <- FALSE
        if (model!="competing.risks") stop("Stacked plot works only for competing risks models.")
        ## if (NROW(newdata)>1) stop("Stacked plot works only for one covariate stratum.")
    }else{
        if (length(cause)==0){
            cause <- attributes(x$model.response)$states[[1]]
        }
        ## if (length(cause)>1){
        ## warning("Currently only the cumulative incidence of a single cause can be plotted in one go. Use argument add=TRUE to add the lines of the other causes. For now I use the first cause")
        ## cause <- cause[1]
        ## }
    }
    startValue=ifelse(type=="surv",1,0)
    if (type=="hazard" && model!="survival")
        stats=list(c("cause.hazard",0))
    else
        stats=list(c(type,startValue))
    if (model=="survival" && type=="risk") {
        startValue=1
        stats=list(c("surv",startValue))
    }
    if (confint==TRUE)
        stats=c(stats,list(c("lower",startValue),c("upper",startValue)))
    if (x$cens.type=="intervalCensored"){
        stop("FIXME: There is no plot method implemented for intervalCensored data.")
    }
    if  (model=="competing.risks"){
        sumX <- lifeTab(x,
                        times=plot.times,
                        cause=cause,
                        newdata=newdata,
                        stats=stats,
                        percent=FALSE,format="dt")
    } else{
        sumX <- lifeTab(x,
                        times=plot.times,
                        newdata=newdata,
                        stats=stats,
                        percent=FALSE,format="dt")
    }
    if (model=="survival" && type=="risk"){
        sumX[,risk:=1-surv]
        if (confint[[1]]==TRUE){
            sumX[,lower.risk:=1-upper]
            sumX[,upper:=1-lower]
            sumX[,lower:=NULL]
            setnames(sumX,"lower.risk","lower")
        }
        estimate <- "risk"
    }else{
        if (model=="survival")
            estimate <- "surv"
        else
            estimate <- "risk"
    }
    xvars  <- data.table::key(sumX)
    if (length(xvars)>0){
        xdata <- sumX[,xvars,with=FALSE]
        # degenerated variables are ignored
        xdegen <- sapply(xdata,function(x)length(unique(x))==1)
        xvars <- xvars[!xdegen]
    }
    if (length(xvars)>0){
        xdata <- xdata[,xvars,with=FALSE]
        if (short.labels[[1]]==TRUE){
            xstrata <- apply(do.call("cbind",lapply(xvars,function(n){
                if(is.numeric(xdata[[n]]))
                    x <- format(xdata[[n]],digits=num.digits)
                else 
                    x <- as.character(xdata[[n]])
                # justify to column width
                format(c(n,x),justify="right")
            })),1,paste,collapse=", ")
            xtitle <- xstrata[[1]]
            xstrata <- xstrata[-1]
        }else{
            xstrata <- apply(do.call("cbind",lapply(xvars,function(n){
                if(is.numeric(xdata[[n]]))
                    x <- paste(n,format(xdata[[n]],digits=2),sep="=")
                else 
                    x <- paste(n,xdata[[n]],sep="=")
                # justify to column width
                format(x,justify="right")
            })),1,paste,collapse=", ")
            xtitle <- ""
        }
        Y <- lapply(unique(xstrata),function(u){sumX[[estimate]][xstrata == u]})
        names(Y) <- unique(xstrata)
        ## Y <- split(sumX[[estimate]],xstrata)
        nlost <- lapply(unique(xstrata),function(u){sumX[["n.lost"]][xstrata == u]})
        names(nlost) <- unique(xstrata)
        ## nlost <- split(sumX[["n.lost"]],xstrata)
        if (confint[[1]]==TRUE){
            ci <- lapply(unique(xstrata),function(u){sumX[,data.table::data.table(time,lower,upper)][xstrata == u]})
            names(ci) <- unique(xstrata)
            ## ci <- split(sumX[,data.table::data.table(time,lower,upper)],xstrata)
        } else{
            ci <- NULL
        }
        names(Y) <- unique(xstrata)
    }else{
        xtitle <- ""
        Y <- list(sumX[[estimate]])
        nlost <- list(sumX[["n.lost"]])
        if (confint[[1]]==TRUE){
            ci <- list(data.table::data.table(time=sumX$time,lower=sumX$lower,upper=sumX$upper))
        }
        else{
            ci <- NULL
        }
    }
    # argument select could be removed:
    if (!missing(select)){Y <- Y[select]}
    nlines <- length(Y)
    # }}}
    # {{{  getting default arguments for plot, atrisk, axes, legend, confint, marktime
    if (missing(xlim)) xlim <- c(min(plot.times), max(plot.times))
    if (!missing(timeconverter)){
        units <- strsplit(tolower(as.character(substitute(timeconverter))),"[ \t]?(2|to)[ \t]?")[[1]]
        conversion <- switch(paste0(units,collapse="-"),
                             "days-years"=1/365.25,
                             "months-years"=1/12,
                             "days-months"=1/30.4368499,
                             "years-days"=365.25,
                             "years-months"=12,
                             "months-days"=30.4368499)
        one <- switch(units[[1]],"years"=1,"months"=12,"days"=365.25)
        xlab <- paste0("Time (", units[[2]],")")
        axis1.DefaultArgs <- list(at=seq(xlim[1],xlim[2],one),
                                  labels=seq(xlim[1],xlim[2],one)*conversion)
        atriskDefaultPosition <- seq(xlim[1],xlim[2],one)
    } else {
        if (missing(xlab)) xlab <- "Time"
        axis1.DefaultArgs <- list()
        atriskDefaultPosition <- seq(min(plot.times),
                                     max(plot.times),
                                     (max(plot.times)-min(plot.times))/10)
    }
    if (missing(ylab)) ylab <- switch(type,
                                      "surv"=ifelse(x$reverse==TRUE,"Censoring probability","Survival probability"),
                                      "risk"="Absolute risk",
                                      "hazard"="Cumulative hazard")
    if (missing(ylim)) ylim <- c(0, 1)
    if (missing(lwd)) lwd <- rep(3,nlines)
    if (missing(col)) {
        cbbPalette <- c("#000000", "#E69F00", "#56B4E9", "#009E73", "#D55E00", "#0072B2", "#CC79A7", "#F0E442")
        if (nlines>length(cbbPalette))
            col <- rainbow(nlines)
        else
            col <- cbbPalette[1:nlines]
    }
    if (missing(lty)) lty <- rep(1, nlines)
    if (length(lwd) < nlines) lwd <- rep(lwd, nlines)
    if (length(lty) < nlines) lty <- rep(lty, nlines)
    if (length(col) < nlines) col <- rep(col, nlines)
  
    background.DefaultArgs <- list(xlim=xlim,
                                   ylim=ylim,
                                   horizontal=seq(ylim[1],ylim[2],diff(ylim)/4),
                                   vertical=NULL,
                                   bg="white",
                                   fg="gray88")
    axis2.DefaultArgs <- list(at=seq(ylim[1],ylim[2],ylim[2]/4),side=2)
    lines.DefaultArgs <- list(type="s")
    plot.DefaultArgs <- list(x=0,y=0,type = "n",ylim = ylim,xlim = xlim,xlab = xlab,ylab = ylab)
    marktime.DefaultArgs <- list(x=Y,nlost=nlost,times=plot.times,pch="I",col=col)
    atrisk.DefaultArgs <- list(x=x,
                               newdata=newdata,
                               interspace=1,
                               dist=1.5,
                               col=col,
                               labelcol=1,
                               titlecol=1,
                               title="",
                               labels="",
                               times=atriskDefaultPosition,show.censored=FALSE,unit='npc')
    if (!missing(select) && (!(model=="competing.risks" && stacked))){
        atrisk.DefaultArgs$newdata <- atrisk.DefaultArgs$newdata[select,,drop=FALSE]
    }
    adapted.legend.cex <- as.numeric(as.character(cut(nlines,breaks=c(-Inf,2:17,Inf),labels=seq(1.5,by=-0.05,length.out=17))))
    legend.DefaultArgs <- list(legend=names(Y),
                               title=xtitle,
                               lwd=lwd,
                               col=col,
                               lty=lty,
                               cex=adapted.legend.cex,
                               bty="n",
                               y.intersp=1.3,
                               x=ifelse(type=="surv","topright","topleft"))
    if (stacked) {
        legend.DefaultArgs$title <- "Competing risks"
        legend.DefaultArgs$x <- "topleft"
    }
    
    confint.DefaultArgs <- list(ci=ci,
                                citype="shadow",
                                density=55,
                                col=col[1:nlines],
                                lwd=rep(2,nlines),
                                lty=rep(3,nlines))

    # }}}
    # {{{  smart argument control
    smartA <- SmartControl(call=  list(...),
                           keys=c("plot","lines","atrisk","legend","confint","background","marktime","axis1","axis2"),
                           ignore=c("x","type","cause","newdata","add","col","lty","lwd","ylim","xlim","xlab","ylab","legend","marktime","confint","automar","atrisk","timeOrigin","percent","axes","atrisk.args","confint.args","legend.args"),
                           defaults=list("plot"=plot.DefaultArgs,"atrisk"=atrisk.DefaultArgs,"lines"=lines.DefaultArgs,"legend"=legend.DefaultArgs,"confint"=confint.DefaultArgs,"marktime"=marktime.DefaultArgs,"background"=background.DefaultArgs,"axis1"=axis1.DefaultArgs,"axis2"=axis2.DefaultArgs),
                           forced=list("plot"=list(axes=FALSE),"axis1"=list(side=1)),
                           ignore.case=TRUE,
                           replaceDefaults=FALSE,
                           verbose=TRUE)

    # }}}
    # {{{  setting margin parameters
    if (atrisk==TRUE){
        oldmar <- par()$mar
        on.exit(par(mar=oldmar)) ## reset
        if (missing(automar) || automar==TRUE){
            ##        bottomMargin =  margin line (in 'mex' units) for xlab
            ##                        + distance of xlab from xaxis
            ##                        + distance of atrisk numbers from xlab
            ##                        + number of atrisk lines
            ##                        + one extra line below the bottom number atrisk line
            ##      leftSideMargin =  margin line + atrisk.lab
            bottommargin.height <- smartA$atrisk$dist+ifelse(clusterp,2,1)*nlines * smartA$atrisk$interspace/max(1,length(cause))
            bottomMargin.start <- par()$mgp[1]
            maxlabellen <- max(strwidth(c(smartA$atrisk$labels,smartA$atrisk$title),
                                        cex=smartA$atrisk$cex,
                                        units="inches"))
            maxlabellen <- pmax(maxlabellen * (par("mar")[2] / par("mai")[2]),par("mar")[2])
            leftMargin <- maxlabellen+2-par("mar")[2]
            newmar <- par()$mar + c(bottomMargin.start+bottommargin.height,leftMargin,0,0)
            par(mar=newmar)
        }
    }
    # }}}
    # {{{  plot and backGround
    if (!add) {
        do.call("plot",smartA$plot)
        if (background==TRUE){
            do.call("backGround",smartA$background)
        }
    }
    # }}}
    # {{{  axes
    if (!add) {
        if (axes){
            do.call("axis",smartA$axis1)
            if (percent & is.null(smartA$axis2$labels))
                smartA$axis2$labels <- paste(100*smartA$axis2$at,"%")
            do.call("axis",smartA$axis2)
        }
    }
    # }}}
    # {{{  pointwise confidence intervals
    if (confint==TRUE) {
        ## if (verbose==TRUE){print(smartA$confint)}
        do.call("confInt",smartA$confint)
    }
    # }}}
    # {{{  adding the lines 
    lines.type <- smartA$lines$type
    if (stacked==TRUE){
        if (length(Y)>1){
            nY <- names(Y)
            Y <- apply(do.call("rbind",Y),2,cumsum)
            Y <- lapply(1:nlines,function(i)Y[i,])
            names(Y) <- nY
        }
        ## names(Y) <- attr(x$model.response,"states")
        nix <- lapply(1:nlines, function(s) {
            yyy <- Y[[s]]
            ppp <- plot.times
            pos.na <- is.na(yyy)
            ppp <- ppp[!pos.na]
            yyy <- yyy[!pos.na]
            lines(x = ppp,y = yyy,type = lines.type,col = col[s],lty = lty[s],lwd = lwd[s])
            cc <- dimColor(col[s],density=55)
            ttt <- ppp
            nt <- length(ttt)
            ttt <- c(ttt,ttt)
            uuu <- c(0,yyy[-nt],yyy)
            if (s==1)
                lll <- rep(0,nt*2)
            else
                lll <- c(0,Y[[s-1]][!pos.na][-nt],Y[[s-1]][!pos.na])
            neworder <- order(ttt)
            uuu <- uuu[neworder]
            lll <- lll[neworder]
            ttt <- sort(ttt)
            polygon(x=c(ttt,rev(ttt)),y=c(lll,rev(uuu)),col=cc,border=NA)
        })
    }else{
        nix <- lapply(1:nlines, function(s) {
            lines(x = plot.times,
                  y = Y[[s]],
                  type = lines.type,
                  col = col[s],
                  lty = lty[s],
                  lwd = lwd[s])
        })
    }
    # }}}
    # {{{  marks at the censored times
    if (marktime==TRUE){
        if (model %in% c("survival","competing.risks")){
            do.call("markTime",smartA$marktime)
        }
        else{
            message("Marking the curves at censored times is not yet available for multi-state models.")
        }
    }
    # }}}
    # {{{  adding the number of subjects at risk and uncensored
    if (atrisk==TRUE && !add){
        if (hit <- match("at",names(smartA$atrisk),nomatch=FALSE)){
            if (match("atrisk.times",names(list(...)),nomatch=FALSE)){
                warning("Atrisk argument clash: remove either 'atrisk.at' or 'atrisk.times'.")
            }
            else{
                names(smartA$atrisk)[hit] <- "times"
                smartA$atrisk <- smartA$atrisk[!duplicated(names(smartA$atrisk))]
            }
        }
        do.call("atRisk",smartA$atrisk)
    }
    # }}}
    # {{{  legend
    if(legend==TRUE && !add && !is.null(names(Y))){
        save.xpd <- par()$xpd
        if (logrank && model=="survival" && length(smartA$legend$legend)>1){
            ## formula.names <- try(all.names(formula),silent=TRUE)
            lrform <- x$call$formula
            if (lrform[[2]][[1]]==as.name("Hist"))
                lrform[[2]][[1]] <- as.name("Surv")
            ## require(survival)
            lrtest <- survival::survdiff(eval(lrform),data=eval(x$call$data),subset=eval(x$call$subset))
            if (length(lrtest$n) == 1) {
                p <- 1 - pchisq(lrtest$chisq, 1)
            } else{
                if (is.matrix(lrtest$obs)) {
                    etmp <- apply(lrtest$exp, 1, sum)
                }
                else {
                    etmp <- lrtest$exp
                }
                df <- (sum(1 * (etmp > 0))) - 1
                p <- 1 - pchisq(lrtest$chisq, df)
            }
            pform <- format.pval(p,digits=logrank,eps=0.0001)
            pbelow <- length(grep("<",pform))>0
            pform <- if (pbelow) paste0("p ",pform) else paste0("p = ",pform)
            if (length(smartA$legend$title)){
                smartA$legend$title <- paste(smartA$legend$title," Log-rank: ",pform)
            } else{
                smartA$legend$title <- paste("Log-rank: ",pform)
            }
        }
        par(xpd=TRUE)
        do.call("legend",smartA$legend)
        par(xpd=save.xpd)
    }
    # }}}
    invisible(x)
}

#' @export 
print.IntIndex <- function(x,...){
  mlist <- split(x$Mindex,rep(1:length(x$Mstrata),diff(c(0,x$Mstrata))))
  p <- x$petoInt[1,]
  q <- x$petoInt[2,]
  pqnames <- paste("(p;q)=",paste("(",p,";",q,"]",sep=""))
  pqnames[p==q] <- paste("(p;q)=",paste("[",p[p==q],";",q[p==q],"]",sep=""))
  names(mlist) <- pqnames
  Mlist <- lapply(mlist,function(u){
    L <- x$obsInt[1,u]
    R <- x$obsInt[2,u]
    out <- paste("(",L,";",R,"]",sep="")
    out[L==R] <- paste("[",L[L==R],";",R[L==R],"]",sep="")
    out
  })
  print(Mlist)
  Ilist <- split(x$Iindex,rep(1:length(x$Istrata),diff(c(0,x$Istrata))))
}

#' Print objects in the prodlim library
#' 
#' Pretty printing of objects created with the functionality of the `prodlim'
#' library.
#' 
#' 
#' @aliases print.prodlim print.neighborhood print.Hist
#' @param x Object of class \code{prodlim}, \code{Hist} and
#' \code{neighborhood}.
#' @param \dots Not used.
#' @author Thomas Gerds <tag@@biostat.ku.dk>
#' @seealso \code{\link{summary.prodlim}}, \code{\link{predict.prodlim}}
#' @keywords survival
#' @export 
"print.prodlim" <- function(x,...) {
    cat("\n")
    cat("Call: ")
    print(x$call)
    cat("\n")
    model <- x$model
    ##   message("Estimation method:")
    if (!(model %in% c("survival","competing.risks"))) stop("Under construction")
    if (model=="survival")
        if (x$cens.type=="intervalCensored"){
            message(switch(x$covariate.type,"NPMLE",
                           "Stratified NPMLE estimator",
                           "Stratified NPMLE estimator",
                           "Stratified NPMLE estimator")," for the",ifelse(x$covariate.type==1," "," conditional "),ifelse(x$reverse==FALSE,"event time ","censoring time "),"survival function")
            message(paste("\nIteration steps:",x$n.iter,"\n"))
            ##   summary(x)
            cat("\n")
        }
        else{
            message(switch(x$covariate.type,"Kaplan-Meier estimator",
                           "Stratified Kaplan-Meier estimator",
                           "Stone-Beran estimator",
                           "Stratified Stone-Beran estimator")," for the",ifelse(x$covariate.type==1," "," conditional "),ifelse(x$reverse==FALSE,"event time ","censoring time "),"survival function")
        }
    cat("\n")
    ##   discrete.predictors <- extract.name.from.special(grep("strata.",names(x$X),value=TRUE),pattern="strata\\.")
    ##   continuous.predictors <- extract.name.from.special(grep("NN.",names(x$X),value=TRUE),pattern="NN\\.")
    discrete.predictors <- x$discrete.predictors
    continuous.predictors <- x$continuous.predictors
    if (!is.null(x$cluster))
        message("\nCluster-correlated data:\n\n cluster variable: ",x$cluster,"\n")
    format.disc <- function(name){
        paste(name," (",
              paste(x$xlevels[[name]],collapse=", ",sep=""),")",
              collapse=", ",sep="")
    }
    message(#"Predictor space:\n\n",
        switch(x$covariate.type,
               "No covariates",{
                   if (length(discrete.predictors)==1){
                       c("Discrete predictor variable: ", format.disc(discrete.predictors))
                   }else{
                       c("Discrete predictor variables:\n", sapply(discrete.predictors,function(x)paste("\n - ",format.disc(x))))
                   }},
               c("Continuous predictors: ",continuous.predictors),
               c("  Discrete predictor variables: ",
                 paste(discrete.predictors,collapse=", "),
                 "\nContinuous predictor variables: ",
                 continuous.predictors)))
    summary(x$model.response,verbose=TRUE)
    if (!is.null(x$na.action)){
        cat("\n",
            length(x$na.action),
            ifelse(length(x$na.action)==1,
                   " observation",
                   " observations")," deleted due to missing values.\n",sep="")
    }
    if (attr(x$model.response,"entry.type")=="leftTruncated")
        message("\nThe data are left truncated (delayed entry). The product-limit estimate uses the convention that
in case of ties, on the same 'day', outcome events occur before censored and censored occur before entry.
Note however that the summary tables include the subjects that enter
at time t in the number at risk at time t.")
}

#  These functions extract the number of subjects atRisk and the number of
#  events at given times from the object and binds it together with 
#  quantities like survival prob, risk/cuminc, standard errors, etc. which can 
#  simply be evaluated at the requested times.

lifeTab <- function(object,...){
  if(NROW(object$model.response)<=0) stop("No response found") #  to avoid seg faults
  dummy <- 1
  class(dummy) <- object$model
  UseMethod("lifeTab",object=dummy)
}






  

#' Quantiles for Kaplan-Meier and Aalen-Johansen estimates.
#' 
#' Quantiles for Kaplan-Meier and Aalen-Johansen estimates.
#' 
#' 
#' @param x Object of class \code{"prodlim"}.
#' @param q Quantiles. Vector of values between 0 and 1.
#' @param cause For competing risks the cause of interest.
#' @param ... not used
#' @author Thomas Alexander Gerds <tag@@biostat.ku.dk>
#' @keywords survival
#' @examples
#' library(lava)
#' set.seed(1)
#' d=SimSurv(30)
#'
#' # Quantiles of the potential followup time
#' g=prodlim(Hist(time,status)~1,data=d,reverse=TRUE)
#' quantile(g)
#' 
#' # survival time
#' f=prodlim(Hist(time,status)~1,data=d)
#' f1=prodlim(Hist(time,status)~X1,data=d)
#' # default: median and IQR
#' quantile(f)
#' quantile(f1)
#' # median alone
#' quantile(f,.5)
#' quantile(f1,.5)
#'
#' # competing risks
#' set.seed(3)
#' dd = SimCompRisk(30)
#' ff=prodlim(Hist(time,event)~1,data=dd)
#' ff1=prodlim(Hist(time,event)~X1,data=dd)
#' ## default: median and IQR
#' quantile(ff)
#' quantile(ff1)
#' 
#' print(quantile(ff1),na.val="NA")
#' print(quantile(ff1),na.val="Not reached")
#' 
#' @export quantile.prodlim
#' @export 
"quantile.prodlim" <- function(x,
                               q,
                               cause=1,
                               ...){
    time <- surv <- lower <- upper <- cuminc <- NULL
    get.quantiles <- function(time,x,lower,upper,model="survival"){
        out <- do.call("cbind",lapply(list(x,lower,upper),function(sumw){
            notna= is.na(sumw) | sumw==0 | sumw ==1
            if (all(notna)) return(as.numeric(NA))
            xxx=as.numeric(sumw[!notna])
            ttt=as.numeric(time[!notna])
            found <- 2+sindex(jump.times=xxx,eval.times=q,comp=ifelse(model=="survival","greater","smaller"),strict=FALSE)
            inner <- c(as.vector(c(0,ttt)[found]))
            inner
        }))
        out <- data.frame(out)
        out <- cbind(q,out)
        if (model=="survival") {
            names(out) <- c("q","quantile","lower","upper")
        }else{
            names(out) <- c("q","quantile","upper","lower")
            out <- out[,c("q","quantile","lower","upper")]
        }
        out
    }
    etype <- attr(x$model.response,"entry.type")
    if (!is.null(etype) && etype=="leftTruncated")
        stop("Don't know how to compute quantiles with delayed entry (left-truncation).")
    if(x$model=="survival"){
        if (missing(q)) q <- c(1,.75,0.5,.25,0)
        q <- 1-q ## since this is a survival function
        sumx <- summary(x,newdata=x$X,times=x$time,verbose=FALSE)
        if (attr(sumx,"cotype")==1) {
            out <- sumx[,get.quantiles(time=time,x=surv,lower=lower,upper=upper)]
        } else{
            out <- sumx[,get.quantiles(time=time,x=surv,lower=lower,upper=upper,model="survival"),by=key(sumx)]
        }
    } else{
        ## absolute risks, cumulative incidence, competing risks
        if (missing(q)) q <- c(0,0.25,0.5,0.75,1)
        sumx <- summary(x,newdata=x$X,times=x$time,verbose=FALSE,cause=cause)
        out <- sumx[,get.quantiles(time=time,x=cuminc,lower=lower,upper=upper,model = "cuminc"),by = key(sumx)]
        out
    }
    attr(out,"model") <- x$model
    attr(out,"covariates") <- key(sumx)
    attr(out,"reverse") <- x$reverse
    attr(out,"cotype") <- attr(sumx,"cotype")
    class(out) <- c("quantile.prodlim","data.table")
    out
}
      

##' Reformulate a terms object such that some specials are stripped off 
##'
##' This function is used to remove special specials, i.e., those
##' which cannot or should not be evaluated. 
##' IMPORTANT: the unstripped terms need to know about all specials including the aliases.
##' See examples.
##' @title Strip special functions from terms 
##' @param terms Terms object
##' @param specials Character vector of specials which should be
##' stripped off
##' @param alias.names Optional. A named list with alias names for the specials. 
##' @param unspecials Optional. A special name for treating all the unspecial terms.
##' @param arguments A named list of arguments, one for each element
##' of specials. Elements are passed to \code{parseSpecialNames}.
##' @param keep.response Keep the response in the resulting object?
##' @return Reformulated terms object with an additional attribute which contains the \code{stripped.specials}.
##' @seealso parseSpecialNames reformulate drop.terms
##' @examples
##' 
##' ## parse a survival formula and identify terms which
##' ## should be treated as proportional or timevarying:
##' f <- Surv(time,status)~age+prop(factor(edema))+timevar(sex,test=0)+prop(bili,power=1)
##' tt <- terms(f,specials=c("prop","timevar"))
##' attr(tt,"specials")
##' st <- strip.terms(tt,specials=c("prop","timevar"),arguments=NULL)
##' formula(st)
##' attr(st,"specials")
##' attr(st,"stripped.specials")
##' 
##' ## provide a default value for argument power of proportional treatment
##' ## and argument test of timevarying treatment: 
##' st2 <- strip.terms(tt,
##'                    specials=c("prop","timevar"),
##'                    arguments=list("prop"=list("power"=0),"timevar"=list("test"=0)))
##' formula(st2)
##' attr(st2,"stripped.specials")
##' attr(st2,"stripped.arguments")
##' 
##' ## treat all unspecial terms as proportional
##' st3 <- strip.terms(tt,
##'                    unspecials="prop",
##'                    specials=c("prop","timevar"),
##'                    arguments=list("prop"=list("power"=0),"timevar"=list("test"=0)))
##' formula(st3)
##' attr(st3,"stripped.specials")
##' attr(st3,"stripped.arguments")
##' 
##' ## allow alias names: strata for timevar and tp, const for prop.
##' ## IMPORTANT: the unstripped terms need to know about
##' ## all specials including the aliases
##' f <- Surv(time,status)~age+const(factor(edema))+strata(sex,test=0)+prop(bili,power=1)+tp(albumin)
##' tt2 <- terms(f,specials=c("prop","timevar","strata","tp","const"))
##' st4 <- strip.terms(tt2,
##'                    specials=c("prop","timevar"),
##'                    unspecials="prop",
##'                    alias.names=list("timevar"="strata","prop"=c("const","tp")),
##'                    arguments=list("prop"=list("power"=0),"timevar"=list("test"=0)))
##' formula(st4)
##' attr(st4,"stripped.specials")
##' attr(st4,"stripped.arguments")
##' 
##' ## test if alias works also without unspecial argument
##' st5 <- strip.terms(tt2,
##'                    specials=c("prop","timevar"),
##'                    alias.names=list("timevar"="strata","prop"=c("const","tp")),
##'                    arguments=list("prop"=list("power"=0),"timevar"=list("test"=0)))
##' formula(st5)
##' attr(st5,"stripped.specials")
##' attr(st5,"stripped.arguments")
##' 
##' library(survival)
##' data(pbc)
##' model.design(st4,data=pbc[1:3,],specialsDesign=TRUE)
##' model.design(st5,data=pbc[1:3,],specialsDesign=TRUE)
##'
##' 
##' @export 
##' @author Thomas A. Gerds <tag@@biostat.ku.dk>
strip.terms <- function(terms,
                        specials,
                        alias.names=NULL,
                        unspecials=NULL,
                        arguments,
                        keep.response=TRUE){
    termLabels <- attr(terms,"term.labels")
    terms.specials <- attr(terms,"specials")
    intercept <- attr(terms, "intercept")
    if (attr(terms,"response") && keep.response)
        response <- terms[[2L]]
    else
        response <- NULL
    # resolve unspecials
    do.unspecials <- length(unspecials)>0
    if (do.unspecials){
        if (length(unlist(terms.specials))>0)
            any <- -(-attr(terms,"response")+unlist(terms.specials))
        else
            any <- 1:length(termLabels)
        if (length(any))
            termLabels[any] <- paste(unspecials,"(",termLabels[any],")",sep="")
    }
    # resolve aliases
    do.alias <- length(alias.names)>0
    if (do.alias){
        for (spc in specials){
            ali <- alias.names[[spc]]
            if (length(ali)>0){
                ## replace alias with real name
                termLabels <- sub(paste("^(",paste(ali,collapse="|"),")\\(",sep=""),
                                  paste(spc,"(",sep=""),
                                  termLabels)
            }
        }
    }
    ## now remove alias specials
    newspecials <- unique(c(specials,names(terms.specials)))
    anames <- unlist(alias.names)
    if (length(anames)>0){
        catch <- match(anames,newspecials,nomatch=0)
        newspecials <- newspecials[-catch]
    }
    if (do.unspecials||do.alias){
        aform <- reformulate(termLabels,response,intercept)
        environment(aform) <- environment(terms)
        if (do.alias)
            terms <- terms(aform,specials=newspecials)
        else
            terms <- terms(aform,specials=specials)
        terms.specials <- attr(terms,"specials")
    }
    ## terms.specials <- specials
    ## remove unused specials
    ## terms.specials <- terms.specials[!sapply(terms.specials,is.null)]
    ## only strip the specials in specials
    found <- match(names(terms.specials),specials,nomatch=0)
    if (any(found>0)){
        stripspecials <- names(terms.specials)[found>0]
        strippedTerms <- vector(mode="list")
        strippedArguments <- vector(mode="list")
        for (s in 1:length(stripspecials)){
            ## outcome counts as 1
            spc <- stripspecials[[s]]
            hit.s <- - attr(terms,"response") + terms.specials[[spc]]
            ps <- parseSpecialNames(termLabels[hit.s],
                                    special=spc,
                                    arguments=arguments[[spc]])
            ## attr(ps,"special.position") <- terms.specials[[spc]]
            terms.s <- terms.specials[spc]
            aps <- list(ps)
            names(aps) <- spc
            strippedArguments <- c(strippedArguments,aps)
            strippedTerms <- c(strippedTerms,terms.s)
            termLabels[hit.s] <- names(ps)
        }
        strippedFormula <- reformulate(termLabels,response,intercept)
        environment(strippedFormula) <- environment(terms)
        out <- terms(strippedFormula, specials = names(terms.specials))
        ## reset specials
        attr(out,"stripped.specials") <- strippedTerms
        attr(out,"stripped.arguments") <- strippedArguments
        return(out)
    }else{
        return(terms)
    }
}

#' Survival model for simulation
#' 
#' Create a survival model to simulate a right censored event time data without
#' covariates
#' 
#' This function requires the \code{lava} package.
#' 
#' @return A structural equation model initialized with three variables: the
#' latent event time, the latent right censored time, and the observed
#' right censored event time.
#' @author Thomas A. Gerds <tag@@biostat.ku.dk>
#' @export 
survModel <- function(){
    ## require(lava)
    sm <- lava::lvm(~eventtime+censtime)
    lava::distribution(sm,"eventtime") <- lava::coxWeibull.lvm(scale=1/100)
    lava::distribution(sm,"censtime") <- lava::coxWeibull.lvm(scale=1/100)
    sm <- lava::eventTime(sm,time~min(eventtime=1,censtime=0),"event")
    sm
}

#' Create an event history object
#' 
#' Functionality for managing censored event history response data. The
#' function can be used as the left hand side of a formula: \code{Hist} serves
#' \code{\link{prodlim}} in a similar way as \code{\link[survival]{Surv}} from the
#' survival package serves `survfit'. \code{Hist} provides the suitable
#' extensions for dealing with right censored and interval censored data from
#' competing risks and other multi state models. Objects generated with
#' \code{Hist} have a print and a plot method.
#' 
#' 
#' *Specification of the event times*
#' 
#' If \code{time} is a numeric vector then the values are interpreted as right
#' censored event times, ie as the minimum of the event times and the censoring
#' times.
#' 
#' If \code{time} is a list with two elements or data frame with two numeric
#' columns The first element (column) is used as the left endpoints of interval
#' censored observations and the second as the corresponding right endpoints.
#' When the two endpoints are equal, then this observation is treated as an
#' exact uncensored observation of the event time.  If the value of the right
#' interval endpoint is either \code{NA} or \code{Inf}, then this observation
#' is treated as a right censored observation. Right censored observations can
#' also be specified by setting the value of \code{event} to \code{cens.code}.
#' This latter specification of right censored event times overwrites the
#' former: if \code{event} equals \code{cens.code} the observation is treated
#' as right censored no matter what the value of the right interval endpoint
#' is.
#' 
#' *Specification of the events*
#' 
#' If \code{event} is a numeric, character or logical vector then the order of
#' the attribute "state" given to the \code{value} of \code{Hist} is determined
#' by the order in which the values appear.  If it is a factor then the order
#' from the levels of the factor is used instead.
#' 
#' **Normal form of a multi state model**
#' 
#' If \code{event} is a list or a data.frame with exactly two elements, then
#' these describe the transitions in a multi state model that occurred at the
#' corresponding \code{time} as follows: The values of the first element are
#' interpreted as the \code{from} states of the transition and values of the
#' second as the corresponding \code{to} states.
#' 
#' **Longitudinal form of a multi state model**
#' 
#' If \code{id} is given then \code{event} must be a vector. In this case two
#' subsequent values of \code{event} belonging to the same value of \code{id}
#' are treated as the \code{from} and \code{to} states of the transitions.
#' 
#' @param time for right censored data a numeric vector of event times -- for
#' interval censored data a list or a data.frame providing two numeric vectors
#' the left and right endpoints of the intervals.  See \code{Details}.
#' @param event A vector or a factor that specifies the events that occurred at
#' the corresponding value of \code{time}. Numeric, character and logical
#' values are recognized.  It can also be a list or a data.frame for the
#' longitudinal form of storing the data of a multi state model -- see
#' \code{Details}.
#' @param entry Vector of delayed entry times (left-truncation) or list of two
#' times when the entry time is interval censored.
#' @param id Identifies the subjects to which multiple events belong for the
#' longitudinal form of storing the data of a multi state model -- see
#' \code{Details}.
#' @param cens.code A character or numeric vector to identify the right
#' censored observations in the values of \code{event}.  Defaults to "0" which
#' is equivalent to 0.
#' @param addInitialState If TRUE, an initial state is added to all ids for the
#' longitudinal input form of a multi-state model.
#' @return An object of class \code{Hist} for which there are print and plot
#' methods.  The object's internal is a matrix with some of the following
#' columns: \item{time}{ the right censored times} \item{L}{the left endpoints
#' of internal censored event times} \item{R}{the right endpoints of internal
#' censored event times} \item{status}{\code{0} for right censored, \code{1}
#' for exact, and \code{2} for interval censored event times.} \item{event}{an
#' integer valued numeric vector that codes the events.} \item{from}{an integer
#' valued numeric vector that codes the \code{from} states of a transition in a
#' multi state model.} \item{to}{an integer valued numeric vector that codes
#' the \code{to} states of a transition in a multi state model.}
#' 
#' Further information is stored in \code{\link{attributes}}.  The key to the
#' official names given to the events and the from and to states is stored in
#' an attribute "states".
#' @author Thomas A. Gerds \email{tag@@biostat.ku.dk}, Arthur Allignol
#' \email{arthur.allignol@@fdm.uni-freiburg.de}
#' @seealso \code{\link{plot.Hist}}, \code{\link{summary.Hist}},
#' \code{\link{prodlim}}
#' @keywords survival
#' @examples
#' 
#' 
#' ## Right censored responses of a two state model
#' ## ---------------------------------------------
#' 
#' Hist(time=1:10,event=c(0,1,0,0,0,1,0,1,0,0))
#' 
#' ## change the code for events and censored observations
#' 
#' Hist(time=1:10,event=c(99,"event",99,99,99,"event",99,"event",99,99),cens.code=99)
#' 
#' TwoStateFrame <- SimSurv(10)
#' SurvHist <- with(TwoStateFrame,Hist(time,status))
#' summary(SurvHist)
#' plot(SurvHist)
#' 
#' ## Right censored data from a competing risk model
#' ## --------------------------------------------------
#' 
#' CompRiskFrame <- data.frame(time=1:10,event=c(1,2,0,3,0,1,2,1,2,1))
#' CRHist <- with(CompRiskFrame,Hist(time,event))
#' summary(CRHist)
#' plot(CRHist)
#' 
#' ## Interval censored data from a survival model
#' icensFrame <- data.frame(L=c(1,1,3,4,6),R=c(2,NA,3,6,9),event=c(1,1,1,2,2))
#' with(icensFrame,Hist(time=list(L,R)))
#' 
#' ## Interval censored data from a competing risk model
#' with(icensFrame,Hist(time=list(L,R),event))
#' 
#' ## Multi state model
#' MultiStateFrame <- data.frame(time=1:10,
#' 			      from=c(1,1,3,1,2,4,1,1,2,1),
#' 			      to=c(2,3,1,2,4,2,3,2,4,4))
#' with(MultiStateFrame,Hist(time,event=list(from,to)))
#' 
#' ##  MultiState with right censored observations
#' 
#' MultiStateFrame1 <- data.frame(time=1:10,
#' 			      from=c(1,1,3,2,1,4,1,1,3,1),
#' 			      to=c(2,3,1,0,2,2,3,2,0,4))
#' with(MultiStateFrame1,Hist(time,event=list(from,to)))
#' 
#' ## Using the longitudinal input method 
#' MultiStateFrame2 <- data.frame(time=c(0,1,2,3,4,0,1,2,0,1),
#' 			      event=c(1,2,3,0,1,2,4,2,1,2),
#' 			      id=c(1,1,1,1,2,2,2,2,3,3))
#'  with(MultiStateFrame2,Hist(time,event=event,id=id))
#'
#' @export
"Hist" <- function(time,
                   event,
                   entry=NULL,
                   id=NULL,
                   cens.code="0",
                   addInitialState=FALSE) {

    ## package Cprob provides a vector, 
    ## to avoid problems we pick the first element 
    cens.code <- as.character(cens.code[[1]])
    
                                        # {{{ resolving the `time' argument
    if (is.matrix(time)) time <- data.frame(time)
    if (inherits(time,"list")){
        if (length(time) !=2 || length(time[[1]])!=length(time[[2]]))
            stop("Argument time has a wrong format")
        time <- data.frame(time)
    }
    if (is.data.frame(time)){
        cens.type <- "intervalCensored"
        L <- time[[1]]
        R <- time[[2]]
        N <- length(L)
        stopifnot(is.numeric(L))
        stopifnot(is.numeric(R))
        wrong <- L>R
        wrong[is.na(R)] <- FALSE
        stopifnot(all(wrong==FALSE))
        rm(wrong)
        status <- rep(2,N)
        status[L==R] <- 1
        status[is.infinite(R) | is.na(R) | (L!=R & as.character(R)==cens.code)] <- 0
        ## the last part of the condition achieves to things:
        ##     1. for multi-state models allow transitions to a censored state
        ##     2. to ignore this, if an event occured exactly at time 0 and 0 is the cens.code
        R[status==0] <- Inf
    }
    else{
        stopifnot(is.numeric(time))
        cens.type <- "rightCensored"
        N <- length(time)
        status <- rep(1,N) ## temporary dummy
    }
                                        # }}}
                                        # {{{ resolving the `entry' argument

  if (is.null(entry))
      entry.type <- ""
  else{
      if (is.matrix(entry)) entry <- data.frame(entry)
      if (inherits(entry,"list")){
          if (length(entry) !=2 || length(entry[[1]])!=length(entry[[2]]))
              stop("Argument entry has a wrong format")
          entry <- data.frame(entry)
      }
      if (is.data.frame(entry)){
          entry.type <-"intervalCensored"
          U <- entry[[1]]
          V <- entry[[2]]
          stopifnot(is.numeric(U))
          stopifnot(is.numeric(V))
          stopifnot(all(!is.na(U))|all(!is.na(V)))
      }
      else{
          stopifnot(is.numeric(entry))
          if (is.null(id))
              entry.type <- "leftTruncated"
          else
              entry.type <- "exact"
      }}
  ## check if entry < exit
  if (cens.type=="intervalCensored"){
      if (entry.type=="intervalCensored"){
          stopifnot(all(V<=L))
      } else{
          stopifnot(is.null(entry) || all(entry<=L))
      }
  } else{
      if (entry.type=="intervalCensored"){
          stopifnot(all(V<=time))
      } else{
          stopifnot(is.null(entry) || all(entry<=time))
      }
  }

  # }}}
  # {{{ resolving the argument `event' 

  if (missing(event)){
      model <- "onejump"
      event <-  rep(1,N)
      warning("Argument event is missing:\nassume observations of a survival model\nand only one event per subject")
  }
  else{
      if (is.matrix(event)) event <- data.frame(event)
      ## event can be an ordered factor
      ## in which case class has two elements
      ## to avoid warnings we need [[1]]
      if ((is.vector(event) & !(inherits(event,"list")))|| is.factor(event))
          stopifnot(length(event)==N)
      if (inherits(event,"list")){
          if (length(event) !=2 || length(event[[1]])!=length(event[[2]]))
              stop("Argument event has a wrong format")
          event <- data.frame(event)
      }
      if (!is.data.frame(event)){
          if (is.null(id)){
              model <- "onejump"
              if (is.logical(event)) event <- as.numeric(event)
              status[is.na(event) | is.infinite(event) | as.character(event)==cens.code] <- 0
          }
          else{
              ## inFormat <- "longitudinal"
              stopifnot(is.numeric(id) || is.factor(id))
              model <- "multi.states"
              if (cens.type=="intervalCensored"){
                  stop("Dont know the order of transitions for interval censored observations.")
              }
              else{
                  if (addInitialState==TRUE){
                      time <- c(rep(0,length(unique(id))),time)
                      if (is.factor(event)){
                          event <- factor(c(rep("initial",length(unique(id))),as.character(event)),levels=c("initial",levels(event)))
                      }
                      else{
                          stopifnot(match("initial",unique(event),nomatch=0)==0)
                          event <- c(rep("initial",length(unique(id))),event)
                      }
                      id <- c(unique(id),id)
                      ## status <- c(rep(cens.code,length(unique(id))),status)
                  }
                  # 1. sort the observations by id and time
                  sorted <- order(id,time)
                  time <- time[sorted]
                  ## status <- status[sorted] consists only of 1's
                  id <- id[sorted]
                  event <- event[sorted]
                  # time <- time[duplicated(id)] ## remove the resp. first time
                  # status <- status[duplicated(id)]
                  if (length(unique(id))!=sum(time==0))
                      stop("There are ",length(unique(id))," different individuals (id's), but the state at time 0 is available for ",sum(time==0)," id's.")
                  initialState <- event[time==0]
                  last.id <- c(diff(id) != 0, 1)
                  first.id <- c(1, diff(id) != 0)
                  from <- factor(event[last.id!=1])
                  to <- factor(event[first.id!=1])
                  id <- id[time!=0]
                  time <- time[time!=0]
                  # 2. get back to the original order
                  ### cannot easily get back since
                  ### length(time) < sorted
                  ## time <- time[sorted]
                  ## id <- id[sorted]
                  ## event <- event[sorted]
                  status <- rep(1,length(to))
                  status[is.na(to) | is.infinite(to) | as.character(to)==cens.code] <- 0
              }
          }
      }
      else{
          ## inFormat <- "from2to"
          model <- "multi.states"
          from <- event[[1]]
          to <- event[[2]]
          status[is.na(to) | is.infinite(to) | as.character(to)==cens.code] <- 0
          if (length(unique(from))==1){
              model <- "onejump"
              event <- to
              if (is.logical(to)) to <- as.numeric(to)
              status[is.na(to) | is.infinite(to) | as.character(event)==cens.code] <- 0
          }
      }
  }
  ## if (all(status==0)) warning("All observations are censored")
  if (all(status==1)) cens.type <- "uncensored"
  
    if(model=="onejump"){
        # }}}
        # {{{  2-state and competing.risks models
        if (is.factor(event)){
            # event <- factor(event) # drop unused levels
            states <- levels(event)
            ## states <- states[match(state.order,states)]
        }
        else{
            states <- sort(as.character(unique(event)))
        }
        states <- as.character(states[states!=cens.code])
        if (length(states)>1)
            model <- "competing.risks"
        else
            model <- "survival"
        if (cens.type=="intervalCensored"){
            if (model=="survival"){
                if (entry.type=="intervalCensored")
                    history <- cbind(U=U,V=V,L=L,R=R,status=status)
                else
                    history <- cbind(entry = entry,L=L,R=R,status=status)
            }
            else{
                if (entry.type=="intervalCensored")
                    history <- cbind(U=U,
                                     V=V,
                                     L=L,
                                     R=R,
                                     status=status,
                                     event=as.integer(factor(event,levels=c(states,cens.code))))
                else
                    history <- cbind(entry = entry,
                                     L=L,
                                     R=R,
                                     status=status,
                                     event=as.integer(factor(event,levels=c(states,cens.code))))
            }
        } else{
            if (model=="survival"){
                if (entry.type=="intervalCensored")
                    history <- cbind(U=U,V=V,time=time,status=status)
                else
                    history <- cbind(entry = entry,time=time,status=status)
            }
            else{
                if (entry.type=="intervalCensored")
                    history <- cbind(U=U,
                                     V=V,
                                     time=time,
                                     status=status,
                                     event=as.integer(factor(event,levels=c(states,cens.code))))
                else{
                    history <- cbind(entry = entry,
                                     time=time,
                                     status=status,
                                     event=as.integer(factor(event,levels=c(states,cens.code))))
                }
            }
        }
    } else{
        # }}}
        # {{{  multi.state models

    if (any(as.character(from)==as.character(to))) stop("Data contain transitions from state x to state x")

    eventISfactor <- as.numeric(is.factor(from)) + as.numeric(is.factor(to))

    if (eventISfactor==1) stop("Components of event have different classes")
    
    if (eventISfactor==2)
      states <- unique(c(levels(from),levels(to)))
    else
      states <- as.character(unique(c(from,to)))
    states <- as.character(states[states!=cens.code])
    ## states <- states[match(state.order,states)]
    if (cens.code %in% levels(from)){
      stop(paste("The Cens.code",
                 cens.code,
                 " identifies censored data, but is found amoung the `from' state of some transitions"))
    } 
    if (cens.type=="intervalCensored"){
      if (entry.type=="intervalCensored")
          history <- cbind(U=U,
                           V=V,
                           L=L,
                           R=R,
                           status=status,
                           from=as.integer(factor(from,levels=c(states,cens.code))),
                           to=as.integer(factor(to,levels=c(states,cens.code))))
      else{
          history <- cbind(entry = entry,
                           L=L,
                           R=R,
                           status=status,
                           from=as.integer(factor(from,levels=c(states,cens.code))),
                           to=as.integer(factor(to,levels=c(states,cens.code))))
      }
    }
    else{
      if (entry.type=="intervalCensored")
          history <- cbind(U=U,
                           V=V,
                           time=time,
                           status=status,
                           from=as.integer(factor(from,levels=c(states,cens.code))),
                           to=as.integer(factor(to,levels=c(states,cens.code))))
      else{
          history <- cbind(entry = entry,
                           time=time,
                           status=status,
                           from=as.integer(factor(from,levels=c(states,cens.code))),
                           to=as.integer(factor(to,levels=c(states,cens.code))))
      }
    }
  }

  # }}}
  # {{{ add id

  if (!is.null(id)) history <- cbind(history,id)
  
  # }}}
  # {{{ class and attributes
  rownames(history) <- NULL
  class(history) <- c("Hist")
  attr(history,"states") <- states
  attr(history,"cens.type") <- cens.type
  attr(history,"cens.code") <- as.character(cens.code)
  attr(history,"model") <- model
  ## print(entry.type)
  attr(history,"entry.type") <- entry.type
  history
  # }}}
}


#' Subset an event history object
#'
#' This method makes sure that the attributes are
#' preserved
#' @export subset.Hist
#' @param x object obtained with \code{\link{Hist}}
#' @param subset passed on to next method
#' @param select passed on to next method
#' @param drop passed on to next method
#' @param ... passed on to the next method
subset.Hist <- function(x,subset,select,drop,...){
  if (missing(select)){
    xx <- x
    class(xx) <- "matrix"
    xx <- subset(xx,subset=subset,drop=drop)
    attr(xx,"class") <- attr(x,"class")
    attr(xx,"states") <- attr(x,"states")
    attr(xx,"model") <- attr(x,"model")
    attr(xx,"cens.type") <- attr(x,"cens.type")
    attr(xx,"cens.code") <- attr(x,"cens.code")
    attr(xx,"entry.type") <- attr(x,"entry.type")
    xx
  }
  else{
    class(x) <- "matrix"
    NextMethod("subset")
  }
}

#' Subset an event history object
#'
#' This method makes sure that the attributes are
#' preserved
#' @param x object obtained with \code{\link{Hist}}
#' @param i passed on to next method
#' @param j passed on to next method
#' @param drop passed on to next method
#' @export '[.Hist'
"[.Hist" <- function(x,i,j,drop=FALSE){
  if (missing(j)){
    xx <- x
    class(xx) <- "matrix"
    xx <- xx[i,,drop=drop]
    class(xx) <- "Hist"    
    attr(xx,"class") <- attr(x,"class")
    attr(xx,"states") <- attr(x,"states")
    attr(xx,"model") <- attr(x,"model")
    attr(xx,"cens.type") <- attr(x,"cens.type")
    attr(xx,"cens.code") <- attr(x,"cens.code")
    attr(xx,"entry.type") <- attr(x,"entry.type")
    xx
  }
  else{
    class(x) <- "matrix"
    ## x[i,j,drop=drop]
    NextMethod("[")
  }
}

#' Check for missing values in event history object
#'
#' The object has usually two or three columns and
#' this function checks each column separately
#' @export is.na.Hist
#' @param x object obtained with \code{\link{Hist}} 
is.na.Hist <- function(x) {
  as.vector( (1* is.na(unclass(x)))%*% rep(1, ncol(x)) >0)
}

#' Check the structure of an event history object
#'
#' The object is evaluated as a matrix
#' @export str.Hist
#' @param x object obtained with \code{\link{Hist}} 
str.Hist <- function(x){
  class(x) <- "matrix"
  utils::str(x)
}

#' Check the head of an event history object
#'
#' The object is evaluated as a matrix
#' @export head.Hist
#' @param x object obtained with \code{\link{Hist}} 
head.Hist <- function(x){
  class(x) <- "matrix"
  utils::head(x)
}

#' Check the tail of an event history object
#'
#' The object is evaluated as a matrix
#' @export tail.Hist
#' @param x object obtained with \code{\link{Hist}} 
tail.Hist <- function(x){
  class(x) <- "matrix"
  utils::tail(x)
}

#' Index for evaluation of step functions.
#' 
#' Returns an index of positions. Intended for evaluating a step function at
#' selected times. The function counts how many elements of a vector, e.g. the
#' jump times of the step function, are smaller or equal to the elements in a
#' second vector, e.g. the times where the step function should be evaluated.
#' 
#' If all \code{jump.times} are greater than a particular \code{eval.time} the
#' sindex returns \code{0}. This must be considered when sindex is used for
#' subsetting, see the Examples below.
#' 
#' @param jump.times Numeric vector: e.g. the unique jump times of a step
#' function.
#' @param eval.times Numeric vector: e.g. the times where the step function
#' should be evaluated
#' @param strict If TRUE make the comparison of jump times and eval times
#' strict
#' @param comp If "greater" count the number of jump times that are greater
#' (greater or equal when strict==FALSE) than the eval times
#' @return Index of the same length as \code{eval.times} containing the numbers
#' of the \code{jump.times} that are smaller than or equal to
#' \code{eval.times}.
#' @author Thomas A. Gerds \email{tag@@biostat.ku.dk}
#' @keywords misc
#' @examples
#' 
#' 
#' test <- list(time = c(1, 1,5,5,2,7,9),
#' 		status = c(1,0,1,0,1,1,0))
#' fit <- prodlim(Hist(time,status)~1,data=test)
#' jtimes <- fit$time
#' etimes <- c(0,.5,2,8,10)
#' fit$surv
#' c(1,fit$surv)[1+sindex(jtimes,etimes)]
#' 
#' @export
"sindex" <- function(jump.times,eval.times,comp="smaller",strict=FALSE) {
  stopifnot(is.numeric(jump.times))
  stopifnot(is.numeric(eval.times))
  N <- length(jump.times)
  if (comp=="greater"){
    N-sindex(jump.times=jump.times,
             eval.times=eval.times,
             comp="smaller",
             strict=!strict)
  }
  else{
    neval <- length(eval.times)
    if (!(neval> 0 && N >0)) stop("missing data")
    new.order <- order(eval.times)
    ind <- .C("sindexSRC",index = integer(neval),as.double(sort(jump.times)),as.double(eval.times[new.order]),as.integer(N),as.integer(neval),as.integer(strict),PACKAGE="prodlim")$index
    ind[order(new.order)]
  }
}

## sindexStrata <- function(jump.times,
                         ## first,
                         ## size,
                         ## eval.times,
                         ## strict=FALSE) {
  ## stopifnot(is.numeric(jump.times))
  ## stopifnot(is.numeric(eval.times))
  ## NK <- length(size)
  ## stopifnot(length(first)==NK)
  ## N <- length(jump.times)
  ## neval <- length(eval.times)
  ## if (!(neval> 0 && N >0)) stop("missing data")
  ## new.order <- order(eval.times)
  ## ind <- .C("sindexStrata",
            ## index = integer(neval),
            ## as.double(sort(jump.times)),
            ## as.double(eval.times[new.order]),
            ## as.integer(N),
            ## as.integer(neval),
            ## as.integer(strict),
            ## DUP=FALSE,
            ## PACKAGE="prodlim")$index
  ## ind[order(new.order)]
## }

model.specials <- function(data,specials,allowInteractions=FALSE){
    names(specials) <- specials
    Terms <- attr(data,"terms")
    spec <- lapply(specials,function(sp){
        if (length(attr(Terms,"specials")[[sp]])){
            untangle <- function (tt, special, order = 1) {
                spc <- attr(tt, "specials")[[special]]
                if (length(spc) == 0) 
                    return(list(vars = character(0), terms = numeric(0)))
                facs <- attr(tt, "factor")
                fname <- dimnames(facs)
                ff <- apply(facs[spc, , drop = FALSE], 2, sum)
                list(vars = (fname[[1]])[spc], terms = seq(ff)[ff & match(attr(tt,"order"), order, nomatch = 0)])
            }
            untangle(Terms,sp,1:10)
        }
        else NULL
    })

  # -------------------------check interactions-------------------------
  
  if (allowInteractions==FALSE){
    lapply(spec[sapply(spec,length)>0],function(sp){
      ord <- attr(Terms, "order")[sp$terms]
      if (any(ord > 1))
        stop(paste(sp," can not be used in an interaction"),call.=FALSE)})
  }
  
    special.frame <- lapply(spec,function(sp){
        if (length(sp)) {
            sp.frame <- data[,sp$vars,drop=FALSE]
            names(sp.frame) <- extract.name.from.special(names(sp.frame))
            sp.frame
        }
        else NULL})
    all.varnames <- all.vars(delete.response(Terms))
    unspecified <- all.varnames[!(all.varnames %in% unlist(lapply(special.frame,names)))]
    special.frame$unspecified <- data[,unspecified,drop=FALSE]
    special.frame
}

#' Extract a column from an event history object.
#' 
#' Extract a column from an event history object, as obtained with the function
#' \code{\link{Hist}}.
#' 
#' Since objects of class \code{"Hist"} are also matrices, all columns are
#' numeric or integer valued. To extract a correctly labeled version, the
#' attribute \code{states} of the object is used to generate factor levels.
#' 
#' @aliases getEvent 
#' @param object Object of class \code{"Hist"}.
#' @param mode Return mode. One of \code{"numeric"}, \code{"character"}, or
#' \code{"factor"}.
#' @param column Name of the column to extract from the object.
#' @author Thomas Alexander Gerds <tag@@biostat.ku.dk>
#' @seealso \code{\link{Hist}}
#' @keywords survival
#' @examples
#' 
#'   dat= data.frame(time=1:5,event=letters[1:5])
#'   x=with(dat,Hist(time,event))
#'   ## inside integer
#'   unclass(x)
#'   ## extract event (the extra level "unknown" is for censored data)
#'   getEvent(x)
#'
#' @export
getEvent <- function(object,mode="factor",column="event"){
    model <- attr(object,"model")
    if (model=="multi.state")
        stop("Dont know how to extract events from a multi.state model")
    ## cens.code <- attr(object,"cens.code")
    states <- attr(object,"states")
    if (match(column,colnames(object),nomatch=0)==0){
        warning("Object '", class(object),"' does not have this element: ",column,". Returning NULL.")
        return(NULL)
    }
    else{
        E <- factor(as.vector(object[,column]),
                    levels=1:(length(states)+1),
                    labels=c(as.character(states),"unknown"))
        switch(mode,"character"=as.character(E),"numeric"=as.numeric(E),E)
    }
}

#' Plotting an illness-death-model.
#' 
#' Plotting an illness-death-model using \code{plot.Hist}.
#' 
#' 
#' @param stateLabels Labels for the three boxes.
#' @param style Either \code{1} or anything else, switches the orientation of
#' the graph. Hard to explain in words, see examples.
#' @param recovery Logical. If \code{TRUE} there will be an arrow from the
#' illness state to the initial state.
#' @param \dots Arguments passed to plot.Hist.
#' @author Thomas Alexander Gerds <tag@@biostat.ku.dk>
#' @seealso \code{\link{plotCompetingRiskModel}}, \code{\link{plot.Hist}}
#' @keywords survival
##' @examples
##' 
##' plotIllnessDeathModel()
##' plotIllnessDeathModel(style=2)
##' plotIllnessDeathModel(style=2,
##'                       stateLabels=c("a","b\nc","d"),
##'                       box1.col="yellow",
##'                       box2.col="green",
##'                       box3.col="red")
 
#' @export
plotIllnessDeathModel <- function(stateLabels,
                                  style=1,
                                  recovery=FALSE,
                                  ...){
  if (missing(stateLabels)) labels <- c("Disease\nfree","Illness","Death")
  if (recovery==TRUE){
    idHist <- Hist(time=1:4,event=list(from=c(1,1,2,2),to=c(2,3,1,3)))
    if (style==1)
      plot(idHist,
           stateLabels=stateLabels,
           box1.row=2,
           box1.column=1,
           box2.row=1,
           box2.column=3,
           ...)
    else{
      plot(idHist,
           stateLabels=stateLabels,
           ...)
    }
  }
  else{
    idHist <- Hist(time=1:3,event=list(from=c(1,1,2),to=c(2,3,3)))
    if (style==1){
      plot(idHist,
           stateLabels=stateLabels,
           box1.row=2,
           box1.column=1,
           box2.row=1,
           box2.column=3,
           ...)
    }
    else{
      plot(idHist,
           stateLabels=stateLabels,
           ...)
    }
  }
}

## Notation
##    subject specific intervals
##                number: N
##         running index: i
##    support (Peto) intervals
##                number: M
##         running index: m
PetoInt<-function(L,R,status){ 
  #Status: 0 right censored, 1 exact time, 2 interval cencored.
  #R[status==0] <- max(R)+1 #to ensure a right endpoint.
  #it is outcomented because this is done in compGMLE...R instead.
  
  names(L)[status!=1] <- 'L'
  names(R)[status!=1] <- 'R'
  names(L)[status==1] <- 'EL'
  names(R)[status==1] <- 'ER'
  peto.intervals <- c(L,R)
  level.int <- factor(names(peto.intervals),levels=c('R','EL','ER','L'))
  right.order <- order(peto.intervals,level.int)
  peto.intervals <- peto.intervals[right.order]
  tmp1 <- as.numeric(factor(names(peto.intervals), levels=c('R','EL','ER','L')))
  int <- grep('^-3$', diff(tmp1)) #finds the intervals
  tmp2 <- as.numeric(factor(names(peto.intervals), levels=c('EL','R','L','ER')))
  exa <- grep('^3$', diff(tmp2)) #finds the exact observations
  obs.no <- c(int,exa)
  tmp <- peto.intervals[sort(c(obs.no,obs.no+1))]
  out  <-  matrix(tmp,nrow=2)
  out
}

### List2Matrix.R --- 
#----------------------------------------------------------------------
## author: Thomas Alexander Gerds
## created: Sep 21 2015 (07:01) 
## Version: 
## last-updated: Sep 29 2015 (06:32) 
##           By: Thomas Alexander Gerds
##     Update #: 6
#----------------------------------------------------------------------
## 
### Commentary: Reduce a list to a matrix or data.frame and add list names as new columns
## 
### Change Log:
#----------------------------------------------------------------------
## 
### Code:
##' This function is used by summary.prodlim to deal with results.
##'
##' Reduction is done with rbind.
##' @title Reduce list to a matrix or data.frame with names as new columns
##' @param list A named list which contains nested lists 
##' @param depth The depth in the list hierarchy until an rbindable object
##' @param names Names for the list variables
##' @return Matrix or data.frame.
##' @examples
##' 
##' x=list(a=data.frame(u=1,b=2,c=3),b=data.frame(u=3,b=4,c=6))
##' List2Matrix(x,depth=1,"X")
##' @export
##' @author Thomas A. Gerds <tag@@biostat.ku.dk>
List2Matrix <- function(list,depth,names){
    if (missing(names)) names <- paste0("D",1:depth)
    switch(as.character(depth),
           "1"={
               dims <- lapply(list,dim)
               cols <- sapply(dims,function(x)x[[2]])
               rows <- sapply(dims,function(x)x[[1]])
               stopifnot(length(unique(cols))==1)
               nl <- names(list)
               M <- do.call("rbind",list)
               rownames(M) <- NULL
               M <- cbind(rep(nl,rows),M)
               colnames(M)[1] <- names[1]
               M},
           "2"={
               List2Matrix(lapply(list,List2Matrix,depth=1,names=names[2]),
                           depth=1,
                           names=names[1])},
           "3"={
               List2Matrix(lapply(list,function(l){
                                      List2Matrix(lapply(l,List2Matrix,depth=1,names[3]),
                                                  depth=1,
                                                  names=names[2])
                                  }), depth=1,names=names[1])},
           stop("Cannot do this depth."))
}
#----------------------------------------------------------------------
### List2Matrix.R ends here

prodlimMulti <- function(response,size.strata,N,NU,cotype,force.multistate){
  ##   original function by Matthias `Wang' Wangler 
  is.event <- response[,"status"]!=0
  if (force.multistate==TRUE){
    to    <- response[,"status"]
    from <- rep(0,length(to))
  }
  else{
    to    <- response[,"event"]
    from  <- response[,"from"]
  }
  state.names <- unique(c(from, to[response[,"status"]!=0]))
  ns <- length(state.names)
  cens <- FALSE
  if(length(to[is.event])>0) cens <- TRUE
  from <- as.integer(factor(from,levels=state.names)) - 1
  from <- as.numeric(from)
  to[is.event] <- as.integer(factor(to[is.event], levels=state.names)) - 1
  to[!is.event] <- ns
  to <- as.numeric(to)
  states <- sort(unique(c(from, to[is.event])))   
  ## possible transitions
  tra <- unique(cbind(from[is.event], to[is.event]))
  sorted <- order(tra[,1],tra[,2])    
  tra <- matrix(tra[sorted,], ncol=2)
  tra <- cbind(0:(length(tra[,1])-1),tra)
  colnames(tra) <- c("row","from", "to")
  ntra <- nrow(tra)
  trow <- match(paste(from,to), paste(tra[,"from"],tra[,"to"]), nomatch=0) - 1
  cens.in <- sort(unique(from[!is.event]))
  nci <- length(cens.in)
  cpos <- match(paste(from,to), paste(cens.in, ns), nomatch = 0) - 1
  ## start distribution (all are starting in state 0 !!!)
  if( cotype > 1 ) {
    #    nr.start <- table(from,co$covariates$strata[,1])[1,]
    nr.start <- size.strata  ## WANG???
  }
  else{nr.start <- length(from[from==0])}
  fit <- .C("prodlim_multistates",
            as.integer(N),
            as.integer(ns),
            as.integer(length(is.event)),
            as.integer(size.strata),
            as.integer(ntra),
            as.integer(tra[,"from"]),
            as.integer(tra[,"to"]),
            as.integer(trow),
            as.integer(nci),
            as.integer(cens.in),
            as.integer(cpos),
            as.double(response[,"time"]),
            as.integer(response[,"status"]),
            as.integer(nr.start),
            time=double(N),
            hazard=double(N*ns*ns),
            prob=double(N*ns*ns),
            nevent=integer(N*ns*ns),
            ncens=integer(N*ns),
            nrisk=integer(N*ns),
            first.strata=integer(NU),
            ntimes.strata=integer(NU),
            PACKAGE="prodlim")
  tra[,"from"] <- state.names[tra[,"from"]+1]
  tra[,"to"] <- state.names[tra[,"to"]+1]
  cens.in <- state.names[cens.in+1]
  NT <- sum(fit$ntimes.strata)
  res <- list("time"=fit$time[1:NT],"hazard"=fit$hazard[1:(NT*ns*ns)],"prob"=fit$prob[1:(NT*ns*ns)],"nevent"=fit$nevent[1:(NT*ns*ns)],"ncens"=fit$ncens[1:(NT*ns)],"nrisk"=nrisk <- fit$nrisk[1:(NT*ns)],"first.strata"=fit$first.strata,"size.strata"=fit$ntimes.strata,"uniquetrans"=tra,"cens.in"=cens.in,"states"=states,"state.names"=state.names,"model"="multi.states")
  res
}

## Notation
##    subject specific intervals
##                number: N
##         running index: i
##    support (Peto) intervals
##                number: M
##         running index: m
IntIndex  <-  function(x,L,R){
  N <- length(L)
  M <- NCOL(x)
  p <- x[1,]
  q <- x[2,]
  res <- .C('IntIndexSRC',as.double(L),as.double(R),as.double(p),as.double(q),as.integer(N),as.integer(M),Iindex=integer(N*M),Mindex=integer(N*M),Istrata=integer(N),Mstrata=integer(M))
  Iindex <- res$Iindex[res$Iindex!=0]
  Istrata <- res$Istrata#[res$Istrata!=0]
  Mindex <- res$Mindex[res$Mindex!=0]
  Mstrata <- res$Mstrata#[res$Mstrata!=0]
  out <- list(Mindex,Mstrata,Iindex,Istrata,rbind(L,R),x)
  names(out) <- c("Mindex","Mstrata","Iindex","Istrata","obsInt","petoInt")
  #class(out) <- "IntIndex"
  out
}

## old version

## IntIndex  <-  function(x,L,R){
##   N <- length(L)
##   M <- NCOL(x)
##   p <- x[1,]
##   q <- x[2,]
##   res <- .C('IntIndex',as.double(L),as.double(R),as.double(p),as.double(q),as.integer(N),as.integer(M),Iindex=integer(N*M),Mindex=integer(N*M),Istrata=integer(N),Mstrata=integer(M),PACKAGE="prodlim")
##   Iindex <- res$Iindex[res$Iindex!=0]
##   Istrata <- res$Istrata[res$Istrata!=0]
##   Mindex <- res$Mindex[res$Mindex!=0]
##   Mstrata <- res$Mstrata[res$Mstrata!=0]
##   out <- list(Mindex,Mstrata,Iindex,Istrata,rbind(L,R),x)
##   names(out) <- c("Mindex","Mstrata","Iindex","Istrata","obsInt","petoInt")
##   class(out) <- "IntIndex"
##   out
## }

stat_prodlim_setup_data <- function(data, params) {
    data <- data.table::as.data.table(data)
    data
}

stat_prodlim_compute_panel <- function(data, scales,
                                       type = "risk",
                                       cause = NULL,
                                       conf_int = FALSE,
                                       percent = TRUE,
                                       times = NULL,
                                       timeconverter = NULL,
                                       cens.code = "0",
                                       conf_int_alpha = 0.2) {
    time <- absolute_risk <- x <- Event <- alpha_ci <- group <- component <- y <- NULL
    if (nrow(data) == 0L) return(data.frame())

    if (!"event" %in% names(data)) {
      stop("stat_prodlim: need 'event' variable.")
    }
    if (is.logical(data$event)) {
      stop("geom_prodlim(): `event` must be coded as numeric/integer or factor/character, not logical.")
    }

    data <- data.table::data.table(data)

    ignore_formula <- c(
      "x", "y", "xmin", "xmax", "ymin", "ymax",
      "PANEL", "group", "event", "alpha", "size",
      "linewidth",
      "component", "Event", "stacked", "alpha_ci"
    )

    model_cols <- setdiff(names(data), ignore_formula)

    ff <- stats::formula(
      paste0(
        "prodlim::Hist(x, event, cens.code = '", cens.code, "') ~ ",
        if (length(model_cols) == 0) "1" else paste(model_cols, collapse = "+")
      )
    )

    fit <- prodlim::prodlim(formula = ff, data = data, type = type)

    summary_args <- list(
      x = fit,
      surv = identical(type, "surv"),
      percent = percent
    )

    stacked <- identical(cause, "stacked")
    multi_cause <- FALSE
    if (attr(fit$model.response, "model") == "competing.risks") {
        if (stacked || length(cause) == 0) {
            if (stacked){
                requested_causes <- attr(fit$model.response, "states")
            } else{
                requested_causes <- attr(fit$model.response, "states")[[1]]
            }
        } else {
            cause <- intersect(cause, attr(fit$model.response, "states"))
            if (length(cause) == 0) {
                stop(paste0("Cannot find cause(s) ", paste0(cause,collapse = ", "),". The values of data$event are ",paste0(attr(fit$model.response, "states"), collapse = ", ")))
            }
            requested_causes <- cause
        }
        summary_args$cause <- requested_causes
    } else {
        requested_causes <- 1
        stacked <- FALSE
    }

    if (length(times) > 0) summary_args$times <- times

    w <- do.call(data.table::as.data.table, summary_args)
    if (!("cause" %in% names(w))) {
        data.table::set(w, j = "cause", value = factor(rep(1, nrow(w))))
    } else {
        multi_cause <- data.table::uniqueN(w$cause) > 1L
        if (!multi_cause) {
            w[, cause := factor(requested_causes)]
        } else {
            data.table::setorderv(w, c(if (length(model_cols)) model_cols else NULL, "cause", "time"))
            w[, cause := factor(cause, levels = requested_causes)]
        }
    }

    if ("cuminc" %in% names(w) && !"absolute_risk" %in% names(w)) {
      data.table::setnames(w, "cuminc", "absolute_risk")
    }

    if (!is.null(timeconverter)) {
      conv <- c(
        days2years = 1 / 365.25,
        months2years = 1 / 12,
        days2months = 1 / 30.4368499,
        years2days = 365.25,
        years2months = 12,
        months2days = 30.4368499
      )
      if (!timeconverter %in% names(conv)) {
        stop("stat_prodlim: unknown timeconverter")
      }
      data.table::set(w, j = "time", value = w$time * unname(conv[timeconverter]))
    }

    y_name <- intersect(c("surv", "absolute_risk"), names(w))

    by_cols <- if (isTRUE(multi_cause)) c(model_cols, "cause") else model_cols
    data.table::setorderv(w, c(by_cols, "time"))

    if (length(by_cols) == 0) {
        data.table::set(w, j = "Event", value = factor(rep("1", nrow(w))))
    } else {
        if (identical(by_cols[[1]], "cause")) {
            data.table::set(w, j = "Event", value = factor(w[["cause"]], levels = requested_causes))
        } else {
            curve_labels <- as.character(w[[by_cols[[1]]]])
            if (length(by_cols) > 1) {
                for (j in 2:length(by_cols)) {
                    curve_labels <- paste(curve_labels, w[[by_cols[[j]]]], sep = ", ")
                }
            }
            data.table::set(w, j = "Event", value = factor(curve_labels))
        }
    }

    if (isTRUE(stacked) && isTRUE(multi_cause)) {
        if (length(model_cols) == 0) {
            stack_dt <- data.table::copy(w)
            data.table::setorder(stack_dt, cause, time)
            stack_dt[, xmax := data.table::shift(time, type = "lead"),by = cause]
            data.table::setorder(stack_dt, time, cause)
            stack_dt[, ymin := data.table::shift(cumsum(absolute_risk), fill=0), by=time]
            stack_dt[, ymax := cumsum(absolute_risk), by=time]
            data.table::setnames(stack_dt, "time", "xmin")
        } else {
            stop("Cannot split stacked plot.")
        }
        stack_dt <- stack_dt[!is.na(xmax)]
        stack_dt[, x := NA_real_]
        stack_dt[, y := NA_real_]
        stack_dt[, component := "stack"]
        stack_dt[, group := as.integer(Event)]
        stack_dt[, alpha_ci := 1]
        return(as.data.frame(stack_dt))
    }
    step_dt <- data.table::copy(w)
    data.table::setnames(step_dt, y_name, "y")
    data.table::setnames(step_dt, "time", "x")
    step_dt[, component := "step"]
    step_dt[, xmin := NA_real_]
    step_dt[, xmax := NA_real_]
    step_dt[, ymin := NA_real_]
    step_dt[, ymax := NA_real_]
    step_dt[, group := as.integer(Event)]
    step_dt[, alpha_ci := conf_int_alpha]

    if (!isTRUE(conf_int)) {
      return(as.data.frame(step_dt))
    }

    ci_dt <- data.table::copy(w)
    if (length(model_cols) == 0) {
      ci_dt[, xmax := data.table::shift(time, type = "lead"), by = "cause"]
    } else {
      ci_dt[, xmax := data.table::shift(time, type = "lead"), by = c(model_cols, "cause")]
    }
    ci_dt <- ci_dt[!is.na(xmax)]
    ci_dt[, xmin := time]
    data.table::setnames(ci_dt, "lower", "ymin")
    data.table::setnames(ci_dt, "upper", "ymax")
    ci_dt[, x := NA_real_]
    ci_dt[, y := NA_real_]
    ci_dt[, component := "ci"]
    ci_dt[, group := as.integer(Event)]
    ci_dt[, alpha_ci := conf_int_alpha]

    out <- data.table::rbindlist(
      list(step_dt, ci_dt),
      use.names = TRUE,
      fill = TRUE
    )
    as.data.frame(out)
}

StatProdlim <- ggplot2::ggproto(
    "StatProdlim",
    ggplot2::Stat,
    required_aes = c("x", "event"),
    extra_params = c("na.rm",
                     "type",
                     "cause",
                     "conf_int",
                     "percent",
                     "times",
                     "timeconverter",
                     "cens.code",
                     "conf_int_alpha"
    ),
    setup_data = stat_prodlim_setup_data,
    compute_panel = stat_prodlim_compute_panel
)

##' @export 
"print.neighborhood" <- function(x,...){
    n <- x$n
    size <- x$size.nbh
    bw <- lapply(x$bandwidth,function(bw)round(bw,3))
    cat("Nearest neighborhoods for kernel smoothing\n\n")
    print(c(bandwidth=as.numeric(bw),kernel=x$kernel,n.obs=x$n,n.values=x$nu),quote=FALSE)
    cat("\n")
    print(c("Number of nbh's" = length(size),
            "Average size"=round(mean(size)),
            "Min size"=round(min(size)),
            "Max size"=round(max(size))))
    #  if (print.it) print(data.frame(Nbh=x$values,First=x$first.nbh,Size=size))
    invisible(x)
}

#' Compute leave-one-out estimates
#'
#' This function is the work-horse for \code{jackknife}
#' @title Compute jackknife pseudo values.
#' @aliases leaveOneOut leaveOneOut.survival leaveOneOut.competing.risks 
#' @author Thomas Alexander Gerds <tag@@biostat.ku.dk>
#' @seealso \code{\link{jackknife}}
#' 
#' @param object Object of class \code{"prodlim"}.
#' @param times time points at which to compute leave-one-out
#' event/survival probabilities.
#' @param cause Character (other classes are converted with \code{as.character}).
#' For competing risks the cause of interest. 
#' @param lag For survival models only. If \code{TRUE} lag the result, i.e. compute
#' S(t-) instead of S(t).
#' @param ... not used
#' @export
leaveOneOut <- function(object,times,cause,lag=FALSE,...){
    if (missing(cause)) cause <- attr(object$model.response,which="states")[[1]]
    else
        cause <- checkCauses(cause,object)
    if (object$model=="survival")
        leaveOneOut.survival(object=object,times=times,lag=lag,...)
    else if (object$model=="competing.risks")
        leaveOneOut.competing.risks(object=object,times=times,cause=cause,...)
    else stop("No method for jackknifing this object.")
}

#' @export 
leaveOneOut.survival <- function(object,times,lag=FALSE,...){
    stopifnot(object$covariate.type==1)
    mr <- object$model.response
    time <- object$time
    Y <- object$n.risk
    D <- object$n.event
    Y <- Y[D>0]
    time <- time[D>0]
    D <- D[D>0]
    NU <- length(time)
    obstimes <- mr[,"time"]
    status <- mr[,"status"]
    N <- length(obstimes)
    S <- predict(object,times=time,newdata=mr)
    ## idea: find the at-risk set for pseudo-value k by
    ##       subtracting 1 in the period where subj k is
    ##       at risk. need the position of obstime.k in time ...
    found <- sindex(jump.times=time,eval.times=times)
    loo <- .C("loo_surv",
              Y = as.double(Y),
              D=as.double(D),
              time=as.double(time),
              obsT=as.double(obstimes),
              status=as.double(status),
              S=double(NU),
              loo=double(N*length(times)),
              N=as.integer(N),
              NT=as.integer(NU),
              NP=as.integer(length(times)),
              pos=as.integer(found),
              lag=as.integer(lag),
              PACKAGE="prodlim")$loo
    matrix(loo,ncol=length(times),byrow=FALSE)
}

#' @export 
leaveOneOut.competing.risks <- function(object,times,cause,...){
    stopifnot(object$covariate.type==1)
    mr <- object$model.response
    states <- attr(mr,"states")
    if (missing(cause)) {
        C <- 1
        cause <- states[1]
    }
    else{
        C <- match(cause,states,nomatch=0)
        if (length(C)>1 || C==0) stop("Cause must match exactly one of the names of object$n.event.")
    }
    D <- object$n.event[[C]]
    Dall <- Reduce("+",object$n.event)
    #  it is sufficient to consider time points where events occur
    time <- object$time[Dall>0]
    Y <- object$n.risk[Dall>0]
    D <- D[Dall>0]
    Dall <- Dall[Dall>0]
    NU <- length(time)
    obstimes <- mr[,"time"]
    status <- mr[,"status"]
    E <- getEvent(mr)
    N <- length(obstimes)
    ## idea: see leaveOneOut.survival
    found <- sindex(jump.times=time,eval.times=times)
    ## print(cbind(D,Dall,time,Y))
    loo <- .C("loo_comprisk",
              Y = as.double(Y),
              D=as.double(D),
              Dall=as.double(Dall),
              time=as.double(time),
              obsT=as.double(obstimes),
              status=as.double(status),
              event=as.double(status*(E==cause)),
              S=double(NU),
              F=double(NU),
              loo=double(N*length(times)),
              N=as.integer(N),
              NT=as.integer(NU),
              NP=as.integer(length(times)),
              pos=as.integer(found),
              PACKAGE="prodlim")$loo
    matrix(loo,ncol=length(times),byrow=FALSE)
}

##' @export 
print.summary.prodlim <- function(x,digits,...){
    if (!is.null(attr(x,"model"))){
        if (missing(digits))
            digits <- ifelse(attr(x,"percent"),1,3)
        print.data.frame(x,digits=digits,...)
    } else{ # old behaviour from before May 2021
        model <- x$model
        cotype <- x$cotype
        sumtable <- x$table
        if (missing(digits))
            digits <- ifelse(x$percent,1,3)
        if (model=="survival"){
            if (cotype==1){
                print(sumtable,digits=digits,quote=FALSE,...)
            } else{
                print.listof(sumtable,digits=digits,quote=FALSE,...)
            }
        } else{
            if (model=="competing.risks"){
                for (cc in 1:length(sumtable)){
                    cat("\n\n----------> Cause: ",names(sumtable)[cc],"\n\n")
                    if (cotype==1){
                        print(sumtable[[cc]],digits=digits,quote=FALSE,...)
                    }
                    else{
                        print.listof(sumtable[[cc]],digits=digits,quote=FALSE,...)
                    }
                }
            }
        }
    }
}

### redist.R --- 
#----------------------------------------------------------------------
## author: Thomas Alexander Gerds
## created: Nov 28 2015 (10:30) 
## Version: 
## last-updated: Nov 28 2015 (10:35) 
##           By: Thomas Alexander Gerds
##     Update #: 2
#----------------------------------------------------------------------
## 
### Commentary: 
## 
### Change Log:
#----------------------------------------------------------------------
## 
### Code:
##' Calculation of Efron's re-distribution to the right algorithm to obtain the
##' Kaplan-Meier estimate.
##'
#' @param time A numeric vector of event times.
#' @param status The event status vector takes the value \code{1} for observed events and 
#' the value \code{0} for right censored times.  
##' @return Calculations needed to 
##' @seealso prodlim 
##' @examples
##' redist(time=c(.35,0.4,.51,.51,.7,.73),status=c(0,1,1,0,0,1))
##' @export 
##' @author Thomas A. Gerds <tag@@biostat.ku.dk>
redist <- function(time,status){
    library(prodlim)
    cat("\nKaplan-Meier estimate via re-distribution to the right algorithm:\n")
    order <- order(time,-status)
    time <- time[order]
    status <- status[order]
    N <- length(time)
    mass <- as.list(rep(1/N,N))
    fractions <- as.list(rep(paste("1/",N,sep=""),N))
    names(mass) <- paste("subject",1:N)
    for (i in 1:N) names(mass[[i]]) <- "own"
    for (i in 1:N) names(fractions[[i]]) <- "own contribution"
    surv <- 1
    for (i in 1:N) {
        cat("\nSubject ",i,":\n---------------------------\nSurvival before = ",round(surv*100,2),"%\n",sep="")
        if (status[i]==0){
            if (i==N){
                cat("Last subject lost to follow-up event free at time = ",time[i],"\n",sep="")
            }
            else{
                cat("No event until time = ",time[i],"\nRe-distribute mass ",signif(sum(mass[[i]]),2)," to remaining ",N-i,ifelse(N-i==1," subject"," subjects"),"\n",sep="")
                for (j in ((i+1):N)){
                    mass[[j]] <- c(mass[[j]],mass[[i]]/(N-i))
                    fractions[[j]] <- c(fractions[[j]],paste(fractions[[i]],"*1/",(N-i),sep=""))
                    names(fractions[[j]])[length(fractions[[j]])-length(mass[[i]])+1] <- paste("from subject ",i,sep="")
                    names(mass[[j]])[length(mass[[j]])] <- paste("from subject ",i,sep="")
                }
            }
            cat("Survival after = ",round(surv*100,2),"%\n",sep="")
        } else{
            cat("Event at time = ",time[i],"\nContribution to Kaplan-Meier estimate:\n\n",sep="")
            contr <- rbind(fractions[[i]],format(mass[[i]],digits=4,nsmall=4))
            rownames(contr) <- c("fractions","decimal")
            contr <- rbind(t(contr),c("sum",format(sum(mass[[i]]),digits=4,nsmall=4)))
            print(contr,quote=FALSE)
            surv.before <- surv
            surv <- surv-sum(mass[[i]])
            cat("\nSurvival after = ",round(100*surv.before,2),"% - (",paste(fractions[[i]],collapse=" + ") ,")",
                "\n               = ",round(100*surv.before,2),"% - ",round(100*sum(mass[[i]]),2) ,"% = ",round(surv*100,2),"%\n",sep="")
        }
    }
    table <- summary(f <- prodlim(Hist(time,status)~1,data=data.frame(time,status)),times=c(0,time),percent=TRUE)
    cat("\nSummary table:\n\n")
    tab <- table$table[,c("time","n.risk","n.event","n.lost","surv")]
    print(tab)
    out <- list(fit=f,table=tab)
    invisible(out)
}



#----------------------------------------------------------------------
### redist.R ends here

##' @export 
print.quantile.prodlim <- function(x,byvars,digits=2,na.val="--",...){
    if (attr(x,"reverse")==FALSE)
        cat("Quantiles of the event time distribution based on the ",
            ifelse(attr(x,"model")=="survival","Kaplan-Meier","Aalen-Johansen"),
            " method.",sep="")
    else
        cat("Quantiles of the potential follow up time distribution based on the Kaplan-Meier method",
            "\napplied to the censored times reversing the roles of event status and censored.")
    cat("\n")
    cat("\nTable of quantiles and corresponding confidence limits:\n")
    tab <- data.table::data.table(do.call("data.frame",x))
    print(tab,digits=digits)
    if(all(c(0.25,0.5,0.75) %in% tab$q)){
        set(tab,j = "quantile",value = sprintf(paste0("%1.",digits,"f"),tab[["quantile"]]))
        byvars = attr(x,"covariates")
        cat("\n")
        cat("\nMedian with interquartile range (IQR):\n")
        if (attr(x,"model")=="survival"){
            stab <- tab[,{
                data.table::data.table("Median (IQR)" = sprintf(fmt = "%s (%s;%s)",quantile[q==0.5],
                                                                quantile[q==0.75],
                                                                quantile[q==0.25]))
            },by = byvars]
        }
        else{
            stab <- tab[,{
                data.table::data.table("Median (IQR)" = sprintf(fmt = "%s (%s;%s)",quantile[q==0.5],
                                                                quantile[q==0.25],
                                                                quantile[q==0.75]))
            },by = byvars]
        }
        print(stab)
    }else stab <- NULL
    invisible(list(tab,stab))
}

### prodlim-package.R --- 
#----------------------------------------------------------------------
## author: Thomas Alexander Gerds
## created: Apr 24 2015 (09:08) 
## Version: 
## last-updated: mar  6 2026 (08:49) 
##           By: Thomas Alexander Gerds
##     Update #: 31
#----------------------------------------------------------------------
## 
### Commentary: 
## 
### Change Log:
#----------------------------------------------------------------------
## 
### Code:
utils::globalVariables(c(
  "xmin","xmax","ymin","ymax","cause",".gg_kind",".gg_groupid",".gg_colour",".gg_fill"
))
#' Kaplan-Meier and Aalen-Johansen method
#'
#' Fast and user friendly implementation
#' of nonparametric estimators for censored
#' event history (survival) analysis.
#' @keywords internal
# "_PACKAGE"
#' @name prodlim-package
#' @aliases prodlim-package
#' @useDynLib prodlim, .registration=TRUE
#' @importFrom survival survdiff Surv cluster
#' @import data.table
#' @importFrom stats quantile
#' @importFrom grDevices rainbow 
#' @import lava
#' @importFrom Rcpp sourceCpp
## --> importFrom KernSmooth dpik
#' @importFrom graphics grconvertX grconvertY abline axis lines mtext par plot points polygon rect segments strheight strwidth text
#' @importFrom stats .getXlevels delete.response drop.terms formula get_all_vars median model.frame model.matrix model.response na.omit pchisq predict qnorm reformulate terms update update.formula
NULL


#----------------------------------------------------------------------
### prodlim-package.R ends here

### ggprodlim.R --- 
#----------------------------------------------------------------------
## Author: Thomas Alexander Gerds
## Created: Mar  3 2025 (14:32) 
## Version: 
## Last-Updated: mar 10 2026 (14:01) 
##           By: Thomas Alexander Gerds
##     Update #: 428
#----------------------------------------------------------------------
## 
### Commentary: 
## 
### Change Log:
#----------------------------------------------------------------------
## 
### Code:
#' ggplot2 support for Kaplan-Meier and Aalen-Johansen estimators
#'
#' A ggplot2-based alternative to \code{\link{plot.prodlim}} that returns a
#' \code{ggplot} object, enabling ggplot theming, facetting, and composition.
#'
#' @title ggplot2 support for prodlim objects
#' @param x An object obtained with \code{\link{prodlim}}.
#' @param type Either \code{"surv"} or \code{"risk"}. Defaults to \code{x$type}.
#' @param cause For competing risk models: a cause (passed to
#'   \code{\link{as.data.table.prodlim}}), \code{"all"} for all causes, or
#'   \code{"stacked"} for a stacked cumulative incidence plot.
#' @param xlim,ylim Limits for the axes.
#' @param x_breaks,y_breaks Breaks for axes.#
#' @param position_atrisk Numeric vector specifying x-positions where numbers at
#'   risk should be printed. If missing, a default grid is chosen.
#' @param atrisk_title Title shown left of the at-risk table.
#' @param atrisk_labels Optional replacement labels for strata/cause groups in
#'   the at-risk table (in plotting order).
#' @param atrisk_title_line Numeric. Position of the at-risk title below the
#'   plotting region, as an offset from \code{min(ylim)} in units of
#'   \code{diff(ylim)}.
#' @param atrisk_title_step Numeric. Vertical distance from the title to the
#'   first at-risk row, in units of \code{diff(ylim)}.
#' @param atrisk_line_step Numeric. Vertical spacing between at-risk rows, in
#'   units of \code{diff(ylim)}.
#' @param atrisk_bottom_margin Numeric. Bottom plot margin in lines when the
#'   at-risk table is shown.
#' @param facet_panel_spacing_y Optional numeric. Vertical spacing between facet
#'   rows in lines. If \code{NULL} and faceting is used together with an at-risk
#'   table, the default is \code{5 + n_colors}. The same value is used as the
#'   default bottom plot margin below the lowest facet row.
#' @param conf_int Logical. If TRUE and \code{lower}/\code{upper} are available,
#'   draw pointwise confidence intervals.
#' @param conf_int_alpha Alpha for CI shading.
#' @param percent Logical. If TRUE, y axis is shown in percent (0-100).
#' @param timeconverter Optional character, see \code{\link{plot.prodlim}} for
#'   supported values (e.g., "years2months").
#' @param color_var Covariate name must be included in \code{colnames(x$X))} and
#' in competing risk models the word \code{"cause"}.
#' @param facet_formula Formula for facet_wrap. May include covariate names \code{colnames(x$X)} and
#' in competing risk models also the word \code{"cause"}.
#' @param palette Optional vector of colors. If provided, applied via
#'   \code{scale_color_manual}/\code{scale_fill_manual}. Default NULL (do not
#'   impose a palette).
#' @param return_data Logical. If TRUE, return a list with plot and data.
#' @param ... Passed to \code{\link{as.data.table.prodlim}}; commonly \code{newdata},
#'   \code{times}, and for competing risks \code{cause}.
#'
#' @return A ggplot object, or a list if \code{return_data=TRUE}.
#' @seealso \code{\link{plot.prodlim}}, \code{\link{as.data.table.prodlim}}
#'
#' @examples
#' data(Melanoma,package = "riskRegression")
#' # marginal Kaplan-Maier
#' km <- prodlim(Hist(time,status != 0)~1,data = Melanoma)
#' ggprodlim(km)
#' # marginal Aalen-Johansen
#' aj <- prodlim(Hist(time,status)~1,data = Melanoma)
#' ggprodlim(aj)
#' ggprodlim(aj,facet_formula = ~cause)
#'
#' ajsex <- prodlim(Hist(time,status)~sex,data = Melanoma)
#' ggprodlim(ajsex)
#' ggprodlim(ajsex,facet_formula = ~cause)
#' ggprodlim(ajsex,facet_formula = ~sex)
#' ggprodlim(ajsex,color_var = "sex")
#' ggprodlim(ajsex,color_var = "cause")
#'
#' # stacking the causes
#' ggprodlim(ajsex,cause = "stacked")
#' # stacked with two covariates 
#'    ajepisex <- prodlim(Hist(time,status)~sex+epicel,data = Melanoma)
#'    ggprodlim(ajepisex,
#'              cause = "stacked",
#'              facet_formula = sex~epicel,
#'              atrisk_title_line = 0.2)
#'
#' @export
ggprodlim <- function(x,
                      type = x$type,
                      cause,
                      xlim,
                      ylim,
                      y_breaks,
                      x_breaks,
                      position_atrisk,
                      atrisk_title = "Number at risk",
                      atrisk_labels = NULL,
                      atrisk_title_line = 0.14,
                      atrisk_title_step = 0.06,
                      atrisk_line_step = 0.04,
                      atrisk_bottom_margin = NULL,
                      facet_panel_spacing_y = NULL,
                      conf_int = TRUE,
                      conf_int_alpha = 0.2,
                      percent = TRUE,
                      timeconverter,
                      color_var,
                      facet_formula,
                      palette = NULL,
                      return_data = FALSE,
                      ...) {
    stopifnot(inherits(x,"prodlim"))
    # ---- imports & NSE guards
    time <- absolute_risk <- surv <- lower <- upper <- n.risk <- n.lost <- ggprodlim_next_time <- ggprodlim_atrisk_labels <- ggprodlim_curve_id <- ggprodlim_stacked_ymax <- ggprodlim_stacked_ymin <- NULL

    # ---- continuous predictor handling: allow if newdata supplied
    dots <- list(...)
    if ("color_vars" %in% names(dots)){
        stop("ggprodlim: 'color_vars' is not an argument. Change this to 'color_var'.")
    }
    has_newdata <- "newdata" %in% names(dots)
    if (length(x$continuous.predictors) > 0 && !has_newdata) {
        stop("ggprodlim: continuous predictors detected. Please supply 'newdata=' (as in plot.prodlim), or use plot.prodlim().")
    }
    if (has_newdata){
        ggprodlim_newdata <- unique(data.table::as.data.table(dots$newdata))
        dots$newdata <- NULL
        covariates <- data.table::copy(colnames(ggprodlim_newdata))
        has_covariates <- TRUE
    }else{
        covariates <- x$discrete.predictors
        if (length(covariates)>0){
            has_covariates <- TRUE
            ggprodlim_newdata <- data.table::as.data.table(x$X)
        }else{
            has_covariates <- FALSE
        }
    }
    if (!missing(color_var)){
        if (length(color_var) != 1 || !(color_var %in% c(covariates,"cause"))) {
            if (x$model == "competing.risks"){
                stop(paste0("ggprodlim: 'color_var' can be the word 'cause' or must match exactly one of the names of the covariates in the fitted object x$X:\n",
                            paste0(colnames(x$X),collapse = ", ")))
            }else{
                stop(paste0("ggprodlim: 'color_var' must match exactly one of the names covariates in the fitted object x$X:\n",
                            paste0(colnames(x$X),collapse = ", ")))
            }
        }
    }else{
        # may change to 'cause' with multiple causes specified
        color_var <- NULL
    }
    if (!missing(facet_formula)){
        facet_vars <- all.vars(facet_formula)
        if (!all(facet_vars %in% c(covariates,"cause"))){
            stop(paste0("ggprodlim: all variables in facet_formula must be one of the covariates in the fitted object:\n",
                        paste0(colnames(x$X),collapse = ", ")))
        }
    }else{
        facet_vars <- NULL
        facet_formula <- NULL
    }

    # ---- cause handling
    # If user wants "all" causes, we just don't pass 'cause' to as.data.table (so it uses defaults).
    # If user wants "stacked", we need all causes.
    stacked <- FALSE

    if (!missing(cause) && length(cause) > 0) {
        if (identical(cause, "stacked")){
            stacked <- TRUE
            dots$cause <- attr(x$model.response,"states")
        } else {
            if (identical(cause, "all")) {
                dots$cause <- attr(x$model.response,"states")
            }else{
                stopifnot(all(cause %in% attr(x$model.response,"states")))
                dots$cause <- cause
            }
        }
    }
    if (length(dots$cause) == 0){
        requested_causes <- attr(x$model.response,"states")
        cause <- requested_causes
    }else{
        requested_causes <- dots$cause
    }

    # ---- type handling
    if (is.null(type) || length(type) == 0) type <- x$type
    if (!type %in% c("surv","risk")) stop("ggprodlim: 'type' must be 'surv' or 'risk'.")
    if (attr(x$model.response,"model") == "competing.risks"){
        type <- "risk"
    }
    # call summary.prodlim via as.data.table.prodlim

    w <- do.call(data.table::as.data.table,c(list(x = x, surv = (type == "surv"), percent = percent),
                                             dots))

    # normalize column names
    if ("cuminc" %in% names(w) && !"absolute_risk" %in% names(w)) {
        data.table::setnames(w, "cuminc", "absolute_risk")
    }

    # determine outcome column
    outcome_type <- if ("absolute_risk" %in% names(w)) "absolute_risk" else "surv"
    if (type == "risk" && outcome_type == "surv") {
        # as.data.table may still provide surv; convert on the fly if possible
        w[, absolute_risk := 1 - surv]
        outcome_type <- "absolute_risk"
    }

    # check for multiple causes
    has_cause <- ("cause" %in% names(w)) && (length(unique(w$cause)) > 1)
    multi_cause <- has_cause && (length(unique(w$cause))>1)

    if (multi_cause){
        if (!"cause"%in%facet_vars){
            if (length(color_var)>0 && !identical(color_var,"cause")){
                # force cause into facets
                if (length(facet_formula)>0){
                    facet_formula <- reformulate(
                        c(attr(terms(facet_formula), "term.labels"), "cause"),
                        response = if (length(facet_formula) == 3) facet_formula[[2]] else NULL
                    )
                }else{
                    facet_formula <- formula(paste0("~cause"))
                }
                facet_vars <- all.vars(facet_formula)
                ## stop("ggprodlim: when multiple causes are requested 'cause' must appear as color_var or in the facet_formula.")  
                ## facet_vars <- c(facet_vars,"cause")
            }else{
                # if not specified as facet use color to distinguish cause 
                color_var <- "cause"
            }
        }
    }
    if (has_covariates){
        # make sure all fitted covariates appear either in color_var or in facet_vars
        unmapped <- setdiff(names(x$X),c(color_var,facet_vars))
        if (length(unmapped) > 0){
            if (length(color_var) == 0) {
                color_var <- unmapped[[1]]
                if (length(unmapped)>1){
                    stop("ggprodlim: object was fitted with covariates ",paste0(unmapped[[-1]],collapse = ", "),". These must occur either as color_var or in facet_formula.")
                }
            }else{
                if (length(facet_formula)>0){
                    facet_formula <- reformulate(
                        c(attr(terms(facet_formula), "term.labels"), unmapped),
                        response = if (length(facet_formula) == 3) facet_formula[[2]] else NULL
                    )
                }else{
                    facet_formula <- formula(paste0("~",paste(unmapped,collapse = "+")))
                }
                facet_vars <- all.vars(facet_formula)
                ## stop("ggprodlim: object was fitted with covariates ",paste0(unmapped,collapse = ", "),". These must occur either as color_var or in facet_formula.")
            }
        }
    }
    # build a strata column for selection/facetting/atrisk
    # make sure that all covariates are factors
    if (has_covariates){
        w[,(covariates) := lapply(.SD,as.factor),.SDcols = covariates]
    }
    if (multi_cause){
        set(w,j = "cause",value = factor(w[["cause"]],levels = requested_causes))
    }
    if (length(color_var)>0){
        # use newdata to detect the order of the strata
        if (color_var == "cause"){
            levels_color <- levels(w[["cause"]])
        }else{
            if (stacked) {
                stop("ggprodlim: cannot specify color_var for 'stacked' plot.")
            }
            levels_color <- levels(ggprodlim_newdata[[color_var]])
        }
        data.table::setorderv(w,color_var)
        n_colors <- length(levels_color)
        set(w,j = "ggprodlim_curve_id",value = as.integer(w[[color_var]]))
    }else{
        levels_color <- "1"
        n_colors <- 1
        set(w,j = "ggprodlim_curve_id",value = rep(1,NROW(w)))
    }
    
    # ---- default axis limits/breaks
    if (missing(ylim)) {
        ylim <- if (percent) c(0, 100) else c(0, 1)
    }
    if (missing(xlim)) {
        xlim <- c(0, max(w$time, na.rm = TRUE))
    }
    if (missing(y_breaks)) {
        y_breaks <- seq(ylim[1], ylim[2], length.out = 5)
    }
    if (missing(x_breaks)) {
        x_breaks <- ggplot2::waiver()
    }

    # aesthetics
    if (!stacked) {
        if (length(color_var)>0) {
            aes_main <- ggplot2::aes(x = time,
                                     y = !!rlang::sym(outcome_type),
                                     colour = !!rlang::sym(color_var),
                                     fill = !!rlang::sym(color_var),
                                     group = ggprodlim_curve_id)
        }else{
            aes_main <- ggplot2::aes(x = time,
                                     y = !!rlang::sym(outcome_type),
                                     group = ggprodlim_curve_id)
        }
        g <- ggplot2::ggplot(data = w, mapping = aes_main)
    } else {
        g <- NULL
    }

    # ---- confidence intervals: draw as rectangles over intervals
    ggprodlim_add_ci <- function(g, w, xlim, alpha) {
        if (!("lower" %in% names(w) && "upper" %in% names(w))) return(g)
        # prepare interval rectangles per curve
        dt <- data.table::copy(w)
        data.table::setkeyv(dt, c(facet_vars,color_var, "time"))
        dt[, ggprodlim_next_time := data.table::shift(time, type = "lead"), by = setdiff(key(dt),"time")]
        dt <- dt[!is.na(ggprodlim_next_time)]
        dt[, xmin := time]
        dt[, xmax := pmin(ggprodlim_next_time, xlim[2])]
        # clip to xlim
        dt <- dt[xmax > xlim[1]]
        dt[, xmin := pmax(xmin, xlim[1])]
        # map CI
        dt[, ymin := lower]
        dt[, ymax := upper]
        # choose fill aesthetic consistent with main plot
        if (length(color_var)>0) {
            rect_aes <- ggplot2::aes(xmin = xmin,
                                     xmax = xmax,
                                     ymin = ymin,
                                     ymax = ymax,
                                     colour = !!rlang::sym(color_var),
                                     fill = !!rlang::sym(color_var),
                                     group = ggprodlim_curve_id)
        }else{
            rect_aes <- ggplot2::aes(xmin = xmin,
                                     xmax = xmax,
                                     ymin = ymin,
                                     ymax = ymax,
                                     group = ggprodlim_curve_id)
        }
        g + ggplot2::geom_rect(data = dt, mapping = rect_aes,
                               inherit.aes = FALSE, alpha = alpha, colour = NA, show.legend = FALSE)
    }
    if (isTRUE(conf_int)) {
        g <- ggprodlim_add_ci(g, w, xlim = xlim, alpha = conf_int_alpha)
    }
    # set default spacing when facetting is used
    if (length(facet_formula) > 0 && is.null(facet_panel_spacing_y)) {
        facet_panel_spacing_y <- 5 + n_colors
    }
    # ---- main geometry
    if (stacked) {
        if (!has_cause) {
            stop("ggprodlim: cause='stacked' requires a competing risks fit (multiple causes).")
        }
        if (!("absolute_risk" %in% names(w))) {
            stop("ggprodlim: cause='stacked' requires 'absolute_risk' column in data.")
        }

        dt <- data.table::copy(w)

        ## define strata for separate stacked curves/panels, excluding cause
        stack_vars <- intersect(c(covariates, facet_vars), names(dt))
        stack_vars <- setdiff(unique(stack_vars), "cause")

        ## stack vertically at each time within each stratum
        data.table::setorderv(dt, c(stack_vars, "time", "cause"))
        dt[, ggprodlim_stacked_ymax := cumsum(absolute_risk),
           by = c(stack_vars, "time")]
        dt[, ggprodlim_stacked_ymin := data.table::shift(ggprodlim_stacked_ymax, fill = 0),
           by = c(stack_vars, "time")]

        ## create rectangles over time intervals within each cause/stratum
        data.table::setorderv(dt, c(stack_vars, "cause", "time"))
        dt[, ggprodlim_next_time := data.table::shift(time, type = "lead"),
           by = c(stack_vars, "cause")]
        dt <- dt[!is.na(ggprodlim_next_time)]

        dt[, xmin := pmax(time, xlim[1])]
        dt[, xmax := pmin(ggprodlim_next_time, xlim[2])]
        dt <- dt[xmax > xmin]

        g <- ggplot2::ggplot(
                          data = dt,
                          ggplot2::aes(xmin = xmin,
                                       xmax = xmax,
                                       ymin = ggprodlim_stacked_ymin,
                                       ymax = ggprodlim_stacked_ymax,
                                       fill = cause)
                      ) +
            ggplot2::geom_rect(colour = NA)

        if (length(facet_formula)>0){
            g <- g + ggplot2::facet_wrap(facet_formula)
            if (!is.null(facet_panel_spacing_y)) {
                g <- g + ggplot2::theme(panel.spacing.y = grid::unit(facet_panel_spacing_y, "lines"))
            }
        }
    }
    # ---- facetting
    if (!stacked) {
        g <- g + ggplot2::geom_step()
        if (length(facet_formula)>0){
            g <- g + ggplot2::facet_wrap(facet_formula)
            if (!is.null(facet_panel_spacing_y)) {
                g <- g + ggplot2::theme(panel.spacing.y = grid::unit(facet_panel_spacing_y, "lines"))
            }
        }
    }

    # ---- optional palette (do NOT force by default)
    if (!is.null(palette)) {
        g <- g + ggplot2::scale_color_manual(values = palette) +
            ggplot2::scale_fill_manual(values = palette)
    }
    # ---- at-risk table
    atrisk_data <- NULL
    has_atrisk <- FALSE
    if (missing(position_atrisk) || is.numeric(position_atrisk)) {
        has_atrisk <- TRUE
        if (missing(position_atrisk)){
            jump_times <- sort(unique(w$time))
            position_atrisk <- seq(min(jump_times),
                                   max(jump_times),
                                   (max(jump_times)-min(jump_times))/10)
        }else{
            position_atrisk <- unique(position_atrisk)
        }
        atrisk_cols <- intersect(c(facet_vars,color_var, "time", "n.risk","n.lost"),names(w))
        # Numbers at-risk are the same for all causes! 
        if (multi_cause){
            first_cause <- w[1]$cause
            atrisk_data <- w[cause == first_cause, atrisk_cols,with = FALSE]
        }else{
            atrisk_data <- w[, atrisk_cols,with = FALSE]
        }
        if (has_covariates){
            # roll number at risk respecting the covariate strata but ignoring 'cause'
            atrisk_strata_vars <- setdiff(c(facet_vars,color_var),"cause")
            data.table::setorderv(atrisk_data,c(atrisk_strata_vars,"time"))
            # turn into integer for rolling
            atrisk_data[,(atrisk_strata_vars) := lapply(.SD,as.integer),.SDcols = atrisk_strata_vars]
            atrisk_times <- unique(atrisk_data[, atrisk_strata_vars, with = FALSE])
            atrisk_times <- atrisk_times[,data.table::data.table(time = position_atrisk),by = atrisk_strata_vars]
            atrisk_data <- atrisk_data[atrisk_times, on = c(atrisk_strata_vars,"time"), roll = TRUE]
            # now return to factor
            for (asv in atrisk_strata_vars){
                set(atrisk_data,j = asv,value = factor(atrisk_data[[asv]],levels = 1:length(levels(ggprodlim_newdata[[asv]])),labels = levels(ggprodlim_newdata[[asv]])))
            }
        }else{
            # roll number at risk 
            atrisk_times <- data.table::CJ(time = position_atrisk)
            atrisk_data <- atrisk_data[atrisk_times, on = c("time"), roll = TRUE]
        }
    }
    # repeat number at risk for each panel
    if (multi_cause && !identical(color_var,"cause")){
        atrisk_data <- data.table::rbindlist(
                                       lapply(requested_causes, function(value) {
                                           data.table::copy(atrisk_data)[, cause := factor(value,levels = requested_causes)]
                                       })
                                   )
    }
    # Add risk table texts
    if (has_atrisk) {
        # Determine group ordering and labels
        # Build labels: default strata (+ cause if no facet by cause)
        if (length(color_var)>0){
            if (color_var[[1]] == "cause"){
                default_labels <- ""
            }else{
                default_labels <- levels(ggprodlim_newdata[[color_var]])
            }
        }
        if (!is.null(atrisk_labels)) {
            if (length(atrisk_labels) != length(default_labels)) {
                warning("ggprodlim: atrisk_labels length does not match number of curves; ignoring atrisk_labels.")
            } else {
                default_labels <- atrisk_labels
            }
        }
        y_title <- min(ylim) - atrisk_title_line * diff(ylim)
        y_first <- y_title - atrisk_title_step * diff(ylim)
        line_step <- atrisk_line_step * diff(ylim)

        for (i in 1:n_colors){
            if (n_colors>1){
                atrisk_strata <- atrisk_data[atrisk_data[[color_var]] == levels_color[[i]]]
            }else{
                atrisk_strata <- atrisk_data
            }
            set(atrisk_strata,
                j = outcome_type,
                value = y_first - (i - 1) * line_step)

            if (length(color_var) == 0 || identical(color_var,"cause")){
                atrisk_aes <- ggplot2::aes(x = time,
                                           y = !!rlang::sym(outcome_type),
                                           label = n.risk,
                                           color = NULL,
                                           fill = NULL)
            }else{
                atrisk_aes <- ggplot2::aes(x = time,
                                           y = !!rlang::sym(outcome_type),
                                           label = n.risk,
                                           color = !!rlang::sym(color_var),
                                           fill = NULL)
            }
            g <- g + ggplot2::geom_text(data = atrisk_strata,
                                        mapping = atrisk_aes,
                                        inherit.aes = FALSE,
                                        show.legend = FALSE)
        }

        g <- g + ggplot2::annotate("text",
                                   x = xlim[1],
                                   y = y_title,
                                   hjust = 1,
                                   label = atrisk_title)

        if (length(facet_formula) > 0) {
            if (is.null(atrisk_bottom_margin)) {
                atrisk_bottom_margin <- facet_panel_spacing_y
            }
        } else {
            if (is.null(atrisk_bottom_margin)) {
                if (identical(color_var,"cause")){
                    atrisk_bottom_margin <- 9
                }else{
                    atrisk_bottom_margin <- n_colors * 2 + 5
                }
            }
        }
        g <- g + ggplot2::theme(plot.margin = ggplot2::unit(c(1, 1, atrisk_bottom_margin, 1), "lines"))
        # space for atrisk data
        g <- g + ggplot2::coord_cartesian(ylim = ylim, xlim = xlim, clip = "off")
        # ---- scales
        if (isTRUE(percent)) {
            g <- g + ggplot2::scale_y_continuous(breaks = y_breaks, labels = function(z) paste0(z, "%"))
        } else {
            g <- g + ggplot2::scale_y_continuous(breaks = y_breaks)
        }
        g <- g + ggplot2::scale_x_continuous(breaks = x_breaks)
    }else{
        # ---- scales
        if (isTRUE(percent)) {
            g <- g + ggplot2::scale_y_continuous(limits = ylim, breaks = y_breaks, labels = function(z) paste0(z, "%"))
        } else {
            g <- g + ggplot2::scale_y_continuous(limits = ylim, breaks = y_breaks)
        }
        g <- g + ggplot2::scale_x_continuous(limits = xlim, breaks = x_breaks)
    }
    # ---- labels
    ylab <- if (outcome_type == "absolute_risk") "Absolute risk" else "Survival probability"
    g <- g + ggplot2::xlab("Time")
    g <- g + ggplot2::ylab(ylab)
    # ---- legend title
    if (isTRUE(return_data)) {
        return(list(plot = g, data = w, atrisk = atrisk_data))
    }
    g
}


######################################################################
### ggprodlim.R ends here

prodlimIcensSurv <- function(response,
                             grid,
                             tol=7,
                             maxiter,
                             ml=FALSE,
                             exact=TRUE){

  # {{{ data
  ntol <- 10^{-tol}
  L <- response[,"L"]
  N <- length(L)
  R <- response[,"R"]
  status <- response[,"status"]
  # }}}
  # {{{ one-step idea

  if (ml==FALSE) {
      # right censored observations
      # are defined by status
      R[status==0] <- L[status==0]
      if (missing(grid))
          grid <- sort(unique(c(L,R)))
      else
          if (exact)
              grid <- sort(unique(c(min(L,R),grid)))
          else
              grid <- sort(unique(grid))
      ## need an extra grid point before the smallest
      ## `L' to catch right censored and exact
      ## event times that match this smallest `L'
    
    stopifnot(all(grid >= 0))
    if (grid[1]==0)
      grid <- c(-1,grid)
    else
      grid <- c(0,grid)
    
    indexR <- sindex(jump.times=grid,eval.times=R)
    indexL <- sindex(jump.times=grid,eval.times=L)

      ## indexR <- match(R,grid)
      ## indexL <- match(L,grid)
      NS <- length(grid)
      Ind <- iindex(L,R,grid)
       ## fit <- list("icens_prodlim",
       ## as.double(L),
       ## as.double(R),
       ## as.double(grid),
       ## as.integer(indexL),
       ## as.integer(indexR),
       ## as.integer(Ind$iindex),
       ## as.integer(c(Ind$imax,0)),
       ## as.integer(status),
       ## as.double(N),
       ## as.double(NS),
       ## nrisk=double(NS),
       ## nevent=double(NS),
       ## ncens=double(NS),
       ## hazard=double(NS),
       ## varhazard=double(NS),
       ## surv=double(NS),
       ## oldsurv=double(NS),
       ## as.double(ntol),
       ## as.integer(maxiter),
       ## n.iter=integer(1),
       ## PACKAGE="prodlim")
      fit <- .C("icens_prodlim",
                as.double(L),
                as.double(R),
                as.double(grid),
                as.integer(indexL),
                as.integer(indexR),
                as.integer(Ind$iindex),
                as.integer(c(Ind$imax,0)),
                as.integer(status),
                as.double(N),
                as.double(NS),
                nrisk=double(NS),
                nevent=double(NS),
                ncens=double(NS),
                hazard=double(NS),
                varhazard=double(NS),
                surv=double(NS),
                oldsurv=double(NS),
                as.double(ntol),
                as.integer(maxiter),
                n.iter=integer(1),
                PACKAGE="prodlim")
      ## rename the extra grid point before the smallest `L'
      ## if it is negative
      if (grid[1]<0) grid[1] <- 0
      res <- list("time"=rbind(c(0,grid[-length(grid)]),c(grid)),
                  "n.risk"=round(pmax(0,fit$nrisk),tol),
                  "n.event"=round(pmax(0,fit$nevent),tol),
                  "n.lost"=round(fit$ncens,tol),
                  "hazard"=round(fit$hazard,tol),
                  "surv"=round(pmax(0,fit$surv),tol),
                  "maxtime"=max(grid),
                  "n.iter"=fit$n.iter,
                  "tol"=ntol,
                  "model"="survival")
      #    res <- list("time"=rbind(c(0,0,grid[-length(grid)]),c(0,grid)),"n.risk"=c(N,round(pmax(0,fit$nrisk),tol)),"n.event"=c(0,round(pmax(0,fit$nevent),tol)),"n.lost"=c(0,round(fit$ncens,tol)),"hazard"=c(0,round(fit$hazard,tol)),"surv"=c(1,round(pmax(0,fit$surv),tol)),"maxtime"=max(grid),"n.iter"=fit$n.iter,"tol"=ntol,"model"="survival")
  }
  else{
      # }}}
      # {{{ npmle 

    
    ## artificial closure of right censored intervals 
    ## R[Rna] <- max(c(L,R)) + 1
    R[status==0] <- max(c(L,R[status!=0])) + 1
    ##     R[status==0] <- max(c(L,R)) + 1
    ##     print(R[status==0])
    peto.intervals  <-  PetoInt(L,R,status)
    indices <- IntIndex(x=peto.intervals,L=L,R=R)
    Mindex <- indices$Mindex
    Mstrata <- indices$Mstrata
    Iindex <- indices$Iindex
    Istrata <- indices$Istrata
    M <- length(Mstrata)
    N <- length(Istrata)
    ## Zsurv <- predictSurv(prodlimIcensSurv(response=response,grid=grid,tol=tol,maxiter=1,ml=FALSE))
    Z <- rep(1/M,M)
    fit  <- .C('GMLE',as.integer(c(0,Mstrata)),as.integer(c(0,Istrata)),as.integer(Mindex),as.integer(Iindex),as.integer(N),as.integer(M),Z=as.double(Z),double(length(Z)),as.double(ntol),as.integer(maxiter),steps=integer(1),PACKAGE="prodlim")
    n.event <- c(0,fit$Z*M)
    surv <- c(1,1-cumsum(fit$Z))
    hazard <- c(0,fit$Z)/surv
    res <- list("time"=cbind(c(0,0),peto.intervals),"n.risk"=N-n.event,"n.event"=n.event,"n.lost"= c(0,rep(0,M)),"hazard"=round(hazard,tol),"surv"=round(surv,tol),"maxtime"=max(c(peto.intervals)),"n.iter"=fit$steps,"tol"=ntol,"model"="survival")
  }
  # }}}

  class(res) <- "prodlim"
  res  
}

draw_key_prodlim <- function(data, params, size) {
  `%||%` <- function(a, b) if (!is.null(a)) a else b

  alpha_val <- data$alpha %||% NA_real_
  if (is.na(alpha_val)) alpha_val <- data$alpha_ci %||% 0.2

  fill_val <- data$fill %||% NA_character_
  col_val  <- data$colour %||% data$color %||% "black"

  if (is.na(fill_val) && !is.na(col_val)) fill_val <- col_val
  if (is.na(fill_val)) fill_val <- "grey70"
  is_stacked <- (length(params$cause) == 1 && params$cause == "stacked")
  if (is_stacked) {
    return(
      grid::rectGrob(
        width = 1, height = 1,
        gp = grid::gpar(
          fill = scales::alpha(fill_val, 1),
          col = NA
        )
      )
    )
  }

  grid::grobTree(
    grid::rectGrob(
      width = 1, height = 1,
      gp = grid::gpar(
        fill = scales::alpha(fill_val, alpha_val),
        col = NA
      )
    ),
    grid::segmentsGrob(
      x0 = 0.1, x1 = 0.9, y0 = 0.5, y1 = 0.5,
      gp = grid::gpar(
        col = col_val,
        lwd = (data$linewidth %||% data$size %||% 0.5) * ggplot2::.pt,
        lty = data$linetype %||% 1
      )
    )
  )
}



GeomProdlim <- ggplot2::ggproto(
  "GeomProdlim",
  ggplot2::Geom,
  required_aes = character(0),

  handle_na = function(data, params) {
    step  <- data[data$component == "step",  , drop = FALSE]
    ci    <- data[data$component == "ci",    , drop = FALSE]
    stack <- data[data$component == "stack", , drop = FALSE]

    if (nrow(step) > 0) {
      step <- step[stats::complete.cases(step[, c("x", "y")]), , drop = FALSE]
    }

    if (nrow(ci) > 0) {
      ci <- ci[
        stats::complete.cases(ci[, c("xmin", "xmax", "ymin", "ymax")]),
        ,
        drop = FALSE
      ]
    }

    if (nrow(stack) > 0) {
      stack <- stack[
        stats::complete.cases(stack[, c("xmin", "xmax", "ymin", "ymax")]),
        ,
        drop = FALSE
      ]
    }

    rbind(step, ci, stack)
  },

  default_aes = ggplot2::aes(
    colour = "black",
    fill = NA,
    linewidth = 0.5,
    linetype = 1,
    alpha = NA
  ),
  draw_key = draw_key_prodlim,

  draw_panel = function(data, panel_params, coord,
                        conf_int = TRUE,
                        conf_int_alpha = 0.2,
                        na.rm = FALSE) {

    `%||%` <- function(a, b) if (!is.null(a)) a else b

    step_data  <- data[data$component == "step",  , drop = FALSE]
    ci_data    <- data[data$component == "ci",    , drop = FALSE]
    stack_data <- data[data$component == "stack", , drop = FALSE]

    grobs <- list()
      
    if (nrow(stack_data) > 0) {
      stack_data$colour <- NA
      stack_data$linewidth <- 0
      if (!("alpha" %in% names(stack_data)) || all(is.na(stack_data$alpha))) {
        stack_data$alpha <- 1
      }
      grobs[[length(grobs) + 1L]] <- ggplot2::GeomRect$draw_panel(
        stack_data,
        panel_params = panel_params,
        coord = coord
      )
      return(do.call(grid::grobTree, grobs))
    }

    if (isTRUE(conf_int) && nrow(ci_data) > 0) {
      if (!("fill" %in% names(ci_data))) {
        ci_data$fill <- ci_data$colour %||% ci_data$color
      } else {
        missing_fill <- is.na(ci_data$fill)
        if (any(missing_fill) && "colour" %in% names(ci_data)) {
          ci_data$fill[missing_fill] <- ci_data$colour[missing_fill]
        }
      }

      ci_data$colour <- NA
      ci_data$linewidth <- 0

      if (!("alpha" %in% names(ci_data)) || all(is.na(ci_data$alpha))) {
        ci_data$alpha <- ci_data$alpha_ci %||% conf_int_alpha
      } else {
        ci_data$alpha[is.na(ci_data$alpha)] <- ci_data$alpha_ci[is.na(ci_data$alpha)]
        ci_data$alpha[is.na(ci_data$alpha)] <- conf_int_alpha
      }

      grobs[[length(grobs) + 1L]] <- ggplot2::GeomRect$draw_panel(
        ci_data,
        panel_params = panel_params,
        coord = coord
      )
    }

    if (nrow(step_data) > 0) {
      grobs[[length(grobs) + 1L]] <- ggplot2::GeomStep$draw_panel(
        step_data,
        panel_params = panel_params,
        coord = coord,
        direction = "hv"
      )
    }

    do.call(grid::grobTree, grobs)
  }
)

#' Background and grid color control.
#' 
#' Some users like background colors, and it may be helpful to have grid lines
#' to read off e.g. probabilities from a Kaplan-Meier graph. Both things can be
#' controlled with this function. However, it mainly serves
#' \code{\link{plot.prodlim}}.
#' 
#' 
#' @param xlim Limits for the xaxis, defaults to par("usr")[1:2].
#' @param ylim Limits for the yaxis, defaults to par("usr")[3:4].
#' @param bg Background color. Can be multiple colors which are then switched
#' at each horizontal line.
#' @param fg Grid line color.
#' @param horizontal Numerical values at which horizontal grid lines are
#' plotted.
#' @param vertical Numerical values at which vertical grid lines are plotted.
#' @param border The color of the border around the background.
#' @author Thomas Alexander Gerds <tag@@biostat.ku.dk>
#' @keywords survival
#' @examples
#' 
#' 
#' plot(0,0)
#' backGround(bg="beige",fg="red",vertical=0,horizontal=0)
#' 
#' plot(0,0)
#' backGround(bg=c("yellow","green"),fg="red",xlim=c(-1,1),ylim=c(-1,1),horizontal=seq(0,1,.1))
#' backGround(bg=c("yellow","green"),fg="red",horizontal=seq(0,1,.1))
#' 
#' @export
backGround <- function(xlim,
                       ylim,
                       bg="white",
                       fg="gray77",
                       horizontal=NULL,
                       vertical=NULL,
                       border="black"){
    U <- par("usr")
    if (missing(xlim))
        xlim <- c(U[1],U[2])
    if (missing(ylim))
        ylim <- c(U[3],U[4])
    # background
    if (!is.null(bg)){
        if (length(bg)==1){
            rect(U[1],U[3],U[2],U[4],col=bg[1], border=border)
        }else{
            if (length(bg)>1){
                if (is.null(horizontal)){
                    xleft <- sort(unique(c(xlim[1],vertical,xlim[2])))
                    NR <- length(xleft)
                    bcol <- rep(bg,length.out=NR)
                    nix <- sapply(1:(NR-1),function(r){
                        polygon(y=c(U[3],U[3],U[4],U[4],U[3]),
                                x=c(xleft[r],xleft[r+1],xleft[r+1],xleft[r],xleft[r]),
                                col=bcol[r],
                                border=FALSE)})
                } else{
                    ybot <- sort(unique(c(ylim[1],horizontal,ylim[2])))
                    NR <- length(ybot)
                    bcol <- rep(bg,length.out=NR)
                    nix <- sapply(1:(NR-1),function(r){
                        ## for (r in 1:(NR-1)){
                        ## rect(xleft=xlim[1],xright=xlim[2],ybottom=ybot[r],ytop=ybot[r+1],col=bcol[r],border=FALSE)
                        ## polygon(x=c(xlim[1],xlim[1],xlim[2],xlim[2],xlim[1]),
                        polygon(x=c(U[1],U[1],U[2],U[2],U[1]),
                                y=c(ybot[r],ybot[r+1],ybot[r+1],ybot[r],ybot[r]),
                                col=bcol[r],
                                border=FALSE)
                        ## do NOT specify: density=100 as this slows this down!
                    })
                }
            }
        }
    }
    # grid 
    if (length(fg)>0){
        if (length(vertical)>0)
            abline(v=vertical,col=fg)
        if (length(horizontal)>0)
            abline(h=horizontal,col=fg)
    }
}

#' Marking product-limit plots at the censored times.
#' 
#' This function is invoked and controlled by \code{plot.prodlim}.
#' 
#' This function should not be called directly. The arguments can be specified
#' as \code{atRisk.arg} in the call to \code{plot.prodlim}.
#' 
#' @param x The values of the curves at \code{times}.
#' @param times The times where there curves are plotted.
#' @param nlost The number of subjects lost to follow-up (censored) at
#' \code{times}.
#' @param pch The symbol used to mark the curves.
#' @param col The color of the symbols.
#' @param ...  Arguments passed to \code{points}.
#' @return Nil
#' @author Thomas Alexander Gerds <tag@@biostat.ku.dk>
#' @seealso \code{\link{plot.prodlim}}, \code{\link{confInt}},
#' \code{\link{atRisk}}
#' @keywords survival
#' @export
markTime <- function(x,times,nlost,pch,col,...){
  mtimeList=lapply(1:length(x),function(i){
    who=nlost[[i]]>0 & !is.na(nlost[[i]])
    mark.x=times[who]
    mark.y=x[[i]][who]
    if (length(col)<length(x)) mcol=col[1] else mcol=col[i]
    if (length(pch)<length(x)) mpch=pch[1] else mpch=pch[i]
    points(x=mark.x,y=mark.y,col=mcol,pch=mpch,...)
    ##       cbind(mark.x,mark.y)
    invisible(NULL)
  })
}

iindex <- function (L,R,grid) {
  stopifnot((length(grid)>0)
            & (length(L)>0)
            & (length(R)>0))
  stopifnot(is.numeric(c(L,R,grid)))
  N <- length(L)
  NS <- length(grid)
  ind <- .C("iindexSRC",
            index = integer(N*NS),
            strata = integer(NS-1),
            as.double(L),
            as.double(R),
            as.double(grid),
            as.integer(N),
            as.integer(NS),
            PACKAGE="prodlim")
  strata <- ind$strata
  index <- ind$index[1:max(strata)]
  list(iindex=index,imax=strata)
}

#' Add point-wise confidence limits to the graphs of Kaplan-Meier and
#' Aalen-Johansen estimates.
#' 
#' This function is invoked and controlled by \code{plot.prodlim}.
#' 
#' This function should not be called directly. The arguments can be specified
#' as \code{Confint.arg} in the call to \code{plot.prodlim}.
#' 
#' @param ci A \code{data.table} with columns \code{time}, \code{lower} and \code{upper}. 
#' @param citype If \code{"shadow"} then confidence limits are drawn as colored
#' shadows.  Otherwise, dotted lines are used to show the upper and lower
#' confidence limits.
#' @param col the colour of the lines.
#' @param lty the line type of the lines.
#' @param lwd the line thickness of the lines.
#' @param density For \code{citype="shadow"}, the density of the shade. Default
#' is 55 percent.
#' @param \dots Further arguments that are passed to the function
#' \code{segments} if \code{type=="bars"} and to \code{lines} else.
#' @return Nil
#' @author Thomas Alexander Gerds <tag@@biostat.ku.dk>
#' @seealso \code{\link{plot.prodlim}}, \code{\link{atRisk}},
#' \code{\link{markTime}}
#' @keywords survival
#' @export
confInt <- function(ci,
                    citype,
                    col,
                    lty,
                    lwd,
                    density=55,
                    ...){
    lower <- upper <- time <- NULL
    nlines <- length(ci)
    nix <- lapply(1:nlines,function(i){
        # ----------remove confidence limits before the first event----------
        CI <- ci[[i]][upper-lower<1]
        if (NROW(ci[[i]])>0){
            switch(citype,
                   "bars"={
                       nix <- CI[,{segments(x0=time,x1=time,y0=lower,y1=upper,lwd=lwd[i],col=col[i],lty=lty[i],...);NULL}]
                   },
                   "shadow"={
                       nix <- CI[,{
                           cc <- dimColor(col[i],density=density)
                           ttt <- time
                           nt <- length(ttt)
                           ttt <- c(ttt,ttt)
                           uuu <- c(0,upper[-nt],upper)
                           lll <- c(0,lower[-nt],lower)
                           neworder <- order(ttt)
                           uuu <- uuu[neworder]
                           lll <- lll[neworder]
                           ttt <- sort(ttt)
                           polygon(x=c(ttt,rev(ttt)),
                                   y=c(lll,rev(uuu)),col=cc,border=NA)
                           NULL}]
                   },{
                       nix <- CI[,{
                           lines(x=time,lower,type="s",lwd=lwd[i],col=col[i],lty=lty[i],...)
                           lines(x=time,upper,type="s",lwd=lwd[i],col=col[i],lty=lty[i],...)
                           NULL
                       }]
                   })
        }
    })
}

#' @export
lifeTab.survival <- function(object,
                             times,
                             newdata,
                             stats,
                             intervals=FALSE,
                             percent=TRUE,
                             format,...){
    # {{{ get the indices
    IndeX <- predict(object,
                     newdata=newdata,
                     level.chaos=0,
                     times=times,
                     type="list")
    # }}}
    # {{{ times
    times <- IndeX$times
    Ntimes <- IndeX$dimensions$time
    pindex <- IndeX$indices$time
    delayed <- attr(object$model.response,"entry.type")=="leftTruncated"
    # }}}
    # {{{ covariate strata
    Nstrata <- IndeX$dimensions$strata
    findex <- IndeX$indices$strata
    # }}}
    # {{{ stats
    if (missing(stats) || ((!missing(stats)) && is.null(stats)))
        stats <- list(c("n.event",0),c("n.lost",0))
    else{
        stats <- c(list(c("n.event",0),c("n.lost",0)),stats)
    }
    # }}}
    # {{{ summary at exact times
    if (intervals==FALSE){
        if (is.null(object$clustervar)){
            ## only one column for n.risk
            xxx <- .C("summary_prodlim",pred.nrisk=integer(Ntimes*Nstrata),pred.nevent=integer(Ntimes*Nstrata),pred.nlost=integer(Ntimes*Nstrata),nrisk=as.integer(object$n.risk),nevent=as.integer(object$n.event),nlost=as.integer(object$n.lost),as.double(times),as.double(object$time),as.integer(object$first.strata[findex]),as.integer(object$size.strata[findex]),as.integer(Nstrata),as.integer(Ntimes),NAOK=FALSE,PACKAGE="prodlim")
            out <- data.frame(n.risk=xxx$pred.nrisk,n.event=xxx$pred.nevent,n.lost=xxx$pred.nlost)
        }
        else{
            xxx <- .C("summary_prodlim",pred.nrisk=integer(Ntimes*Nstrata),pred.nevent=integer(Ntimes*Nstrata),pred.nlost=integer(Ntimes*Nstrata),nrisk=as.integer(object$n.risk[,1]),nevent=as.integer(object$n.event[,1]),nlost=as.integer(object$n.lost[,1]),as.double(times),as.double(object$time),as.integer(object$first.strata[findex]),as.integer(object$size.strata[findex]),as.integer(Nstrata),as.integer(Ntimes),NAOK=FALSE,PACKAGE="prodlim")
            out <- data.frame(n.risk=xxx$pred.nrisk,n.event=xxx$pred.nevent,n.lost=xxx$pred.nlost)
            for (cv in 1:length(object$clustervar)){
                yyy <- .C("summary_prodlim",pred.nrisk=integer(Ntimes*Nstrata),pred.nevent=integer(Ntimes*Nstrata),pred.nlost=integer(Ntimes*Nstrata),nrisk=as.integer(object$n.risk[,1+cv]),nevent=as.integer(object$n.event[,1+cv]),nlost=as.integer(object$n.lost[,1+cv]),as.double(times),as.double(object$time),as.integer(object$first.strata[findex]),as.integer(object$size.strata[findex]),as.integer(Nstrata),as.integer(Ntimes),NAOK=FALSE,PACKAGE="prodlim")
                outCV <- data.frame(n.risk=yyy$pred.nrisk,n.event=yyy$pred.nevent,n.lost=yyy$pred.nlost)
                names(outCV) <- paste(object$clustervar,names(outCV),sep=".")
                out <- cbind(out,outCV)
            }
        }
    }
    # }}}
    # {{{ summary in Intervals
    else{
        #,----
        #| get no. at risk at the left limit of the interval
        #| and count events and censored including the left limit
        #| but excluding the right interval border
        #`----
        start <- min(min(object$time),0)-.1
        lower <- c(start,times[-length(times)])
        upper <- times
        lagTimes <- c(min(min(object$time),0)-.1 , times[-length(times)])
        if (is.null(object$clustervar)){
            ## only one column in n.event and n.risk
            xxx <- .C("life_table",
                      pred.nrisk=integer(Ntimes*Nstrata),
                      pred.nevent=integer(Ntimes*Nstrata),
                      pred.nlost=integer(Ntimes*Nstrata),
                      nrisk=as.integer(object$n.risk),
                      nevent=as.integer(object$n.event),
                      nlost=as.integer(object$n.lost),
                      lower=as.double(lower),
                      upper=as.double(upper),
                      as.double(object$time),
                      as.integer(object$first.strata[findex]),
                      as.integer(object$size.strata[findex]),
                      as.integer(Nstrata),
                      as.integer(Ntimes),
                      as.integer(delayed),
                      NAOK=FALSE,
                      PACKAGE="prodlim")
            out <- data.frame(n.risk=xxx$pred.nrisk,
                              n.event=xxx$pred.nevent,
                              n.lost=xxx$pred.nlost)
        }
        else{
            xxx <- .C("life_table",
                      pred.nrisk=integer(Ntimes*Nstrata),
                      pred.nevent=integer(Ntimes*Nstrata),
                      pred.nlost=integer(Ntimes*Nstrata),
                      nrisk=as.integer(object$n.risk[,1]),
                      nevent=as.integer(object$n.event[,1]),
                      nlost=as.integer(object$n.lost[,1]),
                      lower=as.double(lower),
                      upper=as.double(upper),
                      as.double(object$time),
                      as.integer(object$first.strata[findex]),
                      as.integer(object$size.strata[findex]),
                      as.integer(Nstrata),
                      as.integer(Ntimes),
                      as.integer(delayed),
                      NAOK=FALSE,
                      PACKAGE="prodlim")
            out <- data.frame(n.risk=xxx$pred.nrisk,n.event=xxx$pred.nevent,n.lost=xxx$pred.nlost)
            for (cv in 1:length(object$clustervar)){
                yyy <- .C("life_table",
                          pred.nrisk=integer(Ntimes*Nstrata),
                          pred.nevent=integer(Ntimes*Nstrata),
                          pred.nlost=integer(Ntimes*Nstrata),
                          nrisk=as.integer(object$n.risk[,1+cv]),
                          nevent=as.integer(object$n.event[,1+cv]),
                          nlost=as.integer(object$n.lost[,1+cv]),
                          lower=as.double(lower),
                          upper=as.double(upper),
                          as.double(object$time),
                          as.integer(object$first.strata[findex]),
                          as.integer(object$size.strata[findex]),
                          as.integer(Nstrata),
                          as.integer(Ntimes),
                          as.integer(delayed),
                          NAOK=FALSE,
                          PACKAGE="prodlim")
                outCV <- data.frame(n.risk=yyy$pred.nrisk,n.event=yyy$pred.nevent,n.lost=yyy$pred.nlost)
                names(outCV) <- paste(object$clustervar,names(outCV),sep=".")
                out <- cbind(out,outCV)
            }
        }
    }
    # }}}
    # {{{ percent
    if (!is.null(stats)){
        statsList <- lapply(stats,function(x){
            if (percent==TRUE && length(grep(x[1],c("n.event","n.lost","n.risk"),value=FALSE))==0){
                100*as.numeric(c(x[2],object[[x[1]]])[pindex+1])
            } else{
                as.numeric(c(x[2],object[[x[1]]])[pindex+1])
            }
        })
        names(statsList) <- sapply(stats,function(x)x[[1]])
        add <- do.call("cbind",statsList)
        add <- add[,match(colnames(add),colnames(out),nomatch=FALSE)==0,drop=FALSE]
        if (NROW(out)==1)
            out <- data.frame(cbind(out,add))
        else
            out <- cbind(out,add)
    }
    # }}}
    # {{{ split into list according to covariate strata
    if (intervals==TRUE){
        tt <- data.table::data.table(time0=c(0,round(times[-length(times)],2)),time1=times)
    }else{
        tt <- data.table::data.table(time=times)
    }
    if (!is.null(newdata) || Nstrata > 1) {
        if (format[[1]]=="list"){
            split.cova <- rep(1:Nstrata,rep(Ntimes,Nstrata))
            out <- split(out,split.cova)
            names(out) <- IndeX$names.strata
            out <- lapply(out,function(x){
                x <- cbind(as.matrix(tt),as.matrix(x))
                rownames(x) <- 1:NROW(x)
                x
            })
        }else{
            X <- IndeX$predictors[rep(1:Nstrata,rep(Ntimes,Nstrata)),,drop=FALSE]
            out <- cbind(X,tt,out)
            data.table::setDT(out)
            data.table::setkeyv(out,colnames(X))
        }
    }
    # }}}
    # {{{ univariate case
    else{
        out <- cbind(tt,out)
    }
    # }}}
    out
}

#' Summary of event histories
#' 
#' Describe events and censoring patterns of an event history.
#' 
#' 
#' @param object An object with class `Hist' derived with \code{\link{Hist}}
#' @param verbose Logical. If FALSE any printing is supressed.
#' @param \dots Not used
#' @return \code{NULL} for survival and competing risk models.  For other
#' multi-state models, it is a list with the following entries:
#' \item{states}{the states of the model} \item{transitions}{the transitions
#' between the states} \item{trans.frame}{a data.frame with the from and to
#' states of the transitions}
#' @author Thomas A. Gerds \email{tag@@biostat.ku.dk}
#' @seealso \code{\link{Hist}}, \code{\link{plot.Hist}}
#' @keywords survival
#' @examples
#' 
#' icensFrame <- data.frame(L=c(1,1,3,4,6),R=c(2,NA,3,6,9),event=c(1,1,1,2,2))
#' with(icensFrame,summary(Hist(time=list(L,R))))
#'
#' @export 
summary.Hist <- function(object, verbose=TRUE,...){
    D <- object[,"status",drop=TRUE]
    states <- attr(object,"states")
    cens.code <- attr(object,"cens.code")
    # {{{ resolving events and model states 
    model <- attr(object,"model")
    model.string <- paste("response of a", model,"model")
    if (model=="multi.states"){
        from <- object[,"from"]
        to <- object[,"to"]
        code.from <- getEvent(object,mode="factor",column="from")
        code.to <- getEvent(object,mode="factor",column="to")
        state.types <- factor(as.numeric(match(states,unique(code.from),nomatch=0)!=0) + 2*as.numeric(match(states,unique(code.to),nomatch=0)!=0),levels=c(1,2,3))
        names(state.types) <- states
        levels(state.types) <- c("initial","absorbing","transient")
        state.types <- table(state.types)
    }
    else{
        from <- rep("initial",NROW(object))
        code.to <- getEvent(object,mode="factor",column=ifelse(model=="survival","status","event"))
        code.from <- factor(from)
        state.types <- c(1,length(states))
        names(state.types) <- c("initial","absorbing")
    }
    # }}}
    # {{{ transition frame
    ##   trans.frame <- unique(data.frame(from=code.from,to=code.to),MARGIN=1)
    trans.frame <- data.frame(from=code.from,to=code.to)
    Transitions <- apply(cbind(as.character(code.from),as.character(code.to)),1,paste,collapse=" -> ")
    obnoxious.factor.levels <- unique(Transitions)
    Transitions <- factor(Transitions,obnoxious.factor.levels)
    transitions <- table(Transitions)
    summary.out <- list(states=state.types,transitions=transitions,trans.frame=trans.frame)
    if (verbose==TRUE){
        state.table <- as.matrix(transitions)
        colnames(state.table) <- c("Freq")
    }
    # }}}
    # {{{ resolving the censoring mechanism
    if (verbose==TRUE){
        ## event time
        cens.type <- attr(object,"cens.type")
        ## cens.string <- capitalize(cens.type)
        cens.string <- switch(cens.type,
                              "intervalCensored"="Interval-censored",
                              "rightCensored"="Right-censored",
                              "uncensored"="Uncensored")
        Observations <- switch(cens.type,
                               "intervalCensored"=factor(D,levels=c(1,2,0),labels=c("exact.time","interval-censored","right-censored")),
                               "rightCensored"=factor(D,levels=c(1,0),labels=c("event","right.censored")),
                               "uncensored"=factor(D,labels=c("event")))
        Freq <- table(Observations)
        ## entry time
        entry.type <- attr(object,"entry.type")
        if (entry.type!="")
            entry.string <- paste(" with ",entry.type," entry time",sep="")
        else
            entry.string <- ""
        ## stop time
        stop.time <- attr(object,"stop.time")
        if (is.null(stop.time))
            stop.string <- ""
        else
            stop.string <- paste(" stopped at time ",stop.time,sep="")
        cat("\n",
            cens.string,
            " ",
            model.string,
            entry.string,
            stop.string,
            "\n",
            sep="")
        cat("\nNo.Observations:",NROW(object),"\n\nPattern:\n")
        switch(model,"survival"={
                         prmatrix(cbind(names(Freq),Freq),
                                  quote=FALSE,
                                  rowlab=rep("",NROW(Freq)))},
               "competing.risks"={
                   events <- getEvent(object)
                   prout <- table("Cause"=events,as.character(Observations))
                   print(prout)
               },
               "multi.states"={
                   x=table(Transitions,Observations)
                   aaa=sapply(strsplit(rownames(x)," -> "),function(x)x[1])
                   bbb=sapply(strsplit(rownames(x)," -> "),function(x)x[1])
                   print(x[order(aaa,bbb),,drop=FALSE])
               })
    }
    # }}}
    invisible(summary.out)
}

## capitalize <- function(x) {
  ## s <- strsplit(x, " ")[[1]]
  ## paste(toupper(substring(s, 1,1)), substring(s, 2), sep="", collapse=" ")
## }

#' Drawing numbers of subjects at-risk of experiencing an event below
#' Kaplan-Meier and Aalen-Johansen plots.
#' 
#' This function is invoked and controlled by \code{plot.prodlim}.
#' 
#' This function should not be called directly. The arguments can be specified
#' as \code{atRisk.arg} in the call to \code{plot.prodlim}.
#' 
#' @param x an object of class `prodlim' as returned by the
#' \code{prodlim} function.
#' @param newdata see \code{plot.prodlim}
#' @param times Where to compute the atrisk numbers.
#' @param line Distance of the atrisk numbers from the inner plot.
#' @param col The color of the text.
#' @param labelcol The color for the labels. Defaults to col.
#' @param interspace Distance between rows of atrisk numbers.
#' @param cex Passed on to \code{mtext} for both atrisk numbers and
#' labels.
#' @param labels Labels for the at-risk rows.
#' @param title Title for the at-risk labels
#' @param titlecol The color for the title. Defaults to 1 (black).
#' @param pos The value is passed on to the \code{mtext} argument
#' \code{at} for the labels (not the atrisk numbers).
#' @param adj Passed on to \code{mtext} for the labels (not the atriks
#' numbers).
#' @param dist If \code{line} is missing, the distance of the upper
#' most atrisk row from the inner plotting region: par()$mgp[2].
#' @param xdist Distance in x-axis direction to define the distance between the labels
#' and the numbers at-risk. Deftaults to \code{strwidth("MM",cex=cex)}.
#' @param adjust.labels If \code{TRUE} the labels are left adjusted.
#' @param show.censored If \code{TRUE} the cumulative number of subjects
#'                      lost to follow up is shown in parentheses.
#' @param unit The graphical coordinate systems unit to convert from when line2user is calling \code{grconvertX} and \code{grconvertY}.
#'             Default is \code{'npc'}
#' @param ... Further arguments that are passed to the function
#' \code{mtext}.
#' @return Nil
#' @author Thomas Alexander Gerds <tag@@biostat.ku.dk>
#' @seealso \code{\link{plot.prodlim}}, \code{\link{confInt}},
#' \code{\link{markTime}}
#' @keywords survival
#' @export
# {{{ header 

atRisk <- function(x,
                   newdata,
                   times,
                   line,
                   col,
                   labelcol=NULL,
                   interspace,
                   cex,
                   labels,
                   title="",
                   titlecol=NULL,
                   pos,
                   adj,
                   dist,
                   xdist,
                   adjust.labels=TRUE,
                   show.censored=FALSE,
                   unit='npc',
                   ...){

    # }}} 
    # {{{ start values
    if (missing(times)) times <- seq(0,x$maxtime,x$maxtime/10) else times <- sort(unique(times))
    if (missing(cex)) cex <- 1
    if (missing(xdist)) xdist=strwidth("MM",cex=cex)
    if (missing(pos)) pos <- NULL
    if (missing(adj)) adj <- 0
    clusterp <- length(x$clustervar[[1]])>0
    # }}}
    # {{{ find numbers at risk at given times
    if (x$model=="competing.risks"){
        px <- lifeTab(object=x,times=times,cause=getStates(x)[1],newdata=newdata,stats=NULL,intervals=FALSE,format="dt")
    }
    else if (x$model=="survival"){
        px <- lifeTab(object=x,times=times,newdata=newdata,stats=NULL,intervals=FALSE,format="dt")
    }
    if (clusterp)
        nrisk.vars <- grep("n.risk",colnames(px))
    else
        nrisk.vars <- "n.risk"
    xvars  <- data.table::key(px)
    xvars <- xvars[xvars!="cause"]
    if (length(xvars)>0){
        xdata <- px[,xvars,with=FALSE]
        xdegen <- sapply(xdata,function(x)length(unique(x))==1)
        xvars <- xvars[!xdegen]
    }
    if (length(xvars)>0){
        xdata <- xdata[,xvars,with=FALSE]
        xstrata <- apply(do.call("cbind",lapply(xvars,function(n){xdata[[n]]})),1,paste,collapse=", ")
        if (clusterp){
            number.atrisk <- c(split(px[[nrisk.vars[[1]]]],xstrata),
                               split(px[[nrisk.vars[[2]]]],xstrata))
        } else{
            number.atrisk <- split(px[[nrisk.vars[[1]]]],xstrata)
        }
    } else{
        if (clusterp){
            xdata <- data.table::data.table(c("Subjects: ","Clusters: "))
            colnames(xdata) <- ""
            number.atrisk <- c(list(px[[nrisk.vars[[1]]]]),list(px[[nrisk.vars[[2]]]]))
        }else{
            xdata <- data.table::data.table("Subjects: ")
            colnames(xdata) <- ""
            number.atrisk <- list(px[[nrisk.vars[[1]]]])
        }
    }
    nlines <- length(number.atrisk)*NCOL(number.atrisk[1])
    # }}}
    # {{{ labels 
    if (!missing(labels) && labels[[1]]!="fixme"){
        if (is.numeric(labels) || is.character(labels) || !(data.table::is.data.table(labels)))
            labels <- data.table::data.table(labels)
        if (!is.data.table(labels))
            xdata <- data.table(labels)
        else
            xdata <- labels
        n.columns <- NCOL(xdata)
    }else{
        xdata <- unique(xdata)
        if (clusterp) {
            xdata <- cbind("V1"=rep(c("Subjects: ","Clusters: "),rep(NROW(xdata),2)),data.table::rbindlist(list(xdata,xdata)))
            data.table::setnames(xdata,"V1","")
        }
        n.columns <- NCOL(xdata)
        isnum <- sapply(xdata,is.numeric)
        if (any(isnum)){
            for (j in (1:n.columns)[isnum]){
                data.table::set(xdata,j=j,value=format(xdata[[j]],digits=2))
            }
        }
    }
    # title of labels column(s)
    if (title[[1]]!=FALSE && title[[1]] !="" && length(title)>0) {
        if (length(title)==length(names(xdata)))
            names(xdata) <- title
        else{
            warning(paste0("Argument 'atRisk.title' has the wrong length.\nIt should have length ",length(names(xdata)),", i.e., one label for each column of the 'atrisk.labels'."))
        }
    }
    # distance from plot 
    if (missing(line)){
        line <- par()$mgp[1]+dist+(1:nlines)*c(1,rep(interspace,nlines-1))
    }
    # }}}
    # {{{ the actual numbers below the plot

    ## color of labels for clustered data
    if (clusterp && (length(col)==nlines/2))
        col <- rep(col,rep(2,length(col)))
    atrisk.figures.lines <- lapply(1:nlines,function(y){
        if (show.censored==FALSE){
            atrisk.figures <- as.character(number.atrisk[[y]])
        }else{
            if (x$model=="competing.risks"){
                qx <- lifeTab(object=x,times=times,cause=getStates(x)[1],newdata=newdata,stats=NULL,intervals=TRUE,format="dt")
            }
            else if (x$model=="survival"){
                qx <- lifeTab(object=x,times=times,newdata=newdata,stats=NULL,intervals=TRUE,format="dt")
            }
            if (length(xvars)>0){
                ncens <- split(qx[["n.lost"]],xstrata)
            }else{
                ncens <- list(qx[["n.lost"]])
            }
            if (show.censored=="cumulative"){
                atrisk.figures <- paste0(as.character(number.atrisk[[y]])," (",cumsum(ncens[[y]]),")")
            }else{# in interval
                atrisk.figures <- paste0(as.character(number.atrisk[[y]])," (",ncens[[y]],")")
            }
        }
    })
    lapply(1:nlines,function(y){
        atrisk.figures <- atrisk.figures.lines[[y]]
        text(labels=atrisk.figures,
             adj=c(0.5,0),
             x=times,
             y=rep(line2user(line[y],side=1,unit=unit),length(times)),
             col=rep(col[y],length(times)),
             cex=cex,
             xpd=NA,
             ...)
        if (is.null(labelcol)){
            lcol <- col[y]
        } else {
            if (is.na(labelcol[y]))
                lcol <- labelcol[1]
            else
                lcol <- labelcol[y]
        }
        if (length(pos)==0) pos <- min(times)-xdist
        if(title!=FALSE && title !="" && length(title)>0){
            column.widths <- cumsum(c(0,xdist+rev(sapply(1:n.columns,function(j){
                max(strwidth(c(names(xdata)[[j]],xdata[[j]])))}))))
        } else{
            column.widths <- cumsum(c(0,xdist+rev(sapply(1:n.columns,function(j){
                max(strwidth(xdata[[j]]))}))))
        }
        # reverse column order
        xdata <- xdata[,n.columns:1,with=FALSE]
        for (j in 1:n.columns){
            if (y==1 && title!=FALSE && title !="" && length(title)>0){
                text(labels=c(names(xdata)[[j]],as.character(xdata[y,j,with=FALSE][[1]])),
                     x=rep(pos-column.widths[[j]],2),
                     col=c(ifelse(length(titlecol)==0 || is.na(titlecol[1]),1,titlecol[1]),lcol),
                     y=c(line2user(line[1]-1,side=1,unit=unit),line2user(line[1],side=1,unit=unit)),
                     adj=c(1,0),
                     cex=cex,
                     xpd=NA,
                     ...)
            }else{
                text(labels=as.character(xdata[y,j,with=FALSE][[1]]),
                     x=pos-column.widths[[j]],
                     col=lcol,
                     y=line2user(line[y],side=1,unit=unit),
                     adj=c(1,0),
                     cex=cex,
                     xpd=NA,
                     ...)
            }
        }
    })

    # }}}
}

#' Predict individual survival probabilities
#' 
#' Function to extract the predicted probabilities at the individual event
#' times that have been used for fitting a prodlim object. 
#' 
#' @param object A fitted object of class "prodlim".
#' @param lag Integer. `0' means predictions at the individual times, 1 means
#' just before the individual times, etc.
#' @return A vector of survival probabilities.
#' @author Thomas A. Gerds \email{tag@@biostat.ku.dk}
#' @seealso \code{\link{predict.prodlim}},\code{\link{predictSurv}},
#' @keywords survival
#' @examples
#' 
#'  SurvFrame <- data.frame(time=1:10,status=rbinom(10,1,.5))
#'  x <- prodlim(formula=Hist(time=time,status!=0)~1,data=SurvFrame)
#'  predictSurvIndividual(x,lag=1)
#' 
#' @export
predictSurvIndividual <- function(object,
                                  lag=1){
    obs.times <- as.numeric(object$model.response[,1])
    if (object$covariate.type==1){
        locOBS <- match(obs.times,object$time,nomatch=FALSE)
        if (any(locOBS==FALSE)) stop("Can't locate all individual observation times" )
        psurv <- c(rep(1,lag),object$surv)[locOBS]}
    else{
        N <- length(obs.times)
        if (is.null(object$model.matrix)) stop("Cannot find model.matrix, need to set x=TRUE in call of prodlim.")
        findex <- row.match(object$model.matrix,object$X)
        ## if (any(is.na(findex)))
        ## stop("Cannot identify all rows in object$model.matrix in ")
        psurv <- .C("predict_individual_survival",
                    pred=double(N),
                    as.double(object$surv),
                    as.double(object$time),
                    as.double(obs.times),
                    as.integer(object$first.strata[findex]),
                    as.integer(object$size.strata[findex]),
                    as.integer(N),
                    as.integer(lag),
                    NAOK=FALSE,
                    PACKAGE="prodlim")$pred}
    psurv
}

#' @keywords internal
stat_prodlim <- function(mapping=NULL, data=NULL, geom="step",
                         position="identity", na.rm=FALSE, show.legend=NA,
                         inherit.aes=TRUE,
                         type=c("surv","risk"),
                         cause=NULL,
                         legend=c("auto","show","hide"),
                         conf_int=FALSE,
                         percent=TRUE,
                         timeconverter=NULL,
                         times=NULL,
                         cens.code,
                         ...) {
    ggplot2::layer(
                 stat = StatProdlim,
                 data = data,
                 mapping = mapping,
                 geom = geom,
                 position = position,
                 show.legend = show.legend,
                 inherit.aes = inherit.aes,
                 params = list(
                     na.rm = na.rm,
                     type = match.arg(type),
                     cause = cause,
                     conf_int = conf_int,
                     percent = percent,
                     timeconverter = timeconverter,
                     times = times,
                     cens.code = cens.code,
                     ...)
             )
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/prodlim-package.R
\name{prodlim-package}
\alias{prodlim-package}
\title{Kaplan-Meier and Aalen-Johansen method}
\description{
Fast and user friendly implementation
of nonparametric estimators for censored
event history (survival) analysis.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/model.design.R
\name{model.design}
\alias{model.design}
\title{Extract a design matrix and specials from a model.frame}
\usage{
model.design(
  terms,
  data,
  xlev = NULL,
  dropIntercept = FALSE,
  maxOrder = 1,
  unspecialsDesign = TRUE,
  specialsFactor = FALSE,
  specialsDesign = FALSE
)
}
\arguments{
\item{terms}{terms object as obtained either with function \code{terms} or \code{strip.terms}.}

\item{data}{A data set in which terms are defined.}

\item{xlev}{a named list of character vectors giving the full set of levels to be assumed for the factors.
Can have less elements, in which case the other levels are learned from the \code{data}.}

\item{dropIntercept}{If TRUE drop intercept term from the design
matrix}

\item{maxOrder}{An error is produced if special variables are
involved in interaction terms of order higher than max.order.}

\item{unspecialsDesign}{A logical value: if \code{TRUE} apply
\code{\link{model.matrix}} to unspecial covariates. If
\code{FALSE} extract unspecial covariates from data.}

\item{specialsFactor}{A character vector containing special
variables which should be coerced into a single factor. If
\code{TRUE} all specials are treated in this way, if \code{FALSE}
none of the specials is treated in this way.}

\item{specialsDesign}{A character vector containing special
variables which should be transformed into a design matrix via
\code{\link{model.matrix}}.  If \code{TRUE} all specials are
treated in this way.}
}
\value{
A list which contains
  - the design matrix with the levels of the variables stored in attribute 'levels' 
  - separate data.frames which contain the values of the special variables.
}
\description{
Extract design matrix and data specials from a model.frame
}
\details{
The function separates special terms from the unspecial terms and returns
a list of design matrices, one for unspecial terms and one for each special.
Some special specials cannot or should not be evaluated in
data. E.g., \code{y~a+dummy(x)+strata(v)} the function strata can and should be evaluated,
but in order to have \code{model.frame} also evaluate dummy(x) one would be to define
and export the function \code{dummy}. Still the term \code{dummy(x)} can be used
to identify a special treatment of the variable \code{x}. To deal with this case,
one can specify \code{stripSpecials="dummy"}. In addition,  the data
should include variables \code{strata(z)} and \code{x}, not \code{dummy(x)}.
See examples.
The function \code{untangle.specials} of the survival function does a similar job.
}
\examples{
# specials that are evaluated. here ID needs to be defined 
set.seed(8)
d <- data.frame(y=rnorm(5),x=factor(c("a","b","b","a","c")),z=c(2,2,7,7,7),v=sample(letters)[1:5])
d$z <- factor(d$z,levels=c(1:8))
ID <- function(x)x
f <- formula(y~x+ID(z))
t <- terms(f,special="ID",data=d)
mda <- model.design(terms(t),data=d,specialsFactor=TRUE)
mda$ID
mda$design
## 
mdb <- model.design(terms(t),data=d,specialsFactor=TRUE,unspecialsDesign=FALSE)
mdb$ID
mdb$design

# set x-levels
attr(mdb$ID,"levels")
attr(model.design(terms(t),data=d,xlev=list("ID(z)"=1:10),
     specialsFactor=TRUE)$ID,"levels")

# special specials (avoid define function SP)
f <- formula(y~x+SP(z)+factor(v))
t <- terms(f,specials="SP",data=d)
st <- strip.terms(t,specials="SP",arguments=NULL)
md2a <- model.design(st,data=d,specialsFactor=TRUE,specialsDesign="SP")
md2a$SP
md2b <- model.design(st,data=d,specialsFactor=TRUE,specialsDesign=FALSE)
md2b$SP

# special function with argument
f2 <- formula(y~x+treat(z,power=2)+treat(v,power=-1))
t2 <- terms(f2,special="treat")
st2 <- strip.terms(t2,specials="treat",arguments=list("treat"=list("power")))
model.design(st2,data=d,specialsFactor=FALSE)
model.design(st2,data=d,specialsFactor=TRUE)
model.design(st2,data=d,specialsDesign=TRUE)

library(survival)
data(pbc)
t3 <- terms(Surv(time,status!=0)~factor(edema)*age+strata(I(log(bili)>1))+strata(sex),
            specials=c("strata","cluster"))
st3 <- strip.terms(t3,specials=c("strata"),arguments=NULL)
md3 <- model.design(terms=st3,data=pbc[1:4,])
md3$strata
md3$cluster

f4 <- Surv(time,status)~const(age)+prop(factor(edema),power=2)+
                        strata(sex,test=0)+prop(bili,power=1)+tp(albumin)
t4 <- terms(f4,specials=c("prop","timevar","strata","tp","const"))
st4 <- strip.terms(t4,
                   specials=c("prop","timevar"),
                   unspecials="prop",
                   alias.names=list("timevar"="strata","prop"=c("const","tp")),
                   arguments=list("prop"=list("power"=0),"timevar"=list("test"=0)))
formula(st4)
md4 <- model.design(st4,data=pbc[1:4,],specialsDesign=TRUE)
md4$prop
md4$timevar

}
\seealso{
\code{\link{EventHistory.frame}} model.frame terms model.matrix .getXlevels
}
\author{
Thomas A. Gerds <tag@biostat.ku.dk>
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/meanNeighbors.R
\name{meanNeighbors}
\alias{meanNeighbors}
\title{Helper function to obtain running means for prodlim objects.}
\usage{
meanNeighbors(x, y, ...)
}
\arguments{
\item{x}{Object of class \code{"neighborhood"}.}

\item{y}{Vector of numeric values.}

\item{\dots}{Not used.}
}
\description{
Compute average values of a variable according to neighborhoods.
}
\examples{

meanNeighbors(x=1:10,y=c(1,10,100,1000,1001,1001,1001,1002,1002,1002))

}
\seealso{
\code{\link{neighborhood}}
}
\author{
Thomas Alexander Gerds <tag@biostat.ku.dk>
}
\keyword{survival}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ggprodlim.R
\name{ggprodlim}
\alias{ggprodlim}
\title{ggplot2 support for prodlim objects}
\usage{
ggprodlim(
  x,
  type = x$type,
  cause,
  xlim,
  ylim,
  y_breaks,
  x_breaks,
  position_atrisk,
  atrisk_title = "Number at risk",
  atrisk_labels = NULL,
  atrisk_title_line = 0.14,
  atrisk_title_step = 0.06,
  atrisk_line_step = 0.04,
  atrisk_bottom_margin = NULL,
  facet_panel_spacing_y = NULL,
  conf_int = TRUE,
  conf_int_alpha = 0.2,
  percent = TRUE,
  timeconverter,
  color_var,
  facet_formula,
  palette = NULL,
  return_data = FALSE,
  ...
)
}
\arguments{
\item{x}{An object obtained with \code{\link{prodlim}}.}

\item{type}{Either \code{"surv"} or \code{"risk"}. Defaults to \code{x$type}.}

\item{cause}{For competing risk models: a cause (passed to
\code{\link{as.data.table.prodlim}}), \code{"all"} for all causes, or
\code{"stacked"} for a stacked cumulative incidence plot.}

\item{xlim, ylim}{Limits for the axes.}

\item{x_breaks, y_breaks}{Breaks for axes.#}

\item{position_atrisk}{Numeric vector specifying x-positions where numbers at
risk should be printed. If missing, a default grid is chosen.}

\item{atrisk_title}{Title shown left of the at-risk table.}

\item{atrisk_labels}{Optional replacement labels for strata/cause groups in
the at-risk table (in plotting order).}

\item{atrisk_title_line}{Numeric. Position of the at-risk title below the
plotting region, as an offset from \code{min(ylim)} in units of
\code{diff(ylim)}.}

\item{atrisk_title_step}{Numeric. Vertical distance from the title to the
first at-risk row, in units of \code{diff(ylim)}.}

\item{atrisk_line_step}{Numeric. Vertical spacing between at-risk rows, in
units of \code{diff(ylim)}.}

\item{atrisk_bottom_margin}{Numeric. Bottom plot margin in lines when the
at-risk table is shown.}

\item{facet_panel_spacing_y}{Optional numeric. Vertical spacing between facet
rows in lines. If \code{NULL} and faceting is used together with an at-risk
table, the default is \code{5 + n_colors}. The same value is used as the
default bottom plot margin below the lowest facet row.}

\item{conf_int}{Logical. If TRUE and \code{lower}/\code{upper} are available,
draw pointwise confidence intervals.}

\item{conf_int_alpha}{Alpha for CI shading.}

\item{percent}{Logical. If TRUE, y axis is shown in percent (0-100).}

\item{timeconverter}{Optional character, see \code{\link{plot.prodlim}} for
supported values (e.g., "years2months").}

\item{color_var}{Covariate name must be included in \code{colnames(x$X))} and
in competing risk models the word \code{"cause"}.}

\item{facet_formula}{Formula for facet_wrap. May include covariate names \code{colnames(x$X)} and
in competing risk models also the word \code{"cause"}.}

\item{palette}{Optional vector of colors. If provided, applied via
\code{scale_color_manual}/\code{scale_fill_manual}. Default NULL (do not
impose a palette).}

\item{return_data}{Logical. If TRUE, return a list with plot and data.}

\item{...}{Passed to \code{\link{as.data.table.prodlim}}; commonly \code{newdata},
\code{times}, and for competing risks \code{cause}.}
}
\value{
A ggplot object, or a list if \code{return_data=TRUE}.
}
\description{
ggplot2 support for Kaplan-Meier and Aalen-Johansen estimators
}
\details{
A ggplot2-based alternative to \code{\link{plot.prodlim}} that returns a
\code{ggplot} object, enabling ggplot theming, facetting, and composition.
}
\examples{
data(Melanoma,package = "riskRegression")
# marginal Kaplan-Maier
km <- prodlim(Hist(time,status != 0)~1,data = Melanoma)
ggprodlim(km)
# marginal Aalen-Johansen
aj <- prodlim(Hist(time,status)~1,data = Melanoma)
ggprodlim(aj)
ggprodlim(aj,facet_formula = ~cause)

ajsex <- prodlim(Hist(time,status)~sex,data = Melanoma)
ggprodlim(ajsex)
ggprodlim(ajsex,facet_formula = ~cause)
ggprodlim(ajsex,facet_formula = ~sex)
ggprodlim(ajsex,color_var = "sex")
ggprodlim(ajsex,color_var = "cause")

# stacking the causes
ggprodlim(ajsex,cause = "stacked")
# stacked with two covariates 
   ajepisex <- prodlim(Hist(time,status)~sex+epicel,data = Melanoma)
   ggprodlim(ajepisex,
             cause = "stacked",
             facet_formula = sex~epicel,
             atrisk_title_line = 0.2)

}
\seealso{
\code{\link{plot.prodlim}}, \code{\link{as.data.table.prodlim}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/confInt.R
\name{confInt}
\alias{confInt}
\title{Add point-wise confidence limits to the graphs of Kaplan-Meier and
Aalen-Johansen estimates.}
\usage{
confInt(ci, citype, col, lty, lwd, density = 55, ...)
}
\arguments{
\item{ci}{A \code{data.table} with columns \code{time}, \code{lower} and \code{upper}.}

\item{citype}{If \code{"shadow"} then confidence limits are drawn as colored
shadows.  Otherwise, dotted lines are used to show the upper and lower
confidence limits.}

\item{col}{the colour of the lines.}

\item{lty}{the line type of the lines.}

\item{lwd}{the line thickness of the lines.}

\item{density}{For \code{citype="shadow"}, the density of the shade. Default
is 55 percent.}

\item{\dots}{Further arguments that are passed to the function
\code{segments} if \code{type=="bars"} and to \code{lines} else.}
}
\value{
Nil
}
\description{
This function is invoked and controlled by \code{plot.prodlim}.
}
\details{
This function should not be called directly. The arguments can be specified
as \code{Confint.arg} in the call to \code{plot.prodlim}.
}
\seealso{
\code{\link{plot.prodlim}}, \code{\link{atRisk}},
\code{\link{markTime}}
}
\author{
Thomas Alexander Gerds <tag@biostat.ku.dk>
}
\keyword{survival}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plotCompetingRiskModel.R
\name{plotCompetingRiskModel}
\alias{plotCompetingRiskModel}
\title{Plotting a competing-risk-model.}
\usage{
plotCompetingRiskModel(stateLabels, horizontal = TRUE, ...)
}
\arguments{
\item{stateLabels}{Labels for the boxes.}

\item{horizontal}{The orientation of the plot.}

\item{\dots}{Arguments passed to \code{\link{plot.Hist}}.}
}
\description{
Plotting a competing-risk-model.
}
\examples{

plotCompetingRiskModel()
plotCompetingRiskModel(labels=c("a","b"))
plotCompetingRiskModel(labels=c("a","b","c"))

}
\seealso{
\code{\link{plotIllnessDeathModel}}, \code{\link{plot.Hist}}
}
\author{
Thomas Alexander Gerds <tag@biostat.ku.dk>
}
\keyword{survival}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/SmartControl.R
\name{SmartControl}
\alias{SmartControl}
\title{Function to facilitate the control of arguments passed to subroutines.}
\usage{
SmartControl(
  call,
  keys,
  ignore,
  defaults,
  forced,
  split,
  ignore.case = TRUE,
  replaceDefaults,
  verbose = TRUE
)
}
\arguments{
\item{call}{A list of named arguments, as for example can be obtained via
\code{list(...)}.}

\item{keys}{A vector of names of subroutines.}

\item{ignore}{A list of names which are removed from the argument
\code{call} before processing.}

\item{defaults}{A named list of default argument lists for the subroutines.}

\item{forced}{A named list of forced arguments for the subroutines.}

\item{split}{Regular expression used for splitting keys from arguments.
Default is \code{"\."}.}

\item{ignore.case}{If \code{TRUE} then all matching and splitting is not
case sensitive.}

\item{replaceDefaults}{If \code{TRUE} default arguments are replaced by
given arguments. Can also be a named list with entries for each subroutine.}

\item{verbose}{If \code{TRUE} warning messages are given for arguments in
\code{call} that are not ignored via argument \code{ignore} and that do not
match any \code{key}.}
}
\description{
Many R functions need to pass several arguments to several different
subroutines. Such arguments can are given as part of the three magic dots
"...". The function SmartControl reads the dots together with a list of
default values and returns for each subroutine a list of arguments.
}
\examples{


myPlot = function(...){
 ## set defaults
 plot.DefaultArgs=list(x=0,y=0,type="n")
 lines.DefaultArgs=list(x=1:10,lwd=3)
 ## apply smartcontrol
 x=SmartControl(call=list(...),
           defaults=list("plot"=plot.DefaultArgs, "lines"=lines.DefaultArgs),
        ignore.case=TRUE,keys=c("plot","axis2","lines"),
             forced=list("plot"=list(axes=FALSE),"axis2"=list(side=2)))
 ## call subroutines
 do.call("plot",x$plot)
 do.call("lines",x$lines)
 do.call("axis",x$axis2)
}
myPlot(plot.ylim=c(0,5),plot.xlim=c(0,20),lines.lty=3,axis2.At=c(0,3,4))

}
\seealso{
\code{\link{plot.prodlim}}
}
\author{
Thomas Alexander Gerds <tag@biostat.ku.dk>
}
\keyword{Graphics}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Hist.R
\name{Hist}
\alias{Hist}
\title{Create an event history object}
\usage{
Hist(
  time,
  event,
  entry = NULL,
  id = NULL,
  cens.code = "0",
  addInitialState = FALSE
)
}
\arguments{
\item{time}{for right censored data a numeric vector of event times -- for
interval censored data a list or a data.frame providing two numeric vectors
the left and right endpoints of the intervals.  See \code{Details}.}

\item{event}{A vector or a factor that specifies the events that occurred at
the corresponding value of \code{time}. Numeric, character and logical
values are recognized.  It can also be a list or a data.frame for the
longitudinal form of storing the data of a multi state model -- see
\code{Details}.}

\item{entry}{Vector of delayed entry times (left-truncation) or list of two
times when the entry time is interval censored.}

\item{id}{Identifies the subjects to which multiple events belong for the
longitudinal form of storing the data of a multi state model -- see
\code{Details}.}

\item{cens.code}{A character or numeric vector to identify the right
censored observations in the values of \code{event}.  Defaults to "0" which
is equivalent to 0.}

\item{addInitialState}{If TRUE, an initial state is added to all ids for the
longitudinal input form of a multi-state model.}
}
\value{
An object of class \code{Hist} for which there are print and plot
methods.  The object's internal is a matrix with some of the following
columns: \item{time}{ the right censored times} \item{L}{the left endpoints
of internal censored event times} \item{R}{the right endpoints of internal
censored event times} \item{status}{\code{0} for right censored, \code{1}
for exact, and \code{2} for interval censored event times.} \item{event}{an
integer valued numeric vector that codes the events.} \item{from}{an integer
valued numeric vector that codes the \code{from} states of a transition in a
multi state model.} \item{to}{an integer valued numeric vector that codes
the \code{to} states of a transition in a multi state model.}

Further information is stored in \code{\link{attributes}}.  The key to the
official names given to the events and the from and to states is stored in
an attribute "states".
}
\description{
Functionality for managing censored event history response data. The
function can be used as the left hand side of a formula: \code{Hist} serves
\code{\link{prodlim}} in a similar way as \code{\link[survival]{Surv}} from the
survival package serves `survfit'. \code{Hist} provides the suitable
extensions for dealing with right censored and interval censored data from
competing risks and other multi state models. Objects generated with
\code{Hist} have a print and a plot method.
}
\details{
*Specification of the event times*

If \code{time} is a numeric vector then the values are interpreted as right
censored event times, ie as the minimum of the event times and the censoring
times.

If \code{time} is a list with two elements or data frame with two numeric
columns The first element (column) is used as the left endpoints of interval
censored observations and the second as the corresponding right endpoints.
When the two endpoints are equal, then this observation is treated as an
exact uncensored observation of the event time.  If the value of the right
interval endpoint is either \code{NA} or \code{Inf}, then this observation
is treated as a right censored observation. Right censored observations can
also be specified by setting the value of \code{event} to \code{cens.code}.
This latter specification of right censored event times overwrites the
former: if \code{event} equals \code{cens.code} the observation is treated
as right censored no matter what the value of the right interval endpoint
is.

*Specification of the events*

If \code{event} is a numeric, character or logical vector then the order of
the attribute "state" given to the \code{value} of \code{Hist} is determined
by the order in which the values appear.  If it is a factor then the order
from the levels of the factor is used instead.

**Normal form of a multi state model**

If \code{event} is a list or a data.frame with exactly two elements, then
these describe the transitions in a multi state model that occurred at the
corresponding \code{time} as follows: The values of the first element are
interpreted as the \code{from} states of the transition and values of the
second as the corresponding \code{to} states.

**Longitudinal form of a multi state model**

If \code{id} is given then \code{event} must be a vector. In this case two
subsequent values of \code{event} belonging to the same value of \code{id}
are treated as the \code{from} and \code{to} states of the transitions.
}
\examples{


## Right censored responses of a two state model
## ---------------------------------------------

Hist(time=1:10,event=c(0,1,0,0,0,1,0,1,0,0))

## change the code for events and censored observations

Hist(time=1:10,event=c(99,"event",99,99,99,"event",99,"event",99,99),cens.code=99)

TwoStateFrame <- SimSurv(10)
SurvHist <- with(TwoStateFrame,Hist(time,status))
summary(SurvHist)
plot(SurvHist)

## Right censored data from a competing risk model
## --------------------------------------------------

CompRiskFrame <- data.frame(time=1:10,event=c(1,2,0,3,0,1,2,1,2,1))
CRHist <- with(CompRiskFrame,Hist(time,event))
summary(CRHist)
plot(CRHist)

## Interval censored data from a survival model
icensFrame <- data.frame(L=c(1,1,3,4,6),R=c(2,NA,3,6,9),event=c(1,1,1,2,2))
with(icensFrame,Hist(time=list(L,R)))

## Interval censored data from a competing risk model
with(icensFrame,Hist(time=list(L,R),event))

## Multi state model
MultiStateFrame <- data.frame(time=1:10,
			      from=c(1,1,3,1,2,4,1,1,2,1),
			      to=c(2,3,1,2,4,2,3,2,4,4))
with(MultiStateFrame,Hist(time,event=list(from,to)))

##  MultiState with right censored observations

MultiStateFrame1 <- data.frame(time=1:10,
			      from=c(1,1,3,2,1,4,1,1,3,1),
			      to=c(2,3,1,0,2,2,3,2,0,4))
with(MultiStateFrame1,Hist(time,event=list(from,to)))

## Using the longitudinal input method 
MultiStateFrame2 <- data.frame(time=c(0,1,2,3,4,0,1,2,0,1),
			      event=c(1,2,3,0,1,2,4,2,1,2),
			      id=c(1,1,1,1,2,2,2,2,3,3))
 with(MultiStateFrame2,Hist(time,event=event,id=id))

}
\seealso{
\code{\link{plot.Hist}}, \code{\link{summary.Hist}},
\code{\link{prodlim}}
}
\author{
Thomas A. Gerds \email{tag@biostat.ku.dk}, Arthur Allignol
\email{arthur.allignol@fdm.uni-freiburg.de}
}
\keyword{survival}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/strip.terms.R
\name{strip.terms}
\alias{strip.terms}
\title{Strip special functions from terms}
\usage{
strip.terms(
  terms,
  specials,
  alias.names = NULL,
  unspecials = NULL,
  arguments,
  keep.response = TRUE
)
}
\arguments{
\item{terms}{Terms object}

\item{specials}{Character vector of specials which should be
stripped off}

\item{alias.names}{Optional. A named list with alias names for the specials.}

\item{unspecials}{Optional. A special name for treating all the unspecial terms.}

\item{arguments}{A named list of arguments, one for each element
of specials. Elements are passed to \code{parseSpecialNames}.}

\item{keep.response}{Keep the response in the resulting object?}
}
\value{
Reformulated terms object with an additional attribute which contains the \code{stripped.specials}.
}
\description{
Reformulate a terms object such that some specials are stripped off
}
\details{
This function is used to remove special specials, i.e., those
which cannot or should not be evaluated. 
IMPORTANT: the unstripped terms need to know about all specials including the aliases.
See examples.
}
\examples{

## parse a survival formula and identify terms which
## should be treated as proportional or timevarying:
f <- Surv(time,status)~age+prop(factor(edema))+timevar(sex,test=0)+prop(bili,power=1)
tt <- terms(f,specials=c("prop","timevar"))
attr(tt,"specials")
st <- strip.terms(tt,specials=c("prop","timevar"),arguments=NULL)
formula(st)
attr(st,"specials")
attr(st,"stripped.specials")

## provide a default value for argument power of proportional treatment
## and argument test of timevarying treatment: 
st2 <- strip.terms(tt,
                   specials=c("prop","timevar"),
                   arguments=list("prop"=list("power"=0),"timevar"=list("test"=0)))
formula(st2)
attr(st2,"stripped.specials")
attr(st2,"stripped.arguments")

## treat all unspecial terms as proportional
st3 <- strip.terms(tt,
                   unspecials="prop",
                   specials=c("prop","timevar"),
                   arguments=list("prop"=list("power"=0),"timevar"=list("test"=0)))
formula(st3)
attr(st3,"stripped.specials")
attr(st3,"stripped.arguments")

## allow alias names: strata for timevar and tp, const for prop.
## IMPORTANT: the unstripped terms need to know about
## all specials including the aliases
f <- Surv(time,status)~age+const(factor(edema))+strata(sex,test=0)+prop(bili,power=1)+tp(albumin)
tt2 <- terms(f,specials=c("prop","timevar","strata","tp","const"))
st4 <- strip.terms(tt2,
                   specials=c("prop","timevar"),
                   unspecials="prop",
                   alias.names=list("timevar"="strata","prop"=c("const","tp")),
                   arguments=list("prop"=list("power"=0),"timevar"=list("test"=0)))
formula(st4)
attr(st4,"stripped.specials")
attr(st4,"stripped.arguments")

## test if alias works also without unspecial argument
st5 <- strip.terms(tt2,
                   specials=c("prop","timevar"),
                   alias.names=list("timevar"="strata","prop"=c("const","tp")),
                   arguments=list("prop"=list("power"=0),"timevar"=list("test"=0)))
formula(st5)
attr(st5,"stripped.specials")
attr(st5,"stripped.arguments")

library(survival)
data(pbc)
model.design(st4,data=pbc[1:3,],specialsDesign=TRUE)
model.design(st5,data=pbc[1:3,],specialsDesign=TRUE)


}
\seealso{
parseSpecialNames reformulate drop.terms
}
\author{
Thomas A. Gerds <tag@biostat.ku.dk>
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/print.prodlim.R
\name{print.prodlim}
\alias{print.prodlim}
\alias{print.neighborhood}
\alias{print.Hist}
\title{Print objects in the prodlim library}
\usage{
\method{print}{prodlim}(x, ...)
}
\arguments{
\item{x}{Object of class \code{prodlim}, \code{Hist} and
\code{neighborhood}.}

\item{\dots}{Not used.}
}
\description{
Pretty printing of objects created with the functionality of the `prodlim'
library.
}
\seealso{
\code{\link{summary.prodlim}}, \code{\link{predict.prodlim}}
}
\author{
Thomas Gerds <tag@biostat.ku.dk>
}
\keyword{survival}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/summary.prodlim.R
\name{summary.prodlim}
\alias{summary.prodlim}
\title{Summary method for prodlim objects.}
\usage{
\method{summary}{prodlim}(
  object,
  times,
  newdata,
  max.tables = 20,
  surv = TRUE,
  cause,
  intervals = FALSE,
  percent = FALSE,
  format = "df",
  ...
)
}
\arguments{
\item{object}{An object with class `prodlim' derived with
\code{\link{prodlim}}}

\item{times}{Vector of times at which to return the estimated
probabilities.}

\item{newdata}{A data frame with the same variable names as those
that appear on the right hand side of the 'prodlim' formula.
Defaults to \code{object$X}.}

\item{max.tables}{Integer. If \code{newdata} is not given the value
of \code{max.tables} decides about the maximal number of tables to
be shown.  Defaults to 20.}

\item{surv}{Logical. If FALSE report event probabilities instead of
survival probabilities. Only available for
\code{object$model=="survival"}.}

\item{cause}{For competing risk models. The event of interest for which predictions of the absolute risks are obtained by evaluating the cause-specific cumulative
incidence functions at  \code{times}.}

\item{intervals}{Logical. If TRUE count events and censored in
intervals between the values of \code{times}.}

\item{percent}{Logical. If TRUE all estimated values are multiplied
by 100 and thus interpretable on a percent scale.}

\item{format}{Control format of output. Since May 2021,
the result is a data.table and data.frame with attributes. When there are multiple
covariate strata or competing risks, these are indicated by columns.
Set format to \code{"list"} to get the old behaviour.}

\item{...}{Further arguments that are passed to the print
function.}
}
\value{
A data.frame with the relevant information.
}
\description{
Summarizing the result of the product limit method in life-table format.
Calculates the number of subjects at risk and counts events and censored
observations at specified times or in specified time intervals.
}
\details{
For cluster-correlated data the number of clusters at-risk are are also
given. Confidence intervals are displayed when they are part of the fitted
object.
}
\examples{

library(lava)
set.seed(17)
m <- survModel()
distribution(m,~age) <- uniform.lvm(30,80)
distribution(m,~sex) <- binomial.lvm()
m <- categorical(m,~z,K=3)
regression(m,eventtime~age) <- 0.01
regression(m,eventtime~sex) <- -0.4
d <- sim(m,50)
d$sex <- factor(d$sex,levels=c(0,1),labels=c("female","male"))
d$Z <- factor(d$z,levels=c(1,0,2),labels=c("B","A","C"))

# Univariate Kaplan-Meier
# -----------------------------------------------------------------------------------------
fit0 <- prodlim(Hist(time,event)~1,data=d)
summary(fit0)

## show survival probabilities as percentage and
## count number of events within intervals of a
## given time-grid:
summary(fit0,times=c(1,5,10,12),percent=TRUE,intervals=TRUE)

## the result of summary has a print function
## which passes ... to print and print.listof
sx <- summary(fit0,times=c(1,5,10,12),percent=TRUE,intervals=TRUE)
print(sx,digits=3)

## show absolute risks, i.e., cumulative incidences (1-survival)
summary(fit0,times=c(1,5,10,12),surv=FALSE,percent=TRUE,intervals=TRUE)

# Stratified Kaplan-Meier
# -----------------------------------------------------------------------------------------

fit1 <- prodlim(Hist(time,event)~sex,data=d)
print(summary(fit1,times=c(1,5,10),intervals=TRUE,percent=TRUE),digits=3)
# old behaviour
print(summary(fit1,times=c(1,5,10),intervals=TRUE,percent=TRUE,format="list"),digits=3)

summary(fit1,times=c(1,5,10),intervals=TRUE,percent=TRUE)

fit2 <- prodlim(Hist(time,event)~Z,data=d)
print(summary(fit2,times=c(1,5,10),intervals=TRUE,percent=TRUE),digits=3)

## Continuous strata (Beran estimator)
# -----------------------------------------------------------------------------------------
fit3 <- prodlim(Hist(time,event)~age,data=d)
print(summary(fit3,
              times=c(1,5,10),
              newdata=data.frame(age=c(20,50,70)),
              intervals=TRUE,
              percent=TRUE),digits=3)

## stratified Beran estimator
# -----------------------------------------------------------------------------------------
fit4 <- prodlim(Hist(time,event)~age+sex,data=d)
print(summary(fit4,
              times=c(1,5,10),
              newdata=data.frame(age=c(20,50,70),sex=c("female","male","male")),
              intervals=TRUE,
              percent=TRUE),digits=3)

print(summary(fit4,
              times=c(1,5,10),
              newdata=data.frame(age=c(20,50,70),sex=c("female","male","male")),
              intervals=TRUE,
              percent=TRUE),digits=3)

## assess results from summary
x <- summary(fit4,times=10,newdata=expand.grid(age=c(60,40,50),sex=c("male","female")))
cbind(names(x$table),do.call("rbind",lapply(x$table,round,2)))

x <- summary(fit4,times=10,newdata=expand.grid(age=c(60,40,50),sex=c("male","female")))

## Competing risks: Aalen-Johansen
# -----------------------------------------------------------------------------------------
d <- SimCompRisk(30)
crfit <- prodlim(Hist(time,event)~X1,data=d)
summary(crfit,times=c(1,2,5))
summary(crfit,times=c(1,2,5),cause=1,intervals=TRUE)
summary(crfit,times=c(1,2,5),cause=1)
summary(crfit,times=c(1,2,5),cause=1:2)


# extract the actual tables from the summary 
sumfit <- summary(crfit,times=c(1,2,5),print=FALSE)
sumfit$table[[1]] # cause 1
sumfit$table[[2]] # cause 2


# '
}
\seealso{
\code{\link{prodlim}}, \code{\link{summary.Hist}}
}
\author{
Thomas A. Gerds \email{tag@biostat.ku.dk}
}
\keyword{survival}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/survModel.R
\name{survModel}
\alias{survModel}
\title{Survival model for simulation}
\usage{
survModel()
}
\value{
A structural equation model initialized with three variables: the
latent event time, the latent right censored time, and the observed
right censored event time.
}
\description{
Create a survival model to simulate a right censored event time data without
covariates
}
\details{
This function requires the \code{lava} package.
}
\author{
Thomas A. Gerds <tag@biostat.ku.dk>
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/predict.prodlim.R
\name{predict.prodlim}
\alias{predict.prodlim}
\alias{predictSurv}
\alias{predictAbsrisk}
\alias{predictCuminc}
\title{Predicting event probabilities from product limit estimates}
\usage{
\method{predict}{prodlim}(
  object,
  times,
  newdata,
  level.chaos = 1,
  type = c("surv", "risk", "cuminc", "list"),
  mode = "list",
  bytime = FALSE,
  cause,
  ...
)
}
\arguments{
\item{object}{A fitted object of class "prodlim".}

\item{times}{Vector of times at which to return the estimated probabilities (survival or absolute event risks).}

\item{newdata}{A data frame with the same variable names as those that
appear on the right hand side of the 'prodlim' formula.  If there are
covariates this argument is required.}

\item{level.chaos}{Integer specifying the sorting of the output: `0' sort by
time and newdata; `1' only by time; `2' no sorting at all}

\item{type}{Choice between "surv","risk","cuminc","list":

"surv": predict survival probabilities only survival models

"risk"/"cuminc": predict absolute risk, i.e., cumulative incidence function.

"list": find the indices corresponding to times and newdata. See value.

Defaults to "surv" for two-state models and to "risk" for competing risk
models.}

\item{mode}{Only for \code{type=="surv"} and \code{type=="risk"}. Can
either be "list" or "matrix". For "matrix" the predicted probabilities will
be returned in matrix form.}

\item{bytime}{Logical. If TRUE and \code{mode=="matrix"} the matrix with
predicted probabilities will have a column for each time and a row for each
newdata. Only when \code{object$covariate.type>1} and more than one time is
given.}

\item{cause}{Character (other classes are converted with \code{as.character}).
The cause for predicting the absolute risk of an event, i.e., the cause-specific cumulative
incidence function, in competing risk models. At any time after time zero this is the absolute risk of 
an event of type \code{cause} to occur between time zero and \code{times} .}

\item{\dots}{Only for compatibility reasons.}
}
\value{
\code{type=="surv"} A list or a matrix with survival probabilities
for all times and all newdata.

\code{type=="risk"} or \code{type=="cuminc"} A list or a matrix with cumulative incidences for all
times and all newdata.

\code{type=="list"} A list with the following components:

\item{times}{The argument \code{times} carried forward}

\item{predictors}{The relevant part of the argument \code{newdata}.}
\item{indices}{ A list with the following components

\code{time}: Where to find values corresponding to the requested times
\code{strata}: Where to find values corresponding to the values of the
variables in newdata.  Together time and strata show where to find the
predicted probabilities.  } \item{dimensions}{ a list with the following
components: \code{time} : The length of \code{times} \code{strata} : The
number of rows in \code{newdata} \code{names.strata} : Labels for the
covariate values.  }
}
\description{
Evaluation of estimated survival or event probabilities at given times and
covariate constellations.
}
\details{
Predicted (survival) probabilities are returned that can be plotted,
summarized and used for inverse of probability of censoring weighting.
}
\examples{


dat <- SimSurv(400)
fit <- prodlim(Hist(time,status)~1,data=dat)

## predict the survival probs at selected times 
predict(fit,times=c(3,5,10))

## NA is returned when the time point is beyond the
## range of definition of the Kaplan-Meier estimator:
predict(fit,times=c(-1,0,10,100,1000,10000))

## when there are strata, newdata is required 
## or neighborhoods (i.e. overlapping strata)
mfit <- prodlim(Hist(time,status)~X1+X2,data=dat)
predict(mfit,times=c(-1,0,10,100,1000,10000),newdata=dat[18:21,])

## this can be requested in matrix form
predict(mfit,times=c(-1,0,10,100,1000,10000),newdata=dat[18:21,],mode="matrix")

## and even transposed
predict(mfit,times=c(-1,0,10,100,1000,10000),newdata=dat[18:21,],mode="matrix",bytime=TRUE)

}
\seealso{
\code{\link{predictSurvIndividual}}
}
\author{
Thomas Alexander Gerds <tag@biostat.ku.dk>
}
\keyword{survival}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/crModel.R
\name{crModel}
\alias{crModel}
\title{Competing risks model for simulation}
\usage{
crModel()
}
\value{
A structural equation model initialized with four variables: the
latent event times of two causes, the latent right censored time, and the observed
right censored event time.
}
\description{
Competing risks model for simulation
}
\details{
Create a competing risks model with to causes to simulate a right censored event time data without
covariates

This function requires the \code{lava} package.
}
\examples{
library(lava)
m <- crModel()
d <- sim(m,6)
print(d)

}
\author{
Thomas A. Gerds
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plot.prodlim.R
\name{plot.prodlim}
\alias{plot.prodlim}
\alias{lines.prodlim}
\title{Plotting event probabilities over time}
\usage{
\method{plot}{prodlim}(
  x,
  type,
  cause,
  select,
  newdata,
  add = FALSE,
  col,
  lty,
  lwd,
  ylim,
  xlim,
  ylab,
  xlab = "Time",
  num.digits = 2,
  timeconverter,
  legend = TRUE,
  short.labels = TRUE,
  logrank = FALSE,
  marktime = FALSE,
  confint = TRUE,
  automar,
  atrisk = ifelse(add, FALSE, TRUE),
  timeOrigin = 0,
  axes = TRUE,
  background = TRUE,
  percent = TRUE,
  minAtrisk = 0,
  limit = 10,
  ...
)
}
\arguments{
\item{x}{an object of class `prodlim' as returned by the
\code{prodlim} function.}

\item{type}{Either \code{"surv"} or \code{"risk"} AKA \code{"cuminc"}. Controls what
part of the object is plotted.  Defaults to \code{object$type}.}

\item{cause}{For competing risk models. Character (other classes are converted with \code{as.character}).
The argument \code{cause} determines the event of interest. Currently one cause is allowed at a time, but you can
call the function again with \code{add=TRUE} to add the lines of the other
causes. Also, if \code{cause="stacked"} is specified the absolute risks of all causes are stacked.}

\item{select}{Select which lines to plot. This can be used when
there are many strata or many competing risks to select a
subset of the lines.  However, a more clean way to select
covariate strata is to use the argument \code{newdata}. Another
application is when there are several competing risks and the 
 stacked plot (\code{cause="stacked"}) should only show a selected subset
of the available causes.}

\item{newdata}{a data frame containing covariate strata for which
to show curves. When omitted element \code{X} of object
\code{x} is used.}

\item{add}{if \code{TRUE} curves are added to an existing plot.}

\item{col}{color for curves. Default is \code{1:number(curves)}}

\item{lty}{line type for curves. Default is 1.}

\item{lwd}{line width for all curves. Default is 3.}

\item{ylim}{limits of the y-axis}

\item{xlim}{limits of the x-axis}

\item{ylab}{label for the y-axis}

\item{xlab}{label for the x-axis}

\item{num.digits}{Number of digits when rounding off numerical values for legend and at-risk tables.}

\item{timeconverter}{The following options are supported:
"days2years" (conversion factor: 1/365.25)
"months2years" (conversion factor: 1/12)
"days2months" (conversion factor 1/30.4368499)
"years2days" (conversion factor 365.25)
"years2months" (conversion factor 12)
 "months2days" (conversion factor 30.4368499)}

\item{legend}{if TRUE a legend is plotted by calling the function
legend.  Optional arguments of the function \code{legend} can
be given in the form \code{legend.x=val} where x is the name of
the argument and val the desired value. See also Details.}

\item{short.labels}{Logical. When \code{FALSE} construct labels as cause=1, var1=v1, var2=v2 else as 1, v1, v2.}

\item{logrank}{If TRUE, the logrank p-value will be extracted from
a call to \code{survdiff} and added to the legend. This works
only for survival models, i.e. Kaplan-Meier with discrete
predictors.}

\item{marktime}{if TRUE the curves are tick-marked at right
censoring times by invoking the function
\code{markTime}. Optional arguments of the function
\code{markTime} can be given in the form \code{confint.x=val}
as with legend. See also Details.}

\item{confint}{Logical. If \code{TRUE} pointwise confidence intervals are plotted
by invoking the function \code{confInt}. Optional arguments of
the function \code{confInt} can be given in the form
\code{confint.x=val} as with legend.  See also Details.}

\item{automar}{If TRUE the function trys to find suitable values
for the figure margins around the main plotting region.}

\item{atrisk}{if TRUE display numbers of subjects at risk by
invoking the function \code{atRisk}. Optional arguments of the
function \code{atRisk} can be given in the form
\code{atrisk.x=val} as with legend. See also Details.}

\item{timeOrigin}{Start of the time axis}

\item{axes}{If true axes are drawn. See details.}

\item{background}{If \code{TRUE} the background color and grid
color can be controlled using smart arguments SmartControl,
such as background.bg="yellow" or
background.bg=c("gray66","gray88").  The following defaults are
passed to \code{background} by \code{plot.prodlim}:
horizontal=seq(0,1,.25), vertical=NULL, bg="gray77",
fg="white".  See \code{background} for all arguments, and the
examples below.}

\item{percent}{If true the y-axis is labeled in percent.}

\item{minAtrisk}{Integer. Show the curve only until the number
at-risk is at least \code{minAtrisk}}

\item{limit}{When newdata is not specified and the number of lines
in element \code{X} of object \code{x} exceeds limits, only the
results for covariate constellations of the first, the middle
and the last row in \code{X} are shown. Otherwise all lines of
\code{X} are shown.}

\item{...}{Parameters that are filtered by
\code{\link{SmartControl}} and then passed to the functions
\code{\link{plot}}, \code{\link{legend}}, \code{\link{axis}},
\code{\link{atRisk}}, \code{\link{confInt}},
\code{\link{markTime}}, \code{\link{backGround}}}
}
\value{
The (invisible) object.
}
\description{
Function to plot survival probabilities or absolute risks (cumulative incidence function) against time.
}
\details{
From version 1.1.3 on the arguments legend.args, atrisk.args, confint.args
are obsolete and only available for backward compatibility. Instead
arguments for the invoked functions \code{atRisk}, \code{legend},
\code{confInt}, \code{markTime}, \code{axis} are simply specified as
\code{atrisk.cex=2}. The specification is not case sensitive, thus
\code{atRisk.cex=2} or \code{atRISK.cex=2} will have the same effect.  The
function \code{axis} is called twice, and arguments of the form
\code{axis1.labels}, \code{axis1.at} are used for the time axis whereas
\code{axis2.pos}, \code{axis1.labels}, etc. are used for the y-axis.

These arguments are processed via \code{\dots{}} of \code{plot.prodlim} and
inside by using the function \code{SmartControl}.  Documentation of these
arguments can be found in the help pages of the corresponding functions.
}
\examples{
## simulate right censored data from a two state model 
set.seed(100)
dat <- SimSurv(100)
# with(dat,plot(Hist(time,status)))

### marginal Kaplan-Meier estimator
kmfit <- prodlim(Hist(time, status) ~ 1, data = dat)
plot(kmfit)
plot(kmfit,atrisk.show.censored=1L,atrisk.at=seq(0,12,3))
plot(kmfit,timeconverter="years2months")

# change time range
plot(kmfit,xlim=c(0,4))

# change scale of y-axis
plot(kmfit,percent=FALSE)

# mortality instead of survival
plot(kmfit,type="risk")

# change axis label and position of ticks
plot(kmfit,
     xlim=c(0,10),
     axis1.at=seq(0,10,1),
     axis1.labels=0:10,
     xlab="Years",
     axis2.las=2,
     atrisk.at=seq(0,10,2.5),
     atrisk.title="")

# change background color
plot(kmfit,
     xlim=c(0,10),
     confint.citype="shadow",
     col=1,
     axis1.at=0:10,
     axis1.labels=0:10,
     xlab="Years",
     axis2.las=2,
     atrisk.at=seq(0,10,2.5),
     atrisk.title="",
     background=TRUE,
     background.fg="white",
     background.horizontal=seq(0,1,.25/2),
     background.vertical=seq(0,10,2.5),
     background.bg=c("gray88"))

# change type of confidence limits
plot(kmfit,
     xlim=c(0,10),
     confint.citype="dots",
     col=4,
     background=TRUE,
     background.bg=c("white","gray88"),
     background.fg="gray77",
     background.horizontal=seq(0,1,.25/2),
     background.vertical=seq(0,10,2))


### Kaplan-Meier in discrete strata
kmfitX <- prodlim(Hist(time, status) ~ X1, data = dat)
plot(kmfitX,atrisk.show.censored=1L)
# move legend
plot(kmfitX,legend.x="bottomleft",atRisk.cex=1.3,
     atrisk.title="No. subjects")

## Control the order of strata
## since version 1.5.1 prodlim does  obey the order of
## factor levels
dat$group <- factor(cut(dat$X2,c(-Inf,0,0.5,Inf)),
                    labels=c("High","Intermediate","Low"))
kmfitG <- prodlim(Hist(time, status) ~ group, data = dat)
plot(kmfitG)

## relevel 
dat$group2 <- factor(cut(dat$X2,c(-Inf,0,0.5,Inf)),
                     levels=c("(0.5, Inf]","(0,0.5]","(-Inf,0]"),
                     labels=c("Low","Intermediate","High"))
kmfitG2 <- prodlim(Hist(time, status) ~ group2, data = dat)
plot(kmfitG2)

# add log-rank test to legend
plot(kmfitX,
     atRisk.cex=1.3,
     logrank=TRUE,
     legend.x="topright",
     atrisk.title="at-risk")

# change atrisk labels
plot(kmfitX,
     legend.x="bottomleft",
     atrisk.title="Patients",
     atrisk.cex=0.9,
     atrisk.labels=c("X1=0","X1=1"))

# multiple categorical factors

kmfitXG <- prodlim(Hist(time,status)~X1+group2,data=dat)
plot(kmfitXG,select=1:2)

### Kaplan-Meier in continuous strata
kmfitX2 <- prodlim(Hist(time, status) ~ X2, data = dat)
plot(kmfitX2,xlim=c(0,10))

# specify values of X2 for which to show the curves 
plot(kmfitX2,xlim=c(0,10),newdata=data.frame(X2=c(-1.8,0,1.2)))

### Cluster-correlated data
library(survival)
cdat <- cbind(SimSurv(20),patnr=sample(1:5,size=20,replace=TRUE))
kmfitC <- prodlim(Hist(time, status) ~ cluster(patnr), data = cdat)
plot(kmfitC)
plot(kmfitC,atrisk.labels=c("Units","Patients"))

kmfitC2 <- prodlim(Hist(time, status) ~ X1+cluster(patnr), data = cdat)
plot(kmfitC2)
plot(kmfitC2,atrisk.labels=c("Teeth","Patients","Teeth","Patients"),
     atrisk.col=c(1,1,2,2))


### Cluster-correlated data with strata
n = 50
foo = runif(n)
bar = rexp(n)
baz = rexp(n,1/2)
d = stack(data.frame(foo,bar,baz))
d$cl = sample(10, 3*n, replace=TRUE)
fit = prodlim(Surv(values) ~ ind + cluster(cl), data=d)
plot(fit)


## simulate right censored data from a competing risk model 
datCR <- SimCompRisk(100)
with(datCR,plot(Hist(time,event)))

### marginal Aalen-Johansen estimator
ajfit <- prodlim(Hist(time, event) ~ 1, data = datCR)
plot(ajfit) # same as plot(ajfit,cause=1)
plot(ajfit,atrisk.show.censored=1L)

# cause 2
plot(ajfit,cause=2)

# both in one
plot(ajfit,cause=1)
plot(ajfit,cause=2,add=TRUE,col=2)

### stacked plot

plot(ajfit,cause="stacked",select=2)

### stratified Aalen-Johansen estimator
ajfitX1 <- prodlim(Hist(time, event) ~ X1, data = datCR)
plot(ajfitX1)

## add total number at-risk to a stratified curve
ttt = 1:10
plot(ajfitX1,atrisk.at=ttt,col=2:3)
plot(ajfit,add=TRUE,col=1)
atRisk(ajfit,newdata=datCR,col=1,times=ttt,line=3,labels="Total")


## stratified Aalen-Johansen estimator in nearest neighborhoods
## of a continuous variable
ajfitX <- prodlim(Hist(time, event) ~ X1+X2, data = datCR)
plot(ajfitX,newdata=data.frame(X1=c(1,1,0),X2=c(4,10,10)))
plot(ajfitX,newdata=data.frame(X1=c(1,1,0),X2=c(4,10,10)),cause=2)

## stacked plot

plot(ajfitX,
     newdata=data.frame(X1=0,X2=0.1),
     cause="stacked",
     legend.title="X1=0,X2=0.1",
     legend.legend=paste("cause:",getStates(ajfitX$model.response)),
     plot.main="Subject specific stacked plot")
 
}
\seealso{
\code{\link{plot}}, \code{\link{legend}},
    \code{\link{axis}},
    \code{\link{prodlim}},\code{\link{plot.Hist}},\code{\link{summary.prodlim}},
    \code{\link{neighborhood}}, \code{\link{atRisk}},
    \code{\link{confInt}}, \code{\link{markTime}},
    \code{\link{backGround}}
}
\author{
Thomas Alexander Gerds <tag@biostat.ku.dk>
}
\keyword{survival}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/prodlim.R
\name{prodlim}
\alias{prodlim}
\title{product limit method}
\usage{
prodlim(
  formula,
  data = parent.frame(),
  subset,
  na.action = NULL,
  reverse = FALSE,
  conf.int = 0.95,
  bandwidth = NULL,
  caseweights,
  discrete.level = 3,
  x = TRUE,
  maxiter = 1000,
  grid,
  tol = 7,
  method = c("npmle", "one.step", "impute.midpoint", "impute.right"),
  exact = TRUE,
  type
)
}
\arguments{
\item{formula}{A formula whose left hand side is a \code{Hist}
object. In some special cases it can also be a \code{\link[survival]{Surv}}
response object, see the details section. The right hand side is
as usual a linear combination of covariates which may contain at
most one continuous factor. Whether or not a covariate is
recognized as continuous or discrete depends on its class and on
the argument \code{discrete.level}. The right hand side may also
be used to specify clusters, see the details section.}

\item{data}{A data.frame in which all the variables of
\code{formula} can be interpreted.}

\item{subset}{Passed as argument \code{subset} to function
\code{subset} which applied to \code{data} before the formula is
processed.}

\item{na.action}{All lines in data with any missing values in the
variables of formula are removed.}

\item{reverse}{For right censored data, if reverse=TRUE then the
censoring distribution is estimated.}

\item{conf.int}{The level (between 0 and 1) for two-sided
pointwise confidence intervals. Defaults to 0.95. Remark: only
plain Wald-type confidence limits are available.}

\item{bandwidth}{Smoothing parameter for nearest neighborhoods
based on the values of a continuous covariate. See function
\code{neighborhood} for details.}

\item{caseweights}{Weights applied to the contribution of each
subject to change the number of events and the number at
risk. This can be used for bootstrap and survey analysis. Should
be a vector of the same length and the same order as \code{data}.}

\item{discrete.level}{Numeric covariates are treated as factors
when their number of unique values exceeds not
\code{discrete.level}. Otherwise the product limit method is
applied, in overlapping neighborhoods according to the bandwidth.}

\item{x}{logical value: if \code{TRUE}, the full covariate matrix
with is returned in component \code{model.matrix}.  The reduced
matrix contains unique rows of the full covariate matrix and is
always returned in component \code{X}.}

\item{maxiter}{For interval censored data only.  Maximal number of
iterations to obtain the nonparametric maximum likelihood
estimate.  Defaults to 1000.}

\item{grid}{For interval censored data only. When method=one.step
grid for one-step product limit estimate. Defaults to sorted list
of unique left and right endpoints of the observed intervals.}

\item{tol}{For interval censored data only. Numeric value whose
negative exponential is used as convergence criterion for finding
the nonparametric maximum likelihood estimate.  Defaults to 7
meaning exp(-7).}

\item{method}{For interval censored data only.  If equal to
\code{"npmle"} (the default) use the usual Turnbull algorithm,
else the product limit version of the self-consistent estimate.}

\item{exact}{If TRUE the grid of time points used for estimation
includes all the L and R endpoints of the observed intervals.}

\item{type}{In two state models either \code{"surv"} for the Kaplan-Meier estimate of the survival
function or \code{"risk"} for 1-Kaplan-Meier. Default is \code{"surv"} when \code{reverse==FALSE} and \code{"risk"} when \code{reverse==TRUE}.
In competing risks models it has to be \code{"risk"}
Aalen-Johansen estimate of the cumulative incidence function.}
}
\value{
Object of class "prodlim". See \code{\link{print.prodlim}}, \code{\link{predict.prodlim}}, predict,
\code{\link{summary.prodlim}}, \code{\link{plot.prodlim}}.
}
\description{
Nonparametric estimation in event history analysis. Featuring fast
algorithms and user friendly syntax adapted from the survival package.  The
product limit algorithm is used for right censored data; the
self-consistency algorithm for interval censored data.
}
\details{
The response of \code{formula} (ie the left hand side of the `~' operator)
specifies the model.

In two-state models -- the classical survival case -- the standard
Kaplan-Meier method is applied.  For this the response can be specified as a
\code{\link[survival]{Surv}} or as a \code{\link{Hist}} object. The \code{\link{Hist}}
function allows you to change the code for censored observations, e.g.
\code{Hist(time,status,cens.code="4")}.

Besides a slight gain of computing efficiency, there are some extensions
that are not included in the current version of the survival package:

(0) The Kaplan-Meier estimator for the censoring times \code{reverse=TRUE}
is correctly estimated when there are ties between event and censoring
times.

(1) A conditional version of the kernel smoothed Kaplan-Meier estimator for at most one
continuous predictors using nearest neighborhoods (Beran 1981,
Stute 1984, Akritas 1994).

(2) For cluster-correlated data the right hand side of \code{formula} may
identify a \code{\link[survival]{cluster}} variable. In that case Greenwood's variance
formula is replaced by the formula of Ying and Wei (1994).

(3) Competing risk models can be specified via \code{\link{Hist}} response
objects in \code{formula}.

The Aalen-Johansen estimator is applied for estimating the absolute risk of the competing causes, i.e., the cumulative
incidence functions. 

Under construction:

(U0) Interval censored event times specified via \code{\link{Hist}} are used
to find the nonparametric maximum likelihood estimate. Currently this works
only for two-state models and the results should match with those from the
package `Icens'.

(U1) Extensions to more complex multi-states models

(U2) The nonparametric maximum likelihood estimate for interval censored
observations of competing risks models.
}
\examples{

##---------------------two-state survival model------------
dat <- SimSurv(30)
with(dat,plot(Hist(time,status)))
fit <- prodlim(Hist(time,status)~1,data=dat)
print(fit)
plot(fit)
summary(fit)
quantile(fit)

## Subset
fit1a <- prodlim(Hist(time,status)~1,data=dat,subset=dat$X1==1)
fit1b <- prodlim(Hist(time,status)~1,data=dat,subset=dat$X1==1 & dat$X2>0)

## --------------------clustered data---------------------
library(survival)
cdat <- cbind(SimSurv(30),patnr=sample(1:5,size=30,replace=TRUE))
fit <- prodlim(Hist(time,status)~cluster(patnr),data=cdat)
print(fit)
plot(fit)
summary(fit)


##-----------compare Kaplan-Meier to survival package---------

dat <- SimSurv(30)
pfit <- prodlim(Surv(time,status)~1,data=dat)
pfit <- prodlim(Hist(time,status)~1,data=dat) ## same thing
sfit <- survfit(Surv(time,status)~1,data=dat,conf.type="plain")
##  same result for the survival distribution function 
all(round(pfit$surv,12)==round(sfit$surv,12))
summary(pfit,digits=3)
summary(sfit,times=quantile(unique(dat$time)))

##-----------estimating the censoring survival function----------------

rdat <- data.frame(time=c(1,2,3,3,3,4,5,5,6,7),status=c(1,0,0,1,0,1,0,1,1,0))
rpfit <- prodlim(Hist(time,status)~1,data=rdat,reverse=TRUE)
rsfit <- survfit(Surv(time,1-status)~1,data=rdat,conf.type="plain")
## When there are ties between times at which events are observed
## times at which subjects are right censored, then the convention
## is that events come first. This is not obeyed by the above call to survfit,
## and hence only prodlim delivers the correct reverse Kaplan-Meier:
cbind("Wrong:"=rsfit$surv,"Correct:"=rpfit$surv)
##----------------  quantiles of the potential followup time-----

G=prodlim(Hist(time,status)~X1,data=dat,reverse=TRUE)
quantile(G)

##-------------------stratified Kaplan-Meier---------------------

stratfit <- prodlim(Surv(time,status)~X1,data=dat)
summary(stratfit)
summary(stratfit,intervals=TRUE)
plot(stratfit)

##----------continuous covariate: Stone-Beran estimate------------

SB=prodlim(Surv(time,status)~X2,data=dat)
summary(SB,newdata=data.frame(X2=c(-0.3,0,.5)))

##-------------both discrete and continuous covariates------------

prodlim(Surv(time,status)~X2+X1,data=dat)

##----------------------interval censored data----------------------

dat <- data.frame(L=1:10,R=c(2,3,12,8,9,10,7,12,12,12),status=c(1,1,0,1,1,1,1,0,0,0))
with(dat,Hist(time=list(L,R),event=status))

dat$event=1
npmle.fitml <- prodlim(Hist(time=list(L,R),event)~1,data=dat)

##-------------competing risks-------------------

CompRiskFrame <- data.frame(time=1:100,event=rbinom(100,2,.5),X=rbinom(100,1,.5))
crFit <- prodlim(Hist(time,event)~X,data=CompRiskFrame)
summary(crFit)
plot(crFit)
summary(crFit,cause=2)
plot(crFit,cause=2)


# Changing the cens.code:
dat <- data.frame(time=1:10,status=c(1,2,1,2,5,5,1,1,2,2))
fit <- prodlim(Hist(time,status)~1,data=dat)
print(fit$model.response)
fit <- prodlim(Hist(time,status,cens.code="2")~1,data=dat)
print(fit$model.response)
plot(fit)
plot(fit,cause="5")


##------------delayed entry----------------------

## left-truncated event times with competing risk endpoint 

dat <- data.frame(entry=c(7,3,11,12,11,2,1,7,15,17,3),time=10:20,status=c(1,0,2,2,0,0,1,2,0,2,0))
fitd <- prodlim(Hist(time=time,event=status,entry=entry)~1,data=dat)
summary(fitd)
plot(fitd)

}
\references{
Andersen, Borgan, Gill, Keiding (1993) Springer `Statistical
Models Based on Counting Processes'

Akritas (1994) The Annals of Statistics 22, 1299-1327 Nearest neighbor
estimation of a bivariate distribution under random censoring.

R Beran (1981) http://anson.ucdavis.edu/~beran/paper.html `Nonparametric
regression with randomly censored survival data'

Stute (1984) The Annals of Statistics 12, 917--926 `Asymptotic Normality of
Nearest Neighbor Regression Function Estimates'

Ying, Wei (1994) Journal of Multivariate Analysis 50, 17-29 The Kaplan-Meier
estimate for dependent failure time observations
}
\seealso{
\code{\link{predictSurv}}, \code{\link{predictSurvIndividual}},
\code{\link{predictAbsrisk}}, \code{\link{Hist}}, \code{\link{neighborhood}},
\code{\link[survival]{Surv}}, \code{\link[survival]{survfit}}, \code{\link[survival]{strata}},
}
\author{
Thomas A. Gerds \email{tag@biostat.ku.dk}
}
\keyword{cluster}
\keyword{nonparametric}
\keyword{survival}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/backGround.R
\name{backGround}
\alias{backGround}
\title{Background and grid color control.}
\usage{
backGround(
  xlim,
  ylim,
  bg = "white",
  fg = "gray77",
  horizontal = NULL,
  vertical = NULL,
  border = "black"
)
}
\arguments{
\item{xlim}{Limits for the xaxis, defaults to par("usr")[1:2].}

\item{ylim}{Limits for the yaxis, defaults to par("usr")[3:4].}

\item{bg}{Background color. Can be multiple colors which are then switched
at each horizontal line.}

\item{fg}{Grid line color.}

\item{horizontal}{Numerical values at which horizontal grid lines are
plotted.}

\item{vertical}{Numerical values at which vertical grid lines are plotted.}

\item{border}{The color of the border around the background.}
}
\description{
Some users like background colors, and it may be helpful to have grid lines
to read off e.g. probabilities from a Kaplan-Meier graph. Both things can be
controlled with this function. However, it mainly serves
\code{\link{plot.prodlim}}.
}
\examples{


plot(0,0)
backGround(bg="beige",fg="red",vertical=0,horizontal=0)

plot(0,0)
backGround(bg=c("yellow","green"),fg="red",xlim=c(-1,1),ylim=c(-1,1),horizontal=seq(0,1,.1))
backGround(bg=c("yellow","green"),fg="red",horizontal=seq(0,1,.1))

}
\author{
Thomas Alexander Gerds <tag@biostat.ku.dk>
}
\keyword{survival}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/PercentAxis.R
\name{PercentAxis}
\alias{PercentAxis}
\title{Percentage-labeled axis.}
\usage{
PercentAxis(x, at, ...)
}
\arguments{
\item{x}{Side of the axis}

\item{at}{Positions (decimals) at which to label the axis.}

\item{\dots}{Given to \code{axis}.}
}
\description{
Use percentages instead of decimals to label the an axis with a probability
scale .
}
\examples{

  plot(0,0,xlim=c(0,1),ylim=c(0,1),axes=FALSE)
  PercentAxis(1,at=seq(0,1,.25))
  PercentAxis(2,at=seq(0,1,.25))

}
\seealso{
\code{\link{plot.prodlim}}
}
\author{
Thomas Alexander Gerds
}
\keyword{survival}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sindex.R
\name{sindex}
\alias{sindex}
\title{Index for evaluation of step functions.}
\usage{
sindex(jump.times, eval.times, comp = "smaller", strict = FALSE)
}
\arguments{
\item{jump.times}{Numeric vector: e.g. the unique jump times of a step
function.}

\item{eval.times}{Numeric vector: e.g. the times where the step function
should be evaluated}

\item{comp}{If "greater" count the number of jump times that are greater
(greater or equal when strict==FALSE) than the eval times}

\item{strict}{If TRUE make the comparison of jump times and eval times
strict}
}
\value{
Index of the same length as \code{eval.times} containing the numbers
of the \code{jump.times} that are smaller than or equal to
\code{eval.times}.
}
\description{
Returns an index of positions. Intended for evaluating a step function at
selected times. The function counts how many elements of a vector, e.g. the
jump times of the step function, are smaller or equal to the elements in a
second vector, e.g. the times where the step function should be evaluated.
}
\details{
If all \code{jump.times} are greater than a particular \code{eval.time} the
sindex returns \code{0}. This must be considered when sindex is used for
subsetting, see the Examples below.
}
\examples{


test <- list(time = c(1, 1,5,5,2,7,9),
		status = c(1,0,1,0,1,1,0))
fit <- prodlim(Hist(time,status)~1,data=test)
jtimes <- fit$time
etimes <- c(0,.5,2,8,10)
fit$surv
c(1,fit$surv)[1+sindex(jtimes,etimes)]

}
\author{
Thomas A. Gerds \email{tag@biostat.ku.dk}
}
\keyword{misc}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/row.match.R
\name{row.match}
\alias{row.match}
\title{Identifying rows in a matrix or data.frame}
\usage{
row.match(x, table, nomatch = NA)
}
\arguments{
\item{x}{Vector or matrix whose rows are to be matched}

\item{table}{Matrix or data.frame that contain the rows to be matched
against.}

\item{nomatch}{the value to be returned in the case when no match is found.
Note that it is coerced to 'integer'.}
}
\value{
A vector of the same length as 'x'.
}
\description{
Function for finding matching rows between two matrices or data.frames.
First the matrices or data.frames are vectorized by row wise pasting
together the elements. Then it uses the function match.  Thus the function
returns a vector with the row numbers of (first) matches of its first
argument in its second.
}
\examples{

tab <- data.frame(num=1:26,abc=letters)
x <- c(3,"c")
row.match(x,tab)
x <- data.frame(n=c(3,8),z=c("c","h"))
row.match(x,tab)

}
\seealso{
\code{match}
}
\author{
Thomas A. Gerds
}
\keyword{misc}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/getStates.R
\name{getStates}
\alias{getStates}
\title{States of a multi-state model}
\usage{
getStates(object, ...)
}
\arguments{
\item{object}{Object of class \code{prodlim} or \code{Hist} .}

\item{...}{not used}
}
\value{
A character vector with the states of the model.
}
\description{
Extract the states of a multi-state model
}
\details{
Applying this function to the fit of prodlim means to apply
it to \code{fit$model.response}.
}
\author{
Thomas A. Gerds
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plot.Hist.R
\name{plot.Hist}
\alias{plot.Hist}
\title{Box-arrow diagrams for multi-state models.}
\usage{
\method{plot}{Hist}(
  x,
  nrow,
  ncol,
  box.width,
  box.height,
  box.padding,
  xbox.position,
  ybox.position,
  stateLabels,
  arrowLabels,
  arrowLabelStyle = "symbolic",
  arrowLabelSymbol = "lambda",
  changeArrowLabelSide,
  curved,
  tagBoxes = FALSE,
  startCountZero = TRUE,
  oneFitsAll,
  margin,
  cex,
  rasta = FALSE,
  verbose = FALSE,
  ...
)
}
\arguments{
\item{x}{An object of class \code{Hist}.}

\item{nrow}{the number of graphic rows}

\item{ncol}{the number of graphic columns}

\item{box.width}{the widths of the boxes on the scale from 0 to 100}

\item{box.height}{the heights of the boxes on the scale from 0 to 100}

\item{box.padding}{how much room there should be between the label and the border of a box.
Two values on the scale from 0 to 100: the first for the horizontal
x-direction and the second for the vertical y-direction padding.}

\item{xbox.position}{the x box positions (left lower corner) on the scale from 0 to 100.}

\item{ybox.position}{the y box positions (left lower corner) on the scale from 0 to 100.}

\item{stateLabels}{Vector of names to appear in the boxes (states).
Defaults to attr(x,"state.names").  The boxes can also be
individually labeled by smart arguments of the form
\code{box3.label="diseased"}, see examples.}

\item{arrowLabels}{Vector of labels to appear in the boxes
(states). One for each arrow.  The arrows can also be
individually labeled by smart arguments of the form
\code{arrow1.label=paste(expression(eta(s,u)))}, see examples.}

\item{arrowLabelStyle}{Either "symbolic" for automated symbolic
arrow labels, or "count" for arrow labels that reflect the
number of transitions in the data.}

\item{arrowLabelSymbol}{Symbol for automated symbolic arrow
labels. Defaults to "lambda".}

\item{changeArrowLabelSide}{A vector of mode logical (TRUE,FALSE)
one for each arrow to change the side of the arrow on which the
label is placed.}

\item{curved}{The curvature of curved arrows via diagram::curvedarrow. Experimental. Values between 0 (no curvature) and 1 are meaningful.}

\item{tagBoxes}{Logical. If TRUE the boxes are numbered in the
upper left corner. The size can be controlled with smart
argument boxtags.cex. The default is boxtags.cex=1.28.}

\item{startCountZero}{Control states numbers for symbolic arrow
labels and box tags.}

\item{oneFitsAll}{If \code{FALSE} then boxes have individual size,
depending on the size of the label, otherwise all boxes have
the same size dependent on the largest label.}

\item{margin}{Set the figure margin via
\code{par(mar=margin)}. Less than 4 values are repeated.}

\item{cex}{Initial cex value for the state and the arrow
\code{labels}.}

\item{rasta}{For construction purposes.}

\item{verbose}{If TRUE echo various things.}

\item{\dots}{Smart control of arguments for the subroutines text
(box label), rect (box), arrows, text (arrow label). Thus the
three dots can be used to draw individual boxes with individual
labels, arrows and arrow labels. E.g. arrow2.label="any label"
changes the label of the second arrow.  See examples.}
}
\description{
Automated plotting of the states and transitions that characterize a multi
states model.
}
\note{
Use the functionality of the unix program `dot'
    http://www.graphviz.org/About.php via R package Rgraphviz to
    obtain more complex graphs.
}
\examples{


## A simple survival model

SurvFrame <- data.frame(time=1:10,status=c(0,1,1,0,0,1,0,0,1,0))
SurvHist <- with(SurvFrame,Hist(time,status))
plot(SurvHist)
plot(SurvHist,box2.col=2,box2.label="experienced\nR user")
plot(SurvHist,
     box2.col=2,
     box1.label="newby",
     box2.label="experienced\nR user",
     oneFitsAll=FALSE,
     arrow1.length=.5,
     arrow1.label="",
     arrow1.lwd=4)

## change the cex of all box labels:
plot(SurvHist,
     box2.col=2,
     box1.label="newby",
     box2.label="experienced\nR user",
     oneFitsAll=FALSE,
     arrow1.length=.5,
     arrow1.label="",
     arrow1.lwd=4,
     label.cex=1)

## change the cex of single box labels:
plot(SurvHist,
     box2.col=2,
     box1.label="newby",
     box2.label="experienced\nR user",
     oneFitsAll=FALSE,
     arrow1.length=.5,
     arrow1.label="",
     arrow1.lwd=4,
     label1.cex=1,
     label2.cex=2)


## The pbc data set from the survival package
library(survival)
data(pbc)
plot(with(pbc,Hist(time,status)),
     stateLabels=c("randomized","transplant","dead"),
     arrowLabelStyle="count")

## two competing risks
comprisk.model <- data.frame(time=1:3,status=1:3)
CRHist <- with(comprisk.model,Hist(time,status,cens.code=2))
plot(CRHist)
plot(CRHist,arrow1.label=paste(expression(eta(s,u))))

plot(CRHist,box2.label="This\nis\nstate 2",arrow1.label=paste(expression(gamma[1](t))))
plot(CRHist,box3.label="Any\nLabel",arrow2.label="any\nlabel")

## change the layout
plot(CRHist,
     box1.label="Alive",
     box2.label="Dead\n cause 1",
     box3.label="Dead\n cause 2",
     arrow1.label=paste(expression(gamma[1](t))),
     arrow2.label=paste(expression(eta[2](t))),
     box1.col=2,
     box2.col=3,
     box3.col=4,
     nrow=2,
     ncol=3,
     box1.row=1,
     box1.column=2,
     box2.row=2,
     box2.column=1,
     box3.row=2,
     box3.column=3)

## more competing risks
comprisk.model2 <- data.frame(time=1:4,status=1:4)
CRHist2 <- with(comprisk.model2,Hist(time,status,cens.code=2))
plot(CRHist2,box1.row=2)

## illness-death models
illness.death.frame <- data.frame(time=1:4,
				  from=c("Disease\nfree",
                                      "Disease\nfree",
                                      "Diseased",
                                      "Disease\nfree"),
				  to=c("0","Diseased","Dead","Dead"))
IDHist <- with(illness.death.frame,Hist(time,event=list(from,to)))
plot(IDHist)

## illness-death with recovery
illness.death.frame2 <- data.frame(time=1:5,
from=c("Disease\nfree","Disease\nfree","Diseased","Diseased","Disease\nfree"),
to=c("0","Diseased","Disease\nfree","Dead","Dead"))
IDHist2 <- with(illness.death.frame2,Hist(time,event=list(from,to)))
plot(IDHist2)

## 4 state models
x=data.frame(from=c(1,2,1,3,4),to=c(2,1,3,4,1),time=1:5)
y=with(x,Hist(time=time,event=list(from=from,to=to)))
plot(y)

## moving the label of some arrows

d <- data.frame(time=1:5,from=c(1,1,1,2,2),to=c(2,3,4,3,4))
h <- with(d,Hist(time,event=list(from,to)))
plot(h,box.padding=c(5,2),
tagBoxes=TRUE,
stateLabels=c("Remission\nwithout\nGvHD",
    "Remission\nwith\nGvHD",
    "Relapse",
    "Death\nwithout\nrelapse"),
arrowLabelSymbol='alpha',
arrowlabel3.x=35,
arrowlabel3.y=53,
arrowlabel4.y=54,
arrowlabel4.x=68)

##'
}
\seealso{
\code{\link{Hist}}\code{\link{SmartControl}}
}
\author{
Thomas A Gerds \email{tag@biostat.ku.dk}
}
\keyword{survival}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/parseSpecialNames.R
\name{parseSpecialNames}
\alias{parseSpecialNames}
\title{Parse special terms}
\usage{
parseSpecialNames(x, special, arguments)
}
\arguments{
\item{x}{Vector of character strings.}

\item{special}{A character string: the name of the special argument.}

\item{arguments}{A vector which contains the arguments of the special function}
}
\value{
A named list of parsed arguments. The names of the list are the special variable names, the elements
are lists of arguments.
}
\description{
Extract from a vector of character strings the names of special functions and auxiliary arguments
}
\details{
Signals an error if an element has more arguments than specified by argument arguments.
}
\examples{

## ignore arguments
parseSpecialNames("treat(Z)",special="treat")
## set default to 0
parseSpecialNames(c("log(Z)","a","log(B)"),special="log",arguments=list("base"=0))
## set default to 0
parseSpecialNames(c("log(Z,3)","a","log(B,base=1)"),special="log",arguments=list("base"=0))
## different combinations of order and names
parseSpecialNames(c("log(Z,3)","a","log(B,1)"),
                  special="log",
                  arguments=list("base"=0))
parseSpecialNames(c("log(Z,1,3)","a","log(B,u=3)"),
                  special="log",
                  arguments=list("base"=0,"u"=1))
parseSpecialNames(c("log(Z,u=1,base=3)","a","log(B,u=3)"),
                  special="log",
                  arguments=list("base"=0,"u"=1))
parseSpecialNames(c("log(Z,u=1,base=3)","a","log(B,base=8,u=3)"),
                  special="log",
                  arguments=list("base"=0,"u"=1))
parseSpecialNames("treat(Z,u=2)",
                  special="treat",
                  arguments=list("u"=1,"k"=1))
parseSpecialNames(c("treat(Z,1,u=2)","treat(B,u=2,k=3)"),
                  special="treat",
                  arguments=list("u"=NA,"k"=NULL))
## does not work to set default to NULL:
parseSpecialNames(c("treat(Z,1,u=2)","treat(B,u=2)"),
                  special="treat",
                  arguments=list("u"=NA,"k"=NULL))
}
\seealso{
model.design
}
\author{
Thomas A. Gerds <tag@biostat.ku.dk>
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plotIllnessDeathModel.R
\name{plotIllnessDeathModel}
\alias{plotIllnessDeathModel}
\title{Plotting an illness-death-model.}
\usage{
plotIllnessDeathModel(stateLabels, style = 1, recovery = FALSE, ...)
}
\arguments{
\item{stateLabels}{Labels for the three boxes.}

\item{style}{Either \code{1} or anything else, switches the orientation of
the graph. Hard to explain in words, see examples.}

\item{recovery}{Logical. If \code{TRUE} there will be an arrow from the
illness state to the initial state.}

\item{\dots}{Arguments passed to plot.Hist.}
}
\description{
Plotting an illness-death-model using \code{plot.Hist}.
}
\examples{

plotIllnessDeathModel()
plotIllnessDeathModel(style=2)
plotIllnessDeathModel(style=2,
                      stateLabels=c("a","b\nc","d"),
                      box1.col="yellow",
                      box2.col="green",
                      box3.col="red")
}
\seealso{
\code{\link{plotCompetingRiskModel}}, \code{\link{plot.Hist}}
}
\author{
Thomas Alexander Gerds <tag@biostat.ku.dk>
}
\keyword{survival}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/redist.R
\name{redist}
\alias{redist}
\title{Calculation of Efron's re-distribution to the right algorithm to obtain the
Kaplan-Meier estimate.}
\usage{
redist(time, status)
}
\arguments{
\item{time}{A numeric vector of event times.}

\item{status}{The event status vector takes the value \code{1} for observed events and 
the value \code{0} for right censored times.}
}
\value{
Calculations needed to
}
\description{
Calculation of Efron's re-distribution to the right algorithm to obtain the
Kaplan-Meier estimate.
}
\examples{
redist(time=c(.35,0.4,.51,.51,.7,.73),status=c(0,1,1,0,0,1))
}
\seealso{
prodlim
}
\author{
Thomas A. Gerds <tag@biostat.ku.dk>
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/geom_prodlim.R
\name{geom_prodlim}
\alias{geom_prodlim}
\title{Product-limit geoms for survival and competing risks curves}
\usage{
geom_prodlim(
  mapping = NULL,
  data = NULL,
  position = "identity",
  na.rm = FALSE,
  show.legend = NA,
  inherit.aes = TRUE,
  type = "risk",
  cause = NULL,
  conf_int = TRUE,
  conf_int_alpha = 0.2,
  percent = TRUE,
  timeconverter = NULL,
  times = NULL,
  cens.code = "0",
  ...
)
}
\arguments{
\item{mapping}{Set of aesthetic mappings created by [ggplot2::aes()].}

\item{data}{Data frame.}

\item{position}{Position adjustment.}

\item{na.rm}{If `FALSE`, missing values are removed with a warning.}

\item{show.legend}{Logical. Should this layer be included in the legends?}

\item{inherit.aes}{If `FALSE`, overrides the default aesthetics.}

\item{type}{Passed to [prodlim::prodlim()], usually `"risk"` or `"surv"`.}

\item{cause}{Cause(s) for competing risks. Can be a vector of causes or the
string `"stacked"`.}

\item{conf_int}{Logical. Draw confidence intervals.}

\item{conf_int_alpha}{Alpha level for confidence shadows.}

\item{percent}{Logical. Passed to [prodlim::summary.prodlim()].}

\item{timeconverter}{Optional time conversion string.}

\item{times}{Optional evaluation times passed to
[prodlim::summary.prodlim()].}

\item{cens.code}{Censoring code passed to [prodlim::Hist()].}

\item{...}{Further arguments passed to [ggplot2::layer()].}
}
\description{
Draw Kaplan-Meier or Aalen-Johansen curves with optional confidence bands.
For competing risks with multiple causes, the default display shows one curve
per cause. When `cause = "stacked"`, causes are drawn as stacked filled
rectangles instead of step curves with confidence shadows.
}
\details{
When multiple causes are specified, `fill`/`colour` aesthetics are ignored
and replaced by a cause-based mapping so that all causes are shown in a
single legend. When `cause = "stacked"`, causes are stacked on top of each
other and confidence shadows are not drawn.
}
\examples{
library(riskRegression)
library(data.table)
library(ggplot2)
data(Melanoma)
# Kaplan-Meier
ggplot(data = Melanoma,aes(x = time, event = 1*(status != 0)))+
       geom_prodlim(type = "surv")
# stratified Kaplan-Meier inherited aes
ggplot(data = Melanoma,aes(x = time, event = 1*(status != 0),fill = sex,color = sex))+
       geom_prodlim(type = "surv")
# stratified Kaplan-Meier geom aes
ggplot(data = Melanoma,aes(x = time, event = 1*(status != 0)))+
       geom_prodlim(aes(fill = sex,color = sex),type = "surv")
# facet
ggplot(data = Melanoma,aes(x = time, event = 1*(status != 0)))+
       geom_prodlim(type = "surv")+facet_grid(~sex)
# stratified and facet
ggplot(data = Melanoma,aes(x = time, event = 1*(status != 0),fill = epicel,color = epicel))+
       geom_prodlim(type = "surv")+facet_grid(~sex)

# Aalen-Johansen
ggplot(data = Melanoma,aes(x = time, event = status))+
       geom_prodlim(type = "surv",cause = 1:2)
# stratified Aalen-Johansen inherited aes
ggplot(data = Melanoma,aes(x = time, event = status,fill = sex,color = sex))+
       geom_prodlim(type = "surv")
# stratified Aalen-Johansen geom aes
ggplot(data = Melanoma,aes(x = time, event = status))+
       geom_prodlim(aes(fill = sex,color = sex),type = "surv")
# facet
ggplot(data = Melanoma,aes(x = time, event = status))+
       geom_prodlim(type = "surv")+facet_grid(~sex)
# stratified and facet
ggplot(data = Melanoma,aes(x = time, event = status,fill = epicel,color = epicel))+
       geom_prodlim(type = "surv")+facet_grid(~sex)
# stacked
ggplot(data = Melanoma,aes(x = time, event = status))+
       geom_prodlim(cause = "stacked")+facet_grid(~sex)
# stacked with cens.code option
ggplot(data = Melanoma,aes(x = time, event = event))+
       geom_prodlim(cens.code="censored",cause = "stacked")+facet_grid(~sex)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dimColor.R
\name{dimColor}
\alias{dimColor}
\title{Dim a given color to a specified density}
\usage{
dimColor(col, density = 55)
}
\arguments{
\item{col}{Color name or number passed to \code{\link{col2rgb}}.}

\item{density}{Integer value passed as alpha coefficient to
\code{\link{rgb}} between 0 and 255}
}
\value{
A character vector with the color code. See \code{rgb} for details.
}
\description{
This function calls first \code{\link{col2rgb}} on a color name and then 
uses \code{\link{rgb}} to adjust the intensity of the result.
}
\examples{
dimColor(2,33)
dimColor("green",133)
}
\seealso{
rgb col2rgb
}
\author{
Thomas A. Gerds <tag@biostat.ku.dk>
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/SimSurv.R
\name{SimSurv}
\alias{SimSurv}
\title{Simulate survival data}
\usage{
SimSurv(N, ...)
}
\arguments{
\item{N}{sample size}

\item{...}{do nothing}
}
\value{
data.frame with simulated data
}
\description{
Simulate right censored survival data with two covariates X1 and X2, both have effect exp(1) on the hazard of the unobserved event time.
}
\details{
This function calls \code{survModel}, then adds  covariates and finally calls \code{sim.lvm}.
}
\examples{

SimSurv(10)

}
\references{
Bender, Augustin & Blettner. Generating survival times to simulate Cox proportional hazards models. Statistics in Medicine, 24: 1713-1723, 2005.
}
\author{
Thomas Alexander Gerds
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Hist.R
\name{[.Hist}
\alias{[.Hist}
\title{Subset an event history object}
\usage{
\method{[}{Hist}(x, i, j, drop = FALSE)
}
\arguments{
\item{x}{object obtained with \code{\link{Hist}}}

\item{i}{passed on to next method}

\item{j}{passed on to next method}

\item{drop}{passed on to next method}
}
\description{
This method makes sure that the attributes are
preserved
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/stopTime.R
\name{stopTime}
\alias{stopTime}
\title{Stop the time of an event history object}
\usage{
stopTime(object, stop.time)
}
\arguments{
\item{object}{Event history object as obtained with \code{Hist}}

\item{stop.time}{Time point at which to stop the event history object}
}
\value{
Stopped event history object where all times are censored 
        at \code{stop.time}. All observations with times greater than \code{stop.time}
        are set to \code{stop.time} and the event status is set to \code{attr(object,"cens.code")}.
        A new column \code{"stop.time"} is equal to \code{1} for stopped observations
        and equal to \code{0} for the other observations.
}
\description{
All event times are stopped at a given time point and
corresponding events are censored
}
\examples{

set.seed(29)
d <- SimSurv(10)
h <- with(d,Hist(time,status))
h
stopTime(h,8)
stopTime(h,5)

## works also with Surv objects
library(survival)
s <- with(d,Surv(time,status))
stopTime(s,5)

## competing risks
set.seed(29)
dr <- SimCompRisk(10)
hr <- with(dr,Hist(time,event))
hr
stopTime(hr,8)
stopTime(hr,5)

}
\seealso{
Hist
}
\author{
Thomas A. Gerds <tag@biostat.ku.dk>
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/predictSurvIndividual.R
\name{predictSurvIndividual}
\alias{predictSurvIndividual}
\title{Predict individual survival probabilities}
\usage{
predictSurvIndividual(object, lag = 1)
}
\arguments{
\item{object}{A fitted object of class "prodlim".}

\item{lag}{Integer. `0' means predictions at the individual times, 1 means
just before the individual times, etc.}
}
\value{
A vector of survival probabilities.
}
\description{
Function to extract the predicted probabilities at the individual event
times that have been used for fitting a prodlim object.
}
\examples{

 SurvFrame <- data.frame(time=1:10,status=rbinom(10,1,.5))
 x <- prodlim(formula=Hist(time=time,status!=0)~1,data=SurvFrame)
 predictSurvIndividual(x,lag=1)

}
\seealso{
\code{\link{predict.prodlim}},\code{\link{predictSurv}},
}
\author{
Thomas A. Gerds \email{tag@biostat.ku.dk}
}
\keyword{survival}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Hist.R
\name{is.na.Hist}
\alias{is.na.Hist}
\title{Check for missing values in event history object}
\usage{
\method{is.na}{Hist}(x)
}
\arguments{
\item{x}{object obtained with \code{\link{Hist}}}
}
\description{
The object has usually two or three columns and
this function checks each column separately
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/checkCauses.R
\name{checkCauses}
\alias{checkCauses}
\title{Check availability of a cause in competing risk settings}
\usage{
checkCauses(cause, object)
}
\arguments{
\item{cause}{cause of interest}

\item{object}{object either obtained with \code{Hist} or \code{prodlim}}
}
\description{
For competing risk settings, check if the requested cause is known to the object
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Hist.R
\name{head.Hist}
\alias{head.Hist}
\title{Check the head of an event history object}
\usage{
\method{head}{Hist}(x)
}
\arguments{
\item{x}{object obtained with \code{\link{Hist}}}
}
\description{
The object is evaluated as a matrix
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/EventHistory.frame.R
\name{EventHistory.frame}
\alias{EventHistory.frame}
\title{Event history frame}
\usage{
EventHistory.frame(
  formula,
  data,
  unspecialsDesign = TRUE,
  specials,
  specialsFactor = TRUE,
  specialsDesign = FALSE,
  stripSpecials = NULL,
  stripArguments = NULL,
  stripAlias = NULL,
  stripUnspecials = NULL,
  dropIntercept = TRUE,
  check.formula = TRUE,
  response = TRUE,
  na.action = options()$na.action
)
}
\arguments{
\item{formula}{Formula whose left hand side specifies the event
history, i.e., either via Surv() or Hist().}

\item{data}{Data frame in which the formula is interpreted}

\item{unspecialsDesign}{Passed as is to
\code{\link{model.design}}.}

\item{specials}{Character vector of special function names.
Usually the body of the special functions is function(x)x but
e.g., \code{\link[survival]{strata}} from the survival package does treat
the values}

\item{specialsFactor}{Passed as is to \code{\link{model.design}}.}

\item{specialsDesign}{Passed as is to \code{\link{model.design}}}

\item{stripSpecials}{Passed as \code{specials} to
\code{\link{strip.terms}}}

\item{stripArguments}{Passed as \code{arguments} to
\code{\link{strip.terms}}}

\item{stripAlias}{Passed as \code{alias.names} to
\code{\link{strip.terms}}}

\item{stripUnspecials}{Passed as \code{unspecials} to
\code{\link{strip.terms}}}

\item{dropIntercept}{Passed as is to \code{\link{model.design}}}

\item{check.formula}{If TRUE check if formula is a Surv or Hist
thing.}

\item{response}{If FALSE do not evaluate the left hand side of the formula and do not return the response (event.history).}

\item{na.action}{Decide what to do with missing values.}
}
\value{
A list which contains
- the event.history (see \code{\link{Hist}})
- the design matrix (see \code{\link{model.design}})
- one entry for each special (see \code{\link{model.design}})
}
\description{
Extract event history data and design matrix including specials from call
}
\details{
Obtain a list with the data used for event history regression analysis. This
function cannot be used directly on the user level but inside a function
to prepare data for survival analysis.
}
\examples{

## Here are some data with an event time and no competing risks
## and two covariates X1 and X2.
## Suppose we want to declare that variable X1 is treated differently
## than variable X2. For example, X1 could be a cluster variable, or
## X1 should have a proportional effect on the outcome.
dsurv <- data.frame(time=1:7,
                    status=c(0,1,1,0,0,0,1),
                    X2=c(2.24,3.22,9.59,4.4,3.54,6.81,5.05),
                    X3=c(1,1,1,1,0,0,1),
                    X4=c(44.69,37.41,68.54,38.85,35.9,27.02,41.84),
                    X1=factor(c("a","b","a","c","c","a","b"),
                        levels=c("c","a","b")))
## We pass a formula and the data
e <- EventHistory.frame(Hist(time,status)~prop(X1)+X2+cluster(X3)+X4,
                        data=dsurv,
                        specials=c("prop","cluster"),
                        stripSpecials=c("prop","cluster"))
names(e)
## The first element is the event.history which is result of the left hand
## side of the formula:
e$event.history
## same as
with(dsurv,Hist(time,status))
## to see the structure do 
colnames(e$event.history)
unclass(e$event.history)
## in case of competing risks there will be an additional column called event,
## see help(Hist) for more details

## The other elements are the design, i.e., model.matrix for the non-special covariates
e$design
## and a data.frame for the special covariates
e$prop
## The special covariates can be returned as a model.matrix
e2 <- EventHistory.frame(Hist(time,status)~prop(X1)+X2+cluster(X3)+X4,
                         data=dsurv,
                         specials=c("prop","cluster"),
                         stripSpecials=c("prop","cluster"),
                         specialsDesign=TRUE)
e2$prop
## and the non-special covariates can be returned as a data.frame
e3 <- EventHistory.frame(Hist(time,status)~prop(X1)+X2+cluster(X3)+X4,
                         data=dsurv,
                         specials=c("prop","cluster"),
                         stripSpecials=c("prop","cluster"),
                         specialsDesign=TRUE,
                         unspecialsDesign=FALSE)
e3$design

## the general idea is that the function is used to parse the combination of
## formula and data inside another function. Here is an example with
## competing risks
SampleRegression <- function(formula,data=parent.frame()){
    thecall <- match.call()
    ehf <- EventHistory.frame(formula=formula,
                              data=data,
                              stripSpecials=c("prop","cluster","timevar"),
                              specials=c("prop","timevar","cluster"))
    time <- ehf$event.history[,"time"]
    status <- ehf$event.history[,"status"]
    ## event as a factor
    if (attr(ehf$event.history,"model")=="competing.risks"){
        event <- ehf$event.history[,"event"]
        Event <- getEvent(ehf$event.history)
        list(response=data.frame(time,status,event,Event),X=ehf[-1])
    }
    else{ # no competing risks
        list(response=data.frame(time,status),X=ehf[-1])
    }
}
dsurv$outcome <- c("cause1","0","cause2","cause1","cause2","cause2","0")
SampleRegression(Hist(time,outcome)~prop(X1)+X2+cluster(X3)+X4,dsurv)
ehf <- EventHistory.frame(formula=Hist(time,outcome)~prop(X1)+X2+cluster(X3)+X4,
                          data=dsurv,
                          stripSpecials=c("prop","cluster","timevar"),
                          specials=c("prop","timevar","cluster"))

## let's test if the parsing works
form1 <- Hist(time,outcome!="0")~prop(X1)+X2+cluster(X3)+X4
form2 <- Hist(time,outcome)~prop(X1)+cluster(X3)+X4
ff <- list(form1,form2)
lapply(ff,function(f){SampleRegression(f,dsurv)})


## here is what the riskRegression package uses to
## distinguish between covariates with
## time-proportional effects and covariates with
## time-varying effects:
\dontrun{
library(riskRegression)
data(Melanoma)
f <- Hist(time,status)~prop(thick)+strata(sex)+age+prop(ulcer,power=1)+timevar(invasion,test=1)
## here the unspecial terms, i.e., the term age is treated as prop
## also, strata is an alias for timvar

EHF <- prodlim::EventHistory.frame(formula,
                                   Melanoma[1:10],
                                   specials=c("timevar","strata","prop","const","tp"),
                                   stripSpecials=c("timevar","prop"),
                                   stripArguments=list("prop"=list("power"=0),
                                       "timevar"=list("test"=0)),
                                   stripAlias=list("timevar"=c("strata"),
                                       "prop"=c("tp","const")),
                                   stripUnspecials="prop",
                                   specialsDesign=TRUE,
                                   dropIntercept=TRUE)       
EHF$prop
EHF$timevar
}
}
\seealso{
model.frame model.design Hist
}
\author{
Thomas A. Gerds <tag@biostat.ku.dk>
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/markTime.R
\name{markTime}
\alias{markTime}
\title{Marking product-limit plots at the censored times.}
\usage{
markTime(x, times, nlost, pch, col, ...)
}
\arguments{
\item{x}{The values of the curves at \code{times}.}

\item{times}{The times where there curves are plotted.}

\item{nlost}{The number of subjects lost to follow-up (censored) at
\code{times}.}

\item{pch}{The symbol used to mark the curves.}

\item{col}{The color of the symbols.}

\item{...}{Arguments passed to \code{points}.}
}
\value{
Nil
}
\description{
This function is invoked and controlled by \code{plot.prodlim}.
}
\details{
This function should not be called directly. The arguments can be specified
as \code{atRisk.arg} in the call to \code{plot.prodlim}.
}
\seealso{
\code{\link{plot.prodlim}}, \code{\link{confInt}},
\code{\link{atRisk}}
}
\author{
Thomas Alexander Gerds <tag@biostat.ku.dk>
}
\keyword{survival}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/summary.Hist.R
\name{summary.Hist}
\alias{summary.Hist}
\title{Summary of event histories}
\usage{
\method{summary}{Hist}(object, verbose = TRUE, ...)
}
\arguments{
\item{object}{An object with class `Hist' derived with \code{\link{Hist}}}

\item{verbose}{Logical. If FALSE any printing is supressed.}

\item{\dots}{Not used}
}
\value{
\code{NULL} for survival and competing risk models.  For other
multi-state models, it is a list with the following entries:
\item{states}{the states of the model} \item{transitions}{the transitions
between the states} \item{trans.frame}{a data.frame with the from and to
states of the transitions}
}
\description{
Describe events and censoring patterns of an event history.
}
\examples{

icensFrame <- data.frame(L=c(1,1,3,4,6),R=c(2,NA,3,6,9),event=c(1,1,1,2,2))
with(icensFrame,summary(Hist(time=list(L,R))))

}
\seealso{
\code{\link{Hist}}, \code{\link{plot.Hist}}
}
\author{
Thomas A. Gerds \email{tag@biostat.ku.dk}
}
\keyword{survival}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/followup.R
\name{followup}
\alias{followup}
\title{Estimation of the median potential followup time.}
\usage{
followup(formula, cens.code = 0, data, ...)
}
\arguments{
\item{formula}{A formula whose left hand side is a \code{Hist} or a 
\code{\link[survival]{Surv}} object specifying the event time
and the event type where by default 0=censored, 1=event, 2=competing risk (if any).
Use \code{cens.code} to change the value for censored.}

\item{cens.code}{Value of the event}

\item{data}{A data.frame in which all the variables of
\code{formula} can be interpreted.}

\item{...}{Arguments passed to \code{prodlim.quantile}.}
}
\value{
The estimated median potential followup time with inter quartile ranges.
}
\description{
The reverse Kaplan-Meier method estimates the median potential followup time.
}
\details{
This is merely a wrapper for \code{prodlim} with argument reverse.
}
\examples{
set.seed(8)
d <- SimCompRisk(117)

# overall
followup(Hist(time,event)~1,data=d)

# in strata defined by variable X1 
followup(Hist(time,event)~X1,data=d)
}
\references{
Michael Schemper and Terry L. Smith. A note on quantifying follow-up in studies of failure time. Controlled Clinical Trials, 17(4):343--346, 1996.
}
\seealso{
\code{\link{prodlim}}
}
\author{
Thomas A. Gerds <tag@biostat.ku.dk>
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/parseFormula.R
\name{parseFormula}
\alias{parseFormula}
\title{Parse Formula}
\usage{
parseFormula(
  formula,
  data = NULL,
  specials,
  special.default.values = NULL,
  alias.names = NULL
)
}
\arguments{
\item{formula}{A model formula (e.g., Y ~ X + Z or Y ~ fun(X, param = value)).}

\item{data}{An optional data frame containing the variables referenced in the formula.}

\item{specials}{A character vector of special function names that should be specifically handled.}

\item{special.default.values}{A named list of named lists specifying default argument values for each special.}

\item{alias.names}{A named list mapping special names to their alias(es) (e.g., `list("special" = "alias")`).}
}
\value{
A list where each element corresponds to a special function in the formula and contains the parsed terms and their arguments.

A named list with an element for each special that contains a named list where each element
        corresponds to a formula term that is affected by the special that contains
        the value(s) of the additional argument(s) for this term.
}
\description{
Parse a formula to extract terms and handle special functions
}
\details{
This function parses a model formula, extracts terms, and processes specified 'special' functions and their arguments.
It allows defining default values for special functions and specifying aliases for those names.
}
\examples{
# Basic usage with specials
parseFormula(Y ~ fun(X, a = 1), specials = "fun")

# With defaults and aliases
parseFormula(Y ~ gfun(X, a = 7), 
             specials = "fun", 
             alias.names = list("fun" = "gfun"),
             special.default.values = list("fun" = list("a" = 0)))

# Complex use case with multiple specials
parseFormula(Surv(time, status) ~ age + strata(sex, test = 1) + prop(albumin),
             specials = c("prop", "strata"), 
             special.default.values = list("prop" = list(power = 0), "strata" = list(test = 1)))
parseFormula(Y~fun(X,a=1),specials="fun",special.default.values=list("fun"=list("a"=0)))
parseFormula(Y~fun(X),specials="fun",special.default.values=list("fun"=list("a"=0)))
parseFormula(Y~gfun(X,a=7),specials="fun",
             alias.names=list("fun"="gfun"),
             special.default.values=list("fun"=list("a"=0)))
parseFormula(formula = Surv(time,status)~age
                            +const(factor(edema))
                            +strata(sex,test=0)
                            +prop(alb)+prop(bili,power=1)
                            +tp(albumin),
                  specials = c("prop","timevar","strata","tp","const"),
                  special.default.values =list("prop" = list(power = 0),
                                               "strata" = list(test = 1)))
parseFormula(formula = Surv(time,status)~age
                            +const(factor(edema))
                            +strata(sex,test=0)
                            +prop(alb)+prop(bili,power=1)
                            +tp(albumin),
                  specials = c("prop","timevar","strata","tp"),
                  special.default.values =list("prop" = list(power = 0),
                                               "strata" = list(test = 1)),
                  alias.names=list("prop"="const"))
}
\seealso{
[stats::terms()], [strip.terms()]

model.design strip.terms
}
\author{
Thomas A. Gerds <tag@biostat.ku.dk>
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/List2Matrix.R
\name{List2Matrix}
\alias{List2Matrix}
\title{Reduce list to a matrix or data.frame with names as new columns}
\usage{
List2Matrix(list, depth, names)
}
\arguments{
\item{list}{A named list which contains nested lists}

\item{depth}{The depth in the list hierarchy until an rbindable object}

\item{names}{Names for the list variables}
}
\value{
Matrix or data.frame.
}
\description{
This function is used by summary.prodlim to deal with results.
}
\details{
Reduction is done with rbind.
}
\examples{

x=list(a=data.frame(u=1,b=2,c=3),b=data.frame(u=3,b=4,c=6))
List2Matrix(x,depth=1,"X")
}
\author{
Thomas A. Gerds <tag@biostat.ku.dk>
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/jackknife.R
\name{jackknife}
\alias{jackknife}
\alias{jackknife.survival}
\alias{jackknife.competing.risks}
\title{Compute jackknife pseudo values.}
\usage{
jackknife(object, times, cause, keepResponse = FALSE, ...)
}
\arguments{
\item{object}{Object of class \code{"prodlim"}.}

\item{times}{Time points at which to compute pseudo values.}

\item{cause}{Character (other classes are converted with \code{as.character}).
For competing risks the cause of failure.}

\item{keepResponse}{If \code{TRUE} add the model response,
i.e. event time, event status, etc. to the result.}

\item{...}{not used}
}
\description{
Compute jackknife pseudo values.
}
\details{
Compute jackknife pseudo values based on marginal Kaplan-Meier estimate of
survival, or based on marginal Aalen-Johansen estimate of the absolute risks, i.e., the cumulative
incidence function.
}
\note{
The R-package pseudo does a similar job, and appears to be a little faster in small samples, but much slower in large samples. See examples.
}
\examples{


## pseudo-values for survival models

d=SimSurv(20) 
f=prodlim(Hist(time,status)~1,data=d) 
jackknife(f,times=c(3,5))

## in some situations it may be useful to attach the
## the event time history to the result
jackknife(f,times=c(3,5),keepResponse=TRUE)

# pseudo-values for competing risk models
set.seed(15)
d=SimCompRisk(15) 
f=prodlim(Hist(time,event)~1,data=d) 
jackknife(f,times=c(3,5),cause=1)
jackknife(f,times=c(1,3,5),cause=2)

}
\references{
Andersen PK & Perme MP (2010). Pseudo-observations in survival
analysis Statistical Methods in Medical Research, 19(1), 71-99.
}
\seealso{
\code{\link{prodlim}}
}
\author{
Thomas Alexander Gerds <tag@biostat.ku.dk>
}
\keyword{survival}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Hist.R
\name{subset.Hist}
\alias{subset.Hist}
\title{Subset an event history object}
\usage{
\method{subset}{Hist}(x, subset, select, drop, ...)
}
\arguments{
\item{x}{object obtained with \code{\link{Hist}}}

\item{subset}{passed on to next method}

\item{select}{passed on to next method}

\item{drop}{passed on to next method}

\item{...}{passed on to the next method}
}
\description{
This method makes sure that the attributes are
preserved
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/leaveOneOut.R
\name{leaveOneOut}
\alias{leaveOneOut}
\alias{leaveOneOut.survival}
\alias{leaveOneOut.competing.risks}
\title{Compute jackknife pseudo values.}
\usage{
leaveOneOut(object, times, cause, lag = FALSE, ...)
}
\arguments{
\item{object}{Object of class \code{"prodlim"}.}

\item{times}{time points at which to compute leave-one-out
event/survival probabilities.}

\item{cause}{Character (other classes are converted with \code{as.character}).
For competing risks the cause of interest.}

\item{lag}{For survival models only. If \code{TRUE} lag the result, i.e. compute
S(t-) instead of S(t).}

\item{...}{not used}
}
\description{
Compute leave-one-out estimates
}
\details{
This function is the work-horse for \code{jackknife}
}
\seealso{
\code{\link{jackknife}}
}
\author{
Thomas Alexander Gerds <tag@biostat.ku.dk>
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/neighborhood.R
\name{neighborhood}
\alias{neighborhood}
\title{Nearest neighborhoods for kernel smoothing}
\usage{
neighborhood(x, bandwidth = NULL, kernel = "box")
}
\arguments{
\item{x}{Numeric vector -- typically the observations of a continuous random
variate.}

\item{bandwidth}{Controls the distance between neighbors in a neighborhood.
It can be a decimal, i.e.\ the bandwidth, or the string `"smooth"', in which
case \code{N^{-1/4}} is used, \code{N} being the sample size, or \code{NULL}
in which case the \code{\link[KernSmooth]{dpik}} function of the package KernSmooth is
used to find the optimal bandwidth.}

\item{kernel}{Only the rectangular kernel ("box") is implemented.}
}
\value{
An object of class 'neighborhood'.  The value is a list that
includes the unique values of `x' (\code{values}) for which a neighborhood,
consisting of the nearest neighbors, is defined by the first neighbor
(\code{first.nbh}) of the usually very long vector \code{neighbors} and the
size of the neighborhood (\code{size.nbh}).

Further values are the arguments \code{bandwidth}, \code{kernel}, the total
sample size \code{n} and the number of unique values \code{nu}.
}
\description{
Nearest neighborhoods for the values of a continuous predictor. The result
is used for the conditional Kaplan-Meier estimator and other conditional
product limit estimators.
}
\examples{

d <- SimSurv(20)
neighborhood(d$X2)
}
\references{
Stute, W. "Asymptotic Normality of Nearest Neighbor Regression
Function Estimates", \emph{The Annals of Statistics}, 1984,12,917--926.
}
\seealso{
\code{\link[KernSmooth]{dpik}}, \code{\link{prodlim}}
}
\author{
Thomas Gerds
}
\keyword{smooth}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/atRisk.R
\name{atRisk}
\alias{atRisk}
\title{Drawing numbers of subjects at-risk of experiencing an event below
Kaplan-Meier and Aalen-Johansen plots.}
\usage{
atRisk(
  x,
  newdata,
  times,
  line,
  col,
  labelcol = NULL,
  interspace,
  cex,
  labels,
  title = "",
  titlecol = NULL,
  pos,
  adj,
  dist,
  xdist,
  adjust.labels = TRUE,
  show.censored = FALSE,
  unit = "npc",
  ...
)
}
\arguments{
\item{x}{an object of class `prodlim' as returned by the
\code{prodlim} function.}

\item{newdata}{see \code{plot.prodlim}}

\item{times}{Where to compute the atrisk numbers.}

\item{line}{Distance of the atrisk numbers from the inner plot.}

\item{col}{The color of the text.}

\item{labelcol}{The color for the labels. Defaults to col.}

\item{interspace}{Distance between rows of atrisk numbers.}

\item{cex}{Passed on to \code{mtext} for both atrisk numbers and
labels.}

\item{labels}{Labels for the at-risk rows.}

\item{title}{Title for the at-risk labels}

\item{titlecol}{The color for the title. Defaults to 1 (black).}

\item{pos}{The value is passed on to the \code{mtext} argument
\code{at} for the labels (not the atrisk numbers).}

\item{adj}{Passed on to \code{mtext} for the labels (not the atriks
numbers).}

\item{dist}{If \code{line} is missing, the distance of the upper
most atrisk row from the inner plotting region: par()$mgp[2].}

\item{xdist}{Distance in x-axis direction to define the distance between the labels
and the numbers at-risk. Deftaults to \code{strwidth("MM",cex=cex)}.}

\item{adjust.labels}{If \code{TRUE} the labels are left adjusted.}

\item{show.censored}{If \code{TRUE} the cumulative number of subjects
lost to follow up is shown in parentheses.}

\item{unit}{The graphical coordinate systems unit to convert from when line2user is calling \code{grconvertX} and \code{grconvertY}.
Default is \code{'npc'}}

\item{...}{Further arguments that are passed to the function
\code{mtext}.}
}
\value{
Nil
}
\description{
This function is invoked and controlled by \code{plot.prodlim}.
}
\details{
This function should not be called directly. The arguments can be specified
as \code{atRisk.arg} in the call to \code{plot.prodlim}.
}
\seealso{
\code{\link{plot.prodlim}}, \code{\link{confInt}},
\code{\link{markTime}}
}
\author{
Thomas Alexander Gerds <tag@biostat.ku.dk>
}
\keyword{survival}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/getEvent.R
\name{getEvent}
\alias{getEvent}
\title{Extract a column from an event history object.}
\usage{
getEvent(object, mode = "factor", column = "event")
}
\arguments{
\item{object}{Object of class \code{"Hist"}.}

\item{mode}{Return mode. One of \code{"numeric"}, \code{"character"}, or
\code{"factor"}.}

\item{column}{Name of the column to extract from the object.}
}
\description{
Extract a column from an event history object, as obtained with the function
\code{\link{Hist}}.
}
\details{
Since objects of class \code{"Hist"} are also matrices, all columns are
numeric or integer valued. To extract a correctly labeled version, the
attribute \code{states} of the object is used to generate factor levels.
}
\examples{

  dat= data.frame(time=1:5,event=letters[1:5])
  x=with(dat,Hist(time,event))
  ## inside integer
  unclass(x)
  ## extract event (the extra level "unknown" is for censored data)
  getEvent(x)

}
\seealso{
\code{\link{Hist}}
}
\author{
Thomas Alexander Gerds <tag@biostat.ku.dk>
}
\keyword{survival}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/SimCompRisk.R
\name{SimCompRisk}
\alias{SimCompRisk}
\title{Simulate competing risks data}
\usage{
SimCompRisk(N, ...)
}
\arguments{
\item{N}{sample size}

\item{...}{do nothing.}
}
\value{
data.frame with simulated data
}
\description{
Simulate right censored competing risks data with two covariates X1 and X2. Both covariates have effect exp(1) on the hazards of event 1 and zero effect on the hazard of event 2.
}
\details{
This function calls \code{crModel}, then adds covariates and finally calls \code{sim.lvm}.
}
\examples{

SimCompRisk(10)
            
}
\author{
Thomas Alexander Gerds
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/as.data.frame.prodlim.R
\name{as.data.frame.prodlim}
\alias{as.data.frame.prodlim}
\title{Turn prodlim objects into a data.frame}
\usage{
\method{as.data.frame}{prodlim}(x, ...)
}
\arguments{
\item{x}{object obtained with function \code{\link{prodlim}}}

\item{...}{passed to \code{\link{summary.prodlim}}}
}
\value{
A data.table with the results of the prodlim object
}
\description{
Collect results of a fitted prodlim object in a data.frame
}
\details{
By default object contains results for all fitted time points and all strata.
Use arguments times and newdata of \code{\link{summary.prodlim}} to subset.
}
\examples{
set.seed(8)
d <- SimCompRisk(17)
fit <- prodlim(Hist(time,event)~X1,data=d)
as.data.frame.prodlim(fit)
as.data.frame.prodlim(fit)

}
\seealso{
\code{\link{prodlim}}
}
\author{
Thomas A. Gerds <tag@biostat.ku.dk>
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/quantile.prodlim.R
\name{quantile.prodlim}
\alias{quantile.prodlim}
\title{Quantiles for Kaplan-Meier and Aalen-Johansen estimates.}
\usage{
\method{quantile}{prodlim}(x, q, cause = 1, ...)
}
\arguments{
\item{x}{Object of class \code{"prodlim"}.}

\item{q}{Quantiles. Vector of values between 0 and 1.}

\item{cause}{For competing risks the cause of interest.}

\item{...}{not used}
}
\description{
Quantiles for Kaplan-Meier and Aalen-Johansen estimates.
}
\examples{
library(lava)
set.seed(1)
d=SimSurv(30)

# Quantiles of the potential followup time
g=prodlim(Hist(time,status)~1,data=d,reverse=TRUE)
quantile(g)

# survival time
f=prodlim(Hist(time,status)~1,data=d)
f1=prodlim(Hist(time,status)~X1,data=d)
# default: median and IQR
quantile(f)
quantile(f1)
# median alone
quantile(f,.5)
quantile(f1,.5)

# competing risks
set.seed(3)
dd = SimCompRisk(30)
ff=prodlim(Hist(time,event)~1,data=dd)
ff1=prodlim(Hist(time,event)~X1,data=dd)
## default: median and IQR
quantile(ff)
quantile(ff1)

print(quantile(ff1),na.val="NA")
print(quantile(ff1),na.val="Not reached")

}
\author{
Thomas Alexander Gerds <tag@biostat.ku.dk>
}
\keyword{survival}
